/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "backend-vdcs.h"
#include <ds/asio/test.h>
#include <ds/catch/catch.h>
#include <ds/clock-mock.h>
#include <ds/conf-mock.h>
#include <ds/json.h>
#include <ds/log.h>
#include "businterface-mock.h"
#include "dsm-api-mock.h"
#include "event-mock.h"
#include "event.h"
#include "http-client-mock.h"
#include "model/dsmeter.h"
#include "model/set.h"
#include "propertysystem_common_paths.h"

static ds::LogChannel dsLogChannel("backendVdcsTest");

static const char* TAGS = "[dss][dssBackendVdcs]";

using namespace dss;
using namespace trompeloeil;
using namespace std::literals;

namespace {

TEST_CASE("dssBackendVdcs", TAGS) {
  ds::ClockMocker clock;
  PropertySystem propertySystem;
  BusInterfaceMock busInterface;
  ALLOW_CALL(busInterface, add(_));

  EventSinkMock eventSink;
  ALLOW_CALL(eventSink, pushEvent(_));

  auto pApartment = makeApartment(nullptr, &propertySystem);
  auto& apartment = *pApartment;
  apartment.setBusInterface(&busInterface);
  apartment.setEventSink(eventSink);

  ALLOW_CALL(dsm_api_mock, DsmApiGetDeviceDescription(_, _, _, _, _, _, _)).RETURN(0);

  EventInterpreter eventInterpreter(nullptr);
  eventInterpreter.initialize();
  ds::asio::TestLoopResource loop;

  SimpleHttpClientMock httpClient;
  const auto baseUrl = "test://test";
  const auto dssId = "dssId";
  const auto ospToken = "ospToken";
  decltype(HttpRequest::headers) headers{{"x-ds-dss-id", dssId}, {"x-ds-osp-token", ospToken}};
  propertySystem.createProperty(pp_websvc_mshub_active)->setBooleanValue(true);
  propertySystem.createProperty(pp_sysinfo_dsid)->setStringValue(dssId);
  propertySystem.createProperty(pp_websvc_rc_osptoken)->setStringValue(ospToken);
  propertySystem.createProperty(pp_websvc_private_url)->setStringValue(baseUrl);
  auto getRequest = [&](const std::string& path) {
    return HttpRequest{.url = ds::str(baseUrl, path), .type = HttpMethod::GET, .headers = headers};
  };

  auto backendVdcs = makeBackendVdcs(loop, apartment, eventInterpreter, httpClient);

  SECTION("can be destroyed after no activity") {}

  SECTION("reregister retries on /v1/cloudvdc/vdcs failure") {
    backendVdcs->reregister({.succeeded = [&] { loop.stop(); }, .attemptFailed = [&] { loop.stop(); }});
    {
      REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs")))
          .RETURN(HttpResponse{})
          .LR_SIDE_EFFECT(DS_FAIL_REQUIRE("test failure"));
      loop.run();
    }
    {
      clock.advance(10s);
      REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs")))
          .RETURN(HttpResponse{.code = 200, .data = "{}"});
      loop.run();
    }
    CHECK(apartment.getDSMeters().empty());
  }

  SECTION("reregister retries on /v1/cloudvdc/vdc0/devices failure") {
    backendVdcs->reregister({.succeeded = [&] { loop.stop(); }, .attemptFailed = [&] { loop.stop(); }});
    ALLOW_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs")))
        .RETURN(HttpResponse{.code = 200, .data = R"({"vdc0":{"dSUID": "1111111111111111111111111111111100"}})"});
    {
      REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs/vdc0/devices")))
          .RETURN(HttpResponse{})
          .LR_SIDE_EFFECT(DS_FAIL_REQUIRE("test failure"));
      loop.run();
    }
    {
      clock.advance(10s);
      REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs/vdc0/devices")))
          .RETURN(HttpResponse{.code = 200, .data = R"([{"dSUID": "2222222222222222222222222222222200"}])"});
      loop.run();
    }
    CHECK(!apartment.getDSMeters().empty());
  }

  SECTION("no vdcs, no devices") {
    REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs"))).RETURN(HttpResponse{.code = 200, .data = "{}"});
    backendVdcs->reregister({.succeeded = [&] { loop.stop(); }});
    loop.run();
    CHECK(apartment.getDSMeters().empty());
    CHECK(apartment.getDevices().empty());
  }

  SECTION("vdc fields are propagated to model - smoke") {
    REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs"))).RETURN(HttpResponse{.code = 200, .data = R"({
          "vdc0": {
            "dSUID": "1111111111111111111111111111111100",
            "model": "model0",
            "modelVersion": "modelVersion0",
            "modelUID": "modelUID0",
            "hardwareGuid": "hardwareGuid0",
            "hardwareModelGuid": "hardwareModelGuid0",
            "vendorGuid": "vendorGuid0",
            "oemGuid": "oemGuid0",
            "oemModelGuid": "oemModelGuid0",
            "configURL": "http://example.com/configURL0",
            "name": "name0",
            "authorized": true,
            "displayId": "displayId"
          }
        })"});
    REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs/vdc0/devices")))
        .RETURN(HttpResponse{.code = 200, .data = "[]"});
    backendVdcs->reregister({.succeeded = [&] { loop.stop(); }});
    loop.run();
    auto x = apartment.getDSMeterByDSID(Dsuid{"1111111111111111111111111111111100"});
    CHECK(x->getDSID() == Dsuid{"1111111111111111111111111111111100"});
    CHECK(x->getBusMemberType() == BusMember_backendVdc);
    CHECK(x->getHardwareName() == "model0");
    CHECK(x->getSoftwareVersion() == "modelVersion0");
    CHECK(x->getVdcHardwareGuid() == "hardwareGuid0");
    CHECK(x->getVdcHardwareModelGuid() == "hardwareModelGuid0");
    CHECK(x->getVdcVendorGuid() == "vendorGuid0");
    CHECK(x->getVdcOemGuid() == "oemGuid0");
    CHECK(x->getVdcModelUID() == "modelUID0");
    CHECK(x->getVdcOemModelGuid() == "oemModelGuid0");
    CHECK(x->getVdcConfigURL() == "http://example.com/configURL0");
    CHECK(x->getName() == "name0");
    CHECK(x->getVdcImplementationId() == "vdc0");
    CHECK(x->authorized() == true);
    CHECK(x->isPresent() == true);
    CHECK(x->isConnected() == true);
    CHECK(x->isValid() == true);
  }

  SECTION("Generic device fields are propagated to model - smoke") {
    REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs"))).RETURN(HttpResponse{.code = 200, .data = R"({
          "vdc0": {
            "dSUID": "1111111111111111111111111111111100"
          }
        })"});
    REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs/vdc0/devices")))
        .RETURN(HttpResponse{.code = 200, .data = R"([{
          "displayId": "2822787917022775630",
          "model": "Smoke detector",
          "oemModelGuid": "gs1:(01)7640156794052",
          "primaryGroup": 8,
          "active": true,
          "dSUID": "2222222222222222222222222222222200"
        }])"});
    backendVdcs->reregister({.succeeded = [&] { loop.stop(); }});
    loop.run();
    auto x = apartment.getDeviceByDSID(Dsuid{"2222222222222222222222222222222200"});
    CHECK(x->getDSID() == Dsuid{"2222222222222222222222222222222200"});
    CHECK(x->isVdcDevice() == true);
    CHECK(x->getName() == "");
    CHECK(x->getDSMeterDSID() == Dsuid{"1111111111111111111111111111111100"});
    CHECK(x->isPresent() == true);
    CHECK(x->isConnected() == true);
    CHECK(x->isValid() == true);
    CHECK(x->getActiveGroup() == static_cast<int>(ApplicationType::joker));
    CHECK(x->getGroupIds() == std::vector<int>{static_cast<int>(ApplicationType::joker)});
  }

  SECTION("reregister") {
    SECTION("having vdc0") {
      ALLOW_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs/vdc0/devices")))
          .RETURN(HttpResponse{.code = 200, .data = "[]"});
      ALLOW_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs/vdc1/devices")))
          .RETURN(HttpResponse{.code = 200, .data = "[]"});

      REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs"))).RETURN(HttpResponse{.code = 200, .data = R"({
            "vdc0": {
              "dSUID": "1111111111111111111111111111111100",
              "model": "model0"
            }
          })"});
      backendVdcs->reregister({.succeeded = [&] { loop.stop(); }});
      loop.run();
      auto vdc0 = apartment.getDSMeterByDSID(Dsuid{"1111111111111111111111111111111100"});
      REQUIRE(vdc0->getHardwareName() == "model0");
      REQUIRE(vdc0->isPresent());

      SECTION("add new vdc1") {
        REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs"))).RETURN(HttpResponse{.code = 200, .data = R"({
              "vdc0": {
                "dSUID": "1111111111111111111111111111111100",
                "model": "model0"
              },
              "vdc1": {
                "dSUID": "1111111111111111111111111111111101",
                "model": "model1"
              }
            })"});
        backendVdcs->reregister({.succeeded = [&] { loop.stop(); }});
        loop.run();
        REQUIRE(vdc0->getHardwareName() == "model0");
        REQUIRE(vdc0->isPresent());
        auto vdc1 = apartment.getDSMeterByDSID(Dsuid{"1111111111111111111111111111111101"});
        REQUIRE(vdc1->getHardwareName() == "model1");
        REQUIRE(vdc1->isPresent());
      }

      SECTION("update matching vdc0") {
        REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs"))).RETURN(HttpResponse{.code = 200, .data = R"({
              "vdc0": {
                "dSUID": "1111111111111111111111111111111100",
                "model": "model01"
              }
            })"});
        backendVdcs->reregister({.succeeded = [&] { loop.stop(); }});
        loop.run();
        REQUIRE(vdc0->getHardwareName() == "model01");
        REQUIRE(vdc0->isPresent());
      }

      SECTION("inactivates missing vdc0") {
        REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs"))).RETURN(HttpResponse{.code = 200, .data = R"({
            })"});
        backendVdcs->reregister({.succeeded = [&] { loop.stop(); }});
        loop.run();
        REQUIRE(vdc0->getHardwareName() == "model0");
        REQUIRE(!vdc0->isPresent());
      }

      SECTION("activates matching inactive vdc0") {
        REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs")))
            .RETURN(HttpResponse{.code = 200, .data = R"({})"});
        backendVdcs->reregister({.succeeded = [&] { loop.stop(); }});
        loop.run();
        REQUIRE(!vdc0->isPresent());
        REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs"))).RETURN(HttpResponse{.code = 200, .data = R"({
              "vdc0": {
                "dSUID": "1111111111111111111111111111111100"
              }
            })"});
        backendVdcs->reregister({.succeeded = [&] { loop.stop(); }});
        loop.run();
        REQUIRE(vdc0->isPresent());
      }
    }
  }

  SECTION("reregisterDevices") {
    SECTION("having vdc0 with device0") {
      REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs"))).RETURN(HttpResponse{.code = 200, .data = R"({
            "vdc0": {
              "dSUID": "1111111111111111111111111111111100"
            }
          })"});

      REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs/vdc0/devices")))
          .RETURN(HttpResponse{.code = 200, .data = R"([{
            "dSUID": "2222222222222222222222222222222200",
            "model": "model0",
            "active": true
          }])"});
      backendVdcs->reregister({.succeeded = [&] { loop.stop(); }});
      loop.run();
      auto device0 = apartment.getDeviceByDSID(Dsuid{"2222222222222222222222222222222200"});
      REQUIRE(device0->getVdsdSpec().model == "model0");
      REQUIRE(device0->isPresent());

      SECTION("add new device1") {
        REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs/vdc0/devices")))
            .RETURN(HttpResponse{.code = 200, .data = R"([{
              "dSUID": "2222222222222222222222222222222200",
              "model": "model0",
              "active": true
            },{
              "dSUID": "2222222222222222222222222222222201",
              "model": "model1",
              "active": true
            }])"});
        backendVdcs->reregisterDevices(Dsuid{"1111111111111111111111111111111100"});
        REQUIRE(device0->getVdsdSpec().model == "model0");
        REQUIRE(device0->isPresent());
        auto device1 = apartment.getDeviceByDSID(Dsuid{"2222222222222222222222222222222201"});
        REQUIRE(device1->getVdsdSpec().model == "model1");
        REQUIRE(device1->isPresent());
      }

      SECTION("update matching device0") {
        REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs/vdc0/devices")))
            .RETURN(HttpResponse{.code = 200, .data = R"([{
              "dSUID": "2222222222222222222222222222222200",
              "model": "model01",
              "active": true
            }])"});
        backendVdcs->reregisterDevices(Dsuid{"1111111111111111111111111111111100"});
        REQUIRE(device0->getVdsdSpec().model == "model01");
        REQUIRE(device0->isPresent());
      }

      SECTION("inactivates missing device0") {
        REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs/vdc0/devices")))
            .RETURN(HttpResponse{.code = 200, .data = R"([])"});
        backendVdcs->reregisterDevices(Dsuid{"1111111111111111111111111111111100"});
        REQUIRE(device0->getVdsdSpec().model == "model0");
        REQUIRE(!device0->isPresent());
      }

      SECTION("activates matching inactive device0") {
        REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs/vdc0/devices")))
            .RETURN(HttpResponse{.code = 200, .data = R"([])"});
        backendVdcs->reregisterDevices(Dsuid{"1111111111111111111111111111111100"});
        REQUIRE(!device0->isPresent());
        REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs/vdc0/devices")))
            .RETURN(HttpResponse{.code = 200, .data = R"([{
              "dSUID": "2222222222222222222222222222222200",
              "active": true
            }])"});
        backendVdcs->reregisterDevices(Dsuid{"1111111111111111111111111111111100"});
        REQUIRE(device0->isPresent());
      }
    }
  }

  SECTION("deviceEvent-action:ready") {
    ALLOW_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs"))).RETURN(HttpResponse{.code = 200, .data = R"({
          "vdc0": {
            "dSUID": "1111111111111111111111111111111100"
          }
        })"});
    ALLOW_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs/vdc0/devices")))
        .RETURN(HttpResponse{.code = 200, .data = R"([{
          "dSUID": "2222222222222222222222222222222200",
          "active": true
        }])"});
    DeviceReference deviceReference{Dsuid{"2222222222222222222222222222222200"}, &apartment};
    Event event{EventName::DeviceEvent, deviceReference, Properties{{"action", "ready"}}};

    SECTION("is called once for new device") {
      {
        REQUIRE_CALL(eventSink, pushEvent(_)).WITH(*_1 == event);
        backendVdcs->reregister({.succeeded = [&] { loop.stop(); }});
        loop.run();
      }
      {
        FORBID_CALL(eventSink, pushEvent(_)).WITH(*_1 == event);
        backendVdcs->reregister({.succeeded = [&] { loop.stop(); }});
        loop.run();
      }
      REQUIRE(true);
    }

    SECTION("is called once for device already in model") {
      apartment.allocateDevice(Dsuid{"2222222222222222222222222222222200"});
      {
        REQUIRE_CALL(eventSink, pushEvent(_)).WITH(*_1 == event);
        backendVdcs->reregister({.succeeded = [&] { loop.stop(); }});
        loop.run();
      }
      {
        FORBID_CALL(eventSink, pushEvent(_)).WITH(*_1 == event);
        backendVdcs->reregister({.succeeded = [&] { loop.stop(); }});
        loop.run();
      }
      REQUIRE(true);
    }
  }

  SECTION("deviceEvent-action:added") {
    ALLOW_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs"))).RETURN(HttpResponse{.code = 200, .data = R"({
          "vdc0": {
            "dSUID": "1111111111111111111111111111111100"
          }
        })"});
    ALLOW_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs/vdc0/devices")))
        .RETURN(HttpResponse{.code = 200, .data = R"([{
          "dSUID": "2222222222222222222222222222222200",
          "active": true
        }])"});
    DeviceReference deviceReference{Dsuid{"2222222222222222222222222222222200"}, &apartment};
    Event event{EventName::DeviceEvent, deviceReference, Properties{{"action", "added"}}};

    SECTION("is called once for new device") {
      {
        REQUIRE_CALL(eventSink, pushEvent(_)).WITH(*_1 == event);
        backendVdcs->reregister({.succeeded = [&] { loop.stop(); }});
        loop.run();
      }
      {
        FORBID_CALL(eventSink, pushEvent(_)).WITH(*_1 == event);
        backendVdcs->reregister({.succeeded = [&] { loop.stop(); }});
        loop.run();
      }
      REQUIRE(true);
    }

    SECTION("is not called for device already in model") {
      apartment.allocateDevice(Dsuid{"2222222222222222222222222222222200"});
      {
        FORBID_CALL(eventSink, pushEvent(_)).WITH(*_1 == event);
        backendVdcs->reregister({.succeeded = [&] { loop.stop(); }});
        loop.run();
      }
      REQUIRE(true);
    }
  }

  SECTION("notify vDC") {
      REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs"))).RETURN(HttpResponse{.code = 200, .data = R"({
            "vdc0": {
              "dSUID": "1111111111111111111111111111111100"
            }
          })"});
      REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs/vdc0/devices")))
              .RETURN(HttpResponse{.code = 200, .data = "[]"});

      backendVdcs->reregister({.succeeded = [&] { loop.stop(); }});
      loop.run();
      auto vdc0 = apartment.getDSMeterByDSID(Dsuid{"1111111111111111111111111111111100"});
      CHECK(vdc0->getName() == "");
      CHECK(vdc0->getHardwareName() == "");

      SECTION("PUT notification") {
        const auto& path = "/vdcs/vdc0";
        const auto& params = R"({"dSUID": "1111111111111111111111111111111100", "name": "name0", "model": "model0"})";
        backendVdcs->notify(path, PutResource{ds::fromJson<ds::JsonFragment>(params)});

        CHECK(vdc0->getName() == "name0");
        CHECK(vdc0->getHardwareName() == "model0");
      }

      SECTION("DELETE notification") {
        const auto& path = "/vdcs/vdc0";
        backendVdcs->notify(path, DeleteResource{});

        REQUIRE(!vdc0->isPresent());
      }
    }

  SECTION("notify vDC device") {
      ALLOW_CALL(dsm_api_mock, DsmApiGetDeviceDescription(_, _, _, _, _, _, _)).RETURN(0);

      REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs"))).RETURN(HttpResponse{.code = 200, .data = R"({
            "vdc0": {
              "dSUID": "1111111111111111111111111111111100"
            }
          })"});
      REQUIRE_CALL(httpClient, request(getRequest("/v1/cloudvdc/vdcs/vdc0/devices")))
              .RETURN(HttpResponse{.code = 200, .data = R"([{
        "dSUID": "2222222222222222222222222222222200",
        "model": "model0",
        "active": true
      }])"});

      backendVdcs->reregister({.succeeded = [&] { loop.stop(); }});
      loop.run();
      auto vdc0 = apartment.getDSMeterByDSID(Dsuid{"1111111111111111111111111111111100"});
      auto device0 = apartment.getDeviceByDSID(Dsuid{"2222222222222222222222222222222200"});

      CHECK(device0->getVdsdSpec().model == "model0");
      CHECK(device0->isPresent());

      SECTION("PUT notification") {
        const auto& path = "/vdcs/vdc0/devices/2222222222222222222222222222222200";
        const auto& params = R"({"dSUID": "2222222222222222222222222222222200", "model": "model01"})";
        backendVdcs->notify(path, PutResource{ds::fromJson<ds::JsonFragment>(params)});

        CHECK(device0->getVdsdSpec().model == "model01");
      }

      SECTION("DELETE notification") {
        const auto& path = "/vdcs/vdc0/devices/2222222222222222222222222222222200";
        backendVdcs->notify(path, DeleteResource{});

        REQUIRE(!device0->isPresent());
      }
    }
}

} // namespace
