/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "controller-statuses.h"

#include <ds/log.h>

static ds::LogChannel dsLogChannel("controllerStatuses");

namespace dss {

struct DsaLedInfo {
  struct Dsm {
    std::string dsuid;
    int ledcode;
    std::string status;
  };
  std::vector<Dsm> dsms;
};

ControllerStatuses controllerStatuses(const std::locale& locale, DsaHttpClient& dsa, BackendVdcs& backendVdcs) {
  ControllerStatuses result;
  // Do not log http requests at notice log severity, because this api is polled and would be verbose in logs.
  try {
    auto ledInfo = ds::fromJson<DsaLedInfo>(
        dsa.request(dss::HttpRequest{.url = "/dsa/json/dsm/getLedInfo"}, ds::LogSeverity::INFO));
    for (auto& dsm : ledInfo.dsms) {
      try {
        result.push_back({dsm.dsuid,
                          {.code =
                               [&] {
                                 switch (dsm.ledcode) {
                                   case 0: return ControllerStatus::Code::ready; // LED_green
                                   case 1: return ControllerStatus::Code::error; // LED_red_black
                                   case 2: return ControllerStatus::Code::error; // LED_red_yellow
                                   case 3: return ControllerStatus::Code::error; // LED_red_blue
                                   case 4: return ControllerStatus::Code::error; // LED_magenta
                                   case 5: return ControllerStatus::Code::off;   // LED_off
                                 }
                                 return ControllerStatus::Code::off;
                               }(),
                           .message = std::move(dsm.status)}});
      } catch (const std::exception& e) {
        DS_ERROR("Failed to process controller status.", dsm, e);
      }
    }
  } catch (const std::exception& e) {
    DS_ERROR("Failed to process controller statuses.", e);
  }

  if (backendVdcs.hasVdcs()) {
    try {
      auto statuses = ds::fromJson<std::vector<std::pair<BackendVdcId, ControllerStatus>>>(
          backendVdcs.http(HttpRequest{.url = "/v1/cloudvdc/vdcstatus/"}, locale, ds::LogSeverity::INFO));
      for (auto& status : statuses) {
        try {
          result.emplace_back(backendVdcs.vdcDsuidForId(status.first), std::move(status.second));
        } catch (const std::exception& e) {
          DS_ERROR("Failed to process controller status.", status, e);
        }
      }
    } catch (const std::exception& e) {
      DS_ERROR("Failed to process backend controller statuses.", e);
    }
  }
  DS_INFO_LEAVE(result);
  return result;
}

} // namespace dss

DS_REFLECT_STRUCT((dss), DsaLedInfo::Dsm, (dsuid)(ledcode)(status));
DS_REFLECT_STRUCT((dss), DsaLedInfo, (dsms));
