/*
    Copyright (c) 2020 digitalSTROM.org," " Zurich," " Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation," " either version 3 of the License," " or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful," "
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not," " see <http://www.gnu.org/licenses/>.
*/
#include "installation.h"

#include <ds-api/rest-mock.h>
#include <ds/asio/test.h>
#include <model/apartment-mock.h>
#include <model/dsmeter.h>
#include <model/floor.h>
#include <model/zone.h>

using namespace std::literals;
static const char* TAGS = "[dss][dsApartment][dsApartmentInstallation]";

namespace {

TEST_CASE("dsApartmentInstallation", TAGS) {
  ds::asio::TestLoopResource loop;
  auto propertySystem = dss::PropertySystem{};
  auto configNode = propertySystem.createProperty("/config");
  dss::ApartmentMock apartment{nullptr, &propertySystem};
  const auto apartmentId = dsModel::ApartmentId{"apartmentId-0"};
  ALLOW_CALL(apartment, id()).RETURN(apartmentId);
  apartment.setName("apartment1");
  dsApi::AsyncCalleeMock bus;
  auto installation = dsApartment::makeInstallation(loop, apartment, bus);

  SECTION("floors") {
    auto& floor = *apartment.getFloor(0);
    floor.setName("floor0");
    auto& zone1 = *apartment.allocateZone(1);

    SECTION("floor") {
      CHECK(installation->floor(floor.id()) ==
            dsModel::Floor{.id = dsModel::FloorId{"0"},
                           .type = "floor",
                           .lastChanged = {},
                           .attributes = {.name = "floor0", .orderId = 0, .zones = {zone1.id()}}});
    }

    SECTION("special character") {
      floor.setName("floor&amp;");
      CHECK(installation->floor(floor.id()) ==
            dsModel::Floor{.id = dsModel::FloorId{"0"},
                           .type = "floor",
                           .lastChanged = {},
                           .attributes = {.name = "floor&", .orderId = 0, .zones = {zone1.id()}}});
    }

    SECTION("floors") {
      CHECK(installation->floors() ==
            dsModel::Floors{.floors = {{.id = dsModel::FloorId{"0"},
                                        .type = "floor",
                                        .lastChanged = {},
                                        .attributes = {.name = "floor0", .orderId = 0, .zones = {zone1.id()}}}}});
    }

    SECTION("setFloorName") {
      THEN("success") {
        installation->setFloorName(floor.id(), "floor1x");
        CHECK(floor.getName() == "floor1x");
      }

      THEN("Empty name fails") {
        CHECK_THROWS_FIND(installation->setFloorName(floor.id(), ""), "New name cannot be empty");
      }

      THEN("Invalid zoneId fails") {
        CHECK_THROWS_FIND(installation->setFloorName(dsModel::FloorId{"455"}, "xyz"), "Floor not found");
      }
    }
  }

  SECTION("installation") {
    const auto installationId = dsModel::InstallationId{apartmentId};
    SECTION("No properties") {
      CHECK(installation->installation() ==
            dsModel::Installation{.id = installationId, .type = "installation", .lastChanged = {}, .attributes = {}});
    }
    SECTION("Full properties") {
      auto geodataNode = configNode->createProperty("geodata");
      geodataNode->setChildValue("longitude", "11.5"s);
      geodataNode->setChildValue("latitude", "22.5"s);
      geodataNode->setChildValue("country", "DE"s);
      geodataNode->setChildValue("location", "Berlin"s);
      REQUIRE_CALL(loop.system, timeZone()).RETURN("Europe/Test");

      CHECK(installation->installation() ==
            dsModel::Installation{.id = installationId,
                                  .type = "installation",
                                  .lastChanged = {},
                                  .attributes = {.location = {.longitude = 11.5, .latitude = 22.5},
                                                 .countryCode = "DE",
                                                 .city = "Berlin",
                                                 .timeZone = "Europe/Test"}});
    }
  }

  SECTION("dsServer") {
    const auto serverId = dsModel::DSServerId{"dsuid"};

    auto systemNode = propertySystem.createProperty("/system/");
    systemNode->setChildValue("dSUID", "dsuid"s);
    systemNode->setChildValue("dSID", "dsid"s);
    auto versionNode = propertySystem.createProperty("/system/version/");
    versionNode->setChildValue("distroVersion", "1.0"s);
    versionNode->setChildValue("model", "model"s);
    versionNode->setChildValue("gtin", "gtin"s);

    CHECK(installation->dsServer() ==
          dsModel::DSServer{.id = serverId,
                            .type = "dsServer",
                            .lastChanged = {},
                            .attributes = {.name = "apartment1",
                                           .model = "model",
                                           .gtin = "gtin",
                                           .dsid = "dsid",
                                           .version = "1.0",
                                           .environment = dsModel::Environment::development}});

    SECTION("special character") {
      apartment.setName("Test&amp;Name");
      CHECK(installation->dsServer().attributes.name == "Test&Name");
    }

    SECTION("environment") {
      struct X {
        dss::eState dssValue;
        std::optional<bool> dsValue;
      };
      const auto& x = GENERATE(values(std::initializer_list<std::pair<ds::Release, dsModel::Environment>>{
          {ds::Release::devel, dsModel::Environment::development},
          {ds::Release::testing, dsModel::Environment::alpha},
          {ds::Release::fieldtest, dsModel::Environment::beta},
          {ds::Release::production, dsModel::Environment::production}}));
      REQUIRE_CALL(loop.system, release()).RETURN(x.first);
      CHECK(installation->dsServer().attributes.environment == x.second);
    }
  }

  SECTION("controller(s)") {
    auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
    auto& controller1 = *apartment.allocateDSMeter(dss::Dsuid{controllerId1});
    controller1.setName("controller1");

    auto controllerId0 = dsModel::ControllerId{"1111111111111111111111111111111000"};
    auto& controller0 = *apartment.allocateDSMeter(dss::Dsuid{controllerId0});
    controller0.setName("controller0");

    SECTION("controler(id) succeeds") {
      CHECK(installation->controller(controllerId1) ==
            dsModel::Controller{
                .id = controllerId1, .type = "controller", .lastChanged = {}, .attributes = {.name = "controller1"}});
    }

    SECTION("name is unescaped") {
      controller1.setName("&amp;");
      CHECK(installation->controller(controllerId1).attributes.name == "&");
    }

    SECTION("controler(id) with wrong id fails") {
      CHECK_THROWS_FIND(installation->controller(dsModel::ControllerId{"1111111111111111111111111111111299"}),
                        "Controller not found");
    }

    SECTION("controlers()") {
      CHECK(
          installation->controllers() ==
          dsModel::Controllers{
              .controllers = {
                  {.id = controllerId0, .type = "controller", .lastChanged = {}, .attributes = {.name = "controller0"}},
                  {.id = controllerId1,
                   .type = "controller",
                   .lastChanged = {},
                   .attributes = {.name = "controller1"}}}});
    }
  }
}
} // namespace
