/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <ds/asio/ds-signal.h>
#include <ds/asio/resource.h>
#include <ds/box.h>
#include <ds/clock.h>
#include <ds/json.h>

#include <boost/noncopyable.hpp>
#include <boost/serialization/strong_typedef.hpp>

namespace dsApi {

BOOST_STRONG_TYPEDEF(std::string, EventType);
struct Event {
  EventType type;
  std::string path;
  ds::JsonFragment data;
};

struct Subscription {
  dsApi::EventType type;
  ds::Clock::time_point expiresAt;
};

struct EventSink;

/// EventHub distributes events to subscribed EventSinks in thread safe manner.
struct EventHub : boost::noncopyable {
  virtual ~EventHub() = default;

  /// Emit event to subscribed event sinks. Thread safe.
  virtual void emitEvent(const ds::Box<dsApi::Event> &x) = 0;

  /// Event sink subscribes itself.
  ///
  /// No functionality now, only for mocking.
  /// Later we can notify event owner about subscription
  virtual void subscribe(const Subscription &, EventSink &) = 0;

private:
  virtual ds::Slots<ds::Box<dsApi::Event>> &event() = 0;
  friend struct EventSink;
};

struct EventSinkHooks {
  std::function<void(ds::Box<dsApi::Event> &&)> event;
  // primarily to support unit tests
  std::function<void(ds::Box<dsApi::Event> &&)> discardedEvent = [](auto &&) {};
};

/// EventSink holds subscriptions to events in EventHub
/// and forward subscribed events to event hook.
struct EventSink : boost::noncopyable {
  virtual ~EventSink() = default;
  virtual void subscribe(Subscription &&) = 0;

protected:
  static ds::Slots<ds::Box<dsApi::Event>> &hubEvent(EventHub &hub) { return hub.event(); }
};

std::unique_ptr<EventHub> makeEventHub(ds::asio::ResourceBase &parent);
std::unique_ptr<EventSink> makeEventSink(ds::asio::ResourceBase &parent, EventHub &, EventSinkHooks &&);

struct InsideSubscription {
  dsApi::EventType type;
  ds::DateTime expiresAt;
};

} // namespace dsApi

DS_JSON_DECLARE_STRUCT((dsApi), Event, (type)(path)(data));
DS_REFLECT_DECLARE_STRUCT((dsApi), Subscription, (type)(expiresAt));
DS_JSON_DECLARE_STRUCT((dsApi), InsideSubscription, (type)(expiresAt));
