/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#include "publish.h"

#include <avahi-common/alternative.h>
#include <avahi-common/malloc.h>
#include <ds/log.h>
#include <ds/str.h>

#include <boost/foreach.hpp>

#include "client.h"
#include "util.h"

static ds::LogChannel dsLogChannel("dsAvahiPublish");

namespace dsAvahi {

struct AvahiCharPtrDeleter {
  void operator()(char* x) { avahi_free(x); }
};
typedef std::unique_ptr<char, AvahiCharPtrDeleter> UniqueAvahiCharPtr;

struct AvahiStringListDeleter {
  void operator()(AvahiStringList* x) { avahi_string_list_free(x); }
};
typedef std::unique_ptr<AvahiStringList, AvahiStringListDeleter> UniqueAvahiStringList;

EntryGroup::EntryGroup(Client& client, CollisionMode collisionMode)
    : m_client(client),
      m_setServicesModerator(client.ioService()),
      m_establishTimeout(std::chrono::seconds(30)),
      m_establishTimer(client.ioService()),
      m_collisionMode(collisionMode),
      m_isRunning(false),
      m_isEstablished(false) {
  DS_INFO_ENTER("");
  m_client.addHook(*this);
  DS_DEBUG_LEAVE("");
}
EntryGroup::~EntryGroup() {
  DS_INFO_ENTER("");
  m_client.removeHook(*this);
  DS_DEBUG_LEAVE("");
}

void EntryGroup::setServices(std::vector<PublishedService>&& services) {
  if (m_services == services) {
    return;
  }
  DS_INFO_ENTER(services);
  m_services = std::move(services);

  m_setServicesModerator.expiresNow([this] {
    try {
      publish();
    } catch (const std::exception& e) {
      m_client.reconnectOnError(e);
    }
  });
  setEstablished(false);
}

void EntryGroup::setService(PublishedService&& service) {
  setServices({std::move(service)});
}

void EntryGroup::onRunning() {
  DS_ASSERT(!m_isRunning);
  DS_ASSERT(!m_ptr);
  DS_ASSERT(!m_isEstablished);

  m_isRunning = true;
  DS_INFO_ENTER("");

  m_ptr.reset(avahi_entry_group_new(m_client.ptr(), sCallback, this));
  if (!m_ptr) {
    DS_AVAHI_FAIL_CLIENT_ERROR(m_client, "Failed to create avahi entry group");
  }
  publish();
}

void EntryGroup::freed() {
  DS_INFO_ENTER(m_ptr.get(), m_isRunning);
  m_isRunning = false;
  m_setServicesModerator.cancel();
  m_ptr.release();
  setEstablished(false);
}

void EntryGroup::publish() {
  DS_DEBUG_ENTER(m_isRunning, m_services.size());
  if (!m_isRunning) {
    return;
  }

  setEstablished(false);
  avahi_entry_group_reset(ptr());

  if (m_services.empty()) {
    setEstablished(true);
    // avahi_entry_group_commit() must NOT be called for empty list of services
    return;
  }

  BOOST_FOREACH (const auto& service, m_services) {
    UniqueAvahiStringList txtsList;
    auto&& txts = service.txts;
    if (!txts.empty()) {
      auto it = txts.begin();
      txtsList.reset(avahi_string_list_new(it->c_str(), NULL));
      it++;
      for (; it != txts.end(); ++it) {
        txtsList.reset(avahi_string_list_add(txtsList.release(), it->c_str()));
        DS_ASSERT(txtsList, "");
      }
    }

    const auto& interface = service.interface;
    const auto& protocol = service.protocol;
    const auto& flags = service.flags;
    const auto& name = service.name;
    const auto& type = ds::str(typePrefix(), service.type);
    const auto& domain = service.domain;
    const auto& host = service.host;
    const auto& port = service.port;
    DS_NOTICE("Add service", interface, protocol, flags, name, type, domain, host, port, txts);
    if (auto&& error =
            avahi_entry_group_add_service_strlst(ptr(), interface, protocol, flags, name.c_str(), type.c_str(),
                                                 domain.c_str(), host.c_str(), port, txtsList.get())) {
      if (error == AVAHI_ERR_COLLISION) {
        handleCollision();
        return;
      }
      DS_AVAHI_FAIL(error, "avahi_entry_group_add_service_strlst failed");
      return;
    }
  }
  if (auto&& error = avahi_entry_group_commit(ptr())) {
    DS_AVAHI_FAIL(error, "avahi_entry_group_commit failed");
    return;
  }
  m_establishTimer.expiresFromNow(
      m_establishTimeout, [this] { m_client.reconnectOnError(ds::Exception("Timed out establishing entry group.")); });
}

void EntryGroup::handleCollision() {
  DS_INFO_ENTER("");
  switch (m_collisionMode) {
    case CollisionMode::ALTERNATIVE_NAME: {
      // Let's pick a new name.
      BOOST_FOREACH (auto&& x, m_services) {
        UniqueAvahiCharPtr name(avahi_alternative_service_name(x.name.c_str()));
        x.name = name.get();
        DS_NOTICE("Using alternative service name on collision.", x.name, x.type);
      }
      publish();
      break;
    }
  }
}

void EntryGroup::sCallback(AvahiEntryGroup* g, AvahiEntryGroupState state, void* data) {
  auto&& x = *static_cast<EntryGroup*>(data);
  try {
    x.callback(state);
  } catch (const std::exception& e) {
    x.m_client.reconnectOnError(e);
  }
}

void EntryGroup::callback(AvahiEntryGroupState state) {
  DS_INFO_ENTER(state);
  switch (state) {
    case AVAHI_ENTRY_GROUP_ESTABLISHED:
      // The entries have successfully been established.
      setEstablished(true);
      BOOST_FOREACH (auto&& x, m_services) {
        auto type = ds::str(typePrefix(), x.type);
        DS_NOTICE("Service established.", x.name, type, x.port);
      }
      break;
    case AVAHI_ENTRY_GROUP_COLLISION:
      // A name collision for one of the entries in the group has been detected,
      // the entries have been withdrawn.
      handleCollision();
      break;
    case AVAHI_ENTRY_GROUP_FAILURE: DS_AVAHI_FAIL_CLIENT_ERROR(m_client, "Avahi entry group failure"); break;

    case AVAHI_ENTRY_GROUP_UNCOMMITED:
    case AVAHI_ENTRY_GROUP_REGISTERING:;
  }
}

void EntryGroup::setEstablished(bool isEstablished) {
  // isEstablished == true -> successfully established -> cancel timer
  // isEstablished == false -> establishing is canceled -> cancel timer
  m_establishTimer.cancel();

  if (m_isEstablished == isEstablished) {
    return;
  }
  m_isEstablished = isEstablished;
  DS_INFO_LEAVE(isEstablished);
  if (m_isEstablishedChanged) {
    m_isEstablishedChanged(m_isEstablished);
  }
}

} // namespace dsAvahi

DS_REFLECT_DEFINE_STRUCT((dsAvahi), PublishedService)
