/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include <ds/asio/catch.h>
#include <ds/log.h>
#include <ds/random.h>

#include <boost/algorithm/string/predicate.hpp>
#include <queue>

#include "lookup.h"
#include "publish.h"

namespace {

static const char* TAGS = "[dsAvahi][ds]";

typedef dsAvahi::ResolvedService ResolvedService;
typedef dsAvahi::BrowsedService BrowsedService;
typedef dsAvahi::ServicesBrowser::Event BrowserEvent;
typedef dsAvahi::ServiceResolver::Reply ResolverReply;
typedef dsAvahi::ServicesResolver::Event ResolverEvent;

// Avahi deamon needs more than 1 second to establish entry group.
static constexpr auto TEST_LATENCY = std::chrono::seconds(10);

struct Fixture { // using fixtures instead of sections to be able to run each test independently
  Fixture() : client(ioService), service({}) {
    client.setRandomTypePrefix();
    // fail on unexpected reconnects
    client.setOnReconnectOnError([&](const std::exception& e) {
      REQUIRE(reconnectsOnError.size() > 0);
      reconnectsOnError.front()(e);
      reconnectsOnError.pop();
    });
    service.protocol = AVAHI_PROTO_INET;
    service.name = "dsAvahiTest";
    // Randomly generated service type so that parallel tests don't influence each other.
    service.type = "_ds-avahi-test._tcp";
    service.port = 7123;
    service.txts = {"txt1", "txt2=foo"};
  }
  ds::asio::catch_::IoService ioService;
  std::queue<std::function<void(const std::exception&)>> reconnectsOnError;
  dsAvahi::Client client;
  dsAvahi::PublishedService service;
};

TEST_CASE_METHOD(Fixture, "Empty EntryGroup gets established", TAGS) {
  dsAvahi::EntryGroup entryGroup(client, dsAvahi::CollisionMode::ALTERNATIVE_NAME);
  // Empty group is established synchronously in constructor.
  CHECK(entryGroup.isEstablished());
}

TEST_CASE_METHOD(Fixture, "ServicesBrowser returns empty set without any service", TAGS) {
  dsAvahi::ServicesBrowser browser(client, service.type, [&](BrowserEvent event) {
    CHECK_NOTHROW(std::get<ds::collection::AllForNow>(event));
    ioService.stop();
  });
  CHECK_STOP_RUN_FOR(ioService, TEST_LATENCY);
}

TEST_CASE_METHOD(Fixture, "ServicesResolver returns empty set without published service", TAGS) {
  dsAvahi::ServicesResolver resolver(client, service.type, [&](ResolverEvent event) {
    CHECK_NOTHROW(std::get<ds::collection::AllForNow>(event));
    ioService.stop();
  });
  CHECK_STOP_RUN_FOR(ioService, TEST_LATENCY);
}

struct EstablishedServiceFixture : Fixture {
  EstablishedServiceFixture() : entryGroup(client, dsAvahi::CollisionMode::ALTERNATIVE_NAME) {
    entryGroup.setService(dsAvahi::PublishedService(service));

    entryGroup.setIsEstablishedChanged([&](bool isEstablished) {
      CHECK(isEstablished);
      ioService.stop();
    });
    CHECK_FALSE(entryGroup.isEstablished());
    CHECK_STOP_RUN_FOR(ioService, TEST_LATENCY);
    CHECK(entryGroup.isEstablished());
    entryGroup.setIsEstablishedChanged([&](bool) {});
  }
  dsAvahi::EntryGroup entryGroup;
};

struct EstablishedServiceFixtureWithServicesBrowser : EstablishedServiceFixture {
  EstablishedServiceFixtureWithServicesBrowser()
      : browser(client, service.type, [&](BrowserEvent event) {
          services.applyEvent(std::move(event));
          ioService.stop();
        }) {}

  ds::collection::Vector<BrowsedService> services;
  dsAvahi::ServicesBrowser browser;

  void checkServicesMatchService() {
    CHECK_FALSE(services.empty());
    for (const auto& x : services) {
      const auto& value = x.value;
      CHECK(value.name == service.name);
      CHECK(value.type == service.type);
      CHECK(value.domain == std::string("local"));
    }
  }
};

struct EstablishedServiceFixtureWithServicesResolver : EstablishedServiceFixture {
  EstablishedServiceFixtureWithServicesResolver()
      : resolver(client, service.type, [&](ResolverEvent event) {
          services.applyEvent(std::move(event));
          ioService.stop();
        }) {}

  ds::collection::Vector<ResolvedService> services;
  dsAvahi::ServicesResolver resolver;

  void checkServicesMatchService() {
    CHECK_FALSE(services.empty());
    for (const auto& x : services) {
      const auto& value = x.value;
      CHECK(value.name == service.name);
      CHECK(value.type == service.type);
      CHECK(value.domain == std::string("local"));
      CHECK(value.port == service.port);
      CHECK(value.txts == service.txts);
    }
  }
};

TEST_CASE_METHOD(EstablishedServiceFixture, "Conflicting EntryGroup establishes alternative name", TAGS) {
  dsAvahi::EntryGroup entryGroup2(client, dsAvahi::CollisionMode::ALTERNATIVE_NAME);
  {
    auto service2 = dsAvahi::PublishedService(service);
    service2.port++; // We need real difference to trigger collision code
    entryGroup2.setServices({std::move(service2)});
  }
  entryGroup2.setIsEstablishedChanged([&](bool isEstablished) {
    CHECK(isEstablished);
    ioService.stop();
  });
  CHECK_STOP_RUN_FOR(ioService, TEST_LATENCY);
  CHECK(entryGroup2.isEstablished());
  CHECK(entryGroup2.services().front().name == ds::str(service.name, " #2"));
}

TEST_CASE_METHOD(EstablishedServiceFixture, "ServiceResolver resolves", TAGS) {
  dsAvahi::ServicesBrowser::OnEvent onEvent;
  dsAvahi::ServicesBrowser browser(client, service.type, [&](BrowserEvent x) { onEvent(std::move(x)); });

  BrowsedService browsedService;
  onEvent = [&](BrowserEvent event) {
    CHECK_NOTHROW(std::get<ds::collection::Added<BrowsedService>>(event));
    browsedService = std::move(std::get<ds::collection::Added<BrowsedService>>(event).value);
    ioService.stop();
  };
  CHECK_STOP_RUN_FOR(ioService, TEST_LATENCY);

  onEvent = [&](BrowserEvent) {};
  dsAvahi::ServiceResolver resolver(client, browsedService, browser.flags(), [&](ResolverReply reply) {
    REQUIRE_NOTHROW(std::get_if<ResolvedService>(&reply));
    auto& value = std::get<ResolvedService>(reply);
    CHECK(value.name == service.name);
    CHECK(value.type == service.type);
    CHECK(value.domain == std::string("local"));
    CHECK(value.port == service.port);
    CHECK(value.txts == service.txts);
    ioService.stop();
  });
  CHECK_STOP_RUN_FOR(ioService, TEST_LATENCY);
}

TEST_CASE_METHOD(EstablishedServiceFixtureWithServicesBrowser, "ServicesBrowser follows", TAGS) {
  for (int i = 0; i < 10 && !services.isAllForNow(); i++) {
    CHECK_STOP_RUN_FOR(ioService, TEST_LATENCY);
  }
  auto servicesSize = services.size();
  checkServicesMatchService();

  // removed
  entryGroup.setServices({});

  for (size_t i = 0; i < servicesSize && !services.empty(); i++) {
    CHECK_STOP_RUN_FOR(ioService, TEST_LATENCY);
  }
  CHECK(services.empty());
}

TEST_CASE_METHOD(EstablishedServiceFixtureWithServicesResolver, "ServicesResolver follows", TAGS) {
  for (int i = 0; i < 10 && !services.isAllForNow(); i++) {
    CHECK_STOP_RUN_FOR(ioService, TEST_LATENCY);
  }
  auto servicesSize = services.size();
  checkServicesMatchService();

  // changed
  service.port++;
  entryGroup.setService(dsAvahi::PublishedService(service));

  // expect either Changed, or Removed + Added for every service
  for (size_t i = 0; i < servicesSize * 2; i++) {
    CHECK_STOP_RUN_FOR(ioService, TEST_LATENCY);
    if (services.size() == servicesSize &&
        std::all_of(services.begin(), services.end(),
                    [&](ds::collection::Entry<ResolvedService>& x) { return x.value.port == service.port; })) {
      break; // all services were changed
    }
  }
  checkServicesMatchService();

  // removed
  entryGroup.setServices({});

  for (size_t i = 0; i < servicesSize && !services.empty(); i++) {
    CHECK_STOP_RUN_FOR(ioService, TEST_LATENCY);
  }
  CHECK(services.empty());
}

TEST_CASE_METHOD(EstablishedServiceFixture, "ServiceResolver destroyed after reconnect", TAGS) {
  dsAvahi::ServicesBrowser::OnEvent onEvent;
  dsAvahi::ServicesBrowser browser(client, service.type, [&](BrowserEvent x) { onEvent(std::move(x)); });

  BrowsedService browsedService;
  onEvent = [&](BrowserEvent event) {
    CHECK_NOTHROW(std::get<ds::collection::Added<BrowsedService>>(event));
    browsedService = std::move(std::get<ds::collection::Added<BrowsedService>>(event).value);
    ioService.stop();
  };
  CHECK_STOP_RUN_FOR(ioService, TEST_LATENCY);

  onEvent = [&](BrowserEvent) {};
  dsAvahi::ServiceResolver resolver(client, browsedService, browser.flags(), [&](ResolverReply reply) {
    CHECK(false); // not expected
  });
  reconnectsOnError.push([&](const std::exception& e) { REQUIRE_THAT(e.what(), Catch::Contains("Simulated error")); });
  client.reconnectOnError(std::runtime_error("Simulated error"));
}

TEST_CASE_METHOD(EstablishedServiceFixtureWithServicesBrowser, "ServiceBrowser handles reconnects", TAGS) {
  SECTION("wait for first event") { CHECK_STOP_RUN_FOR(ioService, TEST_LATENCY); }

  SECTION("wait for AllForNow") {
    for (int i = 0; i < 10 && !services.isAllForNow(); i++) {
      CHECK_STOP_RUN_FOR(ioService, TEST_LATENCY);
    }
  }

  reconnectsOnError.push([&](const std::exception& e) { REQUIRE_THAT(e.what(), Catch::Contains("Simulated error")); });
  client.reconnectOnError(std::runtime_error("Simulated error"));
  CHECK_FALSE(services.isAllForNow());

  for (int i = 0; i < 10 && !services.isAllForNow(); i++) {
    CHECK_STOP_RUN_FOR(ioService, TEST_LATENCY);
  }

  CHECK(services.isAllForNow());
  checkServicesMatchService();
}

TEST_CASE_METHOD(EstablishedServiceFixtureWithServicesResolver, "ServiceResolver handles reconnects", TAGS) {
  SECTION("wait for first event") { CHECK_STOP_RUN_FOR(ioService, TEST_LATENCY); }

  SECTION("wait for AllForNow") {
    for (int i = 0; i < 10 && !services.isAllForNow(); i++) {
      CHECK_STOP_RUN_FOR(ioService, TEST_LATENCY);
    }
  }

  reconnectsOnError.push([&](const std::exception& e) { REQUIRE_THAT(e.what(), Catch::Contains("Simulated error")); });
  client.reconnectOnError(std::runtime_error("Simulated error"));
  CHECK_FALSE(services.isAllForNow());

  for (int i = 0; i < 10 && !services.isAllForNow(); i++) {
    CHECK_STOP_RUN_FOR(ioService, TEST_LATENCY);
  }

  CHECK(services.isAllForNow());
  checkServicesMatchService();
}

TEST_CASE_METHOD(Fixture, "Establishing entry group timeout triggers client reconnect", TAGS) {
  dsAvahi::EntryGroup entryGroup(client, dsAvahi::CollisionMode::ALTERNATIVE_NAME);
  // Set small enough timeout to trigger established guard
  // before avahi deamon replies
  entryGroup.setEstablishTimeout(std::chrono::milliseconds(10));

  entryGroup.setService(dsAvahi::PublishedService(service));
  reconnectsOnError.push([&](const std::exception& e) {
    REQUIRE_THAT(e.what(), Catch::Contains("Timed out establishing entry group."));
    ioService.stop();
  });
  CHECK_STOP_RUN_FOR(ioService, TEST_LATENCY);
}

} // namespace
