/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "ds-meterings.h"

#include <ds-api/event-mock.h>
#include <ds-model/apartment-api.h>
#include <ds/asio/test.h>
#include <ds/catch/catch.h>
#include <ds/clock-mock.h>
#include <ds/log.h>

#include "ds-model/meterings.h"
#include "model/apartment-mock.h"

using namespace std::literals;
static ds::LogChannel dsLogChannel("dssDsMeteringTest");
static const char* TAGS = "[dss][dssDsMetering]";

using namespace dss;
using namespace dsModel;
using namespace trompeloeil;

namespace {
SCENARIO("dssDsMetering", TAGS) {
  auto clock = ds::ClockMocker{};
  ApartmentMock apartment{};
  MeteringMock& meteringMock = apartment.metering;

  ds::asio::TestLoopResource loop;
  auto eventHub = dsApi::EventHubMock{};

  std::unique_ptr<dsModel::MeteringsApi> dsMeterings;

  loop.io.post([&] {
    dsMeterings = makeDsMeterings(loop, apartment, eventHub);
    loop.stop();
  });
  loop.run();

  auto matchConsumedElectricity = ds::catch_::any_of({MeteringType::energyMetering, MeteringType::powerMetering});
  auto matchProducedElectricity =
      ds::catch_::any_of({MeteringType::energyProducedMetering, MeteringType::powerProducedMetering});

  SECTION("decode metering id") {
    SECTION("controller") {
      CHECK(decodeId(dsModel::MeteringId{"dsm-x-power"}) ==
            DecodedMeteringId{.origin = dsModel::ControllerMeteringOrigin{.id = dsModel::ControllerId{"x"}},
                              .meteringType = dsModel::MeteringType::powerMetering});
      CHECK(decodeId(dsModel::MeteringId{"dsm-y-energy"}) ==
            DecodedMeteringId{.origin = dsModel::ControllerMeteringOrigin{.id = dsModel::ControllerId{"y"}},
                              .meteringType = dsModel::MeteringType::energyMetering});
      CHECK(decodeId(dsModel::MeteringId{"dsm-x-powerProduced"}) ==
            DecodedMeteringId{.origin = dsModel::ControllerMeteringOrigin{.id = dsModel::ControllerId{"x"}},
                              .meteringType = dsModel::MeteringType::powerProducedMetering});
      CHECK(decodeId(dsModel::MeteringId{"dsm-y-energyProduced"}) ==
            DecodedMeteringId{.origin = dsModel::ControllerMeteringOrigin{.id = dsModel::ControllerId{"y"}},
                              .meteringType = dsModel::MeteringType::energyProducedMetering});

      CHECK_THROWS_FIND(decodeId(dsModel::MeteringId{"meter-x-power"}), "Unsupported measuring device.");
      CHECK_THROWS_FIND(decodeId(dsModel::MeteringId{"dsm-x-phase"}), "Unsupported measuring type.");

      // dsModel::MeteringType::none enum value exists but is invalid metering type
      CHECK_THROWS_FIND(decodeId(dsModel::MeteringId{"dsm-k-none"}), "Unsupported measuring type.");
    }
    SECTION("apartment") {
      CHECK(decodeId(dsModel::MeteringId{"apartment-power"}) ==
            DecodedMeteringId{.origin = dsModel::ApartmentMeteringOrigin{},
                              .meteringType = dsModel::MeteringType::powerMetering});
      CHECK(decodeId(dsModel::MeteringId{"apartment-powerProduced"}) ==
            DecodedMeteringId{.origin = dsModel::ApartmentMeteringOrigin{},
                              .meteringType = dsModel::MeteringType::powerProducedMetering});
      CHECK(decodeId(dsModel::MeteringId{"apartment-energy"}) ==
            DecodedMeteringId{.origin = dsModel::ApartmentMeteringOrigin{},
                              .meteringType = dsModel::MeteringType::energyMetering});
      CHECK(decodeId(dsModel::MeteringId{"apartment-energyProduced"}) ==
            DecodedMeteringId{.origin = dsModel::ApartmentMeteringOrigin{},
                              .meteringType = dsModel::MeteringType::energyProducedMetering});
    }
    SECTION("openMeteringSystemOrigin") {
      CHECK(decodeId(dsModel::MeteringId{"openMeteringSystem-power"}) ==
            DecodedMeteringId{.origin = dsModel::OpenMeteringSystemOrigin{},
                              .meteringType = dsModel::MeteringType::powerMetering});
      CHECK(decodeId(dsModel::MeteringId{"openMeteringSystem-powerProduced"}) ==
            DecodedMeteringId{.origin = dsModel::OpenMeteringSystemOrigin{},
                              .meteringType = dsModel::MeteringType::powerProducedMetering});
      CHECK(decodeId(dsModel::MeteringId{"openMeteringSystem-energy"}) ==
            DecodedMeteringId{.origin = dsModel::OpenMeteringSystemOrigin{},
                              .meteringType = dsModel::MeteringType::energyMetering});
      CHECK(decodeId(dsModel::MeteringId{"openMeteringSystem-energyProduced"}) ==
            DecodedMeteringId{.origin = dsModel::OpenMeteringSystemOrigin{},
                              .meteringType = dsModel::MeteringType::energyProducedMetering});
    }
  }

  SECTION("encode metering id") {
    using dsModel::MeteringId;
    using dsModel::MeteringType;
    SECTION("controller") {
      auto x = GENERATE(table<MeteringType, std::string>({
          {MeteringType::powerMetering, "dsm-x-power"s},
          {MeteringType::energyMetering, "dsm-x-energy"s},
          {MeteringType::powerProducedMetering, "dsm-x-powerProduced"s},
          {MeteringType::energyProducedMetering, "dsm-x-energyProduced"s},
      }));
      using dsModel::ControllerId;
      using dsModel::ControllerMeteringOrigin;
      CHECK(encodeId(ControllerMeteringOrigin{.id = ControllerId{"x"}}, std::get<MeteringType>(x)) ==
            MeteringId{std::get<std::string>(x)});
    }
    SECTION("apartment") {
      auto x = GENERATE(table<MeteringType, std::string>({
          {MeteringType::powerMetering, "apartment-power"s},
          {MeteringType::energyMetering, "apartment-energy"s},
          {MeteringType::powerProducedMetering, "apartment-powerProduced"s},
          {MeteringType::energyProducedMetering, "apartment-energyProduced"s},
      }));
      using dsModel::ApartmentMeteringOrigin;
      CHECK(encodeId(ApartmentMeteringOrigin{}, std::get<MeteringType>(x)) == MeteringId{std::get<std::string>(x)});
    }
    SECTION("invalid") {
      CHECK_THROWS(encodeId(dsModel::MeteringOrigin{}, dsModel::MeteringType::energyMetering));
      CHECK_THROWS(encodeId(dsModel::ApartmentMeteringOrigin{}, dsModel::MeteringType::none));
    }
  }

  SECTION("Energy and power meterings") {
    auto& meter = *apartment.allocateDSMeter(Dsuid{"1234511111111111111111111111111100"});
    meter.setIsPresent(true);
    meter.setCapability_HasMetering(true);
    meter.setIsConnected(true);
    meter.setHardwareName("dsmTest");

    const auto apartmentOrigin = dsModel::ApartmentMeteringOrigin{.type = "apartment"};
    const auto controllerOrigin = dsModel::ControllerMeteringOrigin{
        .type = "controller", .id = dsModel::ControllerId{"1234511111111111111111111111111100"}};
    const auto openMeteringSystemOrigin = dsModel::OpenMeteringSystemOrigin{.serialNumber = "serial-number:1123456"};

    SECTION("meterings") {
      using dsModel::ApartmentMeteringOrigin;
      using dsModel::MeteringOrigin;
      using dsModel::MeteringType;
      REQUIRE_CALL(meteringMock, meterings())
          .RETURN(std::vector{
              std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::energyMetering},
              std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::powerMetering},
              std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::powerMetering},
              std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::energyMetering},
              std::pair<MeteringOrigin, MeteringType>{openMeteringSystemOrigin, MeteringType::powerMetering},
              std::pair<MeteringOrigin, MeteringType>{openMeteringSystemOrigin, MeteringType::energyMetering}});
      CHECK(dsMeterings->meterings() ==
            dsModel::Meterings{
                .meterings = {
                    {.id = dsModel::MeteringId{"apartment-energy"},
                     .type = dsModel::MeteringType::energyMetering,
                     .lastChanged = {},
                     .attributes = {.unit = "Wh", .technicalName = "apartment energy", .origin = apartmentOrigin}},
                    {.id = dsModel::MeteringId{"apartment-power"},
                     .type = dsModel::MeteringType::powerMetering,
                     .lastChanged = {},
                     .attributes = {.unit = "W", .technicalName = "apartment power", .origin = apartmentOrigin}},

                    {.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-energy"},
                     .type = dsModel::MeteringType::energyMetering,
                     .lastChanged = {},
                     .attributes = {.unit = "Wh", .technicalName = "dsmTest energy", .origin = controllerOrigin}},
                    {.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-power"},
                     .type = dsModel::MeteringType::powerMetering,
                     .lastChanged = {},
                     .attributes = {.unit = "W", .technicalName = "dsmTest power", .origin = controllerOrigin}},

                    {.id = dsModel::MeteringId{"openMeteringSystem-energy"},
                     .type = dsModel::MeteringType::energyMetering,
                     .attributes = {.unit = "Wh",
                                    //.technicalName = "duerr&co metering master 3000",
                                    .technicalName = "openMeteringSystem-tariff0",
                                    .origin = openMeteringSystemOrigin}},
                    {.id = dsModel::MeteringId{"openMeteringSystem-power"},
                     .type = dsModel::MeteringType::powerMetering,
                     .attributes = {.unit = "W",
                                    .technicalName = "openMeteringSystem-tariff0",
                                    .origin = openMeteringSystemOrigin}},

                }});
    }
    SECTION("metering") {
      CHECK(dsMeterings->metering(dsModel::MeteringArgs{
                .meteringId = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-energy"}}) ==
            dsModel::Metering{
                .id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-energy"},
                .type = dsModel::MeteringType::energyMetering,
                .lastChanged = {},
                .attributes = {.unit = "Wh", .technicalName = "dsmTest energy", .origin = controllerOrigin}});
      CHECK(
          dsMeterings->metering(dsModel::MeteringArgs{
              .meteringId = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-power"}}) ==
          dsModel::Metering{.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-power"},
                            .type = dsModel::MeteringType::powerMetering,
                            .lastChanged = {},
                            .attributes = {.unit = "W", .technicalName = "dsmTest power", .origin = controllerOrigin}});
      CHECK(dsMeterings->metering(dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"apartment-energy"}}) ==
            dsModel::Metering{
                .id = dsModel::MeteringId{"apartment-energy"},
                .type = dsModel::MeteringType::energyMetering,
                .lastChanged = {},
                .attributes = {.unit = "Wh", .technicalName = "apartment energy", .origin = apartmentOrigin}});
      CHECK(dsMeterings->metering(dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"apartment-power"}}) ==
            dsModel::Metering{
                .id = dsModel::MeteringId{"apartment-power"},
                .type = dsModel::MeteringType::powerMetering,
                .lastChanged = {},
                .attributes = {.unit = "W", .technicalName = "apartment power", .origin = apartmentOrigin}});
    }
    SECTION("meteringValues") {
      auto values = [&] {
        return dsModel::MeteringValues{.values = {
                                           {.id = dsModel::MeteringId{"apartment-energy"},
                                            .type = "meteringValue",
                                            .lastChanged = ds::DateTime{187s},
                                            .attributes = {.value = 55}},
                                           {.id = dsModel::MeteringId{"apartment-power"},
                                            .type = "meteringValue",
                                            .lastChanged = ds::DateTime{187s},
                                            .attributes = {.value = 100}},
                                           {.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-energy"},
                                            .type = "meteringValue",
                                            .lastChanged = ds::DateTime{187s},
                                            .attributes = {.value = 55}},
                                           {.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-power"},
                                            .type = "meteringValue",
                                            .lastChanged = ds::DateTime{187s},
                                            .attributes = {.value = 100}},
                                       }};
      };
      SECTION("polled") {
        REQUIRE_CALL(meteringMock, meterings())
            .RETURN(std::vector{
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::energyMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::powerMetering},
                std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::powerMetering},
                std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::energyMetering}});
        REQUIRE_CALL(meteringMock, getLastReading(_, _, _))
            .TIMES(4)
            .RETURN(ArchiveValue{.power = 100, .energy = 55.0, .lastChanged = ds::DateTime(187s)});
        CHECK(dsMeterings->meteringValues() == values());
      }

      SECTION("pushed every second") {
        const auto event =
            ds::Box{dsApi::Event{.type = dsModel::meteringValues, .path = {}, .data = ds::toJsonFragment(values())}};
        REQUIRE_CALL(meteringMock, meterings())
            .RETURN(std::vector{
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::energyMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::powerMetering},
                std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::powerMetering},
                std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::energyMetering}});
        REQUIRE_CALL(meteringMock, getLastReading(_, _, _))
            .TIMES(4)
            .RETURN(ArchiveValue{.power = 100, .energy = 55.0, .lastChanged = ds::DateTime(187s)});
        REQUIRE_CALL(eventHub, emitEvent(event)).LR_SIDE_EFFECT(loop.stop());
        clock.advance(2s);
        loop.run();

        REQUIRE_CALL(meteringMock, meterings())
            .RETURN(std::vector{
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::energyMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::powerMetering},
                std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::powerMetering},
                std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::energyMetering}});
        REQUIRE_CALL(meteringMock, getLastReading(_, _, _))
            .TIMES(4)
            .RETURN(ArchiveValue{.power = 100, .energy = 55.0, .lastChanged = ds::DateTime(187s)});
        REQUIRE_CALL(eventHub, emitEvent(event)).LR_SIDE_EFFECT(loop.stop());
        clock.advance(2s);
        loop.run();
      }
    }
    SECTION("meteringValue") {
      REQUIRE_CALL(meteringMock, getLastReading(_, _, _))
          .RETURN(ArchiveValue{.power = 100, .energy = 55.0, .lastChanged = ds::DateTime(187s)});
      CHECK(dsMeterings->meteringValue(dsModel::MeteringArgs{
                .meteringId = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-energy"}}) ==
            dsModel::MeteringValue{.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-energy"},
                                   .type = "meteringValue",
                                   .lastChanged = ds::DateTime{187s},
                                   .attributes = {.value = 55}});
      REQUIRE_CALL(meteringMock, getLastReading(_, _, _))
          .RETURN(ArchiveValue{.power = 100, .energy = 55.0, .lastChanged = ds::DateTime(187s)});
      CHECK(dsMeterings->meteringValue(dsModel::MeteringArgs{
                .meteringId = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-power"}}) ==
            dsModel::MeteringValue{.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-power"},
                                   .type = "meteringValue",
                                   .lastChanged = ds::DateTime{187s},
                                   .attributes = {.value = 100}});

      REQUIRE_CALL(meteringMock, getLastReading(_, _, _))
          .RETURN(ArchiveValue{.power = 100, .energy = 55.0, .lastChanged = ds::DateTime(187s)});
      CHECK(dsMeterings->meteringValue(dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"apartment-energy"}}) ==
            dsModel::MeteringValue{.id = dsModel::MeteringId{"apartment-energy"},
                                   .type = "meteringValue",
                                   .lastChanged = ds::DateTime{187s},
                                   .attributes = {.value = 55}});

      REQUIRE_CALL(meteringMock, getLastReading(_, _, _))
          .RETURN(ArchiveValue{.power = 100, .energy = 55.0, .lastChanged = ds::DateTime(187s)});
      CHECK(dsMeterings->meteringValue(dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"apartment-power"}}) ==
            dsModel::MeteringValue{.id = dsModel::MeteringId{"apartment-power"},
                                   .type = "meteringValue",
                                   .lastChanged = ds::DateTime{187s},
                                   .attributes = {.value = 100}});
    }
    /*SECTION("seriesCompatValue") {
      REQUIRE_CALL(meteringMock, getSeries(_, 60s, _, _, _, _))
          .RETURN(ValueRange{.step = 60, .values = std::vector{Value{.value = 13.5, .timeStamp = ds::DateTime{301s}}}});

      CHECK(dsMeterings->seriesCompatValue(dsModel::MeteringSeriesArgs{
                .resolution = 60, .meteringId = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-power"}}) ==
            dsModel::SeriesCompatValue{

                .id = "dsm-1234511111111111111111111111111100-power",
                .unit = "W",
                .resolution = "60",
                .values = {{.timestamp = ds::DateTime{301s}, .value = 13.5}}});
    }*/
  }

  SECTION("Produced energy and power meterings") {
    auto& meter = *apartment.allocateDSMeter(Dsuid{"1234511111111111111111111111111100"});
    meter.setIsPresent(true);
    meter.setCapability_HasMetering(true);
    meter.setCapability_HasMeteringProducer(true);
    meter.setCapability_HasMeteringProducerEnabled(true);
    meter.setIsConnected(true);
    meter.setHardwareName("dsmTest");

    const auto apartmentOrigin = dsModel::ApartmentMeteringOrigin{.type = "apartment"};
    const auto controllerOrigin = dsModel::ControllerMeteringOrigin{
        .type = "controller", .id = dsModel::ControllerId{"1234511111111111111111111111111100"}};

    SECTION("meterings") {
      using dsModel::ApartmentMeteringOrigin;
      using dsModel::MeteringOrigin;
      using dsModel::MeteringType;
      REQUIRE_CALL(meteringMock, meterings())
          .RETURN(std::vector{
              std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::energyMetering},
              std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::energyProducedMetering},
              std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::powerMetering},
              std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::powerProducedMetering},
              std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::powerMetering},
              std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::powerProducedMetering},
              std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::energyMetering},
              std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::energyProducedMetering}});
      CHECK(
          dsMeterings->meterings() ==
          dsModel::Meterings{
              .meterings = {
                  {.id = dsModel::MeteringId{"apartment-energy"},
                   .type = dsModel::MeteringType::energyMetering,
                   .lastChanged = {},
                   .attributes = {.unit = "Wh", .technicalName = "apartment energy", .origin = apartmentOrigin}},
                  {.id = dsModel::MeteringId{"apartment-energyProduced"},
                   .type = dsModel::MeteringType::energyProducedMetering,
                   .lastChanged = {},
                   .attributes = {.unit = "Wh",
                                  .technicalName = "apartment energy produced",
                                  .origin = apartmentOrigin}},

                  {.id = dsModel::MeteringId{"apartment-power"},
                   .type = dsModel::MeteringType::powerMetering,
                   .lastChanged = {},
                   .attributes = {.unit = "W", .technicalName = "apartment power", .origin = apartmentOrigin}},
                  {.id = dsModel::MeteringId{"apartment-powerProduced"},
                   .type = dsModel::MeteringType::powerProducedMetering,
                   .lastChanged = {},
                   .attributes = {.unit = "W", .technicalName = "apartment power produced", .origin = apartmentOrigin}},

                  {.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-energy"},
                   .type = dsModel::MeteringType::energyMetering,
                   .lastChanged = {},
                   .attributes = {.unit = "Wh", .technicalName = "dsmTest energy", .origin = controllerOrigin}},
                  {.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-energyProduced"},
                   .type = dsModel::MeteringType::energyProducedMetering,
                   .lastChanged = {},
                   .attributes = {.unit = "Wh",
                                  .technicalName = "dsmTest energy produced",
                                  .origin = controllerOrigin}},
                  {.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-power"},
                   .type = dsModel::MeteringType::powerMetering,
                   .lastChanged = {},
                   .attributes = {.unit = "W", .technicalName = "dsmTest power", .origin = controllerOrigin}},

                  {.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-powerProduced"},
                   .type = dsModel::MeteringType::powerProducedMetering,
                   .lastChanged = {},
                   .attributes = {
                       .unit = "W", .technicalName = "dsmTest power produced", .origin = controllerOrigin}}}});
    }
    SECTION("meteringValues") {
      auto values = [&] {
        return dsModel::MeteringValues{
            .values = {
                {.id = dsModel::MeteringId{"apartment-energy"},
                 .type = "meteringValue",
                 .lastChanged = ds::DateTime{187s},
                 .attributes = {.value = 55}},
                {.id = dsModel::MeteringId{"apartment-energyProduced"},
                 .type = "meteringValue",
                 .lastChanged = ds::DateTime{188s},
                 .attributes = {.value = 56}},
                {.id = dsModel::MeteringId{"apartment-power"},
                 .type = "meteringValue",
                 .lastChanged = ds::DateTime{187s},
                 .attributes = {.value = 100}},
                {.id = dsModel::MeteringId{"apartment-powerProduced"},
                 .type = "meteringValue",
                 .lastChanged = ds::DateTime{188s},
                 .attributes = {.value = 101}},
                {.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-energy"},
                 .type = "meteringValue",
                 .lastChanged = ds::DateTime{187s},
                 .attributes = {.value = 55}},
                {.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-energyProduced"},
                 .type = "meteringValue",
                 .lastChanged = ds::DateTime{188s},
                 .attributes = {.value = 56}},
                {.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-power"},
                 .type = "meteringValue",
                 .lastChanged = ds::DateTime{187s},
                 .attributes = {.value = 100}},
                {.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-powerProduced"},
                 .type = "meteringValue",
                 .lastChanged = ds::DateTime{188s},
                 .attributes = {.value = 101}},
            }};
      };
      SECTION("polled") {
        REQUIRE_CALL(meteringMock, meterings())
            .RETURN(std::vector{
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::energyMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::powerMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{},
                                                        MeteringType::energyProducedMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::powerProducedMetering},
                std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::powerMetering},
                std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::energyMetering},
                std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::powerProducedMetering},
                std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::energyProducedMetering},
            });
        REQUIRE_CALL(meteringMock, getLastReading(_, matchConsumedElectricity, _))
            .TIMES(2)
            .RETURN(ArchiveValue{.power = 100, .energy = 55.0, .lastChanged = ds::DateTime(187s)});
        REQUIRE_CALL(meteringMock, getLastReading(_, matchProducedElectricity, _))
            .TIMES(2)
            .RETURN(ArchiveValue{.power = 101, .energy = 56.0, .lastChanged = ds::DateTime(188s)});

        REQUIRE_CALL(meteringMock, getLastReading(ApartmentMeteringOrigin{}, matchConsumedElectricity, _))
            .TIMES(2)
            .RETURN(ArchiveValue{.power = 100, .energy = 55.0, .lastChanged = ds::DateTime(187s)});

        REQUIRE_CALL(meteringMock, getLastReading(dsModel::ApartmentMeteringOrigin{}, matchProducedElectricity, _))
            .TIMES(2)
            .RETURN(ArchiveValue{.power = 101, .energy = 56.0, .lastChanged = ds::DateTime(188s)});
        CHECK(dsMeterings->meteringValues() == values());
      }

      SECTION("pushed every second") {
        const auto event =
            ds::Box{dsApi::Event{.type = dsModel::meteringValues, .path = {}, .data = ds::toJsonFragment(values())}};
        REQUIRE_CALL(meteringMock, meterings())
            .RETURN(std::vector{
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::energyMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::powerMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{},
                                                        MeteringType::energyProducedMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::powerProducedMetering},
                std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::powerMetering},
                std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::energyMetering},
                std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::powerProducedMetering},
                std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::energyProducedMetering},
            });
        REQUIRE_CALL(meteringMock, getLastReading(_, matchConsumedElectricity, _))
            .TIMES(2)
            .RETURN(ArchiveValue{.power = 100, .energy = 55.0, .lastChanged = ds::DateTime(187s)});
        REQUIRE_CALL(meteringMock, getLastReading(_, matchProducedElectricity, _))
            .TIMES(2)
            .RETURN(ArchiveValue{.power = 101, .energy = 56.0, .lastChanged = ds::DateTime(188s)});

        REQUIRE_CALL(meteringMock, getLastReading(ApartmentMeteringOrigin{}, matchConsumedElectricity, _))
            .TIMES(2)
            .RETURN(ArchiveValue{.power = 100, .energy = 55.0, .lastChanged = ds::DateTime(187s)});

        REQUIRE_CALL(meteringMock, getLastReading(dsModel::ApartmentMeteringOrigin{}, matchProducedElectricity, _))
            .TIMES(2)
            .RETURN(ArchiveValue{.power = 101, .energy = 56.0, .lastChanged = ds::DateTime(188s)});

        REQUIRE_CALL(eventHub, emitEvent(event)).LR_SIDE_EFFECT(loop.stop());
        clock.advance(2s);
        loop.run();

        REQUIRE_CALL(meteringMock, meterings())
            .RETURN(std::vector{
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::energyMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::powerMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{},
                                                        MeteringType::energyProducedMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::powerProducedMetering},
                std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::powerMetering},
                std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::energyMetering},
                std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::powerProducedMetering},
                std::pair<MeteringOrigin, MeteringType>{controllerOrigin, MeteringType::energyProducedMetering},
            });
        REQUIRE_CALL(meteringMock, getLastReading(_, matchConsumedElectricity, _))
            .TIMES(2)
            .RETURN(ArchiveValue{.power = 100, .energy = 55.0, .lastChanged = ds::DateTime(187s)});
        REQUIRE_CALL(meteringMock, getLastReading(_, matchProducedElectricity, _))
            .TIMES(2)
            .RETURN(ArchiveValue{.power = 101, .energy = 56.0, .lastChanged = ds::DateTime(188s)});

        REQUIRE_CALL(meteringMock, getLastReading(ApartmentMeteringOrigin{}, matchConsumedElectricity, _))
            .TIMES(2)
            .RETURN(ArchiveValue{.power = 100, .energy = 55.0, .lastChanged = ds::DateTime(187s)});
        REQUIRE_CALL(meteringMock, getLastReading(ApartmentMeteringOrigin{}, matchProducedElectricity, _))
            .TIMES(2)
            .RETURN(ArchiveValue{.power = 101, .energy = 56.0, .lastChanged = ds::DateTime(188s)});

        REQUIRE_CALL(eventHub, emitEvent(event)).LR_SIDE_EFFECT(loop.stop());
        clock.advance(2s);
        loop.run();
      }
    }
    SECTION("meteringValue") {
      REQUIRE_CALL(meteringMock, getLastReading(_, _, _))
          .RETURN(ArchiveValue{.power = 101, .energy = 57.0, .lastChanged = ds::DateTime(189s)});
      CHECK(dsMeterings->meteringValue(dsModel::MeteringArgs{
                .meteringId = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-energyProduced"}}) ==
            dsModel::MeteringValue{.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-energyProduced"},
                                   .type = "meteringValue",
                                   .lastChanged = ds::DateTime{189s},
                                   .attributes = {.value = 57}});
      REQUIRE_CALL(meteringMock, getLastReading(_, _, _))
          .RETURN(ArchiveValue{.power = 101, .energy = 57.0, .lastChanged = ds::DateTime(189s)});
      CHECK(dsMeterings->meteringValue(dsModel::MeteringArgs{
                .meteringId = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-powerProduced"}}) ==
            dsModel::MeteringValue{.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-powerProduced"},
                                   .type = "meteringValue",
                                   .lastChanged = ds::DateTime{189s},
                                   .attributes = {.value = 101}});

      REQUIRE_CALL(meteringMock, getLastReading(ApartmentMeteringOrigin{}, MeteringType::energyProducedMetering, _))
          .RETURN(ArchiveValue{.power = 101, .energy = 57.0, .lastChanged = ds::DateTime(189s)});
      CHECK(dsMeterings->meteringValue(
                dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"apartment-energyProduced"}}) ==
            dsModel::MeteringValue{.id = dsModel::MeteringId{"apartment-energyProduced"},
                                   .type = "meteringValue",
                                   .lastChanged = ds::DateTime{189s},
                                   .attributes = {.value = 57}});

      REQUIRE_CALL(meteringMock, getLastReading(ApartmentMeteringOrigin{}, MeteringType::powerProducedMetering, _))
          .RETURN(ArchiveValue{.power = 101, .energy = 57.0, .lastChanged = ds::DateTime(189s)});
      CHECK(dsMeterings->meteringValue(
                dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"apartment-powerProduced"}}) ==
            dsModel::MeteringValue{.id = dsModel::MeteringId{"apartment-powerProduced"},
                                   .type = "meteringValue",
                                   .lastChanged = ds::DateTime{189s},
                                   .attributes = {.value = 101}});
    }
  }
}
} // namespace
