/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "ds-model.h"

#include <ds-apartment/api-revision-mock.h>
#include <ds-api/event-mock.h>
#include <ds-backend/connector-mock.h>
#include <ds-model/meterings.h>
#include <ds/asio/test.h>
#include <ds/catch/catch.h>
#include <ds/clock-mock.h>
#include <ds/conf-mock.h>
#include <ds/json.h>
#include <ds/log.h>
#include <ds/vector.h>

#include <boost/asio/io_service.hpp>
#include <boost/range/adaptor/transformed.hpp>
#include <boost/range/algorithm/sort.hpp>

#include "businterface-mock.h"
#include "cache/apartment-proxy-mock.h"
#include "ds-apartment/ds-devices.h"
#include "dsm-api-mock.h"
#include "event-mock.h"
#include "event/event_create.h"
#include "metering/metering-mock.h"
#include "model/apartment-mock.h"
#include "model/cluster.h"
#include "model/device-mock.h"
#include "model/dsmeter.h"
#include "model/floor.h"
#include "model/group.h"
#include "model/modelconst.h"
#include "model/state.h"
#include "model/zone.h"

using namespace std::literals;
static ds::LogChannel dsLogChannel("dssDsModelTest");
static const char* TAGS = "[dss][dssDsModel]";

using namespace dss;
using namespace trompeloeil;

namespace {

struct DsModelHooksMock {
public:
  MAKE_MOCK0(apartmentStructurePolled, void());
  MAKE_MOCK0(apartmentStatusPolled, void());

  dss::DsModelHooks dsModelHooks() {
    return {
        .pollersActive = false,
        .apartmentStructurePolled = [&] { return apartmentStructurePolled(); },
        .apartmentStatusPolled = [&] { return apartmentStatusPolled(); },
    };
  }
};

SCENARIO("dssDsModel", TAGS) {
  auto clock = ds::ClockMocker{};
  auto propertySystem = PropertySystem{};
  propertySystem.createProperty("/config");
  EventSinkMock eventSink;

  ApartmentMock apartment{nullptr, &propertySystem};
  MeteringMock& meteringMock = apartment.metering;
  auto proxyPtr = std::make_unique<ApartmentProxyMock>();
  auto& proxy = *proxyPtr;
  const auto apartmentId = dsModel::ApartmentId{"apartmentId-0"};
  ALLOW_CALL(apartment, id()).RETURN(apartmentId);
  apartment.setName("TestName");
  apartment.setEventSink(eventSink);
  apartment.setProxy(std::move(proxyPtr));

  BusInterfaceMock busInterface;
  apartment.setBusInterface(&busInterface);

  ds::asio::TestLoopResource loop;
  auto& system = loop.system;
  dsBackend::ConnectorMock connectorBus;
  ALLOW_CALL(connectorBus, setActive(_));
  EventInterpreter eventInterpreter{nullptr};
  eventInterpreter.initialize();
  ds::sqlite3::Database d(":memory:");
  auto hal = makeHardwareAbstraction(&apartment, &d);
  auto eventHub = dsApi::EventHubMock{};
  auto hooks = DsModelHooksMock();
  ALLOW_CALL(hooks, apartmentStructurePolled());
  ALLOW_CALL(hooks, apartmentStatusPolled());
  std::unique_ptr<DsModel> dsModel;
  auto apiRevision = dsApartment::ApiRevisionMock{};
  ALLOW_CALL(apiRevision, additions()).RETURN(2);
  ALLOW_CALL(apiRevision, removals()).RETURN(0);
  ALLOW_CALL(proxy, ventilationUnitLevels()).RETURN(dsModel::Ventilation{});
  ALLOW_CALL(proxy, ventilationUnitStatus()).RETURN(dsModel::VentilationStatus{});
  loop.io.post([&] {
    dsModel = makeDsModel(loop, connectorBus, apartment, eventInterpreter, eventHub, loop.context, apiRevision,
                          hooks.dsModelHooks());
    loop.stop();
  });
  loop.run();

  auto requireAsyncStatusChanged = [&](auto&& f) {
    f();
    REQUIRE_CALL(hooks, apartmentStatusPolled()).LR_SIDE_EFFECT(loop.stop());
    loop.run();
  };

  SECTION("apartment empty included collections are serialized") {
    apartment.removeZone(NEW_DEVICES_ZONE_ID);
    apartment.removeFloor(APARTMENT_FLOOR_ID);
    using dsModel::MeteringOrigin;
    using dsModel::MeteringType;
    REQUIRE_CALL(meteringMock, meterings()).RETURN(std::vector<std::pair<MeteringOrigin, MeteringType>>{});
    CHECK(dsModel->apartment({.include = {}, .includeAll = true}).included ==
          dsModel::ApartmentIncluded{
              .installation = dsModel->installation(),
              .dsDevices = std::vector<dsModel::DSDevice>{},
              .submodules = std::vector<dsModel::Submodule>{},
              .functionBlocks = std::vector<dsModel::FunctionBlock>{},
              .zones = std::vector<dsModel::Zone>{},
              .scenarios = std::vector<dsModel::Scenario>{},
              .userDefinedStates = std::vector<dsModel::UserDefinedState>{},
              .floors = std::vector<dsModel::Floor>{},
              .clusters = std::vector<dsModel::Cluster>{},
              .applications = std::vector<dsModel::Application>{},
              .dsServer = dsModel->dsServer(),
              .controllers = std::vector<dsModel::Controller>{},
              .apiRevision = dsModel->apiRevision(),
              .meterings = {},
          });
    CHECK(dsModel->apartmentStatus({.include = {}, .includeAll = true}).included ==
          dsModel::ApartmentStatusIncluded{.dsDevices = std::vector<dsModel::DSDeviceStatus>{},
                                           .zones = std::vector<dsModel::ZoneStatus>{},
                                           .clusters = std::vector<dsModel::ClusterStatus>{},
                                           .userDefinedStates = std::vector<dsModel::UserDefinedStateStatus>{}});
  }

  SECTION("setZoneName") {
    auto& zone1 = *apartment.allocateZone(1);
    zone1.setName("zone1");

    THEN("success") {
      dsModel->setZoneName({.zoneId = zone1.id()}, "zone1x");
      CHECK(zone1.getName() == "zone1x");
    }

    THEN("Empty name fails") {
      CHECK_THROWS_FIND(dsModel->setZoneName({.zoneId = zone1.id()}, ""), "New name cannot be empty");
    }

    THEN("Invalid zoneId fails") {
      CHECK_THROWS_FIND(dsModel->setZoneName({.zoneId = dsModel::ZoneId{"455"}}, "xyz"), "Zone not found");
    }

    THEN("Zone 0 fails") {
      CHECK_THROWS_FIND(dsModel->setZoneName({.zoneId = dsModel::ZoneId{"0"}}, "xyz"), "Zone not found");
    }
    THEN("New devices zone fails") {
      CHECK_THROWS_FIND(dsModel->setZoneName({.zoneId = dsModel::ZoneId{ds::str(NEW_DEVICES_ZONE_ID)}}, "xyz"),
                        "New devices zone cannot be renamed");
    }
  }

  SECTION("basic apartment information") {
    MeteringMock& meteringMock = apartment.metering;
    using dsModel::MeteringOrigin;
    using dsModel::MeteringType;
    REQUIRE_CALL(meteringMock, meterings()).RETURN(std::vector<std::pair<MeteringOrigin, MeteringType>>{});
    const auto dsApartment = dsModel->apartment({.include = {}, .includeAll = true});
    CHECK(dsApartment.id == apartmentId);
    CHECK(dsApartment.type == "apartment");
    CHECK(dsApartment.attributes.name == apartment.getName());

    SECTION("included are present") {
      const auto& included = dsApartment.included;
      CHECK(included.installation == dsModel->installation());
      CHECK(included.scenarios == dsModel->scenarios().scenarios());
      CHECK(included.userDefinedStates == dsModel->userDefinedStates().userDefinedStates);
      CHECK(included.floors == dsModel->floors().floors);
      CHECK(included.applications == dsModel->applications());
      CHECK(included.apiRevision == dsModel->apiRevision());
    }

    SECTION("special character") {
      apartment.setName("Test&amp;Name");
      CHECK(dsModel->apartment({}).attributes.name == "Test&Name");
    }

    SECTION("api revision") {
      CHECK(dsModel->apiRevision() == dsModel::ApiRevision{.id = dsModel::ApiRevisionId{"apiRevisionId"},
                                                           .type = "apiRevision",
                                                           .lastChanged = {},
                                                           .attributes = {.additions = apiRevision.additions(),
                                                                          .removals = apiRevision.removals()}});
    }
  }

  SECTION("shadeType check") {
    SECTION("device of shade type") {
      GIVEN("GR_KL200") {
        auto& device = *createDevice(apartment, {.hardwareType = HardwareType::GR_KL200});
        THEN("rollerShutter") {
          CHECK(dsModel->device(device.id()) ==
                (dsModel::DSDevice{.id = device.id(),
                                   .type = "dsDevice",
                                   .lastChanged = {},
                                   .attributes = {.name = {},
                                                  .model = {},
                                                  .gtin = {},
                                                  .dsid = ds::str(device.getLegacyDSID()),
                                                  .displayId = device.getDisplayID(),
                                                  .present = 0,
                                                  .firstSeen = {},
                                                  .lastDiscovered = {},
                                                  .inactiveSince = {},
                                                  .shadeAttributes = {.shadeType = dsModel::ShadeType::rollerShutter,
                                                                      .protectionClass = {},
                                                                      .mountOrientation = {}},
                                                  .submodules = {device.submoduleId()},
                                                  .zone = dsModel::ZoneId{"65534"}}}));
        };
      }
    }

    SECTION("non-shade device") {
      GIVEN("GE_KM200") {
        auto& device = *createDevice(apartment, {.hardwareType = HardwareType::GE_KM200});
        THEN("brightness") {
          CHECK(dsModel->device(device.id()) ==
                (dsModel::DSDevice{
                    .id = device.id(),
                    .type = "dsDevice",
                    .lastChanged = {},
                    .attributes = {.name = {},
                                   .model = {},
                                   .gtin = {},
                                   .dsid = ds::str(device.getLegacyDSID()),
                                   .displayId = device.getDisplayID(),
                                   .present = 0,
                                   .firstSeen = {},
                                   .lastDiscovered = {},
                                   .inactiveSince = {},
                                   .shadeAttributes = {.shadeType = {}, .protectionClass = {}, .mountOrientation = {}},
                                   .submodules = {device.submoduleId()},
                                   .zone = dsModel::ZoneId{"65534"}}}));
        };
      }
    }

    SECTION("setShadeType") {
      GIVEN("GR_KL300") {
        auto& device = *createDevice(apartment, {.hardwareType = HardwareType::GR_KL300});
        auto deviceId = dsModel::DSDeviceId{ds::str(device.getDSID())};

        THEN("inital shadeType is none") {
          REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 1, 0x5a)).RETURN(0);
          CHECK(dsModel->device(deviceId).attributes.shadeAttributes ==
                dsModel::ShadeAttributes{
                    .shadeType = dsModel::ShadeType::none, .protectionClass = {}, .mountOrientation = {}});

          THEN("shadeType is cached") {
            CHECK(dsModel->device(deviceId).attributes.shadeAttributes ==
                  dsModel::ShadeAttributes{
                      .shadeType = dsModel::ShadeType::none, .protectionClass = {}, .mountOrientation = {}});
          }
        }

        WHEN("modifying shadeType") {
          REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 1, 0x5a, 1));
          dsModel->setShadeAttributes(deviceId, {.shadeType = dsModel::ShadeType::windowBlind});
          CHECK(dsModel->device(deviceId).attributes.shadeAttributes ==
                dsModel::ShadeAttributes{
                    .shadeType = dsModel::ShadeType::windowBlind, .protectionClass = {}, .mountOrientation = {}});

          THEN("shadeType is cached") {
            CHECK(dsModel->device(deviceId).attributes.shadeAttributes ==
                  dsModel::ShadeAttributes{
                      .shadeType = dsModel::ShadeType::windowBlind, .protectionClass = {}, .mountOrientation = {}});
          }
          WHEN("setting shadeType to none") {
            THEN("nothing changes") {
              dsModel->setShadeAttributes(deviceId, {.shadeType = {}});
              CHECK(dsModel->device(deviceId).attributes.shadeAttributes ==
                    dsModel::ShadeAttributes{
                        .shadeType = dsModel::ShadeType::windowBlind, .protectionClass = {}, .mountOrientation = {}});
            }
          }
        }
      }
      GIVEN("GE_KM200") {
        auto& device = *createDevice(apartment, {.hardwareType = HardwareType::GE_KM200});
        auto deviceId = dsModel::DSDeviceId{ds::str(device.getDSID())};

        THEN("shadeType is none") {
          CHECK(dsModel->device(deviceId).attributes.shadeAttributes ==
                dsModel::ShadeAttributes{
                    .shadeType = dsModel::ShadeType::none, .protectionClass = {}, .mountOrientation = {}});
        }
        WHEN("modifying shadeType") {
          THEN("exception is thrown") {
            CHECK_THROWS(dsModel->setShadeAttributes(deviceId, {.shadeType = dsModel::ShadeType::windowBlind}));
          }
        }
      }
    }
  }

  SECTION("apartment") {
    auto& zone0 = *apartment.allocateZone(0);
    GIVEN("access") {
      THEN("Missing state") {
        CHECK(dsModel->apartmentAccessStatus() == (dsModel::ApartmentAccessStatus{.absent = {}}));
      }

      auto presence = apartment.allocateState(StateType_Apartment, dss::StateName::Presence, "");
      State::ValueRange_t presenceValues;
      presenceValues.push_back("unknown");
      presenceValues.push_back("present");
      presenceValues.push_back("absent");
      presenceValues.push_back("invalid");
      presence->setValueRange(presenceValues);

      THEN("State invalid") {
        presence->setState(CallOrigin::unknown, "invalid");
        CHECK(dsModel->apartmentAccessStatus() == (dsModel::ApartmentAccessStatus{.absent = {}}));
      }
      THEN("State unknown") {
        presence->setState(CallOrigin::unknown, "unknown");
        CHECK(dsModel->apartmentAccessStatus() == (dsModel::ApartmentAccessStatus{.absent = {}}));
      }
      THEN("value == true") {
        presence->setState(CallOrigin::unknown, "absent");
        CHECK(dsModel->apartmentAccessStatus() == (dsModel::ApartmentAccessStatus{.absent = true}));
      }
      THEN("value == false") {
        presence->setState(CallOrigin::unknown, "present");
        CHECK(dsModel->apartmentAccessStatus() == (dsModel::ApartmentAccessStatus{.absent = false}));
      }
    }

    GIVEN("user") {
      THEN("Missing state") { CHECK(dsModel->apartmentUserStatus() == (dsModel::ApartmentUserStatus{.sleeping = {}})); }

      auto hibernation = apartment.allocateState(StateType_Apartment, dss::StateName::Hibernation, "");
      State::ValueRange_t sleepmodeValues;
      sleepmodeValues.push_back("unknown");
      sleepmodeValues.push_back("awake");
      sleepmodeValues.push_back("sleeping");
      sleepmodeValues.push_back("invalid");
      hibernation->setValueRange(sleepmodeValues);

      THEN("State invalid") {
        hibernation->setState(CallOrigin::unknown, "invalid");
        CHECK(dsModel->apartmentUserStatus() == (dsModel::ApartmentUserStatus{.sleeping = {}}));
      }
      THEN("State unknown") {
        hibernation->setState(CallOrigin::unknown, "unknown");
        CHECK(dsModel->apartmentUserStatus() == (dsModel::ApartmentUserStatus{.sleeping = {}}));
      }
      THEN("value == true") {
        hibernation->setState(CallOrigin::unknown, "awake");
        CHECK(dsModel->apartmentUserStatus() == (dsModel::ApartmentUserStatus{.sleeping = false}));
      }
      THEN("value == false") {
        hibernation->setState(CallOrigin::unknown, "sleeping");
        CHECK(dsModel->apartmentUserStatus() == (dsModel::ApartmentUserStatus{.sleeping = true}));
      }
    }

    GIVEN("security") {
      auto& group00 = *zone0.getGroup(0);
      SECTION("set alarm 1 name") {
        dsModel->setApartmentAlarm1Name("testName");
        CHECK(group00.getSceneName(SceneAlarm) == "testName");
      }
      SECTION("set alarm 2 name") {
        dsModel->setApartmentAlarm2Name("testName");
        CHECK(group00.getSceneName(SceneAlarm2) == "testName");
      }
      SECTION("set alarm 3 name") {
        dsModel->setApartmentAlarm3Name("testName");
        CHECK(group00.getSceneName(SceneAlarm3) == "testName");
      }
      SECTION("set alarm 4 name") {
        dsModel->setApartmentAlarm4Name("testName");
        CHECK(group00.getSceneName(SceneAlarm4) == "testName");
      }
    }

    GIVEN("securityStatus") {
      THEN("Missing state") {
        CHECK(dsModel->apartmentSecurityStatus() == (dsModel::ApartmentSecurityStatus{.panic = {},
                                                                                      .fire = {},
                                                                                      .smoke = {},
                                                                                      .gas = {},
                                                                                      .pollution = {},
                                                                                      .alarm1 = {},
                                                                                      .alarm2 = {},
                                                                                      .alarm3 = {},
                                                                                      .alarm4 = {},
                                                                                      .burglaryAlarm = {},
                                                                                      .burglaryAttempt = {}}));
      }

      auto panic = apartment.allocateState(StateType_Apartment, dss::StateName::Panic, "");
      auto fire = apartment.allocateState(StateType_Apartment, dss::StateName::Fire, "");
      auto alarm = apartment.allocateState(StateType_Apartment, dss::StateName::Alarm, "");
      auto alarm2 = apartment.allocateState(StateType_Apartment, dss::StateName::Alarm2, "");
      auto alarm3 = apartment.allocateState(StateType_Apartment, dss::StateName::Alarm3, "");
      auto alarm4 = apartment.allocateState(StateType_Apartment, dss::StateName::Alarm4, "");
      THEN("State invalid") {
        panic->setState(CallOrigin::unknown, State_Invalid);
        fire->setState(CallOrigin::unknown, State_Invalid);
        alarm->setState(CallOrigin::unknown, State_Invalid);
        alarm2->setState(CallOrigin::unknown, State_Invalid);
        alarm3->setState(CallOrigin::unknown, State_Invalid);
        alarm4->setState(CallOrigin::unknown, State_Invalid);
        CHECK(dsModel->apartmentSecurityStatus() == (dsModel::ApartmentSecurityStatus{.panic = {},
                                                                                      .fire = {},
                                                                                      .smoke = {},
                                                                                      .gas = {},
                                                                                      .pollution = {},
                                                                                      .alarm1 = {},
                                                                                      .alarm2 = {},
                                                                                      .alarm3 = {},
                                                                                      .alarm4 = {},
                                                                                      .burglaryAlarm = {},
                                                                                      .burglaryAttempt = {}}));
      }
      THEN("State unknown") {
        panic->setState(CallOrigin::unknown, State_Unknown);
        fire->setState(CallOrigin::unknown, State_Unknown);
        alarm->setState(CallOrigin::unknown, State_Unknown);
        alarm2->setState(CallOrigin::unknown, State_Unknown);
        alarm3->setState(CallOrigin::unknown, State_Unknown);
        alarm4->setState(CallOrigin::unknown, State_Unknown);
        CHECK(dsModel->apartmentSecurityStatus() == (dsModel::ApartmentSecurityStatus{.panic = {},
                                                                                      .fire = {},
                                                                                      .smoke = {},
                                                                                      .gas = {},
                                                                                      .pollution = {},
                                                                                      .alarm1 = {},
                                                                                      .alarm2 = {},
                                                                                      .alarm3 = {},
                                                                                      .alarm4 = {},
                                                                                      .burglaryAlarm = {},
                                                                                      .burglaryAttempt = {}}));
      }
      THEN("value == true") {
        panic->setState(CallOrigin::unknown, State_Active);
        fire->setState(CallOrigin::unknown, State_Active);
        alarm->setState(CallOrigin::unknown, State_Active);
        alarm2->setState(CallOrigin::unknown, State_Active);
        alarm3->setState(CallOrigin::unknown, State_Active);
        alarm4->setState(CallOrigin::unknown, State_Active);
        CHECK(dsModel->apartmentSecurityStatus() == (dsModel::ApartmentSecurityStatus{.panic = true,
                                                                                      .fire = true,
                                                                                      .smoke = {},
                                                                                      .gas = {},
                                                                                      .pollution = {},
                                                                                      .alarm1 = true,
                                                                                      .alarm2 = true,
                                                                                      .alarm3 = true,
                                                                                      .alarm4 = true,
                                                                                      .burglaryAlarm = {},
                                                                                      .burglaryAttempt = {}}));
      }
      THEN("value == false") {
        panic->setState(CallOrigin::unknown, State_Inactive);
        fire->setState(CallOrigin::unknown, State_Inactive);
        alarm->setState(CallOrigin::unknown, State_Inactive);
        alarm2->setState(CallOrigin::unknown, State_Inactive);
        alarm3->setState(CallOrigin::unknown, State_Inactive);
        alarm4->setState(CallOrigin::unknown, State_Inactive);
        CHECK(dsModel->apartmentSecurityStatus() == (dsModel::ApartmentSecurityStatus{.panic = false,
                                                                                      .fire = false,
                                                                                      .smoke = {},
                                                                                      .gas = {},
                                                                                      .pollution = {},
                                                                                      .alarm1 = false,
                                                                                      .alarm2 = false,
                                                                                      .alarm3 = false,
                                                                                      .alarm4 = false,
                                                                                      .burglaryAlarm = {},
                                                                                      .burglaryAttempt = {}}));
      }
    }

    SECTION("temperature") {
      auto settingsNode = propertySystem.createProperty("/scripts/system-addon-heating-controller/generalSettings");
      auto setSettingsChildValue = [&](auto&& path, auto&& value) {
        REQUIRE_CALL(hooks, apartmentStatusPolled());
        settingsNode->setChildValue(std::move(path), std::move(value));
      };
      auto ecoNode =
          propertySystem.createProperty("/usr/addon-states/system-addon-heating-controller/operationmode.0.Eco");
      auto setEcoValue = [&](auto&& value) {
        REQUIRE_CALL(hooks, apartmentStatusPolled());
        ecoNode->setChildValue("value", std::move(value));
      };

      GIVEN("no config") {
        THEN("heating not available") {
          CHECK(dsModel->apartmentTemperatureStatus() ==
                (dsModel::ApartmentTemperatureStatus{.mode = dsModel::TemperatureMode::off, .ecoPreferred = false}));
        }
      }

      GIVEN("heating enabled") {
        setSettingsChildValue("heatingEnabled", true);

        GIVEN("no forced settings") {
          THEN("mode = automatic") {
            CHECK(dsModel->apartmentTemperatureStatus() ==
                  (dsModel::ApartmentTemperatureStatus{.mode = dsModel::TemperatureMode::automatic,
                                                       .ecoPreferred = false}));
          }
          GIVEN("eco preferred") {
            setEcoValue(1);

            THEN("eco = true") {
              CHECK(dsModel->apartmentTemperatureStatus() ==
                    (dsModel::ApartmentTemperatureStatus{.mode = dsModel::TemperatureMode::automatic,
                                                         .ecoPreferred = true}));
            }
          }
        }

        GIVEN("forced heating") {
          setSettingsChildValue("overrideSystemModeAsHeating", true);

          THEN("mode = heating") {
            CHECK(dsModel->apartmentTemperatureStatus() ==
                  (dsModel::ApartmentTemperatureStatus{.mode = dsModel::TemperatureMode::heating,
                                                       .ecoPreferred = false}));
          }
        }

        GIVEN("forced cooling") {
          setSettingsChildValue("overrideSystemModeAsCooling", true);

          THEN("mode = cooling") {
            CHECK(dsModel->apartmentTemperatureStatus() ==
                  (dsModel::ApartmentTemperatureStatus{.mode = dsModel::TemperatureMode::cooling,
                                                       .ecoPreferred = false}));
          }
        }
      }
    }

    SECTION("weather") {
      GIVEN("Missing state") {
        THEN("no information") { CHECK(dsModel->apartmentWeatherStatus() == (dsModel::ApartmentWeatherStatus{})); }
      }

      GIVEN("states") {
        auto wind = apartment.allocateState(StateType_Apartment, dss::StateName::Wind, "");
        auto rain = apartment.allocateState(StateType_Apartment, dss::StateName::Rain, "");
        auto hail = apartment.allocateState(StateType_Apartment, dss::StateName::Hail, "");
        auto frost = apartment.allocateState(StateType_Apartment, dss::StateName::Frost, "");
        SECTION("State invalid") {
          wind->setState(CallOrigin::unknown, State_Invalid);
          rain->setState(CallOrigin::unknown, State_Invalid);
          hail->setState(CallOrigin::unknown, State_Invalid);
          frost->setState(CallOrigin::unknown, State_Invalid);
          CHECK(dsModel->apartmentWeatherStatus() ==
                (dsModel::ApartmentWeatherStatus{.wind = {}, .rain = {}, .hail = {}, .frost = {}}));
        }
        SECTION("State unknown") {
          wind->setState(CallOrigin::unknown, State_Unknown);
          rain->setState(CallOrigin::unknown, State_Unknown);
          hail->setState(CallOrigin::unknown, State_Unknown);
          frost->setState(CallOrigin::unknown, State_Unknown);
          CHECK(dsModel->apartmentWeatherStatus() ==
                (dsModel::ApartmentWeatherStatus{.wind = {}, .rain = {}, .hail = {}, .frost = {}}));
        }
        SECTION("value == true") {
          wind->setState(CallOrigin::unknown, State_Active);
          rain->setState(CallOrigin::unknown, State_Active);
          hail->setState(CallOrigin::unknown, State_Active);
          frost->setState(CallOrigin::unknown, State_Active);
          CHECK(dsModel->apartmentWeatherStatus() ==
                (dsModel::ApartmentWeatherStatus{.wind = true, .rain = true, .hail = true, .frost = true}));
        }
        SECTION("value == false") {
          wind->setState(CallOrigin::unknown, State_Inactive);
          rain->setState(CallOrigin::unknown, State_Inactive);
          hail->setState(CallOrigin::unknown, State_Inactive);
          frost->setState(CallOrigin::unknown, State_Inactive);
          CHECK(dsModel->apartmentWeatherStatus() ==
                (dsModel::ApartmentWeatherStatus{.wind = false, .rain = false, .hail = false, .frost = false}));
        }
      }
      SECTION("dayNight, condition") {
        const auto condition = dsModel::WeatherConditionId{"800"};
        SECTION("day") {
          apartment.setWeatherInformation("10d", "800", "", dss::DateTime{system.now()});
          CHECK(dsModel->apartmentWeatherStatus() == dsModel::ApartmentWeatherStatus{.dayNight = dsModel::DayNight::day,
                                                                                     .condition = condition,
                                                                                     .icon = "10d"});
        }
        SECTION("night") {
          apartment.setWeatherInformation("10n", "800", "", dss::DateTime{system.now()});
          CHECK(dsModel->apartmentWeatherStatus() ==
                dsModel::ApartmentWeatherStatus{
                    .dayNight = dsModel::DayNight::night, .condition = condition, .icon = "10n"});
        }
        SECTION("neighter day, neighter night -> DayNight::none") {
          apartment.setWeatherInformation("10", "800", "", dss::DateTime{system.now()});
          CHECK(dsModel->apartmentWeatherStatus() ==
                dsModel::ApartmentWeatherStatus{.condition = condition, .icon = "10"});
        }
        SECTION("condition is passed") {
          apartment.setWeatherInformation("", "800", "", dss::DateTime{system.now()});
          CHECK(dsModel->apartmentWeatherStatus() == dsModel::ApartmentWeatherStatus{.condition = condition});
        }
        SECTION("outdated is not passed") {
          apartment.setWeatherInformation("10d", "800", "", dss::DateTime{system.now() - 2h});
          CHECK(dsModel->apartmentWeatherStatus() == dsModel::ApartmentWeatherStatus{});
        }
      }
    }

    SECTION("apartmentMeasurementStatus") {
      SECTION("No values set") {
        CHECK(dsModel->apartmentMeasurementStatus() == dsModel::ApartmentMeasurementStatus{});
      }

      SECTION("Values set") {
        DateTime now;
        apartment.setTemperature(15.5, now);
        apartment.setHumidityValue(50, now);
        apartment.setBrightnessValue(25, now);
        apartment.setAirPressure(900, now);
        apartment.setPrecipitation(5, now);
        // collecting gustSpeed + gustDirection internally
        apartment.setWindSpeed(6, now);
        apartment.setWindGust(13, now);

        // publishing gustSpeed as windGust, gustDirection not planned to be published
        CHECK(dsModel->apartmentMeasurementStatus() == dsModel::ApartmentMeasurementStatus{.temperature = 15.5,
                                                                                           .humidity = 50,
                                                                                           .brightness = 25,
                                                                                           .airPressure = 900,
                                                                                           .precipitation = 5,
                                                                                           .windSpeed = 6,
                                                                                           .windGust = 13});
      }
    }

    SECTION("apartmentEnvironmentStatus") {
      SECTION("No values set") {
        CHECK(dsModel->apartmentMeasurementStatus() == dsModel::ApartmentMeasurementStatus{});
      }

      SECTION("Values set") {
        propertySystem.createProperty("/config/geodata/sunrise")->setValue("07:03:04"s);
        propertySystem.createProperty("/config/geodata/sunset")->setValue("17:03:11"s);
        propertySystem.createProperty("/config/geodata/civil_dawn")->setValue("06:53:11"s);
        propertySystem.createProperty("/config/geodata/civil_dusk")->setValue("17:13:00"s);
        /// mocked system defaults to
        /// * now(): 2017-09-18T10:37:00.777Z
        /// * timeZoneOffset(): 1h

        CHECK(dsModel->apartmentEnvironmentStatus() ==
              dsModel::ApartmentEnvironmentStatus{.sunrise = ds::toDateTime("2017-09-18T06:03:04.000Z"),
                                                  .sunset = ds::toDateTime("2017-09-18T16:03:11.000Z"),
                                                  .dawn = ds::toDateTime("2017-09-18T05:53:11.000Z"),
                                                  .dusk = ds::toDateTime("2017-09-18T16:13:00.000Z")});
      }
    }
  }

  SECTION("zoneApplicationStatus") {
    auto& zone1 = *apartment.allocateZone(1);
    auto& shadeDevice = *apartment.allocateDevice(Dsuid{"1111111111111111111111111111111100"});
    auto& ventilationDevice = *apartment.allocateDevice(Dsuid{"3111111111111111111111111111111100"});
    auto& recirculationDevice = *apartment.allocateDevice(Dsuid{"4111111111111111111111111111111100"});
    auto& temperatureDevice = *apartment.allocateDevice(Dsuid{"5111111111111111111111111111111100"});

    DeviceSpec spec;
    spec.FunctionID = 0x10; // This means the device has output
    spec.activeGroup = 2;

    ALLOW_CALL(dsm_api_mock, DsmApiGetDeviceDescription(_, _, _, _, _, _, _)).RETURN(0);
    shadeDevice.setPartiallyFromSpec(spec); // Device has output, therefore it is not just button

    auto& awningDevice = *createDevice(apartment, {.hardwareType = HardwareType::GR_KL210});
    awningDevice.setActiveGroup(ApplicationType::blinds);

    ventilationDevice.setActiveGroup(ApplicationType::ventilation);
    recirculationDevice.setActiveGroup(ApplicationType::recirculation);
    temperatureDevice.setActiveGroup(ApplicationType::temperature);

    GIVEN("shades in zone") {
      zone1.addDevice({shadeDevice.sharedFromThis(), &apartment});
      REQUIRE_CALL(proxy, shadesStatus(zone1.id()))
          .RETURN(dsModel::ShadesStatus{.status = dsModel::ShadesStatusStatus::open});
      CHECK(
          dsModel->zoneApplicationsStatus(zone1.id()) ==
          dsModel::ZoneApplicationsStatus{.lights = {},
                                          .shades = dsModel::ShadesStatus{.status = dsModel::ShadesStatusStatus::open},
                                          .awnings = {},
                                          .audio = {},
                                          .ventilation = {},
                                          .recirculation = {},
                                          .temperature = {}});
    }
    GIVEN("awnings in zone") {
      zone1.addDevice({awningDevice.sharedFromThis(), &apartment});
      REQUIRE_CALL(proxy, shadesStatus(zone1.id()))
          .RETURN(dsModel::ShadesStatus{.status = dsModel::ShadesStatusStatus::open});
      CHECK(
          dsModel->zoneApplicationsStatus(zone1.id()) ==
          dsModel::ZoneApplicationsStatus{.lights = {},
                                          .shades = {},
                                          .awnings = dsModel::AwningsStatus{.status = dsModel::AwningsStatusStatus::out},
                                          .audio = {},
                                          .ventilation = {},
                                          .recirculation = {},
                                          .temperature = {}});
    }
    GIVEN("shades and awnings in zone") {
      zone1.addDevice({shadeDevice.sharedFromThis(), &apartment});
      zone1.addDevice({awningDevice.sharedFromThis(), &apartment});
      ALLOW_CALL(proxy, shadesStatus(zone1.id()))
          .RETURN(dsModel::ShadesStatus{.status = dsModel::ShadesStatusStatus::open});
      CHECK(
          dsModel->zoneApplicationsStatus(zone1.id()) ==
          dsModel::ZoneApplicationsStatus{.lights = {},
                                          .shades = dsModel::ShadesStatus{.status = dsModel::ShadesStatusStatus::open},
                                          .awnings = dsModel::AwningsStatus{.status = dsModel::AwningsStatusStatus::out},
                                          .audio = {},
                                          .ventilation = {},
                                          .recirculation = {},
                                          .temperature = {}});
    }

    GIVEN("ventilation, recirculation and temperature in zone") {
      auto zoneVentilationNode =
          propertySystem.createProperty("/scripts/system-addon-ventilation-controller/zone-ventilation/1");
      auto zoneRecirculationNode =
          propertySystem.createProperty("/scripts/system-addon-ventilation-controller/zone-recirculation/1");
      zone1.addDevice({ventilationDevice.sharedFromThis(), &apartment});
      zone1.addDevice({recirculationDevice.sharedFromThis(), &apartment});
      zone1.addDevice({temperatureDevice.sharedFromThis(), &apartment});

      zone1.setTemperatureConfig([] {
        TemperatureConfig result{};
        result.m_mode = HeatingControlMode::PID; // PID mode required for temperature application to exist
        return result;
      }());

      REQUIRE_CALL(proxy, ventilationStatus(zone1.id()))
          .RETURN(dsModel::VentilationStatus{
              .level = dsModel::VentilationLevel::off, .malfunction = true, .serviceRequired = false});
      REQUIRE_CALL(proxy, recirculationStatus(zone1.id()))
          .RETURN(dsModel::VentilationStatus{
              .level = dsModel::VentilationLevel::level1, .malfunction = false, .serviceRequired = true});

      CHECK(dsModel->zoneApplicationsStatus(zone1.id()) ==
            dsModel::ZoneApplicationsStatus{
                .lights = {},
                .shades = {},
                .awnings = {},
                .audio = {},
                .ventilation =
                    dsModel::VentilationStatus{
                        .level = dsModel::VentilationLevel::off, .malfunction = true, .serviceRequired = false},
                .recirculation =
                    dsModel::VentilationStatus{
                        .level = dsModel::VentilationLevel::level1, .malfunction = false, .serviceRequired = true},
                .temperature = dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::off,
                                                          .status = dsModel::TemperatureStatusStatus::off,
                                                          .emergency = {},
                                                          .setpoint = {},
                                                          .controlValue = {},
                                                          .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                                          .setbackRequested = false,
                                                          .ecoPreferred = false,
                                                          .passiveCooling = false}});
    }
  }

  SECTION("temperatureStatus") {
    auto& zone1 = *apartment.allocateZone(1);
    DateTime now;

    auto settingsNode = propertySystem.createProperty("/scripts/system-addon-heating-controller/generalSettings");
    auto setSettingsChildValue = [&](auto&& path, auto&& value) {
      REQUIRE_CALL(hooks, apartmentStatusPolled());
      settingsNode->setChildValue(std::move(path), std::move(value));
    };
    auto configNode = propertySystem.createProperty("/scripts/system-addon-heating-controller/controllerConfig/1");
    auto setConfigChildValue = [&](auto&& path, auto&& value) {
      REQUIRE_CALL(hooks, apartmentStructurePolled());
      REQUIRE_CALL(hooks, apartmentStatusPolled());
      configNode->setChildValue(std::move(path), std::move(value));
    };

    GIVEN("no temperature device") {
      THEN("temperatureStatus throws") {
        CHECK_THROWS_FIND(dsModel->temperatureStatus(zone1.id()), "Temperature application does not exist.");
      }
      THEN("applications zoneApplicationsStatus() is empty") {
        CHECK(dsModel->zoneApplicationsStatus(zone1.id()) == dsModel::ZoneApplicationsStatus{});
      }
    }

    auto& device = *apartment.allocateDevice(Dsuid{"5111111111111111111111111111111100"});
    device.setActiveGroup(ApplicationType::temperature);
    zone1.addDevice({device.sharedFromThis(), &apartment});

    GIVEN("no PID controller or passive cooling in zone") {
      THEN("temperatureStatus throws") {
        const auto& lastSceneInt = GENERATE(values<int>({
            HeatingOperationModeIDOff,
            HeatingOperationModeIDComfort,
            HeatingOperationModeIDEconomy,
            HeatingOperationModeIDNotUsed,
            HeatingOperationModeIDNight,
            HeatingOperationModeIDHoliday,
            HeatingOperationModeIDCoolOff,
            HeatingOperationModeIDCoolComfort,
            HeatingOperationModeIDCoolEconomy,
            HeatingOperationModeIDCoolNotUsed,
            HeatingOperationModeIDCoolNight,
            HeatingOperationModeIDCoolHoliday,
        }));
        zone1.setHeatingOperationMode(lastSceneInt);
        CHECK_THROWS_FIND(dsModel->temperatureStatus(zone1.id()), "Temperature application does not exist.");
      }
      THEN("applications zoneApplicationsStatus() is empty") {
        CHECK(dsModel->zoneApplicationsStatus(zone1.id()) == dsModel::ZoneApplicationsStatus{});
      }
    }

    GIVEN("no PID controller, but passive cooling possible") {
      const auto& [lastSceneInt, lastSceneEnum] = GENERATE(values<std::pair<int, dsModel::TemperatureScene>>(
          {{HeatingOperationModeIDPassiveCoolingOn, dsModel::TemperatureScene::passiveCoolingActive},
           {HeatingOperationModeIDPassiveCoolingOff, dsModel::TemperatureScene::passiveCoolingInactive}}));
      zone1.setHeatingOperationMode(lastSceneInt);
      THEN("temperatureStatus available, passive cooling active") {
        CHECK(dsModel->temperatureStatus(zone1.id()) ==
              dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::off,
                                         .status = dsModel::TemperatureStatusStatus::off,
                                         .emergency = {},
                                         .setpoint = {},
                                         .controlValue = {},
                                         .lastCalledScene = lastSceneEnum,
                                         .setbackRequested = false,
                                         .ecoPreferred = false,
                                         .passiveCooling = false});
      }
    }

    zone1.setTemperatureConfig([] {
      TemperatureConfig result{};
      result.m_mode = HeatingControlMode::PID; // PID mode required for temperature application to exist
      return result;
    }());

    GIVEN("heating disabled") {
      setSettingsChildValue("heatingEnabled", false);
      THEN("all is off") {
        CHECK(dsModel->temperatureStatus(zone1.id()) ==
              dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::off,
                                         .status = dsModel::TemperatureStatusStatus::off,
                                         .emergency = {},
                                         .setpoint = {},
                                         .controlValue = {},
                                         .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                         .setbackRequested = false,
                                         .ecoPreferred = false,
                                         .passiveCooling = false});
      }
      THEN("applications zoneApplicationsStatus() containes temperature status") {
        CHECK(dsModel->zoneApplicationsStatus(zone1.id()) ==
              dsModel::ZoneApplicationsStatus{.lights = {},
                                              .shades = {},
                                              .awnings = {},
                                              .audio = {},
                                              .ventilation = {},
                                              .recirculation = {},
                                              .temperature = dsModel->temperatureStatus(zone1.id())});
      }
    }

    GIVEN("apartment heating enabled") {
      setSettingsChildValue("heatingEnabled", true);

      THEN("zone heating/cooling not possible") {
        setConfigChildValue("heatingPossible", false);
        setConfigChildValue("coolingPossible", false);
        setConfigChildValue("passiveCoolingPossible", false);

        CHECK(dsModel->temperatureStatus(zone1.id()) ==
              dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::off,
                                         .status = dsModel::TemperatureStatusStatus::off,
                                         .emergency = {},
                                         .setpoint = {},
                                         .controlValue = {},
                                         .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                         .setbackRequested = false,
                                         .ecoPreferred = false,
                                         .passiveCooling = false});
      }

      GIVEN("zone heating, cooling, passiveCooling possible") {
        setConfigChildValue("heatingPossible", true);
        setConfigChildValue("coolingPossible", true);
        setConfigChildValue("passiveCoolingPossible", true);

        THEN("heating") {
          zone1.setControlValue(25, now);

          CHECK(dsModel->temperatureStatus(zone1.id()) ==
                dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::automatic,
                                           .status = dsModel::TemperatureStatusStatus::heating,
                                           .emergency = false,
                                           .setpoint = {},
                                           .controlValue = 25,
                                           .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                           .setbackRequested = false,
                                           .ecoPreferred = false,
                                           .passiveCooling = false});
        }

        THEN("cooling") {
          zone1.setControlValue(-50, now);

          CHECK(dsModel->temperatureStatus(zone1.id()) ==
                dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::automatic,
                                           .status = dsModel::TemperatureStatusStatus::cooling,
                                           .emergency = false,
                                           .setpoint = {},
                                           .controlValue = -50,
                                           .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                           .setbackRequested = false,
                                           .ecoPreferred = false,
                                           .passiveCooling = false});
        }

        THEN("neutral") {
          zone1.setControlValue(0, now);

          CHECK(dsModel->temperatureStatus(zone1.id()) ==
                dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::automatic,
                                           .status = dsModel::TemperatureStatusStatus::neutral,
                                           .emergency = false,
                                           .setpoint = {},
                                           .controlValue = 0,
                                           .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                           .setbackRequested = false,
                                           .ecoPreferred = false,
                                           .passiveCooling = false});
        }

        SECTION("emergency") {
          zone1.setHeatingControlState(HeatingControlStateIDEmergency);

          CHECK(dsModel->temperatureStatus(zone1.id()) ==
                dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::automatic,
                                           .status = dsModel::TemperatureStatusStatus::off,
                                           .emergency = true,
                                           .setpoint = {},
                                           .controlValue = {},
                                           .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                           .setbackRequested = false,
                                           .ecoPreferred = false,
                                           .passiveCooling = false});
        }
      }

      GIVEN("heating forced") {
        setSettingsChildValue("overrideSystemModeAsHeating", true);

        THEN("heating not possible") {
          setConfigChildValue("heatingPossible", false);
          CHECK(dsModel->temperatureStatus(zone1.id()) ==
                dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::off,
                                           .status = dsModel::TemperatureStatusStatus::off,
                                           .emergency = {},
                                           .setpoint = {},
                                           .controlValue = {},
                                           .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                           .setbackRequested = false,
                                           .ecoPreferred = false,
                                           .passiveCooling = false});
        }

        THEN("heating possible") {
          setConfigChildValue("heatingPossible", true);
          CHECK(dsModel->temperatureStatus(zone1.id()) ==
                dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::heating,
                                           .status = dsModel::TemperatureStatusStatus::off,
                                           .emergency = false,
                                           .setpoint = {},
                                           .controlValue = {},
                                           .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                           .setbackRequested = false,
                                           .ecoPreferred = false,
                                           .passiveCooling = false});
        }

        GIVEN("cooling forced") {
          setSettingsChildValue("overrideSystemModeAsHeating", false);
          setSettingsChildValue("overrideSystemModeAsCooling", true);

          THEN("cooling and passive cooling not possible") {
            setConfigChildValue("coolingPossible", false);
            setConfigChildValue("passiveCoolingPossible", false);
            CHECK(dsModel->temperatureStatus(zone1.id()) ==
                  dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::off,
                                             .status = dsModel::TemperatureStatusStatus::off,
                                             .emergency = {},
                                             .setpoint = {},
                                             .controlValue = {},
                                             .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                             .setbackRequested = false,
                                             .ecoPreferred = false,
                                             .passiveCooling = false});
          }

          THEN("cooling possible") {
            setConfigChildValue("coolingPossible", true);
            setConfigChildValue("passiveCoolingPossible", false);
            CHECK(dsModel->temperatureStatus(zone1.id()) ==
                  dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::cooling,
                                             .status = dsModel::TemperatureStatusStatus::off,
                                             .emergency = false,
                                             .setpoint = {},
                                             .controlValue = {},
                                             .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                             .setbackRequested = false,
                                             .ecoPreferred = false,
                                             .passiveCooling = false});
          }

          THEN("passive cooling possible") {
            setConfigChildValue("coolingPossible", false);
            setConfigChildValue("passiveCoolingPossible", true);
            propertySystem.createProperty("/usr/addon-states/system-addon-heating-controller/passiveCooling.1/value")->setValue(1);

            CHECK(dsModel->temperatureStatus(zone1.id()) ==
                  dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::cooling,
                                             .status = dsModel::TemperatureStatusStatus::off,
                                             .emergency = false,
                                             .setpoint = {},
                                             .controlValue = {},
                                             .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                             .setbackRequested = false,
                                             .ecoPreferred = false,
                                             .passiveCooling = true});
          }
        }

        GIVEN("lastCalledScene and setpoint") {
          requireAsyncStatusChanged([&] { zone1.setNominalValue(20, DateTime()); });
          requireAsyncStatusChanged([&] { zone1.setHeatingOperationMode(HeatingOperationModeIDEconomy); });

          THEN("check") {
            CHECK(dsModel->temperatureStatus(zone1.id()) ==
                  dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::off,
                                             .status = dsModel::TemperatureStatusStatus::off,
                                             .emergency = {},
                                             .setpoint = 20,
                                             .controlValue = {},
                                             .lastCalledScene = dsModel::TemperatureScene::heatingEconomy,
                                             .setbackRequested = false,
                                             .ecoPreferred = false,
                                             .passiveCooling = false});
          }
        }

        GIVEN("eco and setback set") {
          auto ecoNode =
              propertySystem.createProperty("/usr/addon-states/system-addon-heating-controller/operationmode.1.Eco");
          auto setEcoValue = [&](auto&& value) {
            REQUIRE_CALL(hooks, apartmentStatusPolled());
            ecoNode->setChildValue("value", std::move(value));
          };
          auto setbackNode = propertySystem.createProperty(
              "/usr/addon-states/system-addon-heating-controller/operationmode.1.NotUsed");
          auto setSetbackValue = [&](auto&& value) {
            REQUIRE_CALL(hooks, apartmentStatusPolled());
            setbackNode->setChildValue("value", std::move(value));
          };

          setEcoValue(1);
          setSetbackValue(1);

          THEN("check") {
            CHECK(dsModel->temperatureStatus(zone1.id()) ==
                  dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::off,
                                             .status = dsModel::TemperatureStatusStatus::off,
                                             .emergency = {},
                                             .setpoint = {},
                                             .controlValue = {},
                                             .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                             .setbackRequested = true,
                                             .ecoPreferred = true,
                                             .passiveCooling = false});
          }
        }
      }
    }
  }

  SECTION("Set apartment temperature status eco preferred") {
    const auto& value = GENERATE(true, false);
    REQUIRE_CALL(eventSink, pushEvent(_))
        .LR_SIDE_EFFECT(CHECK(
            *_1 ==
            Event{"heating-controller.operation-mode",
                  Properties{{"actions", "setOperationMode"}, {"zoneID", "0"}, {"operationMode", value ? "2" : "1"}}}));
    dsModel->setApartmentTemperatureStatusEcoPreferred({}, value);
  }

  SECTION("Set temperature status setpoint") {
    auto& zone1 = *apartment.allocateZone(1);
    const auto value = 25.4;
    REQUIRE_CALL(busInterface.action,
                 pushSensor(zone1.getGroup(GroupIDBroadcast).get(), Dsuid{}, SensorType::RoomTemperatureSetpoint, value,
                            Caller{CallOrigin::json, SceneAccessCategory::manual, ""s}));
    dsModel->setTemperatureStatusSetpoint({.zoneId = zone1.id()}, value);
  }

  SECTION("Set temperature status setbackRequested") {
    THEN("setbackRequested true") {
      REQUIRE_CALL(eventSink, pushEvent(_))
          .LR_SIDE_EFFECT(CHECK(
              *_1 == Event{"heating-controller.operation-mode",
                           Properties{{"actions", "setOperationMode"}, {"zoneID", "1"}, {"operationMode", "3"}}}));
      const dsModel::TemperatureStatusArgs& args = {.zoneId = dsModel::ZoneId{"1"}};
      dsModel->setTemperatureStatusSetbackRequested(args, true);
    }
    THEN("setbackRequested false") {
      REQUIRE_CALL(eventSink, pushEvent(_))
          .LR_SIDE_EFFECT(CHECK(
              *_1 == Event{"heating-controller.operation-mode",
                           Properties{{"actions", "resetOperationMode"}, {"zoneID", "1"}, {"operationMode", "3"}}}));
      const dsModel::TemperatureStatusArgs& args = {.zoneId = dsModel::ZoneId{"1"}};
      dsModel->setTemperatureStatusSetbackRequested(args, false);
    }
  }

  SECTION("Set temperature status ecoPreferred") {
    THEN("ecoPreferred true") {
      REQUIRE_CALL(eventSink, pushEvent(_))
          .LR_SIDE_EFFECT(CHECK(
              *_1 == Event{"heating-controller.operation-mode",
                           Properties{{"actions", "setOperationMode"}, {"zoneID", "1"}, {"operationMode", "2"}}}));
      const dsModel::TemperatureStatusArgs& args = {.zoneId = dsModel::ZoneId{"1"}};
      dsModel->setTemperatureStatusEcoPreferred(args, true);
    }
    THEN("ecoPreferred false") {
      REQUIRE_CALL(eventSink, pushEvent(_))
          .LR_SIDE_EFFECT(CHECK(
              *_1 == Event{"heating-controller.operation-mode",
                           Properties{{"actions", "setOperationMode"}, {"zoneID", "1"}, {"operationMode", "1"}}}));
      const dsModel::TemperatureStatusArgs& args = {.zoneId = dsModel::ZoneId{"1"}};
      dsModel->setTemperatureStatusEcoPreferred(args, false);
    }
  }

  SECTION("Set temperature status passiveCooling") {
    THEN("passiveCooling true") {
      REQUIRE_CALL(eventSink, pushEvent(_))
          .LR_SIDE_EFFECT(CHECK(
              *_1 == Event{"heating-controller.passive-cooling", Properties{{"actions", "enable"}, {"zoneID", "1"}}}));
      const dsModel::TemperatureStatusArgs& args = {.zoneId = dsModel::ZoneId{"1"}};
      dsModel->setTemperatureStatusPassiveCooling(args, true);
    }
    THEN("passiveCooling false") {
      REQUIRE_CALL(eventSink, pushEvent(_))
          .LR_SIDE_EFFECT(CHECK(
              *_1 == Event{"heating-controller.passive-cooling", Properties{{"actions", "disable"}, {"zoneID", "1"}}}));
      const dsModel::TemperatureStatusArgs& args = {.zoneId = dsModel::ZoneId{"1"}};
      dsModel->setTemperatureStatusPassiveCooling(args, false);
    }
  }

  SECTION("zoneStatus") {
    auto& zone1 = *apartment.allocateZone(1);
    SECTION("zoneMeasurementsStatus") {
      THEN("no measurements") {
        CHECK(dsModel->zoneMeasurementsStatus(dsModel::ZoneId{"1"}) == dsModel::ZoneMeasurementsStatus{});
      }

      GIVEN("states") {
        auto& motion = *apartment.allocateState(StateType_Service, zoneStateName(StateName::Motion, zone1.getID()), {});
        auto& presence =
            *apartment.allocateState(StateType_Service, zoneStateName(StateName::Presence, zone1.getID()), {});
        motion.setState(CallOrigin::unknown, State_Invalid);
        presence.setState(CallOrigin::unknown, State_Invalid);

        SECTION("no measurements") {
          CHECK(dsModel->zoneMeasurementsStatus(dsModel::ZoneId{"1"}) == dsModel::ZoneMeasurementsStatus{});
        }

        SECTION("zone stored measurements") {
          zone1.setTemperature(11, {});
          zone1.setHumidityValue(12, {});
          zone1.setBrightnessValue(13, {});
          zone1.setCO2ConcentrationValue(14, {});
          CHECK(dsModel->zoneMeasurementsStatus(dsModel::ZoneId{"1"}) ==
                dsModel::ZoneMeasurementsStatus{
                    .temperature = 11, .humidity = 12, .brightness = 13, .carbonDioxide = 14});
        }

        SECTION("state represented measurements") {
          struct X {
            eState dssValue;
            std::optional<bool> dsValue;
          };
          const auto& x = GENERATE(values(std::initializer_list<X>{
              {State_Invalid, {}}, {State_Active, true}, {State_Inactive, false}, {State_Unknown, {}}}));
          INFO(x.dssValue);
          motion.setState(CallOrigin::unknown, x.dssValue);
          presence.setState(CallOrigin::unknown, x.dssValue);
          CHECK(dsModel->zoneMeasurementsStatus(dsModel::ZoneId{"1"}) ==
                dsModel::ZoneMeasurementsStatus{.temperature = {},
                                                .humidity = {},
                                                .brightness = {},
                                                .carbonDioxide = {},
                                                .motion = x.dsValue,
                                                .presence = x.dsValue});
        }
      }
    }
  }

  SECTION("decode application id") {
    CHECK(decodeId(dsModel::ApplicationId{"temperature-z1"}) == DecodedApplicationId{.zoneId = 1});

    CHECK_THROWS_FIND(decodeId(dsModel::ApplicationId{"temperature-1"}), "Wrong applicationId format.");
    CHECK_THROWS_FIND(decodeId(dsModel::ApplicationId{"ventilation-z1"}), "Unsupported application type.");
  }

  SECTION("applications") {
    auto& zone1 = *apartment.allocateZone(1);
    auto& device = *apartment.allocateDevice(Dsuid{"1111111111111111111111111111111100"});
    DeviceReference devRef(device.sharedFromThis(), &apartment);
    zone1.addDevice(devRef);

    SECTION("applications") {
      GIVEN("No device") {
        THEN("No applications") { CHECK(dsModel->applications() == std::vector<dsModel::Application>{}); }
      }
    }

    SECTION("temperature") {
      SECTION("read settings") {
        GIVEN("No temperature application device in zone") {
          THEN("No temperature returned") {
            CHECK_THROWS_FIND(dsModel->temperatureApplication(dsModel::ZoneId{ds::str(zone1.getID())}),
                              "Temperature application does not exist.");
          }
        }

        GIVEN("Temperature application device present in zone") {
          GIVEN("No PID controller ") {
            THEN("No temperature returned") {
              CHECK_THROWS_FIND(dsModel->temperatureApplication(dsModel::ZoneId{ds::str(zone1.getID())}),
                                "Temperature application does not exist.");
            }
          }

          zone1.setTemperatureConfig([] {
            TemperatureConfig result{};
            result.m_mode = HeatingControlMode::PID; // PID mode required for temperature application to exist
            return result;
          }());

          device.setActiveGroup(ApplicationType::temperature);
          auto configNode =
              propertySystem.createProperty("/scripts/system-addon-heating-controller/controllerConfig/1");
          auto setConfigChildValue = [&](auto&& path, auto&& value) {
            REQUIRE_CALL(hooks, apartmentStructurePolled());
            REQUIRE_CALL(hooks, apartmentStatusPolled());
            configNode->setChildValue(std::move(path), std::move(value));
          };
          TemperatureConfig config{};
          config.targetTemperatures = {10, 11, 12, 13, 14, 15, 0, 0, 0, 20, 21, 22, 23, 24, 25};
          zone1.setHeatingSetpoints(config);

          THEN("Default temperature") {
            CHECK(dsModel->temperatureApplication(dsModel::ZoneId{ds::str(zone1.getID())}) ==
                  dsModel::Application{.id = dsModel::ApplicationId{"temperature-z1"},
                                       .attributes = dsModel::TemperatureApplicationAttributes{
                                           .applicationType = dsModel::ZoneApplicationType::temperature,
                                           .zone = dsModel::ZoneId{ds::str(zone1.getID())},
                                           .availableLevels = {},
                                           .setpoints = {}}});
          }

          GIVEN("Heating possible") {
            setConfigChildValue("heatingPossible", true);
            THEN("Heating levels available") {
              setConfigChildValue("heatingPossible", true);
              const std::vector<dsModel::TemperatureLevel> levels = {
                  dsModel::TemperatureLevel::heatingComfort, dsModel::TemperatureLevel::heatingEconomy,
                  dsModel::TemperatureLevel::heatingSetback, dsModel::TemperatureLevel::heatingNight,
                  dsModel::TemperatureLevel::heatingVacation};
              const std::vector<dsModel::TemperatureSetpoint> setpoints = {
                  {.id = dsModel::TemperatureLevel::heatingComfort, .value = 11},
                  {.id = dsModel::TemperatureLevel::heatingEconomy, .value = 12},
                  {.id = dsModel::TemperatureLevel::heatingSetback, .value = 13},
                  {.id = dsModel::TemperatureLevel::heatingNight, .value = 14},
                  {.id = dsModel::TemperatureLevel::heatingVacation, .value = 15}};
              CHECK(dsModel->temperatureApplication(dsModel::ZoneId{ds::str(zone1.getID())}) ==
                    dsModel::Application{.id = dsModel::ApplicationId{"temperature-z1"},
                                         .attributes = dsModel::TemperatureApplicationAttributes{
                                             .applicationType = dsModel::ZoneApplicationType::temperature,
                                             .zone = dsModel::ZoneId{ds::str(zone1.getID())},
                                             .availableLevels = levels,
                                             .setpoints = setpoints}});
            }
          }

          GIVEN("Cooling possible") {
            setConfigChildValue("coolingPossible", true);
            THEN("cooling levels available") {
              const std::vector<dsModel::TemperatureLevel> levels = {
                  dsModel::TemperatureLevel::coolingComfort, dsModel::TemperatureLevel::coolingEconomy,
                  dsModel::TemperatureLevel::coolingSetback, dsModel::TemperatureLevel::coolingNight,
                  dsModel::TemperatureLevel::coolingVacation};
              const std::vector<dsModel::TemperatureSetpoint> setpoints = {
                  {.id = dsModel::TemperatureLevel::coolingComfort, .value = 21},
                  {.id = dsModel::TemperatureLevel::coolingEconomy, .value = 22},
                  {.id = dsModel::TemperatureLevel::coolingSetback, .value = 23},
                  {.id = dsModel::TemperatureLevel::coolingNight, .value = 24},
                  {.id = dsModel::TemperatureLevel::coolingVacation, .value = 25}};
              CHECK(dsModel->temperatureApplication(dsModel::ZoneId{ds::str(zone1.getID())}) ==
                    dsModel::Application{.id = dsModel::ApplicationId{"temperature-z1"},
                                         .attributes = dsModel::TemperatureApplicationAttributes{
                                             .applicationType = dsModel::ZoneApplicationType::temperature,
                                             .zone = dsModel::ZoneId{ds::str(zone1.getID())},
                                             .availableLevels = levels,
                                             .setpoints = setpoints}});
            }
          }

          GIVEN("Passive cooling possible") {
            auto modesNode = propertySystem.createProperty("/scripts/system-addon-heating-controller/modes/1");
            auto setPassiveCoolingThreshold = [&](auto&& value) {
              REQUIRE_CALL(hooks, apartmentStructurePolled());
              modesNode->setChildValue("minimalTemperatur", ds::str(value));
            };

            setPassiveCoolingThreshold(8.0);
            setConfigChildValue("passiveCoolingPossible", true);
            THEN("Passive cooling levels available") {
              const std::vector<dsModel::TemperatureLevel> levels = {
                  dsModel::TemperatureLevel::passiveCoolingThreshold};
              const std::vector<dsModel::TemperatureSetpoint> threshold = {
                  {.id = dsModel::TemperatureLevel::passiveCoolingThreshold, .value = 8}};
              CHECK(dsModel->temperatureApplication(dsModel::ZoneId{ds::str(zone1.getID())}) ==
                    dsModel::Application{.id = dsModel::ApplicationId{"temperature-z1"},
                                         .attributes = dsModel::TemperatureApplicationAttributes{
                                             .applicationType = dsModel::ZoneApplicationType::temperature,
                                             .zone = dsModel::ZoneId{ds::str(zone1.getID())},
                                             .availableLevels = levels,
                                             .setpoints = threshold}});
            }
          }
        }
      }
      SECTION("write settings") {
        GIVEN("zone does not exist") {
          THEN("set new setpoint value throws") {
            const dsModel::TemperatureSetpointArgs& args = {
                .applicationId = dsModel::ApplicationId{"temperature-z11"},
                .temperatureLevel = dsModel::TemperatureLevel::heatingComfort};
            CHECK_THROWS_FIND(dsModel->setTemperatureSetpoint(args, 20), "Zone not found");
          }
        }
        GIVEN("Temperature application device present in zone") {
          device.setActiveGroup(ApplicationType::temperature);
          THEN("set new setpoint value") {
            struct X {
              dsModel::TemperatureLevel level;
              double value;
              std::string levelName;
            };
            const auto& x = GENERATE(values(std::initializer_list<X>{
                {.level = dsModel::TemperatureLevel::heatingComfort, .value = 22.3, .levelName = "Comfort"},
                {.level = dsModel::TemperatureLevel::heatingEconomy, .value = 23.4, .levelName = "Economy"},
                {.level = dsModel::TemperatureLevel::heatingSetback, .value = 24.5, .levelName = "NotUsed"},
                {.level = dsModel::TemperatureLevel::heatingNight, .value = 25, .levelName = "Night"},
                {.level = dsModel::TemperatureLevel::heatingVacation, .value = 26, .levelName = "Holiday"},
                {.level = dsModel::TemperatureLevel::coolingComfort, .value = 28.2, .levelName = "Cooling"},
                {.level = dsModel::TemperatureLevel::coolingEconomy, .value = 29.3, .levelName = "CoolingEconomy"},
                {.level = dsModel::TemperatureLevel::coolingSetback, .value = 30, .levelName = "CoolingNotUsed"},
                {.level = dsModel::TemperatureLevel::coolingNight, .value = 31.7, .levelName = "CoolingNight"},
                {.level = dsModel::TemperatureLevel::coolingVacation, .value = 32, .levelName = "CoolingHoliday"},
            }));
            REQUIRE_CALL(eventSink, pushEvent(_))
                .LR_SIDE_EFFECT(CHECK(*_1 == Event{"heating-controller.operation-mode",
                                                   Properties{{"actions", "setTemperatureValues"},
                                                              {"zoneID", "1"},
                                                              {x.levelName, ds::str(x.value)}}}));
            const dsModel::TemperatureSetpointArgs& args = {.applicationId = dsModel::ApplicationId{"temperature-z1"},
                                                            .temperatureLevel = x.level};
            dsModel->setTemperatureSetpoint(args, x.value);
          }
          THEN("set new passive cooling threshold value") {
            REQUIRE_CALL(eventSink, pushEvent(_))
                .LR_SIDE_EFFECT(CHECK(
                    *_1 == Event{"heating-controller.passive-cooling",
                                 Properties{{"actions", "setMinimalTemperature"}, {"zoneID", "1"}, {"value", "25.3"}}}));
            const dsModel::TemperatureSetpointArgs& args = {
                .applicationId = dsModel::ApplicationId{"temperature-z1"},
                .temperatureLevel = dsModel::TemperatureLevel::passiveCoolingThreshold};
            dsModel->setTemperatureSetpoint(args, 25.3);
          }
        }
      }
    }

    SECTION("apartment applications") {
      SECTION("apartment ventilation") {
        THEN("no apartment ventilation application in apartment") {
          REQUIRE_CALL(proxy, ventilationUnitLevels()).RETURN(dsModel::Ventilation{});
          REQUIRE_CALL(proxy, ventilationUnitStatus()).RETURN(dsModel::VentilationStatus{});
          CHECK(dsModel->apartment({}).attributes.applications2.ventilationUnit == dsModel::Ventilation{});
          CHECK(dsModel->apartmentStatus({}).attributes.applications2.ventilationUnit == dsModel::VentilationStatus{});
        }
        THEN("apartment ventilation application present in apartment") {
          const auto levels =
              dsModel::Ventilation{.availableLevels = std::vector<dsModel::VentilationLevel>{
                                           dsModel::VentilationLevel::off, dsModel::VentilationLevel::level1,
                                           dsModel::VentilationLevel::boost}};
          const auto status = dsModel::VentilationStatus{
              .level = dsModel::VentilationLevel::level1, .malfunction = false, .serviceRequired = true};
          REQUIRE_CALL(proxy, ventilationUnitLevels()).RETURN(levels);
          REQUIRE_CALL(proxy, ventilationUnitStatus()).RETURN(status);
          CHECK(dsModel->apartment({}).attributes.applications2.ventilationUnit == levels);
          CHECK(dsModel->apartmentStatus({}).attributes.applications2.ventilationUnit == status);
        }
      }
    }
  }

  GIVEN("device in zone") {
    auto& zone = *apartment.allocateZone(1);
    const auto zoneId = zone.id();

    const auto deviceId = dsModel::DSDeviceId{"1111111111111111111111111111111100"s};
    auto& device = *apartment.allocateDevice(Dsuid{deviceId});
    device.setButtonInputMode(ButtonInputMode::ONE_WAY);
    device.setButtonInputCount(1);
    zone.addDevice({device.sharedFromThis(), &apartment});

    /// hidden TNY device
    auto hiddenDeviceId = dsModel::DSDeviceId{"3333333333333333333333333333333300"};
    auto& hiddenDevice = *apartment.allocateDevice(Dsuid{hiddenDeviceId});
    hiddenDevice.setPartiallyFromSpec({.ShortAddress = {}, .FunctionID = {}, .ProductID = (10 << 10)}); // TNY
    hiddenDevice.setVisibility(false);
    zone.addDevice({hiddenDevice.sharedFromThis(), &apartment});

    THEN("check device()") {
      CHECK(dsModel->device(deviceId) == dsModel::DSDevice{.id = deviceId,
                                                           .type = "dsDevice",
                                                           .lastChanged = {},
                                                           .attributes = {.name = {},
                                                                          .model = {},
                                                                          .gtin = {},
                                                                          .dsid = {},
                                                                          .displayId = "11111111…",
                                                                          .present = false,
                                                                          .firstSeen = {},
                                                                          .lastDiscovered = {},
                                                                          .inactiveSince = {},
                                                                          .shadeAttributes = {},
                                                                          .submodules = {device.submoduleId()},
                                                                          .zone = zoneId}});
    }

    THEN("special character") {
      device.setName("device&amp;Name");
      CHECK(dsModel->device(deviceId) == dsModel::DSDevice{.id = deviceId,
                                                           .type = "dsDevice",
                                                           .lastChanged = {},
                                                           .attributes = {.name = "device&Name",
                                                                          .model = {},
                                                                          .gtin = {},
                                                                          .dsid = {},
                                                                          .displayId = "11111111…",
                                                                          .present = false,
                                                                          .firstSeen = {},
                                                                          .lastDiscovered = {},
                                                                          .inactiveSince = {},
                                                                          .shadeAttributes = {},
                                                                          .submodules = {device.submoduleId()},
                                                                          .zone = zoneId}});
    }

    THEN("check devices()") {
      CHECK(dsModel->devices() == std::vector<dsModel::DSDevice>{dsModel->device(device.id())});
    }

    THEN("apartment links devices") {
      const auto dsApartment = dsModel->apartment({.include = {"dsDevices"}});
      CHECK(dsApartment.attributes.dsDevices == std::vector<dsModel::DSDeviceId>{deviceId});
      CHECK(dsApartment.included.dsDevices == dsModel->devices());
    }

    THEN("check zone()") {
      CHECK(dsModel->zone(zoneId) == (dsModel::Zone{.id = zoneId,
                                                    .type = "zone",
                                                    .lastChanged = {},
                                                    .attributes = {.name = {},
                                                                   .floor = dsModel::FloorId{"0"},
                                                                   .orderId = 0,
                                                                   .submodules = {device.submoduleId()},
                                                                   .applications = {},
                                                                   .applications2 = {},
                                                                   .applicationTypes = {}}}));
    }

    THEN("special character") {
      zone.setName("zone&amp;Name");
      CHECK(dsModel->zone(zoneId) == (dsModel::Zone{.id = zoneId,
                                                    .type = "zone",
                                                    .lastChanged = {},
                                                    .attributes = {.name = "zone&Name",
                                                                   .floor = dsModel::FloorId{"0"},
                                                                   .orderId = 0,
                                                                   .submodules = {device.submoduleId()},
                                                                   .applications = {},
                                                                   .applications2 = {},
                                                                   .applicationTypes = {}}}));
    }

    THEN("check zones()") {
      CHECK(dsModel->zones() ==
            std::vector<dsModel::Zone>{dsModel->zone(zoneId), dsModel->zone(dsModel::ZoneId{"65534"})});
    }

    THEN("apartment links zones") {
      const auto dsApartment = dsModel->apartment({.include = {"zones"}});
      CHECK(dsApartment.attributes.zones == std::vector<dsModel::ZoneId>{zoneId, dsModel::ZoneId{"65534"}});
      CHECK(dsApartment.included.zones == dsModel->zones());
    }

    WHEN("set device parameters") {
      device.setName("test device");
      device.setActiveGroup(ApplicationType::lights);

      THEN("check submodule()") {
        CHECK(dsModel->submodule(device.submoduleId()) ==
              (dsModel::Submodule{.id = device.submoduleId(),
                                  .type = "submodule",
                                  .lastChanged = {},
                                  .attributes = {.name = "test device",
                                                 .technicalName = {},
                                                 .dsDevice = {device.id()},
                                                 .functionBlocks = {device.functionBlockId()},
                                                 .zone = zoneId,
                                                 .application = dsModel::ZoneApplicationType::lights}}));
      }

      THEN("special character") {
        device.setName("device&amp;Name");
        CHECK(dsModel->submodule(device.submoduleId()) ==
              (dsModel::Submodule{.id = device.submoduleId(),
                                  .type = "submodule",
                                  .lastChanged = {},
                                  .attributes = {.name = "device&Name",
                                                 .technicalName = {},
                                                 .dsDevice = {device.id()},
                                                 .functionBlocks = {device.functionBlockId()},
                                                 .zone = zoneId,
                                                 .application = dsModel::ZoneApplicationType::lights}}));
      }

      THEN("check submodules()") {
        CHECK(dsModel->submodules() == std::vector<dsModel::Submodule>{dsModel->submodule(device.submoduleId())});
      }

      THEN("apartment includes submodules") {
        const auto dsApartment = dsModel->apartment({.include = {"submodules"}});
        CHECK(dsApartment.included.submodules == dsModel->submodules());
      }

      THEN("check functionBlock()") {
        REQUIRE_CALL(proxy, getDeviceThresholds(Dsuid{deviceId})).RETURN(std::vector<double>{});
        auto button = getButtons(device);
        CHECK(
            dsModel->functionBlock(device.functionBlockId()) ==
            (dsModel::FunctionBlock{.id = device.functionBlockId(),
                                    .type = "functionBlock",
                                    .lastChanged = {},
                                    .attributes = {
                                        .name = "test device",
                                        .technicalName = "",
                                        .active = 1,
                                        .outputs = {},
                                        .buttonInputs = button,
                                        .sensorInputs = {},
                                        .submodule = device.submoduleId(),
                                        .deviceAdapter = dsModel::DeviceAdapterId{"1111111111111111111111111111111100"},
                                    }}));
      }

      THEN("special character") {
        device.setName("device&amp;Name");
        REQUIRE_CALL(proxy, getDeviceThresholds(Dsuid{deviceId})).RETURN(std::vector<double>{});
        auto button = getButtons(device);
        CHECK(
            dsModel->functionBlock(device.functionBlockId()) ==
            (dsModel::FunctionBlock{.id = device.functionBlockId(),
                                    .type = "functionBlock",
                                    .lastChanged = {},
                                    .attributes = {
                                        .name = "device&Name",
                                        .technicalName = "",
                                        .active = 1,
                                        .outputs = {},
                                        .buttonInputs = button,
                                        .sensorInputs = {},
                                        .submodule = device.submoduleId(),
                                        .deviceAdapter = dsModel::DeviceAdapterId{"1111111111111111111111111111111100"},
                                    }}));
      }

      THEN("check functionBlocks") {
        ALLOW_CALL(proxy, getDeviceThresholds(_)).RETURN(std::vector<double>{});
        CHECK(dsModel->functionBlocks() ==
              std::vector<dsModel::FunctionBlock>{dsModel->functionBlock(device.functionBlockId())});
      }

      THEN("apartment links functionBlocks") {
        ALLOW_CALL(proxy, getDeviceThresholds(_)).RETURN(std::vector<double>{});
        const auto dsApartment = dsModel->apartment({.include = {"functionBlocks"}});
        CHECK(dsApartment.included.functionBlocks == dsModel->functionBlocks());
      }

      THEN("legacy awnings") {
        device.setActiveGroup(ApplicationType::blinds);
        DeviceSpec spec;
        spec.FunctionID = 0x10;
        device.setPartiallyFromSpec(spec);
        device.setActiveGroup(2);
        CHECK(dsModel->submodule(device.submoduleId()) ==
              (dsModel::Submodule{.id = device.submoduleId(),
                                  .type = "submodule",
                                  .lastChanged = {},
                                  .attributes = {.name = "test device",
                                                 .technicalName = {},
                                                 .dsDevice = {device.id()},
                                                 .functionBlocks = {device.functionBlockId()},
                                                 .zone = zoneId,
                                                 .application = dsModel::ZoneApplicationType::shades}}));
        spec.FunctionID = 0;
        device.setPartiallyFromSpec(spec);
        device.setHardwareType(HardwareType::GR_KL210);
        CHECK(dsModel->submodule(device.submoduleId()) ==
              (dsModel::Submodule{.id = device.submoduleId(),
                                  .type = "submodule",
                                  .lastChanged = {},
                                  .attributes = {.name = "test device",
                                                 .technicalName = {},
                                                 .dsDevice = {device.id()},
                                                 .functionBlocks = {device.functionBlockId()},
                                                 .zone = zoneId,
                                                 .application = dsModel::ZoneApplicationType::awnings}}));
      }
    }

    WHEN("device buttonID set") {
      device.setButtonID(1);
      device.setButtonUsage(ButtonUsage::used);
      THEN("check buttons") {
        CHECK(getButtons(device)[0] == (dsModel::Button{.id = dsModel::ButtonId{"button0"},
                                                        .attributes = {
                                                            .name = "",
                                                            .technicalName = "Pushbutton#0",
                                                            .type = dsModel::ButtonType::area1,
                                                            .mode = dsModel::ButtonMode::button1way,
                                                            .channel = {},
                                                        }}));
      }
    }

    WHEN("device buttonID set but unused") {
      device.setButtonID(1);
      device.setButtonUsage(ButtonUsage::auto_unused);
      THEN("check buttons") {
        CHECK(getButtons(device)[0] == (dsModel::Button{.id = dsModel::ButtonId{"button0"},
                                                        .attributes = {
                                                            .name = "",
                                                            .technicalName = "Pushbutton#0",
                                                            .type = dsModel::ButtonType::area1,
                                                            .mode = dsModel::ButtonMode::disabled,
                                                            .channel = {},
                                                        }}));
      }
    }

    WHEN("device activeGroup and buttonID set") {
      device.setButtonID(14);
      device.setButtonUsage(ButtonUsage::used);

      WHEN("activeGroup in range 16 .. 39") {
        device.setActiveGroup(20);

        THEN("check buttons") {
          CHECK(getButtons(device)[0] == (dsModel::Button{.id = dsModel::ButtonId{"button0"},
                                                          .attributes = {
                                                              .name = "",
                                                              .technicalName = "Pushbutton#0",
                                                              .type = dsModel::ButtonType::group,
                                                              .mode = dsModel::ButtonMode::button1way,
                                                              .channel = {},
                                                          }}));
        }
      }

      WHEN("activeGroup in range 64 .. 239") {
        device.setActiveGroup(100);

        THEN("check buttons") {
          CHECK(getButtons(device)[0] == (dsModel::Button{.id = dsModel::ButtonId{"button0"},
                                                          .attributes = {
                                                              .name = "",
                                                              .technicalName = "Pushbutton#0",
                                                              .type = dsModel::ButtonType::application,
                                                              .mode = dsModel::ButtonMode::button1way,
                                                              .channel = {},
                                                          }}));
        }
      }
    }

    WHEN("device.activeGroup=light") {
      device.setActiveGroup(ApplicationType::lights);

      THEN("check dsModel") {
        CHECK(dsModel->zone(zoneId) ==
              (dsModel::Zone{.id = zoneId,
                             .type = "zone",
                             .lastChanged = {},
                             .attributes = {.name = {},
                                            .floor = dsModel::FloorId{"0"},
                                            .orderId = 0,
                                            .submodules = {device.submoduleId()},
                                            .applications = {dsModel::ZoneApplicationType::lights},
                                            .applications2 = {},
                                            .applicationTypes = {dsModel::ZoneApplicationType::lights}}}));
      }
    }

    WHEN("Sonos device in zone") {
      zone.getGroup(ColorIDCyan)->addConnectedDevice();

      THEN("check dsModel") {
        CHECK(dsModel->zone(zoneId) ==
              (dsModel::Zone{.id = zoneId,
                             .type = "zone",
                             .lastChanged = {},
                             .attributes = {.name = {},
                                            .floor = dsModel::FloorId{"0"},
                                            .orderId = 0,
                                            .submodules = {device.submoduleId()},
                                            .applications = {dsModel::ZoneApplicationType::audio},
                                            .applications2 = {},
                                            .applicationTypes = {dsModel::ZoneApplicationType::audio}}}));
      }
    }

    WHEN("ventilation in zone") {
      auto& ventilationDevice = *apartment.allocateDevice(Dsuid{"3111111111111111111111111111111100"});
      ventilationDevice.setActiveGroup(ApplicationType::ventilation);
      zone.addDevice({ventilationDevice.sharedFromThis(), &apartment});

      std::vector<dsModel::VentilationLevel> availableLevels{
          dsModel::VentilationLevel::off, dsModel::VentilationLevel::level1, dsModel::VentilationLevel::level2,
          dsModel::VentilationLevel::level3, dsModel::VentilationLevel::boost};

      THEN("check dsModel") {
        REQUIRE_CALL(proxy, ventilationLevels(zoneId)).RETURN(dsModel::Ventilation{availableLevels});

        CHECK(dsModel->zone(zoneId) ==
              (dsModel::Zone{.id = zoneId,
                             .type = "zone",
                             .lastChanged = {},
                             .attributes = {.name = {},
                                            .floor = dsModel::FloorId{"0"},
                                            .orderId = 0,
                                            .submodules = {device.submoduleId(), ventilationDevice.submoduleId()},
                                            .applications = {dsModel::ZoneApplicationType::ventilation},
                                            .applications2 = {.ventilation = dsModel::Ventilation{.availableLevels =
                                                                                                      availableLevels}},
                                            .applicationTypes = {dsModel::ZoneApplicationType::ventilation}}}));
      }
    }


    WHEN("recirculation in zone") {
      auto& recirculationDevice = *apartment.allocateDevice(Dsuid{"4111111111111111111111111111111100"});
      recirculationDevice.setActiveGroup(ApplicationType::recirculation);
      zone.addDevice({recirculationDevice.sharedFromThis(), &apartment});

      std::vector<dsModel::VentilationLevel> availableLevels{
          dsModel::VentilationLevel::off, dsModel::VentilationLevel::level1, dsModel::VentilationLevel::level2,
          dsModel::VentilationLevel::boost};

      THEN("check dsModel") {
        REQUIRE_CALL(proxy, recirculationLevels(zoneId)).RETURN(dsModel::Ventilation{availableLevels});
        CHECK(dsModel->zone(zoneId) ==
              (dsModel::Zone{
                  .id = zoneId,
                  .type = "zone",
                  .lastChanged = {},
                  .attributes = {
                      .name = {},
                      .floor = dsModel::FloorId{"0"},
                      .orderId = 0,
                      .submodules = {device.submoduleId(), recirculationDevice.submoduleId()},
                      .applications = {dsModel::ZoneApplicationType::recirculation},
                      .applications2 = {.recirculation = dsModel::Ventilation{.availableLevels = availableLevels}},
                      .applicationTypes = {dsModel::ZoneApplicationType::recirculation}}}));
      }
    }

    WHEN("temperature in zone") {
      auto& temperatureDevice = *apartment.allocateDevice(Dsuid{"5111111111111111111111111111111100"});
      temperatureDevice.setActiveGroup(ApplicationType::temperature);
      zone.addDevice({temperatureDevice.sharedFromThis(), &apartment});
      struct X {
        HeatingControlMode mode{};
        bool exists{};
      };
      const auto& x = GENERATE(values(std::initializer_list<X>{
          // PID mode required for temperature application to exist
          {.mode = HeatingControlMode::OFF, .exists = false},
          {.mode = HeatingControlMode::PID, .exists = true},
          {.mode = HeatingControlMode::ZONE_FOLLOWER, .exists = false},
          {.mode = HeatingControlMode::FIXED, .exists = false},
          {.mode = HeatingControlMode::MANUAL, .exists = false},
      }));
      INFO(x.mode);
      zone.setTemperatureConfig(TemperatureConfig{.m_mode = x.mode});
      zone.setHeatingSetpoints(
          TemperatureConfig{.targetTemperatures = {10, 11, 12, 13, 14, 15, 0, 0, 0, 0, 0, 0, 0, 0, 0}});
      auto configNode = propertySystem.createProperty("/scripts/system-addon-heating-controller/controllerConfig/1");
      configNode->setChildValue("heatingPossible", true);


      THEN("check dsModel") {
        if (x.exists) {
          CHECK(
              dsModel->zone(zoneId) ==
              (dsModel::Zone{
                  .id = zoneId,
                  .type = "zone",
                  .lastChanged = {},
                  .attributes = {
                      .name = {},
                      .floor = dsModel::FloorId{"0"},
                      .orderId = 0,
                      .submodules = {device.submoduleId(), temperatureDevice.submoduleId()},
                      .applications = {dsModel::ZoneApplicationType::temperature},
                      .applications2 = {.temperature =
                                            dsModel::Temperature{
                                                .availableLevels = {dsModel::TemperatureLevel::heatingComfort,
                                                                    dsModel::TemperatureLevel::heatingEconomy,
                                                                    dsModel::TemperatureLevel::heatingSetback,
                                                                    dsModel::TemperatureLevel::heatingNight,
                                                                    dsModel::TemperatureLevel::heatingVacation},
                                                .setpoints =
                                                    {{.id = dsModel::TemperatureLevel::heatingComfort, .value = 11},
                                                     {.id = dsModel::TemperatureLevel::heatingEconomy, .value = 12},
                                                     {.id = dsModel::TemperatureLevel::heatingSetback, .value = 13},
                                                     {.id = dsModel::TemperatureLevel::heatingNight, .value = 14},
                                                     {.id = dsModel::TemperatureLevel::heatingVacation, .value = 15}},
                                            }},
                      .applicationTypes = {dsModel::ZoneApplicationType::temperature}}}));
        } else {
          CHECK(dsModel->zone(zoneId) ==
                (dsModel::Zone{.id = zoneId,
                               .type = "zone",
                               .lastChanged = {},
                               .attributes = {.name = {},
                                              .floor = dsModel::FloorId{"0"},
                                              .orderId = 0,
                                              .submodules = {device.submoduleId(), temperatureDevice.submoduleId()},
                                              .applications = {},
                                              .applications2 = {},
                                              .applicationTypes = {}}}));
        }
      }
    }

    SECTION("device states") {
      REQUIRE_CALL(proxy, dsDeviceStatus(deviceId)).RETURN(dsModel::DSDeviceStatus{});
      std::vector<dsModel::State> states{};
      std::vector<dsModel::StateStatus> stateStatuses{};

      dsModel::State state{
          .id = dsModel::StateId{"stateId"}, .options = {}, .name = {.domain = {}, .message = "stateName"}};
      dsModel::StateOption options{.id = dsModel::StateOptionId{"optionId"},
                                   .name = {.domain = {}, .message = "optionName"}};
      state.options.push_back(options);
      dsModel::StateStatus stateStatus{.id = dsModel::StateId{"stateStatusId"},
                                       .value = dsModel::StateOptionId{"stateStatusValue"}};

      states.push_back(state);
      stateStatuses.push_back(stateStatus);
      device.setStates(std::move(states));
      device.setStateStatuses(std::move(stateStatuses));

      CHECK(dsModel->device(deviceId) == dsModel::DSDevice{.id = deviceId,
                                                           .type = "dsDevice",
                                                           .lastChanged = {},
                                                           .attributes = {.name = {},
                                                                          .model = {},
                                                                          .gtin = {},
                                                                          .dsid = {},
                                                                          .displayId = "11111111…",
                                                                          .present = false,
                                                                          .firstSeen = {},
                                                                          .lastDiscovered = {},
                                                                          .inactiveSince = {},
                                                                          .shadeAttributes = {},
                                                                          .submodules = {device.submoduleId()},
                                                                          .zone = zoneId,
                                                                          .scenarios = {},
                                                                          .states = device.states()}});

      CHECK(dsModel->deviceStatus(deviceId) ==
            dsModel::DSDeviceStatus{.id = {},
                                    .attributes = {.functionBlocks = {}, .submodules = {}, .states = device.stateStatuses()}});
    }

    SECTION("outputs smoke test") {}

    SECTION("output ids sorted, empty&deduplicates removed") {
      device.setOutputMode(16);
      device.setOutputChannels({dss::ChannelType::brightness, dss::ChannelType::brightness,
                                dss::ChannelType::brightness, dss::ChannelType::brightness});
      device.setOutputChannelInfo(
          0, dss::Channel{.index = 0, .id = "id1", .technicalName = {}, .type = dss::ChannelType::brightness});
      device.setOutputChannelInfo(
          1, dss::Channel{.index = 1, .id = "", .technicalName = {}, .type = dss::ChannelType::brightness});
      device.setOutputChannelInfo(
          2, dss::Channel{.index = 2, .id = "id0", .technicalName = {}, .type = dss::ChannelType::brightness});
      device.setOutputChannelInfo(
          3, dss::Channel{.index = 3, .id = "id1", .technicalName = {}, .type = dss::ChannelType::brightness});
      CHECK(dss::getOutputs(device, {}) == std::vector<dsModel::Output>{
                                               {.id = dsModel::OutputId{"id0"},
                                                .attributes = {.name = {},
                                                               .technicalName = {},
                                                               .type = dsModel::OutputType::lightBrightness,
                                                               .function = {},
                                                               .min = 0,
                                                               .max = 100,
                                                               .resolution = 0.1,
                                                               .mode = dsModel::OutputMode::switched,
                                                               .levels = 1}},
                                               {.id = dsModel::OutputId{"id1"},
                                                .attributes = {.name = {},
                                                               .technicalName = {},
                                                               .type = dsModel::OutputType::lightBrightness,
                                                               .function = {},
                                                               .min = 0,
                                                               .max = 100,
                                                               .resolution = 0.1,
                                                               .mode = dsModel::OutputMode::switched,
                                                               .levels = 1}},
                                           });
    }

    SECTION("output types") {
      struct X {
        ChannelType channelType;
        std::vector<dsModel::Output> dsOutputs;
      };
      const auto& x = GENERATE(values(std::initializer_list<X>{
          {ChannelType::brightness,
           {{.id = dsModel::OutputId{"brightness"},
             .attributes = {.name = {},
                            .technicalName = "brightness",
                            .type = dsModel::OutputType::lightBrightness,
                            .function = {},
                            .min = 0,
                            .max = 100,
                            .resolution = 0.1,
                            .mode = dsModel::OutputMode::switched,
                            .levels = 1}}}},
          {ChannelType::hue,
           {{.id = dsModel::OutputId{"hue"},
             .attributes = {.name = {},
                            .technicalName = "hue",
                            .type = dsModel::OutputType::lightHue,
                            .function = {},
                            .min = 0,
                            .max = 360,
                            .resolution = 0.1,
                            .mode = dsModel::OutputMode::switched,
                            .levels = 1}}}},
          {ChannelType::saturation,
           {{.id = dsModel::OutputId{"saturation"},
             .attributes = {.name = {},
                            .technicalName = "saturation",
                            .type = dsModel::OutputType::lightSaturation,
                            .function = {},
                            .min = 0,
                            .max = 100,
                            .resolution = 0.1,
                            .mode = dsModel::OutputMode::switched,
                            .levels = 1}}}},
          {ChannelType::colortemp,
           {{.id = dsModel::OutputId{"colortemp"},
             .attributes = {.name = {},
                            .technicalName = "colortemp",
                            .type = dsModel::OutputType::lightTemperature,
                            .function = {},
                            .min = 100,
                            .max = 1000,
                            .resolution = 1.0,
                            .mode = dsModel::OutputMode::switched,
                            .levels = 1}}}},
          {ChannelType::x,
           {{.id = dsModel::OutputId{"x"},
             .attributes = {.name = {},
                            .technicalName = "x",
                            .type = dsModel::OutputType::lightCieX,
                            .function = {},
                            .min = 0,
                            .max = 1,
                            .resolution = 0.01,
                            .mode = dsModel::OutputMode::switched,
                            .levels = 1}}}},
          {ChannelType::y,
           {{.id = dsModel::OutputId{"y"},
             .attributes = {.name = {},
                            .technicalName = "y",
                            .type = dsModel::OutputType::lightCieY,
                            .function = {},
                            .min = 0,
                            .max = 1,
                            .resolution = 0.01,
                            .mode = dsModel::OutputMode::switched,
                            .levels = 1}}}},
          {ChannelType::shadePositionOutside,
           {{.id = dsModel::OutputId{"shadePositionOutside"},
             .attributes = {.name = {},
                            .technicalName = "shadePositionOutside",
                            .type = dsModel::OutputType::shadePositionOutside,
                            .function = {},
                            .min = 0,
                            .max = 100,
                            .resolution = 0.1,
                            .mode = dsModel::OutputMode::switched,
                            .levels = 1}}}},
          {ChannelType::shadePositionIndoor,
           {{.id = dsModel::OutputId{"shadePositionIndoor"},
             .attributes = {.name = {},
                            .technicalName = "shadePositionIndoor",
                            .type = dsModel::OutputType::shadePositionIndoor,
                            .function = {},
                            .min = 0,
                            .max = 100,
                            .resolution = 0.1,
                            .mode = dsModel::OutputMode::switched,
                            .levels = 1}}}},
          {ChannelType::shadeOpeningAngleOutside,
           {{.id = dsModel::OutputId{"shadeOpeningAngleOutside"},
             .attributes = {.name = {},
                            .technicalName = "shadeOpeningAngleOutside",
                            .type = dsModel::OutputType::shadeOpeningAngleOutside,
                            .function = {},
                            .min = 0,
                            .max = 100,
                            .resolution = 0.1,
                            .mode = dsModel::OutputMode::switched,
                            .levels = 1}}}},
          {ChannelType::shadeOpeningAngleIndoor,
           {{.id = dsModel::OutputId{"shadeOpeningAngleIndoor"},
             .attributes = {.name = {},
                            .technicalName = "shadeOpeningAngleIndoor",
                            .type = dsModel::OutputType::shadeOpeningAngleIndoor,
                            .function = {},
                            .min = 0,
                            .max = 100,
                            .resolution = 0.1,
                            .mode = dsModel::OutputMode::switched,
                            .levels = 1}}}},
          {ChannelType::transparency,
           {{.id = dsModel::OutputId{"transparency"},
             .attributes = {.name = {},
                            .technicalName = "transparency",
                            .type = dsModel::OutputType::shadeTransparency,
                            .function = {},
                            .min = 0,
                            .max = 100,
                            .resolution = 0.1,
                            .mode = dsModel::OutputMode::switched,
                            .levels = 1}}}},
          {ChannelType::airFlowIntensity,
           {{.id = dsModel::OutputId{"airFlowIntensity"},
             .attributes = {.name = {},
                            .technicalName = "airFlowIntensity",
                            .type = dsModel::OutputType::airFlowIntensity,
                            .function = {},
                            .min = 0,
                            .max = 100,
                            .resolution = 0.1,
                            .mode = dsModel::OutputMode::switched,
                            .levels = 1}}}},
          // ChannelType::airFlowDirection, // max incorrect. enum instead of number?
          {ChannelType::airFlapPosition,
           {{.id = dsModel::OutputId{"airFlapPosition"},
             .attributes = {.name = {},
                            .technicalName = "airFlapPosition",
                            .type = dsModel::OutputType::airFlapOpeningAngle,
                            .function = {},
                            .min = 0,
                            .max = 100,
                            .resolution = 0.1,
                            .mode = dsModel::OutputMode::switched,
                            .levels = 1}}}},
          {ChannelType::airLouverPosition,
           {{.id = dsModel::OutputId{"airLouverPosition"},
             .attributes = {.name = {},
                            .technicalName = "airLouverPosition",
                            .type = dsModel::OutputType::ventilationLouverPosition,
                            .function = {},
                            .min = 0,
                            .max = 100,
                            .resolution = 0.1,
                            .mode = dsModel::OutputMode::switched,
                            .levels = 1}}}},
          // ChannelType::heatingPower,// max incorrect. enum instead of number?
          {ChannelType::coolingCapacity,
           {{.id = dsModel::OutputId{"coolingCapacity"},
             .attributes = {.name = {},
                            .technicalName = "coolingCapacity",
                            .type = dsModel::OutputType::coolingCapacity,
                            .function = {},
                            .min = 0,
                            .max = 100,
                            .resolution = 0.1,
                            .mode = dsModel::OutputMode::switched,
                            .levels = 1}}}},
          {ChannelType::audioVolume,
           {{.id = dsModel::OutputId{"audioVolume"},
             .attributes = {.name = {},
                            .technicalName = "audioVolume",
                            .type = dsModel::OutputType::audioVolume,
                            .function = {},
                            .min = 0,
                            .max = 100,
                            .resolution = 0.1,
                            .mode = dsModel::OutputMode::switched,
                            .levels = 1}}}},
          // ChannelType::powerState,// max incorrect. enum instead of number?
          // ChannelType::airLouverAuto,// max incorrect. enum instead of number?
          // ChannelType::airFlowAuto,// max incorrect. enum instead of number?
          {ChannelType::waterTemperature,
           {{.id = dsModel::OutputId{"waterTemperature"},
             .attributes = {.name = {},
                            .technicalName = "waterTemperature",
                            .type = dsModel::OutputType::waterTemperature,
                            .function = {},
                            .min = 0,
                            .max = 150,
                            .resolution = 0.1,
                            .mode = dsModel::OutputMode::switched,
                            .levels = 1}}}},
          {ChannelType::waterFlowRate,
           {{.id = dsModel::OutputId{"waterFlowRate"},
             .attributes = {.name = {},
                            .technicalName = "waterFlowRate",
                            .type = dsModel::OutputType::waterFlowRate,
                            .function = {},
                            .min = 0,
                            .max = 100,
                            .resolution = 0.1,
                            .mode = dsModel::OutputMode::switched,
                            .levels = 1}}}},
          {ChannelType::powerLevel,
           {{.id = dsModel::OutputId{"powerLevel"},
             .attributes = {.name = {},
                            .technicalName = "powerLevel",
                            .type = dsModel::OutputType::powerLevel,
                            .function = {},
                            .min = 0,
                            .max = 100,
                            .resolution = 0.1,
                            .mode = dsModel::OutputMode::switched,
                            .levels = 1}}}},
      }));
      INFO(x.channelType);
      device.setOutputChannels({x.channelType});
      device.setOutputMode(16);
      CHECK(getOutputs(device, {}) == x.dsOutputs);
    }

    SECTION("output modes") {
      device.setOutputChannels({ChannelType::brightness});
      struct X {
        std::vector<int> dssModes;
        dsModel::OutputMode dsMode;
      };
      for (const auto& x : std::initializer_list<X>{
               {.dssModes = {255, 1}, .dsMode = {}},
               {.dssModes = {16, 25, 26, 27, 32, 34, 35, 36, 37, 38, 39, 40, 41, 43, 65, 66, 67, 68},
               .dsMode = dsModel::OutputMode::switched},
               {.dssModes = {33, 42}, .dsMode = dsModel::OutputMode::positional},
               {.dssModes = {17, 18, 19, 20, 21, 22, 23, 24, 31, 48, 49, 50, 51, 64},
               .dsMode = dsModel::OutputMode::gradual},
               {.dssModes = {0}, .dsMode = dsModel::OutputMode::disabled},
           }) {
        for (const auto& dssMode : x.dssModes) {
          INFO(dssMode);
          INFO(x.dsMode);
          device.setOutputMode(dssMode);
          CHECK(getOutputs(device, {}).at(0).attributes.mode == x.dsMode);
        }
      }
    }

    SECTION("setOutputMode propagates to bus") {
      device.setOutputChannels(
          {ChannelType::brightness, ChannelType::hue, ChannelType::colortemp, ChannelType::shadePositionOutside});
      device.setOutputMode(16);
      SECTION("minimal brightness") {
        REQUIRE_CALL(proxy, setOutputValue(_,ChannelType::brightness,_,_));
        REQUIRE_CALL(busInterface.device, setDeviceOutputChannelValue(_, 0, 8, 0, true)).LR_WITH(_1 == device);
        dsModel->setOutputValue({.deviceId = deviceId, .outputId = "brightness"}, 0);
      }
      SECTION("maximal brightness") {
        REQUIRE_CALL(proxy, setOutputValue(_,ChannelType::brightness,_,_));
        REQUIRE_CALL(busInterface.device, setDeviceOutputChannelValue(_, 0, 8, 255, true)).LR_WITH(_1 == device);
        dsModel->setOutputValue({.deviceId = deviceId, .outputId = "brightness"}, 100);
      }
      SECTION("minimal hue") {
        REQUIRE_CALL(proxy, setOutputValue(_,ChannelType::hue,_,_));
        REQUIRE_CALL(busInterface.device, setDeviceOutputChannelValue(_, 1, 8, 0, true)).LR_WITH(_1 == device);
        dsModel->setOutputValue({.deviceId = deviceId, .outputId = "hue"}, 0);
      }
      SECTION("maximal hue") {
        REQUIRE_CALL(proxy, setOutputValue(_,ChannelType::hue,_,_));
        REQUIRE_CALL(busInterface.device, setDeviceOutputChannelValue(_, 1, 8, 255, true)).LR_WITH(_1 == device);
        dsModel->setOutputValue({.deviceId = deviceId, .outputId = "hue"}, 360);
      }
      SECTION("minimal colortemp") {
        REQUIRE_CALL(proxy, setOutputValue(_,ChannelType::colortemp,_,_));
        REQUIRE_CALL(busInterface.device, setDeviceOutputChannelValue(_, 2, 8, 0, true)).LR_WITH(_1 == device);
        dsModel->setOutputValue({.deviceId = deviceId, .outputId = "colortemp"}, 100);
      }
      SECTION("maximal colortemp") {
        REQUIRE_CALL(proxy, setOutputValue(_,ChannelType::colortemp,_,_));
        REQUIRE_CALL(busInterface.device, setDeviceOutputChannelValue(_, 2, 8, 255, true)).LR_WITH(_1 == device);
        dsModel->setOutputValue({.deviceId = deviceId, .outputId = "colortemp"}, 1000);
      }
      SECTION("minimal shadePositionOutside") {
        REQUIRE_CALL(proxy, setOutputValue(_,ChannelType::shadePositionOutside,_,_));
        REQUIRE_CALL(busInterface.device, setDeviceOutputChannelValue(_, 3, 16, 0, true)).LR_WITH(_1 == device);
        dsModel->setOutputValue({.deviceId = deviceId, .outputId = "shadePositionOutside"}, 0);
      }
      SECTION("maximal shadePositionOutside") {
        REQUIRE_CALL(proxy, setOutputValue(_,ChannelType::shadePositionOutside,_,_));
        REQUIRE_CALL(busInterface.device, setDeviceOutputChannelValue(_, 3, 16, 65535, true)).LR_WITH(_1 == device);
        dsModel->setOutputValue({.deviceId = deviceId, .outputId = "shadePositionOutside"}, 100);
      }
    }

    SECTION("setOutputValue on shade device is blocked with active operation lock") {
      auto shadeCluster = apartment.getCluster(16);
      shadeCluster->setName("shadeCluster");
      shadeCluster->setApplicationType(ApplicationType::blinds);
      const auto shadeClusterId = dsModel::ClusterId{"16"};

      DeviceSpec spec;
      spec.activeGroup = 2;
      device.setPartiallyFromSpec(spec);

      device.setOutputChannels({ChannelType::shadePositionOutside});
      device.setOutputMode(16);
      device.addToGroup(GroupIDAppUserMin);

      GIVEN("cluster has operation lock active") {
        shadeCluster->setOperationLock(true, CallOrigin::unknown);
        THEN("setOutputValue is blocked") {
          CHECK_THROWS_FIND(dsModel->setOutputValue({.deviceId = deviceId, .outputId = "shadePositionOutside"}, 100),
                            "Shade device is in cluster with active operation lock.");
        }
        THEN("check status") {
          CHECK(dsModel->clusterStatus(shadeClusterId) ==
                dsModel::ClusterStatus{.id = shadeClusterId,
                                       .type = "clusterStatus",
                                       .lastChanged = {},
                                       .attributes = {.weather = {}, .operationsLocked = true}});

          dsModel::DSDeviceStatus deviceStatus {
            .id = {}, .attributes = {
              .functionBlocks = {},
              .submodules = {dsModel::SubmoduleStatus{.id = dsModel::SubmoduleId{deviceId}, .operationsLocked = true}},
              .states = {}
            }
          };
          REQUIRE_CALL(proxy, dsDeviceStatus(deviceId)).RETURN(deviceStatus);
          CHECK(dsModel->deviceStatus(deviceId) == deviceStatus);
        }
      }
    }

    SECTION("setOutputValue on joker configured as light device") {
      SECTION("joker as light") {
        device.setHardwareType(HardwareType::SW_KL200);
        device.setActiveGroup(ColorIDYellow);
        device.setOutputChannels({});

        SECTION("minimal brightness") {
          REQUIRE_CALL(busInterface.device, setValue(_, 0)).LR_WITH(_1 == device);
          dsModel->setOutputValue({.deviceId = deviceId, .outputId = "brightness"}, 0);
        }
        SECTION("maximal brightness") {
          REQUIRE_CALL(busInterface.device, setValue(_, 255)).LR_WITH(_1 == device);
          dsModel->setOutputValue({.deviceId = deviceId, .outputId = "brightness"}, 100);
        }
      }
      SECTION("joker as shade") {
        device.setHardwareType(HardwareType::SW_KL200);
        device.setActiveGroup(ColorIDGray);
        device.setOutputChannels({});

        SECTION("try to set brightness channel type output") {
          CHECK_THROWS_FIND(dsModel->setOutputValue({.deviceId = deviceId, .outputId = "brightness"}, 0),
                            "Could not find item");
        }
        SECTION("try to set shadePositionOutside channel type output") {
          CHECK_THROWS_FIND(dsModel->setOutputValue({.deviceId = deviceId, .outputId = "shadePositionOutside"}, 0),
                            "Could not find item");
        }
      }
    }

    SECTION("powerline output devices have output channels") {
      enum Ds485 {
        setOutputValue,
        setDeviceConfig64_3_2,
        setDeviceConfig64_4,
      };
      struct Output {
        dsModel::Output ds;
        Ds485 ds485;
      };
      static const Output lightBrightness{.ds = {.id = dsModel::OutputId{"brightness"},
                                                 .attributes = {.name = {},
                                                                .technicalName = "brightness",
                                                                .type = dsModel::OutputType::lightBrightness,
                                                                .function = {},
                                                                .min = 0,
                                                                .max = 100,
                                                                .resolution = 0.1,
                                                                .mode = dsModel::OutputMode::switched,
                                                                .levels = 1}},
                                          .ds485 = Ds485::setOutputValue};
      static const Output shadePositionOutside{.ds = {.id = dsModel::OutputId{"shadePositionOutside"},
                                                      .attributes = {.name = {},
                                                                     .technicalName = "shadePositionOutside",
                                                                     .type = dsModel::OutputType::shadePositionOutside,
                                                                     .function = {},
                                                                     .min = 0,
                                                                     .max = 100,
                                                                     .resolution = 0.1,
                                                                     .mode = dsModel::OutputMode::switched,
                                                                     .levels = 1}},
                                               .ds485 = Ds485::setDeviceConfig64_3_2};
      static const Output shadeOpeningAngleOutside{
          .ds = {.id = dsModel::OutputId{"shadeOpeningAngleOutside"},
                 .attributes = {.name = {},
                                .technicalName = "shadeOpeningAngleOutside",
                                .type = dsModel::OutputType::shadeOpeningAngleOutside,
                                .function = {},
                                .min = 0,
                                .max = 100,
                                .resolution = 0.1,
                                .mode = dsModel::OutputMode::switched,
                                .levels = 1}},
          .ds485 = Ds485::setDeviceConfig64_4};

      struct X {
        HardwareType hardwareType;
        ApplicationType application;
        std::vector<Output> outputs;
      };
      const auto lights = ApplicationType::lights;
      const auto blinds = ApplicationType::blinds;
      const auto& x = GENERATE(values(std::initializer_list<X>{
          // light devices
          {.hardwareType = HardwareType::GE_KM200, .application = lights, .outputs = {lightBrightness}},
          {.hardwareType = HardwareType::GE_TKM210, .application = lights, .outputs = {lightBrightness}},
          {.hardwareType = HardwareType::GE_KM300, .application = lights, .outputs = {lightBrightness}},
          {.hardwareType = HardwareType::GE_TKM300, .application = lights, .outputs = {lightBrightness}},
          {.hardwareType = HardwareType::GE_SDM200, .application = lights, .outputs = {lightBrightness}},
          {.hardwareType = HardwareType::GE_SDM200JS, .application = lights, .outputs = {lightBrightness}},
          {.hardwareType = HardwareType::GE_SDM200CS, .application = lights, .outputs = {lightBrightness}},
          {.hardwareType = HardwareType::GE_SDS200CW, .application = lights, .outputs = {lightBrightness}},
          {.hardwareType = HardwareType::GE_SDS200CS, .application = lights, .outputs = {lightBrightness}},
          {.hardwareType = HardwareType::GE_SDS220CT, .application = lights, .outputs = {lightBrightness}},
          {.hardwareType = HardwareType::GE_KL200, .application = lights, .outputs = {lightBrightness}},
          {.hardwareType = HardwareType::GE_UMV200, .application = lights, .outputs = {lightBrightness}},
          {.hardwareType = HardwareType::GR_KL200, .application = blinds, .outputs = {shadePositionOutside}},
          {.hardwareType = HardwareType::GR_KL210, .application = blinds, .outputs = {shadePositionOutside}},
          // blinds devices
          {.hardwareType = HardwareType::GR_KL220,
           .application = blinds,
           .outputs = {shadePositionOutside, shadeOpeningAngleOutside}},
          {.hardwareType = HardwareType::GR_HKL230,
           .application = blinds,
           .outputs = {shadePositionOutside, shadeOpeningAngleOutside}},
          // joker devices configured as light. Other applications are tested elsewhere.
          {.hardwareType = HardwareType::SW_KL200, .application = lights, .outputs = {lightBrightness}},
          {.hardwareType = HardwareType::SW_ZWS200, .application = lights, .outputs = {lightBrightness}},
          {.hardwareType = HardwareType::SW_ZWS200F, .application = lights, .outputs = {lightBrightness}},
          {.hardwareType = HardwareType::SW_ZWS200EF, .application = lights, .outputs = {lightBrightness}},
          {.hardwareType = HardwareType::SW_ZWS205, .application = lights, .outputs = {lightBrightness}},
          {.hardwareType = HardwareType::SW_SSL200JS, .application = lights, .outputs = {lightBrightness}},
          {.hardwareType = HardwareType::SW_SSL200FS, .application = lights, .outputs = {lightBrightness}},
          // TODO(someday): {.hardwareType = HardwareType::SW_UMR200, .application = lights, .outputs =
          // {lightBrightness}},
      }));
      INFO(x.hardwareType);
      device.setOutputMode(16);
      device.setHardwareType(x.hardwareType);
      device.setActiveGroup(x.application);
      device.setOutputChannels({});
      THEN("Outputs are matching") {
        auto dsOutputs = std::vector<dsModel::Output>{};
        for (const auto& output : x.outputs) {
          dsOutputs.push_back(output.ds);
        }
        boost::sort(dsOutputs, [](const auto& a, const auto& b) { return a.id < b.id; });
        CHECK(getOutputs(device, {}) == dsOutputs);
      }
      THEN("setOutputValue propagates to ds485 bus") {
        ALLOW_CALL(proxy, setOutputValue(_, ChannelType::brightness, _, _));
        ALLOW_CALL(proxy, setOutputValue(_, ChannelType::shadePositionOutside, _, _));
        ALLOW_CALL(proxy, setOutputValue(_, ChannelType::shadeOpeningAngleOutside, _, _));
        for (const auto& output : x.outputs) {
          const auto& dsOutput = output.ds;
          INFO(dsOutput);
          auto value = (*dsOutput.attributes.min + *dsOutput.attributes.max * 2) / 3;
          switch (output.ds485) {
            case Ds485::setOutputValue: {
              REQUIRE_CALL(busInterface.device, setValue(_, 170)).LR_WITH(_1 == device);
              dsModel->setOutputValue({.deviceId = deviceId, .outputId = dsOutput.id}, value);
              break;
            }
            case Ds485::setDeviceConfig64_3_2: {
              REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 64, 3, 170)).LR_WITH(_1 == device);
              REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 64, 2, 170)).LR_WITH(_1 == device);
              dsModel->setOutputValue({.deviceId = deviceId, .outputId = dsOutput.id}, value);
              break;
            }
            case Ds485::setDeviceConfig64_4: {
              REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 64, 4, 170)).LR_WITH(_1 == device);
              dsModel->setOutputValue({.deviceId = deviceId, .outputId = dsOutput.id}, value);
              break;
            }
          }
        }
      }
    }

    SECTION("Joker powerline output devices have matching channel") {
      struct X {
        ApplicationType applicationType;
        std::vector<ChannelType> channelTypes;
      };
      const auto& x = GENERATE(values(std::initializer_list<X>{
          {ApplicationType::lights, {ChannelType::brightness}},
          {ApplicationType::blinds, {ChannelType::powerLevel}},
          {ApplicationType::heating, {ChannelType::heatingPower}},
          {ApplicationType::audio, {ChannelType::powerLevel}},
          {ApplicationType::video, {ChannelType::powerLevel}},
          {ApplicationType::security, {ChannelType::powerLevel}},
          {ApplicationType::access, {ChannelType::powerLevel}},
          {ApplicationType::joker, {ChannelType::powerLevel}},
          {ApplicationType::cooling, {ChannelType::coolingCapacity}},
          {ApplicationType::ventilation, {ChannelType::airFlowIntensity}},
          {ApplicationType::window, {ChannelType::powerLevel}},
          {ApplicationType::recirculation, {ChannelType::airFlowIntensity}},
          {ApplicationType::temperature, {}},
          {ApplicationType::apartmentVentilation, {ChannelType::airFlowIntensity}},
          {ApplicationType::awnings, {ChannelType::powerLevel}},
          {ApplicationType::apartmentRecirculation, {}},
      }));
      INFO(x.applicationType);
      device.setHardwareType(HardwareType::SW_KL200);
      device.setActiveGroup(x.applicationType);
      device.setOutputChannels({});
      THEN("Outputs are matching") {
        const auto outputs = getOutputs(device, {});
        INFO(outputs);
        INFO(x.channelTypes);
        REQUIRE(ds::makeVector(outputs | boost::adaptors::transformed([&](auto&& x) { return x.attributes.type; })) ==
                ds::makeVector(x.channelTypes | boost::adaptors::transformed(
                                                    [&](auto&& x) { return ds::toEnum<dsModel::OutputType>(x); })));
        // remaining attributes tested with output tests

        THEN("setOutputValue propagates to ds485 bus") {
          for (size_t i = 0; i < x.channelTypes.size(); i++) {
            ALLOW_CALL(proxy, setOutputValue(_, x.channelTypes[i], _, _));
            // hardcoded for min:0, max:100
            REQUIRE_CALL(busInterface.device, setValue(_, 168)).LR_WITH(_1 == device);
            dsModel->setOutputValue({.deviceId = deviceId, .outputId = outputs[i].id}, 66);
          }
        }
      }
    }
  }

  GIVEN("userDefinedStates") {
    auto udssNode = propertySystem.createProperty("/usr/addon-states/system-addon-user-defined-states/");

    const auto stringId = "id0"s;
    const auto id = dsModel::UserDefinedStateId{stringId};
    const auto name = "name0"s;
    const auto setName = "set"s;
    const auto resetName = "reset"s;
    REQUIRE_CALL(hooks, apartmentStructurePolled());
    auto udsNode = udssNode->createProperty(stringId);
    udsNode->setChildValue("name", stringId);
    udsNode->setChildValue("displayName", name);
    udsNode->setChildValue("setName", setName);
    udsNode->setChildValue("resetName", resetName);
    udsNode->setChildValue("showOnPhone", true);
    auto& state = *apartment.allocateState(StateType_Script, stringId, "system-addon-user-defined-states");

    SECTION("structure") {
      const auto& [type, typeAPI] =
          GENERATE(values<std::pair<std::string, std::string>>({{"custom-states", "manualState"},
                                                                {"zone-sensor-states", "measurementState"},
                                                                {"consumption-states", "consumptionState"},
                                                                {"apartment-power-states", "consumptionState"},
                                                                {"open-metering-system", "consumptionState"},
                                                                {"power-states", "circuitConsumptionState"},
                                                                {"triggered-states", "triggerState"},
                                                                {"combined-states", "combinedState"},
                                                                {"window-states", "windowState"}}));
      REQUIRE_CALL(hooks, apartmentStructurePolled());
      udsNode->setChildValue("type", type);
      CHECK(dsModel->userDefinedStates().userDefinedStates ==
            std::vector<dsModel::UserDefinedState>{
                {.id = id,
                 .type = typeAPI,
                 .lastChanged = {},
                 .attributes = dsModel::UserDefinedStateAttributes{
                     .name = name, .setName = setName, .resetName = resetName, .visibleForUsers = true}}});
    }

    SECTION("special characters") {
      const auto name1 = "display&gt;Name"s;
      const auto setName1 = "set&lt;Name"s;
      const auto resetName1 = "reset&#039;Name"s;
      const auto type1 = "custom-states"s;
      udsNode->setChildValue("displayName", name1);
      udsNode->setChildValue("setName", setName1);
      udsNode->setChildValue("resetName", resetName1);
      REQUIRE_CALL(hooks, apartmentStructurePolled());
      udsNode->setChildValue("type", type1);
      CHECK(dsModel->userDefinedStates().userDefinedStates ==
            std::vector<dsModel::UserDefinedState>{
                {.id = id,
                 .type = "manualState",
                 .lastChanged = {},
                 .attributes = dsModel::UserDefinedStateAttributes{.name = "display>Name",
                                                                   .setName = "set<Name",
                                                                   .resetName = "reset'Name",
                                                                   .visibleForUsers = true}}});
    }

    SECTION("user defined state is removed") {
      REQUIRE_CALL(hooks, apartmentStructurePolled());
      udssNode->removeChild(udsNode);
      CHECK(dsModel->userDefinedStates() == dsModel::UserDefinedStates{});
    }

    SECTION("state") {
      const auto& [active, activeEnum] = GENERATE(values<std::pair<std::string, dsModel::UserDefinedStateStatusStatus>>(
          {{"active", dsModel::UserDefinedStateStatusStatus::active},
           {"inactive", dsModel::UserDefinedStateStatusStatus::inactive},
           {"undefined", dsModel::UserDefinedStateStatusStatus::undefined}}));
      udsNode->setChildValue("state", active);
      CHECK(dsModel->userDefinedStateStatuses() ==
            std::vector<dsModel::UserDefinedStateStatus>{
                {.id = id, .type = "userDefinedStateStatus", .lastChanged = {}, .attributes = {.status = activeEnum}}});
    }

    SECTION("change state") {
      SECTION("success") {
        udsNode->setChildValue("type", "custom-states"s);
        dsModel->setUserDefinedStateStatus({.userDefinedStateId = id}, "active");
        CHECK(state.getState() == State_Active);
        dsModel->setUserDefinedStateStatus({.userDefinedStateId = id}, "inactive");
        CHECK(state.getState() == State_Inactive);
      }
      SECTION("wrong value") {
        udsNode->setChildValue("type", "custom-states"s);
        CHECK_THROWS_FIND(dsModel->setUserDefinedStateStatus({.userDefinedStateId = id}, "balab"),
                          "Requested state value is illegal");
      }
      SECTION("non-existing") {
        CHECK_THROWS_FIND(dsModel->setUserDefinedStateStatus({.userDefinedStateId = "wrongId"}, "active"),
                          "State is unknown");
      }
      SECTION("no manual state") {
        const auto& type = GENERATE(values<std::string>({"zone-sensor-states", "consumption-states", "power-states",
                                                         "triggered-states", "combined-states", "window-states"}));
        udsNode->setChildValue("type", type);
        CHECK_THROWS_FIND(dsModel->setUserDefinedStateStatus({.userDefinedStateId = id}, "active"),
                          "Only a manualState can be set");
      }
      SECTION("poll apartment status") {
        // State does not raise event in this mocked setup without DSS instance
        // We similate the state event directly
        REQUIRE_CALL(hooks, apartmentStatusPolled());
        dsModel->handleEvent(*createStateChangeEvent(state.shared_from_this(), {}, {}));
      }
    }
  }

  SECTION("clusters") {
    auto cluster16 = apartment.getCluster(16);
    auto cluster17 = apartment.getCluster(17);
    cluster16->setName("cluster16");
    cluster17->setName("cluster17");
    cluster16->setApplicationType(ApplicationType::blinds);
    cluster17->setApplicationType(ApplicationType::lights);
    const auto id16 = dsModel::ClusterId{"16"};
    const auto id17 = dsModel::ClusterId{"17"};

    SECTION("get cluster") {
      const auto cluster = dsModel::Cluster{.id = id16,
                                            .type = "cluster",
                                            .attributes = {.name = "cluster16",
                                                           .orderId = {},
                                                           .applicationType = dsModel::ClusterApplicationType::shades,
                                                           .submodules = {}}};
      CHECK(dsModel->cluster(id16) == cluster);
    }

    SECTION("get awnings cluster -- legacy awning device") {
      auto& deviceAwning = *createDevice(apartment, {.hardwareType = HardwareType::GR_KL210});

      deviceAwning.addToGroup(GroupIDGray);
      deviceAwning.addToGroup(GroupIDAppUserMin);

      // ->isLegacyAwning:true, up/down inverted for blinds/awnings
      // https://redmine.digitalstrom.org/issues/23064
      deviceAwning.setActiveGroup(static_cast<int>(ApplicationType::blinds));

      const auto cluster = dsModel::Cluster{.id = id16,
                                            .type = "cluster",
                                            .attributes = {.name = "cluster16",
                                                           .orderId = {},
                                                           .applicationType = dsModel::ClusterApplicationType::awnings,
                                                           .submodules = {}}};
      CHECK(dsModel->cluster(id16) == cluster);
    }

    SECTION("special character") {
      cluster16->setName("cluster&amp;16");
      const auto cluster = dsModel::Cluster{.id = id16,
                                            .type = "cluster",
                                            .attributes = {.name = "cluster&16",
                                                           .orderId = {},
                                                           .applicationType = dsModel::ClusterApplicationType::shades,
                                                           .submodules = {}}};
      CHECK(dsModel->cluster(id16) == cluster);
    }

    SECTION("get cluster status") {
      THEN("no alarms in status") {
        CHECK(dsModel->clusterStatus(id16) ==
              dsModel::ClusterStatus{
                  .id = id16, .type = "clusterStatus", .lastChanged = {}, .attributes = {.weather = {}, .operationsLocked = false}});
      }
      GIVEN("wind and hail alarm set") {
        auto wind = apartment.allocateState(StateType_Apartment, "wind.group16", "");
        auto hail = apartment.allocateState(StateType_Apartment, "hail.group16", "");
        wind->setState(CallOrigin::unknown, State_Active);
        hail->setState(CallOrigin::unknown, State_Active);
        THEN("alarms in status") {
          CHECK(dsModel->clusterStatus(id16) ==
                dsModel::ClusterStatus{.id = dsModel::ClusterId{"16"},
                                       .type = "clusterStatus",
                                       .lastChanged = {},
                                       .attributes = {.weather = {.wind = true, .hail = true}, .operationsLocked = false}});
        }
      }
    }

    SECTION("all clusters") {
      std::vector<dsModel::Cluster> clusters;
      clusters.push_back(dsModel::Cluster{.id = id16,
                                          .type = "cluster",
                                          .attributes = {.name = "cluster16",
                                                         .orderId = {},
                                                         .applicationType = dsModel::ClusterApplicationType::shades,
                                                         .submodules = {}}});
      clusters.push_back(dsModel::Cluster{.id = id17,
                                          .type = "cluster",
                                          .attributes = {.name = "cluster17",
                                                         .orderId = {},
                                                         .applicationType = dsModel::ClusterApplicationType::lights,
                                                         .submodules = {}}});
      CHECK(dsModel->clusters() == clusters);
    }

    SECTION("all cluster statuses") {
      std::vector<dsModel::ClusterStatus> clusterStatuses;
      clusterStatuses.push_back(dsModel::ClusterStatus{
          .id = id16, .type = "clusterStatus", .lastChanged = {}, .attributes = {.weather = {}, .operationsLocked = false}});
      clusterStatuses.push_back(dsModel::ClusterStatus{
          .id = id17, .type = "clusterStatus", .lastChanged = {}, .attributes = {.weather = {}, .operationsLocked = false}});
      CHECK(dsModel->clusterStatuses() == clusterStatuses);
    }

    SECTION("set cluster name") {
      const auto args = dsModel::ClusterArgs{.clusterId = id16};
      const auto newName = "{newName}";
      dsModel->setClusterName(args, newName);
      CHECK(dsModel->cluster(id16) ==
            dsModel::Cluster{.id = id16,
                             .type = "cluster",
                             .attributes = {.name = "{newName}",
                                            .orderId = {},
                                            .applicationType = dsModel::ClusterApplicationType::shades,
                                            .submodules = {}}});
    }
  }
}

} // namespace
