#include "buffered-message-writer.h"

#include <deque>

#include "session-base.h"

static ds::LogChannel dsLogChannel("dsNetMessageStream");
using boost::system::error_code;

namespace dsNet {
namespace {
struct BufferedMessageWriterImpl final : dsNet::Session<BufferedMessageWriterImpl, BufferedMessageWriterConfig>,
                                         BufferedMessageWriter,
                                         dsNet::SessionStateBase {
  using This = BufferedMessageWriterImpl;
  dsNet::MessageStream& m_stream;
  std::deque<std::string> m_queue{};
  size_t m_size{0};
  size_t m_maxCount{0};
  size_t m_maxSize{0};
  bool m_activeWrite{false};
  BufferedMessageWriterHooks m_hooks;

  BufferedMessageWriterImpl(ds::asio::ResourceBase& parent, dsNet::MessageStream& stream,
                            BufferedMessageWriterConfig&& config, BufferedMessageWriterHooks&& hooks)
      : dsNet::Session<BufferedMessageWriterImpl, BufferedMessageWriterConfig>{parent, "bufferedMessageWriter",
                                                                               std::move(config)},
        m_stream{stream},
        m_hooks{std::move(hooks)} {
    setOpening();
    setOpen();
  }

  size_t count() const override { return m_queue.size(); }
  size_t size() const override { return m_size; }

  void close() override { setClosed(); }

  void enqueue(std::string&& message) override {
    DS_INFO_ENTER(message, count(), size());
    assertSynchronized();
    DS_REQUIRE(count() + 1 <= config().countCapacity, "Buffer message count overflow.", count(), size(), message.size(),
               config());
    DS_REQUIRE(m_size + message.size() <= config().sizeCapacity, "Buffer total size overflow.", count(), size(),
               message.size(), config());
    m_size += message.size();
    m_queue.push_back(message);
    m_maxSize = std::max(m_maxSize, m_size);
    m_maxCount = std::max(m_maxCount, count());
    if (m_activeWrite) {
      m_hooks.enqueueWritePostponed();
    } else {
      m_activeWrite = true;
      DS_DEBUG("Starting write loop.");
      asyncWrite();
    }
  }

  const ds::asio::ResourceInside& inside() override {
    using This = BufferedMessageWriterImpl;
    static constexpr ds::asio::ResourceInside instance{
        .description = "Buffered message writer.",
        .properties = {
            ds::asio::ResourceProperty{
                .id = "count",
                .description = "Number of buffered messages.",
                .value = [](const ResourceBase& x) { return ds::toJsonFragment(static_cast<const This&>(x).count()); }},
            ds::asio::ResourceProperty{
                .id = "size",
                .description = "Total size of buffered messages.",
                .value = [](const ResourceBase& x) { return ds::toJsonFragment(static_cast<const This&>(x).size()); }},
            ds::asio::ResourceProperty{
                .id = "maxCount",
                .description = "Maximal number of buffered messages reached.",
                .value =
                    [](const ResourceBase& x) { return ds::toJsonFragment(static_cast<const This&>(x).m_maxCount); }},
            ds::asio::ResourceProperty{
                .id = "maxSize",
                .description = "Maximal total size of buffered messages reached.",
                .value =
                    [](const ResourceBase& x) { return ds::toJsonFragment(static_cast<const This&>(x).m_maxSize); }},
            ds::asio::ResourceProperty{
                .id = "state",
                .description = "State.",
                .value = [](const ResourceBase& x) { return ds::toJsonFragment(static_cast<const This&>(x).state()); }},
        }};
    return instance;
  }

  void asyncWrite() {
    DS_DEBUG_ENTER(count(), size(), m_activeWrite);
    assertSynchronized();
    try {
      if (m_queue.empty()) {
        m_activeWrite = false;
        DS_DEBUG_LEAVE("Stopping write loop.");
        return;
      }
      auto message = std::move(m_queue.front());
      m_size -= message.size();
      m_queue.pop_front();
      DS_DEBUG("Write", message);
      m_activeWrite = true;
      m_stream.asyncWrite(std::move(message), weakCallback(&This::asyncWritten));
    } catch (const std::exception& e) {
      DS_ERROR("Write failed.", e, *this);
      m_hooks.failed(e);
    }
  }

  void asyncWritten(error_code ec) {
    try {
      DS_REQUIRE(!ec, "Write failed.", ec, *this);
      asyncWrite();
    } catch (std::exception& e) {
      DS_ERROR("Write failed.", e, *this);
      m_hooks.failed(e);
    }
  }
};

} // namespace

dsNet::ClosePtr<BufferedMessageWriter> makeBufferedMessageWriter(ds::asio::ResourceBase& parent,
                                                                 dsNet::MessageStream& stream,
                                                                 BufferedMessageWriterConfig&& config,
                                                                 BufferedMessageWriterHooks&& hooks) {
  return std::make_shared<BufferedMessageWriterImpl>(parent, stream, std::move(config), std::move(hooks));
}

} // namespace dsNet

DS_JSON_DEFINE_STRUCT((dsNet), BufferedMessageWriterConfig);
