/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "ds-scenarios.h"

#include <ds-model/scenarios.h>
#include <ds/asio/test.h>
#include <ds/catch/catch.h>
#include <ds/clock-mock.h>

#include "businterface-mock.h"
#include "event-mock.h"
#include "model/apartment-mock.h"
#include "model/cluster.h"
#include "model/device-mock.h"
#include "model/group.h"
#include "model/modelmaintenance-mock.h"
#include "model/state.h"
#include "model/zone.h"

using namespace std::literals;
static ds::LogChannel dsLogChannel("dssDsScenariosTest");
static const char* TAGS = "[dss][dssDsScenarios]";

using namespace dss;
using namespace trompeloeil;

namespace {

TEST_CASE("Scenario id", TAGS) {
  auto checkDecodeEncode = [&](const dsModel::ScenarioId& id, const dss::ScenarioId& decodedId) {
    INFO(id);
    INFO(decodedId);
    CHECK(decodeId(id) == decodedId);
    CHECK(encodeId(decodedId) == id);
  };
  SECTION("applicationZone") {
    checkDecodeEncode(dsModel::ScenarioId{"applicationZone-z1-g3-s5"},
                      encodeId(ApplicationZoneScenarioId{.zone = 1, .group = 3, .scene = 5}));

    CHECK_THROWS_FIND(decodeDecodeId(dsModel::ScenarioId{"applicationZone"}), "Wrong scenario ID format");
    CHECK_THROWS_FIND(decodeDecodeId(dsModel::ScenarioId{"applicationZone-z0-g3-s5"}),
                      "Cannot use scenario ID for zone 0");
    CHECK_THROWS_FIND(decodeDecodeId(dsModel::ScenarioId{"applicationZone-z2-g0-s5"}),
                      "Cannot use scenario ID for group 0");
    CHECK_THROWS_FIND(decodeDecodeId(dsModel::ScenarioId{"applicationZone--z1-g2-s3"}), "Wrong scenario ID format");
    CHECK_THROWS_FIND(decodeDecodeId(dsModel::ScenarioId{"applicationZone-z1--g2-s3"}), "Wrong scenario ID format");
    CHECK_THROWS_FIND(decodeDecodeId(dsModel::ScenarioId{"applicationZone-z1-g2-/s3"}), "Wrong scenario ID format");
    CHECK_THROWS_FIND(decodeDecodeId(dsModel::ScenarioId{"applicationZone-1-32-23"}), "Wrong scenario ID format");
    CHECK_THROWS_FIND(decodeDecodeId(dsModel::ScenarioId{"applicationZone-c1-g32-s23"}), "Wrong scenario ID format");
    CHECK_THROWS_FIND(decodeDecodeId(dsModel::ScenarioId{"applicationZone-s1-z32-g23"}), "Wrong scenario ID format");
    CHECK_THROWS_FIND(decodeDecodeId(dsModel::ScenarioId{"applicationZone-z1-g32-d23"}), "Wrong scenario ID format");
    CHECK_THROWS_FIND(decodeDecodeId(dsModel::ScenarioId{"applicationZone-z1-g2-s3-"}),
                      "strToInt: Could not parse value: '3-'");
    CHECK_THROWS_FIND(decodeDecodeId(dsModel::ScenarioId{"applicationZone-z1g2-s3"}),
                      "strToInt: Could not parse value: '1g2'");
    SECTION("Non-common scenario") {
      CHECK_THROWS_FIND(decodeDecodeId(dsModel::ScenarioId{"applicationZone-z1-g1-s1"}),
                        "Wrong scenario ID - not a common scene");
    }
  }

  SECTION("userDefined") {
    checkDecodeEncode(dsModel::ScenarioId{"userDefined-32"}, encodeId(UserDefinedActionId{"32"}));
    CHECK_THROWS_FIND(decodeDecodeId(dsModel::ScenarioId{"userDefined"}), "Wrong scenario ID format");
  }

  SECTION("device") {
    checkDecodeEncode(dsModel::ScenarioId{"device-123-action"},
                      encodeId(DeviceScenarioId{.device = dsModel::DSDeviceId{"123"},
                                                .action = dsModel::DeviceScenarioActionId{"action"}}));
    CHECK_THROWS_FIND(decodeDecodeId(dsModel::ScenarioId{"dev-123-action"}), "Wrong scenario ID type");
    CHECK_THROWS_FIND(decodeDecodeId(dsModel::ScenarioId{"device-123"}), "Wrong scenario ID format");
    CHECK_THROWS_FIND(decodeDecodeId(dsModel::ScenarioId{"device-action"}), "Wrong scenario ID format");
  }

  CHECK_THROWS_FIND(decodeDecodeId(dsModel::ScenarioId{"foo-3"}), "Wrong scenario ID type");
}

SCENARIO("dssDsScenarios", TAGS) {
  auto clock = ds::ClockMocker{};
  auto propertySystem = PropertySystem{};
  EventSinkMock eventSink;
  ModelMaintenanceMock modelMaintenance;

  ApartmentMock apartment{nullptr, &propertySystem};
  const auto apartmentId = dsModel::ApartmentId{"apartmentId-0"};
  ALLOW_CALL(apartment, id()).RETURN(apartmentId);
  apartment.setName("TestName");
  apartment.setEventSink(eventSink);
  apartment.setModelMaintenance(&modelMaintenance);

  BusInterfaceMock busInterface;
  apartment.setBusInterface(&busInterface);

  ds::asio::TestLoopResource loop;
  auto dsScenarios = makeDsScenarios(loop, apartment);

  GIVEN("applicationZoneScenario") {
    auto& zone1 = *apartment.allocateZone(1);
    auto& group1 = *zone1.getGroup(1);
    const auto sceneId = 19;
    const auto id = ApplicationZoneScenarioId{.zone = zone1.getID(), .group = group1.getID(), .scene = sceneId};
    const auto encodedId = encodeId(id);
    const auto args = dsModel::ScenarioArgs{.scenarioId = encodeId(encodedId)};

    const auto nonExistingArgs =
        dsModel::ScenarioArgs{.scenarioId = encodeId(encodeId(ApplicationZoneScenarioId{
                                  .zone = zone1.getID(), .group = group1.getID(), .scene = 18}))}; // preset3

    group1.setSceneName(sceneId, "name1");

    SECTION("structure") {
      CHECK(dsScenarios->applicationZoneScenarios() == std::vector<dsModel::Scenario>{dsModel::ApplicationZoneScenario{
                                                           .id = encodedId,
                                                           .type = "applicationZoneScenario",
                                                           .attributes = {
                                                               .name = "name1",
                                                               .actionId = dsModel::CommonActionId::preset4,
                                                               .context = "applicationZone",
                                                               .submodules = {},
                                                               .dsDevices = {},
                                                               .zone = zone1.id(),
                                                               .application = dsModel::ZoneApplicationType::lights,
                                                           }}});
    }

    SECTION("special characters") {
      group1.setSceneName(sceneId, "name&lt;&#039;&amp;1");

      CHECK(dsScenarios->applicationZoneScenarios() == std::vector<dsModel::Scenario>{dsModel::ApplicationZoneScenario{
                                                           .id = encodedId,
                                                           .type = "applicationZoneScenario",
                                                           .attributes = {
                                                               .name = "name<'&1",
                                                               .actionId = dsModel::CommonActionId::preset4,
                                                               .context = "applicationZone",
                                                               .submodules = {},
                                                               .dsDevices = {},
                                                               .zone = zone1.id(),
                                                               .application = dsModel::ZoneApplicationType::lights,
                                                           }}});
    }

    SECTION("invoke") {
      SECTION("success") {
        REQUIRE_CALL(busInterface.action,
                     callScene(&group1, sceneId, Caller{CallOrigin::json, SceneAccessCategory::manual, ""}, false));
        dsScenarios->invokeScenario(args.scenarioId);
      }
      SECTION("non-existing") {
        CHECK_THROWS_FIND(dsScenarios->invokeScenario(nonExistingArgs.scenarioId), "Invalid ApplicationZoneScenarioId");
      }
    }

    SECTION("rename") {
      SECTION("success") {
        REQUIRE_CALL(busInterface.structureModify, sceneSetName(zone1.getID(), group1.getID(), sceneId, "name2"));
        dsScenarios->setScenarioName(args.scenarioId, "name2");
      }
      SECTION("zone application scenario") {
        CHECK(
            zoneApplicationScenarioIds(zone1) ==
            std::vector<ApplicationZoneScenarioId>{{.zone = zone1.getID(), .group = group1.getID(), .scene = sceneId}});
      }
      SECTION("zone scenario id") {
        CHECK(zoneScenarioIds(zone1) == std::vector<dsModel::ScenarioId>{{dsModel::ScenarioId{"applicationZone-z1-g1-s19"}}});
      }
      SECTION("fail on empty name") {
        CHECK_THROWS_FIND(dsScenarios->setScenarioName(args.scenarioId, ""), "New name cannot be empty.");
      }
      SECTION("non existing") {
        CHECK_THROWS_FIND(dsScenarios->setScenarioName(nonExistingArgs.scenarioId, "name2"),
                          "Invalid ApplicationZoneScenarioId");
      }
    }

    SECTION("persist") {
      SECTION("success") {
        REQUIRE_CALL(busInterface.action, saveScene(&group1, sceneId, Caller{CallOrigin::json}));
        dsScenarios->persistScenario(args.scenarioId);
      }
      SECTION("non existing") {
        CHECK_THROWS_FIND(dsScenarios->persistScenario(nonExistingArgs.scenarioId),
                          "Invalid ApplicationZoneScenarioId");
      }
    }

    SECTION("delete") {
      SECTION("success") {
        REQUIRE_CALL(busInterface.structureModify, sceneSetName(zone1.getID(), group1.getID(), sceneId, ""));
        dsScenarios->deleteScenario(args.scenarioId);
        CHECK(dsScenarios->applicationZoneScenarios() == std::vector<dsModel::Scenario>{});
      }
      SECTION("success on non existing") { dsScenarios->deleteScenario(nonExistingArgs.scenarioId); }
    }

    SECTION("create") {
      dsModel::Scenario args2 =
          dsModel::ApplicationZoneScenario{.id = {},
                                           .attributes = {.name = "newName",
                                                          .actionId = dsModel::CommonActionId::preset2,
                                                          .context = {},
                                                          .submodules = {},
                                                          .dsDevices = {},
                                                          .zone = dsModel::ZoneId{"1"},
                                                          .application = dsModel::ZoneApplicationType::lights}};
      REQUIRE_CALL(busInterface.structureModify,
                   sceneSetName(zone1.getID(), group1.getID(), 17 /*preset2*/, "newName"));
      dsScenarios->createScenario(args2);

      SECTION("fail on empty name") {
        dsModel::Scenario args3 =
            dsModel::ApplicationZoneScenario{.id = {},
                                             .attributes = {.name = "",
                                                            .actionId = dsModel::CommonActionId::preset2,
                                                            .context = {},
                                                            .submodules = {},
                                                            .dsDevices = {},
                                                            .zone = dsModel::ZoneId{"1"},
                                                            .application = dsModel::ZoneApplicationType::lights}};
        CHECK_THROWS_FIND(dsScenarios->createScenario(args3), "New name cannot be empty.");
      }
    }

    SECTION("create scenario without actionId") {
      // Presets choice starts from area4 first, and goes downwards
      SECTION("successfully create scenario in first free preset") {
        dsModel::Scenario args1 =
            dsModel::ApplicationZoneScenario{.id = {},
                                             .attributes = {.name = "TestNamePreset44",
                                                            .actionId = {},
                                                            .context = {},
                                                            .submodules = {},
                                                            .dsDevices = {},
                                                            .zone = dsModel::ZoneId{"1"},
                                                            .application = dsModel::ZoneApplicationType::lights}};
        REQUIRE_CALL(busInterface.structureModify, sceneSetName(zone1.getID(), group1.getID(), Scene44, "TestNamePreset44"));
        dsScenarios->createScenario(args1);

        GIVEN("preset42 already named") {
          THEN("new scenario created at preset43") {
            dsModel::Scenario args2 =
                dsModel::ApplicationZoneScenario{.id = {},
                                                 .attributes = {.name = "TestNamePreset43",
                                                                .actionId = {},
                                                                .context = {},
                                                                .submodules = {},
                                                                .dsDevices = {},
                                                                .zone = dsModel::ZoneId{"1"},
                                                                .application = dsModel::ZoneApplicationType::lights}};
            REQUIRE_CALL(busInterface.structureModify,
                         sceneSetName(zone1.getID(), group1.getID(), Scene43, "TestNamePreset43"));
            dsScenarios->createScenario(args2);
          }
        }
      }

      SECTION("creating scenario works for applications other than lights and shades") {
        auto& group3 = *zone1.getGroup(3);
        auto& group10 = *zone1.getGroup(10);
        SECTION("creating scenario works for heating") {
          dsModel::Scenario args1 =
              dsModel::ApplicationZoneScenario{.id = {},
                                               .attributes = {.name = "TestNamePreset44",
                                                              .actionId = {},
                                                              .context = {},
                                                              .submodules = {},
                                                              .dsDevices = {},
                                                              .zone = dsModel::ZoneId{"1"},
                                                              .application = dsModel::ZoneApplicationType::heating}};
          REQUIRE_CALL(busInterface.structureModify,
                       sceneSetName(zone1.getID(), group3.getID(), Scene44, "TestNamePreset44"));
          dsScenarios->createScenario(args1);
        }
        SECTION("creating scenario works for ventilation") {
          dsModel::Scenario args1 = dsModel::ApplicationZoneScenario{
              .id = {},
              .attributes = {.name = "TestNamePreset44",
                             .actionId = {},
                             .context = {},
                             .submodules = {},
                             .dsDevices = {},
                             .zone = dsModel::ZoneId{"1"},
                             .application = dsModel::ZoneApplicationType::ventilation}};
          REQUIRE_CALL(busInterface.structureModify,
                       sceneSetName(zone1.getID(), group10.getID(), Scene44, "TestNamePreset44"));
          dsScenarios->createScenario(args1);
        }
      }

      SECTION("all presets already named") {
        group1.setSceneName(Scene2, "name");
        group1.setSceneName(Scene3, "name");
        group1.setSceneName(Scene4, "name");
        group1.setSceneName(Scene12, "name");
        group1.setSceneName(Scene13, "name");
        group1.setSceneName(Scene14, "name");
        group1.setSceneName(Scene22, "name");
        group1.setSceneName(Scene23, "name");
        group1.setSceneName(Scene24, "name");
        group1.setSceneName(Scene32, "name");
        group1.setSceneName(Scene33, "name");
        group1.setSceneName(Scene34, "name");
        group1.setSceneName(Scene42, "name");
        group1.setSceneName(Scene43, "name");
        group1.setSceneName(Scene44, "name");
        dsModel::Scenario args1 =
            dsModel::ApplicationZoneScenario{.id = {},
                                             .attributes = {.name = "TestName",
                                                            .actionId = {},
                                                            .context = {},
                                                            .submodules = {},
                                                            .dsDevices = {},
                                                            .zone = dsModel::ZoneId{"1"},
                                                            .application = dsModel::ZoneApplicationType::lights}};
        CHECK_THROWS_FIND(dsScenarios->createScenario(args1), "Could not assign scenario to preset in group");
      }

      SECTION("button device configured") {
        GIVEN("button in area4") {
          const auto deviceId = dsModel::DSDeviceId{"1111111111111111111111111111111100"s};
          auto& device = *apartment.allocateDevice(Dsuid{deviceId});
          device.setButtonID(ButtonId_Zone_Extended4);
          device.addToGroup(1);
          zone1.addDevice({device.sharedFromThis(), &apartment});
          THEN("preset chosen from area3 range") {
            dsModel::Scenario args1 =
                dsModel::ApplicationZoneScenario{.id = {},
                                                 .attributes = {.name = "TestNamePreset34",
                                                                .actionId = {},
                                                                .context = {},
                                                                .submodules = {},
                                                                .dsDevices = {},
                                                                .zone = dsModel::ZoneId{"1"},
                                                                .application = dsModel::ZoneApplicationType::lights}};
            REQUIRE_CALL(busInterface.structureModify,
                         sceneSetName(zone1.getID(), group1.getID(), Scene34, "TestNamePreset34"));
            dsScenarios->createScenario(args1);
          }
        }
        GIVEN("button in zone and each area") {
          const auto deviceId1 = dsModel::DSDeviceId{"1111111111111111111111111111111100"s};
          const auto deviceId2 = dsModel::DSDeviceId{"2111111111111111111111111111111100"s};
          const auto deviceId3 = dsModel::DSDeviceId{"3111111111111111111111111111111100"s};
          const auto deviceId4 = dsModel::DSDeviceId{"4111111111111111111111111111111100"s};
          const auto deviceId5 = dsModel::DSDeviceId{"5111111111111111111111111111111100"s};
          auto& device1 = *apartment.allocateDevice(Dsuid{deviceId1});
          auto& device2 = *apartment.allocateDevice(Dsuid{deviceId2});
          auto& device3 = *apartment.allocateDevice(Dsuid{deviceId3});
          auto& device4 = *apartment.allocateDevice(Dsuid{deviceId4});
          auto& device5 = *apartment.allocateDevice(Dsuid{deviceId5});
          device1.addToGroup(1);
          device2.addToGroup(1);
          device3.addToGroup(1);
          device4.addToGroup(1);
          device5.addToGroup(1);
          device1.setButtonID(ButtonId_Zone_Extended4);
          device2.setButtonID(ButtonId_Zone_Extended3);
          device3.setButtonID(ButtonId_Zone_Extended2);
          device4.setButtonID(ButtonId_Zone_Extended1);
          device5.setButtonID(ButtonId_Zone);
          zone1.addDevice({device1.sharedFromThis(), &apartment});
          zone1.addDevice({device2.sharedFromThis(), &apartment});
          zone1.addDevice({device3.sharedFromThis(), &apartment});
          zone1.addDevice({device4.sharedFromThis(), &apartment});
          zone1.addDevice({device5.sharedFromThis(), &apartment});

          THEN("scenario can not be assigned") {
            dsModel::Scenario args1 =
                dsModel::ApplicationZoneScenario{.id = {},
                                                 .attributes = {.name = "TestName",
                                                                .actionId = {},
                                                                .context = {},
                                                                .submodules = {},
                                                                .dsDevices = {},
                                                                .zone = dsModel::ZoneId{"1"},
                                                                .application = dsModel::ZoneApplicationType::lights}};
            CHECK_THROWS_FIND(dsScenarios->createScenario(args1), "Could not assign scenario to preset in group");
          }
        }
      }
    }
  }

  GIVEN("userDefinedAction") {
    auto udasNode = propertySystem.createProperty("/scripts/system-addon-user-defined-actions");

    const auto id = "id0"s;
    const auto name = "name0"s;
    auto udaNode = udasNode->createProperty(id);
    udaNode->setChildValue("id", id);
    udaNode->setChildValue("name", name);

    const auto args = dsModel::ScenarioArgs{.scenarioId = encodeId(encodeId(UserDefinedActionId{id}))};
    const auto nonExistingArgs = dsModel::ScenarioArgs{.scenarioId = encodeId(encodeId(UserDefinedActionId{"idx"}))};

    SECTION("structure") {
      CHECK(dsScenarios->userDefinedActions() ==
            std::vector<dsModel::Scenario>{dsModel::UserDefinedAction{
                .id = encodeId(UserDefinedActionId{id}), .type = "userDefinedAction", .attributes = {.name = name}}});
    }

    SECTION("special characters") {
      const auto id1 = "id&amp;&gt;"s;
      const auto name1 = "name&lt;&#039;"s;
      auto udaNode = udasNode->createProperty(id1);
      udaNode->setChildValue("id", id1);
      udaNode->setChildValue("name", name1);

      CHECK(
          dsScenarios->userDefinedActions() ==
          std::vector<dsModel::Scenario>{
              dsModel::UserDefinedAction{.id = dsModel::UserDefinedActionId{"userDefined-id&>"},
                                         .type = "userDefinedAction",
                                         .attributes = {.name = "name<'"}},
              dsModel::UserDefinedAction{
                  .id = encodeId(UserDefinedActionId{id}), .type = "userDefinedAction", .attributes = {.name = name}}});
    }

    SECTION("invoke") {
      SECTION("success") {
        REQUIRE_CALL(eventSink, pushEvent(_))
            .LR_SIDE_EFFECT(CHECK(*_1 == Event{EventName::HighLevelEvent, Properties{{"id", id}}}));
        dsScenarios->invokeScenario(args.scenarioId);
      }
      SECTION("non-existing") {
        CHECK_THROWS_FIND(dsScenarios->invokeScenario(nonExistingArgs.scenarioId), "Invalid UserDefinedActionId");
      }
    }

    SECTION("rename") {
      SECTION("success") {
        SECTION("success TODO") { CHECK_THROWS_FIND(dsScenarios->setScenarioName(args.scenarioId, "name2"), "TODO"); }
      }
      SECTION("fail on empty name") {
        CHECK_THROWS_FIND(dsScenarios->setScenarioName(args.scenarioId, ""), "New name cannot be empty.");
      }
      SECTION("non-existing") {
        CHECK_THROWS_FIND(dsScenarios->invokeScenario(nonExistingArgs.scenarioId), "Invalid UserDefinedActionId");
      }
    }

    SECTION("persist") {
      SECTION("always fails") {
        CHECK_THROWS_FIND(dsScenarios->persistScenario(args.scenarioId),
                          "Persist of user defined action is not possible.");
      }
    }

    SECTION("delete") {
      SECTION("success TODO") { CHECK_THROWS_FIND(dsScenarios->deleteScenario(args.scenarioId), "TODO"); }
      SECTION("success on non existing") { dsScenarios->deleteScenario(nonExistingArgs.scenarioId); }
    }
  }

  GIVEN("deviceScenario") {
    auto& device = *apartment.allocateDevice(Dsuid{"1111111111111111111111111111111100"});
    const auto actionId = "actionId"s;
    ALLOW_CALL(modelMaintenance, addModelEvent(_)).LR_SIDE_EFFECT(delete _1);

    std::vector<dsModel::DeviceScenario> scenarios{};
    scenarios.push_back(
        dsModel::DeviceScenario{.id = dsModel::DeviceScenarioId{"device-1111111111111111111111111111111100-actionId"},
                                .attributes = {.name = {.domain = {}, .message = "name"},
                                               .dsDevice = dsModel::DSDeviceId{"1111111111111111111111111111111100"},
                                               .actionId = dsModel::DeviceScenarioActionId{actionId}}});
    device.setScenarios(std::move(scenarios));

    SECTION("structure") {
      CHECK(dsScenarios->deviceScenarios() ==
            std::vector<dsModel::Scenario>{dsModel::DeviceScenario{
                .id = encodeId(DeviceScenarioId{dsModel::DSDeviceId{"1111111111111111111111111111111100"},
                                                dsModel::DeviceScenarioActionId{actionId}}),
                .attributes = {.name = {.domain = {}, .message = "name"},
                               .dsDevice = dsModel::DSDeviceId{"1111111111111111111111111111111100"},
                               .actionId = dsModel::DeviceScenarioActionId{actionId}}}});
    }
  }

  SECTION("invoke standard scenario") {
    auto deviceShade = createDevice(apartment, {.hardwareType = HardwareType::GR_KL200});

    ALLOW_CALL(modelMaintenance, addModelEvent(_)).LR_SIDE_EFFECT(delete _1);
    deviceShade->addToGroup(GroupIDGray);
    deviceShade->setActiveGroup(ApplicationType::blinds);

    DeviceReference devShadeRef(deviceShade, &apartment);

    auto& zone0 = *apartment.allocateZone(0);
    auto& zone3 = *apartment.allocateZone(3);
    auto& group00 = *zone0.getGroup(0);
    auto& group3Lights = *zone3.getGroup(GroupIDYellow);
    auto& group3Shades = *zone3.getGroup(GroupIDGray);

    SECTION("Mixed shades and legacy awnings in apartment context") {
      auto deviceAwning = createDevice(apartment, {.hardwareType = HardwareType::GR_KL210});

      deviceAwning->addToGroup(GroupIDGray);
      deviceAwning->setActiveGroup(ApplicationType::blinds);

      DeviceReference devAwningRef(deviceAwning, &apartment);

      auto& zone5 = *apartment.allocateZone(5);
      auto& zoneNewDevices = *apartment.allocateZone(65534);
      auto& group5Shades = *zone5.getGroup(GroupIDGray);
      auto& group5NewDevices = *zoneNewDevices.getGroup(GroupIDGray);

      auto checkRequire = [&](const dsModel::ActionId& actionId, std::optional<int> scene) {
        REQUIRE_CALL(busInterface.action, callScene(&group5Shades, *scene,
                                                    Caller{CallOrigin::json, SceneAccessCategory::manual, ""}, false));
        dsScenarios->invokeStandardScenario({.context = dsModel::InvokeContextId{"applicationApartment"},
                                             .actionId = actionId,
                                             .application = dsModel::ApplicationId{"shades"},
                                             .dsDevice = {},
                                             .zone = {},
                                             .area = {}});
      };
      auto checkRequireAwnings = [&](const dsModel::ActionId& actionId, std::optional<int> scene) {
        REQUIRE_CALL(busInterface.action, callScene(&group5Shades, *scene,
                                                    Caller{CallOrigin::json, SceneAccessCategory::manual, ""}, false));
        dsScenarios->invokeStandardScenario({.context = dsModel::InvokeContextId{"applicationApartment"},
                                             .actionId = actionId,
                                             .application = dsModel::ApplicationId{"awnings"},
                                             .dsDevice = {},
                                             .zone = {},
                                             .area = {}});
      };
      auto checkForbid = [&](const dsModel::ActionId& actionId, std::optional<int> scene) {
        REQUIRE_CALL(busInterface.action, callScene(&group5NewDevices, *scene,
                                                    Caller{CallOrigin::json, SceneAccessCategory::manual, ""}, false));
        FORBID_CALL(busInterface.action,
                    callScene(&group5Shades, *scene, Caller{CallOrigin::json, SceneAccessCategory::manual, ""}, false));
        dsScenarios->invokeStandardScenario({.context = dsModel::InvokeContextId{"applicationApartment"},
                                             .actionId = actionId,
                                             .application = dsModel::ApplicationId{"shades"},
                                             .dsDevice = {},
                                             .zone = {},
                                             .area = {}});
      };
      auto checkForbidAwnings = [&](const dsModel::ActionId& actionId, std::optional<int> scene) {
        FORBID_CALL(busInterface.action,
                    callScene(&group5Shades, *scene, Caller{CallOrigin::json, SceneAccessCategory::manual, ""}, false));
        dsScenarios->invokeStandardScenario({.context = dsModel::InvokeContextId{"applicationApartment"},
                                             .actionId = actionId,
                                             .application = dsModel::ApplicationId{"awnings"},
                                             .dsDevice = {},
                                             .zone = {},
                                             .area = {}});
      };

      SECTION("Only shades in zone") {
        zone5.addDevice(devShadeRef);

        checkRequire(dsModel::ActionId{"up"}, 5);
        checkRequire(dsModel::ActionId{"down"}, 0);
        checkRequire(dsModel::ActionId{"stepUp"}, 12);
        checkRequire(dsModel::ActionId{"stepDown"}, 11);
        checkRequire(dsModel::ActionId{"stop"}, 15);
        checkRequire(dsModel::ActionId{"sunProtection"}, 56);
      }
      SECTION("Both awnings and shades in zone") {
        zone5.addDevice(devAwningRef);
        zone5.addDevice(devShadeRef);

        checkRequire(dsModel::ActionId{"up"}, 5);
        checkRequire(dsModel::ActionId{"down"}, 0);
        checkRequire(dsModel::ActionId{"stepUp"}, 12);
        checkRequire(dsModel::ActionId{"stepDown"}, 11);
        checkRequire(dsModel::ActionId{"stop"}, 15);
        checkRequire(dsModel::ActionId{"sunProtection"}, 56);
        checkForbidAwnings(dsModel::ActionId{"out"}, 5);
        checkForbidAwnings(dsModel::ActionId{"in"}, 0);
        checkForbidAwnings(dsModel::ActionId{"stepOut"}, 12);
        checkForbidAwnings(dsModel::ActionId{"stepIn"}, 11);
        checkForbidAwnings(dsModel::ActionId{"stop"}, 15);
        checkForbidAwnings(dsModel::ActionId{"sunProtection"}, 56);
      }
      SECTION("Only awnings in zone") {
        zone5.addDevice(devAwningRef);

        checkForbid(dsModel::ActionId{"up"}, 5);
        checkForbid(dsModel::ActionId{"down"}, 0);
        checkForbid(dsModel::ActionId{"stepUp"}, 12);
        checkForbid(dsModel::ActionId{"stepDown"}, 11);
        checkForbid(dsModel::ActionId{"stop"}, 15);
        checkForbid(dsModel::ActionId{"sunProtection"}, 56);
        checkRequireAwnings(dsModel::ActionId{"out"}, 5);
        checkRequireAwnings(dsModel::ActionId{"in"}, 0);
        checkRequireAwnings(dsModel::ActionId{"stepOut"}, 12);
        checkRequireAwnings(dsModel::ActionId{"stepIn"}, 11);
        checkRequireAwnings(dsModel::ActionId{"stop"}, 15);
        checkRequireAwnings(dsModel::ActionId{"sunProtection"}, 56);
      }

      SECTION("Check for zoneApplication") {
        auto checkThrowAwningsZone = [&](const dsModel::ActionId& actionId, std::optional<int> scene) {
          CHECK_THROWS(dsScenarios->invokeStandardScenario({.context = dsModel::InvokeContextId{"applicationZone"},
                                                            .actionId = actionId,
                                                            .application = dsModel::ApplicationId{"awnings"},
                                                            .dsDevice = {},
                                                            .zone = dsModel::ZoneId{"5"},
                                                            .area = {}}));
        };
        auto checkRequireAwningsZone = [&](const dsModel::ActionId& actionId, std::optional<int> scene) {
          REQUIRE_CALL(
              busInterface.action,
              callScene(&group5Shades, *scene, Caller{CallOrigin::json, SceneAccessCategory::manual, ""}, false));
          dsScenarios->invokeStandardScenario({.context = dsModel::InvokeContextId{"applicationZone"},
                                               .actionId = actionId,
                                               .application = dsModel::ApplicationId{"awnings"},
                                               .dsDevice = {},
                                               .zone = dsModel::ZoneId{"5"},
                                               .area = {}});
        };

        SECTION("Mixed shades and legacy awnings") {
          zone5.addDevice(devAwningRef);
          zone5.addDevice(devShadeRef);

          checkThrowAwningsZone(dsModel::ActionId{"out"}, 5);
          checkThrowAwningsZone(dsModel::ActionId{"in"}, 0);
          checkThrowAwningsZone(dsModel::ActionId{"stepOut"}, 12);
          checkThrowAwningsZone(dsModel::ActionId{"stepIn"}, 11);
          checkThrowAwningsZone(dsModel::ActionId{"stop"}, 15);
          checkThrowAwningsZone(dsModel::ActionId{"sunProtection"}, 56);
        }
        SECTION("Only awnings") {
          zone5.addDevice(devAwningRef);

          checkRequireAwningsZone(dsModel::ActionId{"out"}, 5);
          checkRequireAwningsZone(dsModel::ActionId{"in"}, 0);
          checkRequireAwningsZone(dsModel::ActionId{"stepOut"}, 12);
          checkRequireAwningsZone(dsModel::ActionId{"stepIn"}, 11);
          checkRequireAwningsZone(dsModel::ActionId{"stop"}, 15);
          checkRequireAwningsZone(dsModel::ActionId{"sunProtection"}, 56);
        }
      }
    }

    struct Fail {};
    struct CallScene {
      int scene{};
      bool force = false;
    };
    struct UndoScene {
      int scene{};
    };
    using Operation = std::variant<Fail, CallScene, UndoScene>;
    auto check = [&](AddressableModelItem& target, dsModel::InvokeStandardScenarioArgs args,
                     const Operation& operation) {
      INFO(args);
      std::visit(ds::Visitor{
                     [&](const Fail& x) { CHECK_THROWS(dsScenarios->invokeStandardScenario(std::move(args))); },
                     [&](const CallScene& x) {
                       INFO(x.scene);
                       INFO(x.force);
                       REQUIRE_CALL(busInterface.action,
                                    callScene(&target, x.scene,
                                              Caller{CallOrigin::json, SceneAccessCategory::manual, ""}, x.force));
                       dsScenarios->invokeStandardScenario(std::move(args));
                     },
                     [&](const UndoScene& x) {
                       INFO(x.scene);
                       REQUIRE_CALL(
                           busInterface.action,
                           undoScene(&target, x.scene, Caller{CallOrigin::json, SceneAccessCategory::manual, ""}));
                       dsScenarios->invokeStandardScenario(std::move(args));
                     },
                 },
                 operation);
    };
    auto checkFail = [&](AddressableModelItem& target, dsModel::InvokeStandardScenarioArgs args,
                         const Operation& operation) {
      INFO(args);
      CHECK_THROWS(dsScenarios->invokeStandardScenario(std::move(args)));
    };

    struct CommonX {
      dsModel::CommonActionId actionId{};
      Operation operation{};
    };
    const std::vector<CommonX> commonActions{{{}, Fail{}},
                                             {dsModel::CommonActionId::preset0, CallScene{SceneOff}},
                                             {dsModel::CommonActionId::preset1, CallScene{Scene1}},
                                             {dsModel::CommonActionId::preset2, CallScene{Scene2}},
                                             {dsModel::CommonActionId::preset3, CallScene{Scene3}},
                                             {dsModel::CommonActionId::preset4, CallScene{Scene4}},
                                             {dsModel::CommonActionId::preset10, CallScene{SceneOffE1}},
                                             {dsModel::CommonActionId::preset11, CallScene{SceneOnE1}},
                                             {dsModel::CommonActionId::preset12, CallScene{Scene12}},
                                             {dsModel::CommonActionId::preset13, CallScene{Scene13}},
                                             {dsModel::CommonActionId::preset14, CallScene{Scene14}},
                                             {dsModel::CommonActionId::preset20, CallScene{SceneOffE2}},
                                             {dsModel::CommonActionId::preset21, CallScene{SceneOnE2}},
                                             {dsModel::CommonActionId::preset22, CallScene{Scene22}},
                                             {dsModel::CommonActionId::preset23, CallScene{Scene23}},
                                             {dsModel::CommonActionId::preset24, CallScene{Scene24}},
                                             {dsModel::CommonActionId::preset30, CallScene{SceneOffE3}},
                                             {dsModel::CommonActionId::preset31, CallScene{SceneOnE3}},
                                             {dsModel::CommonActionId::preset32, CallScene{Scene32}},
                                             {dsModel::CommonActionId::preset33, CallScene{Scene33}},
                                             {dsModel::CommonActionId::preset34, CallScene{Scene34}},
                                             {dsModel::CommonActionId::preset40, CallScene{SceneOffE4}},
                                             {dsModel::CommonActionId::preset41, CallScene{SceneOnE4}},
                                             {dsModel::CommonActionId::preset42, CallScene{Scene42}},
                                             {dsModel::CommonActionId::preset43, CallScene{Scene43}},
                                             {dsModel::CommonActionId::preset44, CallScene{Scene44}}};
    struct LightsX {
      dsModel::LightsActionId actionId{};
      Operation operation{};
    };
    const std::vector<LightsX> lightsActions{{{}, Fail{}},
                                             {dsModel::LightsActionId::off, CallScene{SceneOff}},
                                             {dsModel::LightsActionId::on, CallScene{Scene1}},
                                             {dsModel::LightsActionId::brighter, CallScene{SceneInc}},
                                             {dsModel::LightsActionId::darker, CallScene{SceneDec}},
                                             {dsModel::LightsActionId::stop, CallScene{SceneStop}},
                                             {dsModel::LightsActionId::forcedOff, CallScene{SceneOff, true}},
                                             {dsModel::LightsActionId::fadeOff, CallScene{SceneAutoOff}}};
    struct ShadesX {
      dsModel::ShadesActionId actionId{};
      Operation operation{};
    };
    const std::vector<ShadesX> shadesActions{{{}, Fail{}},
                                             {dsModel::ShadesActionId::down, CallScene{SceneOff}},
                                             {dsModel::ShadesActionId::up, CallScene{Scene1}},
                                             {dsModel::ShadesActionId::stepDown, CallScene{SceneDec}},
                                             {dsModel::ShadesActionId::stepUp, CallScene{SceneInc}},
                                             {dsModel::ShadesActionId::stop, CallScene{SceneStop}},
                                             {dsModel::ShadesActionId::sunProtection, CallScene{SceneSunProtection}}};
    struct AwningsX {
      dsModel::AwningsActionId actionId{};
      Operation operation{};
    };
    const std::vector<AwningsX> awningsActions{
        {{}, Fail{}},
        {dsModel::AwningsActionId::out, CallScene{Scene1}},
        {dsModel::AwningsActionId::in, CallScene{SceneOff}},
        {dsModel::AwningsActionId::stepOut, CallScene{SceneInc}},
        {dsModel::AwningsActionId::stepIn, CallScene{SceneDec}},
        {dsModel::AwningsActionId::stop, CallScene{SceneStop}},
        {dsModel::AwningsActionId::sunProtection, CallScene{SceneSunProtection}}};
    struct AudioX {
      dsModel::AudioActionId actionId{};
      Operation operation{};
    };
    const std::vector<AudioX> audioActions{{{}, Fail{}},
                                           {dsModel::AudioActionId::stop, CallScene{SceneStop}},
                                           {dsModel::AudioActionId::play, CallScene{Scene1}},
                                           {dsModel::AudioActionId::pause, CallScene{SceneOff}},
                                           {dsModel::AudioActionId::mute, CallScene{SceneDecA3}},
                                           {dsModel::AudioActionId::unmute, CallScene{SceneIncA3}},
                                           {dsModel::AudioActionId::volumeUp, CallScene{SceneInc}},
                                           {dsModel::AudioActionId::volumeDown, CallScene{SceneDec}}};
    struct VideoX {
      dsModel::VideoActionId actionId{};
      Operation operation{};
    };
    const std::vector<VideoX> videoActions{{{}, Fail{}},
                                           {dsModel::VideoActionId::stop, CallScene{SceneStop}},
                                           {dsModel::VideoActionId::play, CallScene{Scene1}},
                                           {dsModel::VideoActionId::pause, CallScene{SceneOff}},
                                           {dsModel::VideoActionId::mute, CallScene{SceneDecA3}},
                                           {dsModel::VideoActionId::unmute, CallScene{SceneIncA3}},
                                           {dsModel::VideoActionId::volumeUp, CallScene{SceneInc}},
                                           {dsModel::VideoActionId::volumeDown, CallScene{SceneDec}}};
    struct VentilationX {
      dsModel::VentilationActionId actionId{};
      Operation operation{};
    };
    const std::vector<VentilationX> ventilationActions{{{}, Fail{}},
                                                       {dsModel::VentilationActionId::off, CallScene{SceneOff}},
                                                       {dsModel::VentilationActionId::level1, CallScene{Scene1}},
                                                       {dsModel::VentilationActionId::level2, CallScene{Scene2}},
                                                       {dsModel::VentilationActionId::level3, CallScene{Scene3}},
                                                       {dsModel::VentilationActionId::level4, CallScene{Scene4}},
                                                       {dsModel::VentilationActionId::boost, CallScene{SceneA11}}};
    struct RecirculationX {
      dsModel::RecirculationActionId actionId{};
      Operation operation{};
    };
    const std::vector<RecirculationX> recirculationActions{
        {{}, Fail{}},
        {dsModel::RecirculationActionId::off, CallScene{SceneOff}},
        {dsModel::RecirculationActionId::level1, CallScene{Scene1}},
        {dsModel::RecirculationActionId::level2, CallScene{Scene2}},
        {dsModel::RecirculationActionId::level3, CallScene{Scene3}},
        {dsModel::RecirculationActionId::level4, CallScene{Scene4}},
        {dsModel::RecirculationActionId::boost, CallScene{SceneA11}}};

    SECTION("applicationApartment") {
      struct X {
        dsModel::ApartmentActionType actionId{};
        dsModel::ApplicationId application;
        Operation operation{};
      };
      for (const auto& x : std::initializer_list<X>{
               {{}, {}, Fail{}},
               {dsModel::ApartmentActionType::absent, {}, CallScene{SceneAbsent}},
               {dsModel::ApartmentActionType::absent, dsModel::ApplicationId{"access"}, CallScene{SceneAbsent}}, // "access" is deprecated
               {dsModel::ApartmentActionType::present, {}, CallScene{ScenePresent}},
               {dsModel::ApartmentActionType::present, dsModel::ApplicationId{"access"}, CallScene{ScenePresent}}, // "access" is deprecated
               {dsModel::ApartmentActionType::sleeping, {}, CallScene{SceneSleeping}},
               {dsModel::ApartmentActionType::wakeup, {}, CallScene{SceneWakeUp}},
               {dsModel::ApartmentActionType::doorBell, {}, CallScene{SceneBell}},
               {dsModel::ApartmentActionType::panic, {}, CallScene{ScenePanic}},
               {dsModel::ApartmentActionType::panicEnd, {}, UndoScene{ScenePanic}},
               {dsModel::ApartmentActionType::fire, {}, CallScene{SceneFire}},
               {dsModel::ApartmentActionType::fireEnd, {}, UndoScene{SceneFire}},
               {dsModel::ApartmentActionType::alarm1, {}, CallScene{SceneAlarm}},
               {dsModel::ApartmentActionType::alarm2, {}, CallScene{SceneAlarm2}},
               {dsModel::ApartmentActionType::alarm3, {}, CallScene{SceneAlarm3}},
               {dsModel::ApartmentActionType::alarm4, {}, CallScene{SceneAlarm4}},
               {dsModel::ApartmentActionType::alarm1End, {}, UndoScene{SceneAlarm}},
               {dsModel::ApartmentActionType::alarm2End, {}, UndoScene{SceneAlarm2}},
               {dsModel::ApartmentActionType::alarm3End, {}, UndoScene{SceneAlarm3}},
               {dsModel::ApartmentActionType::alarm4End, {}, UndoScene{SceneAlarm4}},
               {dsModel::ApartmentActionType::smoke, {}, CallScene{SceneSmoke}},
               {dsModel::ApartmentActionType::smokeEnd, {}, UndoScene{SceneSmoke}},
               {dsModel::ApartmentActionType::water, {}, CallScene{SceneWater}},
               {dsModel::ApartmentActionType::waterEnd, {}, UndoScene{SceneWater}},
               {dsModel::ApartmentActionType::gas, {}, CallScene{SceneGas}},
               {dsModel::ApartmentActionType::gasEnd, {}, UndoScene{SceneGas}},
               {dsModel::ApartmentActionType::wind, {}, CallScene{SceneWindActive}},
               {dsModel::ApartmentActionType::windEnd, {}, CallScene{SceneWindInactive}},
               {dsModel::ApartmentActionType::rain, {}, CallScene{SceneRainActive}},
               {dsModel::ApartmentActionType::rainEnd, {}, CallScene{SceneRainInactive}},
               {dsModel::ApartmentActionType::hail, {}, CallScene{SceneHailActive, true}},
               {dsModel::ApartmentActionType::hailEnd, {}, CallScene{SceneHailInactive, true}},
               {dsModel::ApartmentActionType::pollution, {}, CallScene{SceneHazardousAir}},
               {dsModel::ApartmentActionType::pollutionEnd, {}, UndoScene{SceneHazardousAir}},
           }) {
        const auto& actionId = x.actionId;
        check(group00,
              {.context = dsModel::InvokeContextId{"applicationApartment"},
               .actionId = actionId != decltype(actionId){} ? dsModel::ActionId{ds::str(actionId)} : dsModel::ActionId{},
               .application = x.application},
              x.operation);
      }
      SECTION("frost") {
        REQUIRE_CALL(eventSink, pushEvent(_))
            .LR_SIDE_EFFECT(CHECK(*_1 == Event{EventName::FrostProtection, Properties{{"value", "1"}}}));
        dsScenarios->invokeStandardScenario({.context = dsModel::InvokeContextId{"applicationApartment"},
                                             .actionId = dsModel::ActionId{"frost"},
                                             .application = {}});
      }
      SECTION("frostEnd") {
        REQUIRE_CALL(eventSink, pushEvent(_))
            .LR_SIDE_EFFECT(CHECK(*_1 == Event{EventName::FrostProtection, Properties{{"value", "2"}}}));
        dsScenarios->invokeStandardScenario({.context = dsModel::InvokeContextId{"applicationApartment"},
                                             .actionId = dsModel::ActionId{"frostEnd"},
                                             .application = {}});
      }
      SECTION("fireMute") {
        REQUIRE_CALL(modelMaintenance, muteFire(_));
        dsScenarios->invokeStandardScenario({.context = dsModel::InvokeContextId{"applicationApartment"},
                                             .actionId = dsModel::ActionId{"fireMute"},
                                             .application = {}});
      }
      SECTION("zone, area must not be passed") {
        check(group00,
              {.context = dsModel::InvokeContextId{"applicationApartment"},
               .actionId = dsModel::ActionId{"absent"},
               .application = dsModel::ApplicationId{"access"},
               .dsDevice = {},
               .zone = dsModel::ZoneId{"0"},
               .area = {}},
              {});
        check(group00,
              {.context = dsModel::InvokeContextId{"applicationApartment"},
               .actionId = dsModel::ActionId{"absent"},
               .application = dsModel::ApplicationId{"access"},
               .dsDevice = {},
               .zone = {},
               .area = dsModel::AreaId{"1"}},
              {});
      }
      SECTION("apartment ventilation") {
        for (const auto& x : ventilationActions) {
          check(*zone0.getGroup(GroupIDGlobalAppDsVentilation),
                {.context = dsModel::InvokeContextId{"applicationApartment"},
                 .actionId = dsModel::ActionId{ds::str(x.actionId)},
                 .application = dsModel::ApplicationId{"ventilationUnit"},
                 .dsDevice = {},
                 .zone = {}},
                x.operation);
        }
      }
    }
    SECTION("applicationZone") {
      // all zone applications can be invoked also in the context of the whole apartment
      auto checkZoneApplication = [&](int groupId, dsModel::ZoneApplicationType application, auto actionId,
                                      const Operation& operation) {
        auto impl = [&](auto&& context, auto&& zone) {
          check(*zone.getGroup(groupId),
                {.context = context,
                 .actionId = actionId != decltype(actionId){} ? dsModel::ActionId{ds::str(actionId)} : dsModel::ActionId{},
                 .application = application != decltype(application){} ? dsModel::ApplicationId{ds::str(application)} : dsModel::ApplicationId{},
                 .dsDevice = {},
                 .zone = zone.getID() ? dsModel::ZoneId{ds::str(zone.getID())} : dsModel::ZoneId{}},
                operation);
        };
        impl(dsModel::InvokeContextId{"applicationZone"}, zone3);
        impl(dsModel::InvokeContextId{"applicationApartment"}, zone0);
      };
      auto checkZoneApplicationCommonAction = [&](int groupId, dsModel::ZoneApplicationType application) {
        for (const auto& x : commonActions) {
          checkZoneApplication(groupId, application, x.actionId, x.operation);
        }
      };
      SECTION("broadcast") {
        struct X {
          dsModel::ZoneActionId actionId{};
          Operation operation{};
        };
        for (const auto& x : std::initializer_list<X>{
                 {{}, Fail{}},
                 {dsModel::ZoneActionId::deepOff, CallScene{SceneDeepOff}},
                 {dsModel::ZoneActionId::standby, CallScene{SceneStandBy}},
             }) {
          checkZoneApplication(GroupIDBroadcast, {}, x.actionId, x.operation);
        }
      }
      SECTION("lights application") {
        SECTION("common action") {
          checkZoneApplicationCommonAction(GroupIDYellow, dsModel::ZoneApplicationType::lights);
        }
        SECTION("lights action") {
          for (const auto& x : lightsActions) {
            checkZoneApplication(GroupIDYellow, dsModel::ZoneApplicationType::lights, x.actionId, x.operation);
          }
        }
      }
      SECTION("shades application") {
        zone3.addDevice(devShadeRef);
        SECTION("common action") {
          checkZoneApplicationCommonAction(GroupIDGray, dsModel::ZoneApplicationType::shades);
        }
        SECTION("shades action") {
          for (const auto& x : shadesActions) {
            checkZoneApplication(GroupIDGray, dsModel::ZoneApplicationType::shades, x.actionId, x.operation);
          }
        }
      }
      SECTION("audio application") {
        SECTION("common action") { checkZoneApplicationCommonAction(GroupIDCyan, dsModel::ZoneApplicationType::audio); }
        SECTION("audio action") {
          for (const auto& x : audioActions) {
            checkZoneApplication(GroupIDCyan, dsModel::ZoneApplicationType::audio, x.actionId, x.operation);
          }
        }
      }
      SECTION("video application") {
        SECTION("common action") {
          checkZoneApplicationCommonAction(GroupIDViolet, dsModel::ZoneApplicationType::video);
        }
        SECTION("video action") {
          for (const auto& x : videoActions) {
            checkZoneApplication(GroupIDViolet, dsModel::ZoneApplicationType::video, x.actionId, x.operation);
          }
        }
      }
      SECTION("ventilation application") {
        SECTION("common action") {
          checkZoneApplicationCommonAction(GroupIDVentilation, dsModel::ZoneApplicationType::ventilation);
        }
        SECTION("ventilation action") {
          for (const auto& x : ventilationActions) {
            checkZoneApplication(GroupIDVentilation, dsModel::ZoneApplicationType::ventilation, x.actionId,
                                 x.operation);
          }
        }
      }
      SECTION("recirculation application") {
        SECTION("common action") {
          checkZoneApplicationCommonAction(GroupIDRecirculation, dsModel::ZoneApplicationType::recirculation);
        }
        SECTION("recirculation action") {
          for (const auto& x : recirculationActions) {
            checkZoneApplication(GroupIDRecirculation, dsModel::ZoneApplicationType::recirculation, x.actionId,
                                 x.operation);
          }
        }
      }
      SECTION("area must not be passed") {
        check(group00,
              {.context = dsModel::InvokeContextId{"applicationZone"},
               .actionId = dsModel::ActionId{"absent"},
               .application = dsModel::ApplicationId{"access"},
               .dsDevice = {},
               .zone = {},
               .area = dsModel::AreaId{"1"}},
              Fail{});
      }
    }
    SECTION("application area") {
      auto checkAreaApplication = [&](Group& group, int area, dsModel::ZoneApplicationType application, auto actionId,
                                      const Operation& operation) {
        check(group,
              {.context = dsModel::InvokeContextId{"applicationArea"},
               .actionId = actionId != decltype(actionId){} ? dsModel::ActionId{ds::str(actionId)} : dsModel::ActionId{},
               .application = application != decltype(application){} ? dsModel::ApplicationId{ds::str(application)} : dsModel::ApplicationId{},
               .dsDevice = {},
               .zone = dsModel::ZoneId{ds::str(group.getZoneID())},
               .area = dsModel::AreaId{ds::str(area)}},
              operation);
      };
      SECTION("lights application") {
        struct X {
          dsModel::LightsActionId actionId{};
          int area{};
          Operation operation{};
        };
        for (const auto& x : std::initializer_list<X>{
                 {{}, {}, Fail{}},
                 {dsModel::LightsActionId::off, 0, Fail{}},
                 {dsModel::LightsActionId::off, 1, CallScene{SceneOffA1}},
                 {dsModel::LightsActionId::off, 2, CallScene{SceneOffA2}},
                 {dsModel::LightsActionId::off, 3, CallScene{SceneOffA3}},
                 {dsModel::LightsActionId::off, 4, CallScene{SceneOffA4}},
                 {dsModel::LightsActionId::on, 0, Fail{}},
                 {dsModel::LightsActionId::on, 1, CallScene{SceneA11}},
                 {dsModel::LightsActionId::on, 2, CallScene{SceneA21}},
                 {dsModel::LightsActionId::on, 3, CallScene{SceneA31}},
                 {dsModel::LightsActionId::on, 4, CallScene{SceneA41}},
                 {dsModel::LightsActionId::brighter, 0, Fail{}},
                 {dsModel::LightsActionId::brighter, 1, CallScene{SceneIncA1}},
                 {dsModel::LightsActionId::brighter, 2, CallScene{SceneIncA2}},
                 {dsModel::LightsActionId::brighter, 3, CallScene{SceneIncA3}},
                 {dsModel::LightsActionId::brighter, 4, CallScene{SceneIncA4}},
                 {dsModel::LightsActionId::darker, 0, Fail{}},
                 {dsModel::LightsActionId::darker, 1, CallScene{SceneDecA1}},
                 {dsModel::LightsActionId::darker, 2, CallScene{SceneDecA2}},
                 {dsModel::LightsActionId::darker, 3, CallScene{SceneDecA3}},
                 {dsModel::LightsActionId::darker, 4, CallScene{SceneDecA4}},
                 {dsModel::LightsActionId::stop, 0, Fail{}},
                 {dsModel::LightsActionId::stop, 1, CallScene{SceneStopA1}},
                 {dsModel::LightsActionId::stop, 2, CallScene{SceneStopA2}},
                 {dsModel::LightsActionId::stop, 3, CallScene{SceneStopA3}},
                 {dsModel::LightsActionId::stop, 4, CallScene{SceneStopA4}},
                 {dsModel::LightsActionId::forcedOff, 0, Fail{}},
                 {dsModel::LightsActionId::forcedOff, 1, Fail{}},
                 {dsModel::LightsActionId::forcedOff, 2, Fail{}},
                 {dsModel::LightsActionId::forcedOff, 3, Fail{}},
                 {dsModel::LightsActionId::forcedOff, 4, Fail{}},
                 {dsModel::LightsActionId::fadeOff, 0, Fail{}},
                 {dsModel::LightsActionId::fadeOff, 1, Fail{}},
                 {dsModel::LightsActionId::fadeOff, 2, Fail{}},
                 {dsModel::LightsActionId::fadeOff, 3, Fail{}},
                 {dsModel::LightsActionId::fadeOff, 4, Fail{}},
             }) {
          checkAreaApplication(group3Lights, x.area, dsModel::ZoneApplicationType::lights, x.actionId, x.operation);
        }
      }
      SECTION("shades application") {
        struct X {
          dsModel::ShadesActionId actionId{};
          int area{};
          Operation operation{};
        };
        for (const auto& x : std::initializer_list<X>{
                 {{}, {}, Fail{}},
                 {dsModel::ShadesActionId::down, 0, Fail{}},
                 {dsModel::ShadesActionId::down, 1, CallScene{SceneOffA1}},
                 {dsModel::ShadesActionId::down, 2, CallScene{SceneOffA2}},
                 {dsModel::ShadesActionId::down, 3, CallScene{SceneOffA3}},
                 {dsModel::ShadesActionId::down, 4, CallScene{SceneOffA4}},
                 {dsModel::ShadesActionId::up, 0, Fail{}},
                 {dsModel::ShadesActionId::up, 1, CallScene{SceneA11}},
                 {dsModel::ShadesActionId::up, 2, CallScene{SceneA21}},
                 {dsModel::ShadesActionId::up, 3, CallScene{SceneA31}},
                 {dsModel::ShadesActionId::up, 4, CallScene{SceneA41}},
                 {dsModel::ShadesActionId::stepDown, 0, Fail{}},
                 {dsModel::ShadesActionId::stepDown, 1, CallScene{SceneDecA1}},
                 {dsModel::ShadesActionId::stepDown, 2, CallScene{SceneDecA2}},
                 {dsModel::ShadesActionId::stepDown, 3, CallScene{SceneDecA3}},
                 {dsModel::ShadesActionId::stepDown, 4, CallScene{SceneDecA4}},
                 {dsModel::ShadesActionId::stepUp, 0, Fail{}},
                 {dsModel::ShadesActionId::stepUp, 1, CallScene{SceneIncA1}},
                 {dsModel::ShadesActionId::stepUp, 2, CallScene{SceneIncA2}},
                 {dsModel::ShadesActionId::stepUp, 3, CallScene{SceneIncA3}},
                 {dsModel::ShadesActionId::stepUp, 4, CallScene{SceneIncA4}},
                 {dsModel::ShadesActionId::stop, 0, Fail{}},
                 {dsModel::ShadesActionId::stop, 1, CallScene{SceneStopA1}},
                 {dsModel::ShadesActionId::stop, 2, CallScene{SceneStopA2}},
                 {dsModel::ShadesActionId::stop, 3, CallScene{SceneStopA3}},
                 {dsModel::ShadesActionId::stop, 4, CallScene{SceneStopA4}},
                 {dsModel::ShadesActionId::sunProtection, 0, Fail{}},
                 {dsModel::ShadesActionId::sunProtection, 1, Fail{}},
                 {dsModel::ShadesActionId::sunProtection, 2, Fail{}},
                 {dsModel::ShadesActionId::sunProtection, 3, Fail{}},
                 {dsModel::ShadesActionId::sunProtection, 4, Fail{}},
             }) {
          checkAreaApplication(group3Shades, x.area, dsModel::ZoneApplicationType::shades, x.actionId, x.operation);
        }
      }
      SECTION("common action fails") {
        for (const auto& application : {dsModel::AreaApplicationType::none, dsModel::AreaApplicationType::lights,
                                        dsModel::AreaApplicationType::shades}) {
          for (const auto& actionOption : dsEnum(dsModel::CommonActionId{}).options) {
            const auto actionId = actionOption.first;
            INFO(application);
            INFO(actionId);
            CHECK_THROWS(dsScenarios->invokeStandardScenario(
                {.context = dsModel::InvokeContextId{"applicationArea"},
                 .actionId = actionId != decltype(actionId){} ? dsModel::ActionId{ds::str(actionId)} : dsModel::ActionId{},
                 .application = application != decltype(application){} ? dsModel::ApplicationId{ds::str(application)} : dsModel::ApplicationId{},
                 .dsDevice = {},
                 .zone = dsModel::ZoneId{"3"},
                 .area = dsModel::AreaId{"2"}}));
          }
        }
      }
    }
    SECTION("applicationCluster") {
      // Adding one awning device to apartment to make sure that calling cluster scene does not
      // invoke broadcast zone by zone, because clusters are assigned to zone 0.

      auto deviceAwning = createDevice(apartment, {.hardwareType = HardwareType::GR_KL210});

      deviceAwning->addToGroup(GroupIDGray);
      deviceAwning->setActiveGroup(ApplicationType::blinds);

      DeviceReference devAwningRef(deviceAwning, &apartment);

      auto cluster16 = apartment.getCluster(16);
      cluster16->setName("cluster16");
      const auto id16 = dsModel::ClusterId{"16"};

      auto checkClusterApplication = [&](dsModel::ClusterId clusterId, dsModel::ZoneApplicationType application,
                                         auto actionId, const Operation& operation) {
        check(*apartment.cluster(clusterId),
              {.context = dsModel::InvokeContextId{"applicationCluster"},
               .actionId = actionId != decltype(actionId){} ? dsModel::ActionId{ds::str(actionId)} : dsModel::ActionId{},
               .application = application != decltype(application){} ? dsModel::ApplicationId{ds::str(application)} : dsModel::ApplicationId{},
               .dsDevice = {},
               .zone = {},
               .area = {},
               .cluster = clusterId},
              operation);
      };
      auto checkClusterApplicationFail = [&](dsModel::ClusterId clusterId, dsModel::ZoneApplicationType application,
                                             auto actionId, const Operation& operation) {
        checkFail(
            *apartment.cluster(clusterId),
            {.context = dsModel::InvokeContextId{"applicationCluster"},
             .actionId = actionId != decltype(actionId){} ? dsModel::ActionId{ds::str(actionId)} : dsModel::ActionId{},
             .application = application != decltype(application){} ? dsModel::ApplicationId{ds::str(application)}
                                                                   : dsModel::ApplicationId{},
             .dsDevice = {},
             .zone = {},
             .area = {},
             .cluster = clusterId},
            operation);
      };
      auto checkClusterApplicationCommonAction = [&](dsModel::ClusterId clusterId,
                                                     dsModel::ZoneApplicationType application) {

        for (const auto& x : commonActions) {
          checkClusterApplication(clusterId, application, x.actionId, x.operation);
        }
      };
      SECTION("lights application") {
        cluster16->setApplicationType(ApplicationType::lights);
        SECTION("common action") { checkClusterApplicationCommonAction(id16, dsModel::ZoneApplicationType::lights); }
        SECTION("lights action") {
          for (const auto& x : lightsActions) {
            checkClusterApplication(id16, dsModel::ZoneApplicationType::lights, x.actionId, x.operation);
          }
        }
      }
      SECTION("shades application") {
        deviceShade->addToGroup(GroupIDAppUserMin);
        cluster16->setApplicationType(ApplicationType::blinds);
        SECTION("common action") { checkClusterApplicationCommonAction(id16, dsModel::ZoneApplicationType::shades); }
        SECTION("shades action") {
          for (const auto& x : shadesActions) {
            checkClusterApplication(id16, dsModel::ZoneApplicationType::shades, x.actionId, x.operation);
          }
        }
      }
      SECTION("awnings application") {
        deviceAwning->addToGroup(GroupIDAppUserMin);
        cluster16->setApplicationType(ApplicationType::blinds);
        SECTION("common action") { checkClusterApplicationCommonAction(id16, dsModel::ZoneApplicationType::awnings); }
        SECTION("awnings action") {
          struct X {
            dsModel::AwningsActionId actionId{};
            Operation operation{};
          };
          for (const auto& x : awningsActions) {
            checkClusterApplication(id16, dsModel::ZoneApplicationType::awnings, x.actionId, x.operation);
          }
        }
      }
      SECTION("audio application") {
        cluster16->setApplicationType(ApplicationType::audio);
        SECTION("common action") { checkClusterApplicationCommonAction(id16, dsModel::ZoneApplicationType::audio); }
        SECTION("audio action") {
          for (const auto& x : audioActions) {
            checkClusterApplication(id16, dsModel::ZoneApplicationType::audio, x.actionId, x.operation);
          }
        }
      }
      SECTION("video application") {
        cluster16->setApplicationType(ApplicationType::video);
        SECTION("common action") { checkClusterApplicationCommonAction(id16, dsModel::ZoneApplicationType::video); }
        SECTION("video action") {
          for (const auto& x : videoActions) {
            checkClusterApplication(id16, dsModel::ZoneApplicationType::video, x.actionId, x.operation);
          }
        }
      }
      SECTION("ventilation application") {
        cluster16->setApplicationType(ApplicationType::ventilation);
        SECTION("common action") {
          checkClusterApplicationCommonAction(id16, dsModel::ZoneApplicationType::ventilation);
        }
        SECTION("ventilation action") {
          for (const auto& x : ventilationActions) {
            checkClusterApplication(id16, dsModel::ZoneApplicationType::ventilation, x.actionId, x.operation);
          }
        }
      }
      SECTION("recirculation application") {
        cluster16->setApplicationType(ApplicationType::recirculation);
        SECTION("common action") {
          checkClusterApplicationCommonAction(id16, dsModel::ZoneApplicationType::recirculation);
        }
        SECTION("recirculation action") {
          for (const auto& x : recirculationActions) {
            checkClusterApplication(id16, dsModel::ZoneApplicationType::recirculation, x.actionId, x.operation);
          }
        }
      }
      SECTION("incorrect application") {
        cluster16->setApplicationType(ApplicationType::blinds);
        SECTION("lights action") {
          for (const auto& x : lightsActions) {
            checkClusterApplicationFail(id16, dsModel::ZoneApplicationType::lights, x.actionId, x.operation);
          }
        }
        SECTION("shades action - awnings cluster") {
          // Only 1 awning device in cluster, therefore this cluster's application is set to awnings
          deviceAwning->addToGroup(GroupIDAppUserMin);
          for (const auto& x : shadesActions) {
            checkClusterApplicationFail(id16, dsModel::ZoneApplicationType::shades, x.actionId, x.operation);
          }
        }
        SECTION("awnings action - mixed cluster") {
          // Mixed devices in cluster, therefore this cluster's application is set to shades
          deviceAwning->addToGroup(GroupIDAppUserMin);
          deviceShade->addToGroup(GroupIDAppUserMin);
          for (const auto& x : awningsActions) {
            checkClusterApplicationFail(id16, dsModel::ZoneApplicationType::awnings, x.actionId, x.operation);
          }
        }
      }
      SECTION("area must not be passed") {
        check(*cluster16,
              {.context = dsModel::InvokeContextId{"applicationCluster"},
               .actionId = dsModel::ActionId{"absent"},
               .application = dsModel::ApplicationId{"access"},
               .dsDevice = {},
               .zone = {},
               .area = dsModel::AreaId{"1"}},
              Fail{});
      }
      SECTION("zone must not be passed") {
        check(*cluster16,
              {.context = dsModel::InvokeContextId{"applicationCluster"},
               .actionId = dsModel::ActionId{"absent"},
               .application = dsModel::ApplicationId{"access"},
               .dsDevice = {},
               .zone = dsModel::ZoneId{"2"},
               .area = {}},
              Fail{});
      }
    }
    SECTION("applicationDevice") {
      auto& device = *apartment.allocateDevice(Dsuid{"1111111111111111111111111111111100"});
      struct X {
        dsModel::DeviceActionId actionId{};
        Operation operation{};
      };
      for (const auto& x : std::initializer_list<X>{
               {{}, Fail{}},
               {dsModel::DeviceActionId::off, CallScene{SceneMin}},
               {dsModel::DeviceActionId::on, CallScene{SceneMax}},
               {dsModel::DeviceActionId::localOff, CallScene{SceneLocalOff}},
               {dsModel::DeviceActionId::localOn, CallScene{SceneLocalOn}},
               {dsModel::DeviceActionId::stop, CallScene{SceneStop}},
               {dsModel::DeviceActionId::impulse, CallScene{SceneImpulse}},
               {dsModel::DeviceActionId::inc, CallScene{SceneInc}},
               {dsModel::DeviceActionId::dec, CallScene{SceneDec}},
           }) {
        const auto& actionId = x.actionId;
        check(device,
              {.context = dsModel::InvokeContextId{"applicationDevice"},
               .actionId = actionId != decltype(actionId){} ? dsModel::ActionId{ds::str(actionId)} : dsModel::ActionId{},
               .application = {},
               .dsDevice = dsModel::DSDeviceId{ds::str(device.getDSID())}},
              x.operation);
      }

      SECTION("application, zone, area must not be passed") {
        check(group00,
              {.context = dsModel::InvokeContextId{"applicationDevice"},
               .actionId = dsModel::ActionId{"off"},
               .application = dsModel::ApplicationId{"access"},
               .dsDevice = dsModel::DSDeviceId{ds::str(device.getDSID())},
               .zone = {},
               .area = {}},
              Fail{});
        check(group00,
              {.context = dsModel::InvokeContextId{"applicationDevice"},
               .actionId = dsModel::ActionId{"off"},
               .application = {},
               .dsDevice = dsModel::DSDeviceId{ds::str(device.getDSID())},
               .zone = dsModel::ZoneId{"0"},
               .area = {}},
              Fail{});
        check(group00,
              {.context = dsModel::InvokeContextId{"applicationDevice"},
               .actionId = dsModel::ActionId{"off"},
               .application = {},
               .dsDevice = dsModel::DSDeviceId{ds::str(device.getDSID())},
               .zone = {},
               .area = dsModel::AreaId{"1"}},
              Fail{});
      }
    }
  }
}

} // namespace
