/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "ds-scenarios.h"

#include <ds-model/scenarios.h>
#include <ds/asio/resource.h>
#include <ds/log.h>
#include <ds/vector.h>

#include <boost/algorithm/string/predicate.hpp>
#include <boost/range/adaptor/transformed.hpp>
#include <boost/range/algorithm/copy.hpp>
#include <boost/range/algorithm/sort.hpp>

#include "cache/apartment-proxy.h"
#include "model/cluster.h"
#include "model/device.h"
#include "model/group.h"
#include "model/state.h"
#include "model/zone.h"
#include "structuremanipulator.h"
#include "util.h"

using namespace std::literals;
static ds::LogChannel dsLogChannel("dssDsScenarios");

template <>
struct DSToEnum<dsModel::CommonActionId, int, void> {
  static std::optional<dsModel::CommonActionId> tryTo(int x) {
    switch (x) {
      case dss::SceneOff: return dsModel::CommonActionId::preset0;
      case dss::Scene1: return dsModel::CommonActionId::preset1;
      case dss::Scene2: return dsModel::CommonActionId::preset2;
      case dss::Scene3: return dsModel::CommonActionId::preset3;
      case dss::Scene4: return dsModel::CommonActionId::preset4;
      case dss::SceneOffE1: return dsModel::CommonActionId::preset10;
      case dss::SceneOnE1: return dsModel::CommonActionId::preset11;
      case dss::Scene12: return dsModel::CommonActionId::preset12;
      case dss::Scene13: return dsModel::CommonActionId::preset13;
      case dss::Scene14: return dsModel::CommonActionId::preset14;
      case dss::SceneOffE2: return dsModel::CommonActionId::preset20;
      case dss::SceneOnE2: return dsModel::CommonActionId::preset21;
      case dss::Scene22: return dsModel::CommonActionId::preset22;
      case dss::Scene23: return dsModel::CommonActionId::preset23;
      case dss::Scene24: return dsModel::CommonActionId::preset24;
      case dss::SceneOffE3: return dsModel::CommonActionId::preset30;
      case dss::SceneOnE3: return dsModel::CommonActionId::preset31;
      case dss::Scene32: return dsModel::CommonActionId::preset32;
      case dss::Scene33: return dsModel::CommonActionId::preset33;
      case dss::Scene34: return dsModel::CommonActionId::preset34;
      case dss::SceneOffE4: return dsModel::CommonActionId::preset40;
      case dss::SceneOnE4: return dsModel::CommonActionId::preset41;
      case dss::Scene42: return dsModel::CommonActionId::preset42;
      case dss::Scene43: return dsModel::CommonActionId::preset43;
      case dss::Scene44: return dsModel::CommonActionId::preset44;
    }
    return {};
  }
};

namespace dsModel {

std::string_view idView(const dsModel::Scenario& scenario) {
  return std::visit(ds::Visitor{[](const std::monostate&) { return std::string_view{}; },
                                [](const auto& x) { return std::string_view{static_cast<const std::string&>(x.id)}; }},
                    scenario);
}

dsModel::ScenarioId id(const dsModel::Scenario& scenario) {
  return dsModel::ScenarioId{std::string{idView(scenario)}};
}

} // namespace dsModel

namespace dss {
namespace {

static bool isCommonScenario(int scene) {
  switch (scene) {
    case dss::SceneOff:
    case dss::Scene1:
    case dss::Scene2:
    case dss::Scene3:
    case dss::Scene4:
    case dss::SceneOffE1:
    case dss::SceneOnE1:
    case dss::Scene12:
    case dss::Scene13:
    case dss::Scene14:
    case dss::SceneOffE2:
    case dss::SceneOnE2:
    case dss::Scene22:
    case dss::Scene23:
    case dss::Scene24:
    case dss::SceneOffE3:
    case dss::SceneOnE3:
    case dss::Scene32:
    case dss::Scene33:
    case dss::Scene34:
    case dss::SceneOffE4:
    case dss::SceneOnE4:
    case dss::Scene42:
    case dss::Scene43:
    case dss::Scene44: return true;
    default: return false;
  }
}

static int zoneScene(dsModel::ZoneActionId actionId) {
  switch (actionId) {
    case dsModel::ZoneActionId::deepOff: return SceneDeepOff;
    case dsModel::ZoneActionId::standby: return SceneStandBy;
    case dsModel::ZoneActionId::none:;
  }
  DS_FAIL_REQUIRE("No scene for zone action.", actionId);
}

static int lightsScene(dsModel::LightsActionId actionId) {
  switch (actionId) {
    case dsModel::LightsActionId::off: return SceneOff;
    case dsModel::LightsActionId::forcedOff: return SceneOff;
    case dsModel::LightsActionId::on: return Scene1;
    case dsModel::LightsActionId::brighter: return SceneInc;
    case dsModel::LightsActionId::darker: return SceneDec;
    case dsModel::LightsActionId::stop: return SceneStop;
    case dsModel::LightsActionId::fadeOff: return SceneAutoOff;
    case dsModel::LightsActionId::none:
    default: DS_FAIL_REQUIRE("No scene for ligths action.", actionId);
  }
}

static int lightsAreaScene(dsModel::LightsActionId actionId, int area) {
  DS_REQUIRE(area >= 1 && area <= 4, "Invalid area. ", area);
  switch (actionId) {
    case dsModel::LightsActionId::off:
      switch (area) {
        case 1: return SceneOffA1;
        case 2: return SceneOffA2;
        case 3: return SceneOffA3;
        case 4: return SceneOffA4;
      }
      break;
    case dsModel::LightsActionId::forcedOff: break; // not available with areas
    case dsModel::LightsActionId::on:
      switch (area) {
        case 1: return SceneA11;
        case 2: return SceneA21;
        case 3: return SceneA31;
        case 4: return SceneA41;
      }
      break;
    case dsModel::LightsActionId::brighter:
      switch (area) {
        case 1: return SceneIncA1;
        case 2: return SceneIncA2;
        case 3: return SceneIncA3;
        case 4: return SceneIncA4;
      }
      break;
    case dsModel::LightsActionId::darker:
      switch (area) {
        case 1: return SceneDecA1;
        case 2: return SceneDecA2;
        case 3: return SceneDecA3;
        case 4: return SceneDecA4;
      }
      break;
    case dsModel::LightsActionId::stop:
      switch (area) {
        case 1: return SceneStopA1;
        case 2: return SceneStopA2;
        case 3: return SceneStopA3;
        case 4: return SceneStopA4;
      }
      break;
    case dsModel::LightsActionId::fadeOff: break; // Not available for areas
    case dsModel::LightsActionId::none:; break;
  }
  DS_FAIL_REQUIRE("No scene for lights area action.", actionId, area);
}

static int shadesAreaScene(dsModel::ShadesActionId actionId, int area) {
  DS_REQUIRE(area >= 0 && area <= 4, "Invalid area. ", area);
  switch (actionId) {
    case dsModel::ShadesActionId::down:
      switch (area) {
        case 1: return SceneOffA1;
        case 2: return SceneOffA2;
        case 3: return SceneOffA3;
        case 4: return SceneOffA4;
      }
      break;
    case dsModel::ShadesActionId::up:
      switch (area) {
        case 1: return SceneA11;
        case 2: return SceneA21;
        case 3: return SceneA31;
        case 4: return SceneA41;
      }
      break;
    case dsModel::ShadesActionId::stepDown:
      switch (area) {
        case 1: return SceneDecA1;
        case 2: return SceneDecA2;
        case 3: return SceneDecA3;
        case 4: return SceneDecA4;
      }
      break;
    case dsModel::ShadesActionId::stepUp:
      switch (area) {
        case 1: return SceneIncA1;
        case 2: return SceneIncA2;
        case 3: return SceneIncA3;
        case 4: return SceneIncA4;
      }
      break;
    case dsModel::ShadesActionId::stop:
      switch (area) {
        case 1: return SceneStopA1;
        case 2: return SceneStopA2;
        case 3: return SceneStopA3;
        case 4: return SceneStopA4;
      }
      break;
    case dsModel::ShadesActionId::sunProtection: break; // Not available for areas
    case dsModel::ShadesActionId::none:; break;
  }
  DS_FAIL_REQUIRE("No scene for shades area action.", actionId, area);
}

static int audioScene(dsModel::AudioActionId actionId) {
  switch (actionId) {
    case dsModel::AudioActionId::stop: return SceneStop;
    case dsModel::AudioActionId::play: return Scene1;
    case dsModel::AudioActionId::pause: return SceneOff;
    case dsModel::AudioActionId::mute: return SceneDecA3;
    case dsModel::AudioActionId::unmute: return SceneIncA3;
    case dsModel::AudioActionId::volumeUp: return SceneInc;
    case dsModel::AudioActionId::volumeDown: return SceneDec;
    case dsModel::AudioActionId::none:;
  }
  DS_FAIL_REQUIRE("No scene for audio action.", actionId);
}

static int videoScene(dsModel::VideoActionId actionId) {
  switch (actionId) {
    case dsModel::VideoActionId::stop: return SceneStop;
    case dsModel::VideoActionId::play: return Scene1;
    case dsModel::VideoActionId::pause: return SceneOff;
    case dsModel::VideoActionId::mute: return SceneDecA3;
    case dsModel::VideoActionId::unmute: return SceneIncA3;
    case dsModel::VideoActionId::volumeUp: return SceneInc;
    case dsModel::VideoActionId::volumeDown: return SceneDec;
    case dsModel::VideoActionId::none:;
  }
  DS_FAIL_REQUIRE("No scene for video action.", actionId);
}

static int ventilationScene(dsModel::VentilationActionId actionId) {
  switch (actionId) {
    case dsModel::VentilationActionId::off: return SceneOff;
    case dsModel::VentilationActionId::level1: return Scene1;
    case dsModel::VentilationActionId::level2: return Scene2;
    case dsModel::VentilationActionId::level3: return Scene3;
    case dsModel::VentilationActionId::level4: return Scene4;
    case dsModel::VentilationActionId::boost: return SceneA11;
    case dsModel::VentilationActionId::none:;
  }
  DS_FAIL_REQUIRE("No scene for ventilation action.", actionId);
}

static int recirculationScene(dsModel::RecirculationActionId actionId) {
  switch (actionId) {
    case dsModel::RecirculationActionId::off: return SceneOff;
    case dsModel::RecirculationActionId::level1: return Scene1;
    case dsModel::RecirculationActionId::level2: return Scene2;
    case dsModel::RecirculationActionId::level3: return Scene3;
    case dsModel::RecirculationActionId::level4: return Scene4;
    case dsModel::RecirculationActionId::boost: return SceneA11;
    case dsModel::RecirculationActionId::none:;
  }
  DS_FAIL_REQUIRE("No scene for recirculation action.", actionId);
}

static int deviceScene(dsModel::DeviceActionId actionId) {
  switch (actionId) {
    case dsModel::DeviceActionId::off: return SceneMin;
    case dsModel::DeviceActionId::on: return SceneMax;
    case dsModel::DeviceActionId::localOff: return SceneLocalOff;
    case dsModel::DeviceActionId::localOn: return SceneLocalOn;
    case dsModel::DeviceActionId::stop: return SceneStop;
    case dsModel::DeviceActionId::impulse: return SceneImpulse;
    case dsModel::DeviceActionId::inc: return SceneInc;
    case dsModel::DeviceActionId::dec: return SceneDec;
    case dsModel::DeviceActionId::none:;
  }
  DS_FAIL_REQUIRE("No scene for device action.", actionId);
}

static int commonScene(dsModel::CommonActionId actionId) {
  switch (actionId) {
    case dsModel::CommonActionId::preset0: return SceneOff;
    case dsModel::CommonActionId::preset1: return Scene1;
    case dsModel::CommonActionId::preset2: return Scene2;
    case dsModel::CommonActionId::preset3: return Scene3;
    case dsModel::CommonActionId::preset4: return Scene4;
    case dsModel::CommonActionId::preset10: return SceneOffE1;
    case dsModel::CommonActionId::preset11: return SceneOnE1;
    case dsModel::CommonActionId::preset12: return Scene12;
    case dsModel::CommonActionId::preset13: return Scene13;
    case dsModel::CommonActionId::preset14: return Scene14;
    case dsModel::CommonActionId::preset20: return SceneOffE2;
    case dsModel::CommonActionId::preset21: return SceneOnE2;
    case dsModel::CommonActionId::preset22: return Scene22;
    case dsModel::CommonActionId::preset23: return Scene23;
    case dsModel::CommonActionId::preset24: return Scene24;
    case dsModel::CommonActionId::preset30: return SceneOffE3;
    case dsModel::CommonActionId::preset31: return SceneOnE3;
    case dsModel::CommonActionId::preset32: return Scene32;
    case dsModel::CommonActionId::preset33: return Scene33;
    case dsModel::CommonActionId::preset34: return Scene34;
    case dsModel::CommonActionId::preset40: return SceneOffE4;
    case dsModel::CommonActionId::preset41: return SceneOnE4;
    case dsModel::CommonActionId::preset42: return Scene42;
    case dsModel::CommonActionId::preset43: return Scene43;
    case dsModel::CommonActionId::preset44: return Scene44;
    case dsModel::CommonActionId::none:;
  }
  DS_FAIL_REQUIRE("No scene for common action.", actionId);
}

static int groupNumber(dsModel::ZoneApplicationType application) {
  switch (application) {
    case dsModel::ZoneApplicationType::lights: return GroupIDYellow;
    case dsModel::ZoneApplicationType::shades: [[fallthrough]];
    case dsModel::ZoneApplicationType::awnings: return GroupIDGray;
    case dsModel::ZoneApplicationType::audio: return GroupIDCyan;
    case dsModel::ZoneApplicationType::video: return GroupIDViolet;
    case dsModel::ZoneApplicationType::security: return GroupIDRed;
    case dsModel::ZoneApplicationType::access: return GroupIDGreen;
    case dsModel::ZoneApplicationType::heating: return GroupIDHeating;
    case dsModel::ZoneApplicationType::cooling: return GroupIDCooling;
    case dsModel::ZoneApplicationType::temperature: return GroupIDControlTemperature;
    case dsModel::ZoneApplicationType::ventilation: return GroupIDVentilation;
    case dsModel::ZoneApplicationType::recirculation: return GroupIDRecirculation;
    case dsModel::ZoneApplicationType::window: return GroupIDWindow;
    case dsModel::ZoneApplicationType::joker: return GroupIDBlack;
    case dsModel::ZoneApplicationType::none:
    default: break;
  }
  DS_FAIL_REQUIRE("No group for application type.", application);
}

std::string shadesActionId(dsModel::ShadesActionId actionId) {
  if (actionId == dsModel::ShadesActionId::up) {
    return "app.moveUp";
  } else if (actionId == dsModel::ShadesActionId::down) {
    return "app.moveDown";
  } else if (actionId == dsModel::ShadesActionId::stepUp) {
    return "app.stepUp";
  } else if (actionId == dsModel::ShadesActionId::stepDown) {
    return "app.stepDown";
  } else if (actionId == dsModel::ShadesActionId::stop) {
    return "app.stop";
  } else if (actionId == dsModel::ShadesActionId::sunProtection) {
    return "app.sunProtection";
  } else {
    DS_FAIL_REQUIRE("No scene for shades action.", actionId);
  }
}

std::string awningsActionId(dsModel::AwningsActionId actionId) {
  if (actionId == dsModel::AwningsActionId::out) {
    return "app.moveOut";
  } else if (actionId == dsModel::AwningsActionId::in) {
    return "app.moveIn";
  } else if (actionId == dsModel::AwningsActionId::stepOut) {
    return "app.stepOut";
  } else if (actionId == dsModel::AwningsActionId::stepIn) {
    return "app.stepIn";
  } else if (actionId == dsModel::AwningsActionId::stop) {
    return "app.stop";
  } else if (actionId == dsModel::AwningsActionId::sunProtection) {
    return "app.sunProtection";
  } else {
    DS_FAIL_REQUIRE("No scene for awnings action.", actionId);
  }
}

class DsScenarios final : public dsModel::ScenariosApi, public ds::asio::Resource<> {
  Apartment& m_apartment;
  StructureManipulator m_structureManipulator{m_apartment};
  PropertySystem& m_propertySystem = *m_apartment.getPropertySystem();
  PropertyNodePtr m_udasNode{m_propertySystem.createProperty("/scripts/system-addon-user-defined-actions/")};

public:
  DsScenarios(ds::asio::ResourceBase& parent, Apartment& apartment)
      : ds::asio::Resource<>{parent, "scenarios"}, m_apartment{apartment} {}

  // TODO(someday): move to Apartment
  auto group(const ApplicationZoneScenarioId& x) { return m_apartment.getZone(x.zone)->getGroup(x.group); }
  auto requireUdaNode(const UserDefinedActionId& x) {
    auto result = m_udasNode->getProperty(static_cast<std::string>(x));
    DS_REQUIRE(result, "Invalid UserDefinedActionId - unknown id.");
    return result;
  }
  auto requireDeviceScenario(const DeviceScenarioId& x) {
    const auto scenarioId = encodeId(x);
    auto& device = *m_apartment.device(x.device);
    const auto& scenarioIds = device.scenarioIds();
    if (std::find(scenarioIds.begin(), scenarioIds.end(), scenarioId) != scenarioIds.end()) {
      return x;
    }
    DS_FAIL_REQUIRE("Invalid DeviceScenarioId", x);
  }

  DecodedScenarioId requireScenarioId(const dsModel::ScenarioId& id) {
    const auto decodedId = decodeDecodeId(id);
    std::visit(ds::Visitor{
                   [&](const ApplicationZoneScenarioId& x) {
                     DS_REQUIRE(group(x)->isNamedScene(x.scene),
                                "Invalid ApplicationZoneScenarioId - not a named scene.", id);
                   },
                   [&](const UserDefinedActionId& x) { requireUdaNode(x); },
                   [&](const DeviceScenarioId& x) { requireDeviceScenario(x); },
               },
               decodedId);
    return decodedId;
  }

  [[nodiscard]] std::vector<dsModel::Scenario> applicationZoneScenarios() const override {
    std::vector<dsModel::Scenario> result{};
    for (const auto& zone : m_apartment.getZones()) {
      if (zone->getID() == 0) {
        continue;
      }
      for (const auto& scenario : zoneApplicationScenarioIds(*zone)) {
        auto scenarioDeviceList =
            m_apartment.proxy().scenarioDeviceList(*zone->getGroup(scenario.group), scenario.scene);
        result.push_back(dsModel::ApplicationZoneScenario{
            .id = encodeId(scenario),
            .type = "applicationZoneScenario",
            .lastChanged = {},
            .attributes = {
                .name = unescapeHTML(zone->getGroup(scenario.group)->getSceneName(scenario.scene)),
                .actionId = ds::toEnum<dsModel::CommonActionId>(scenario.scene),
                .context = "applicationZone", // only "applicationZone" is supported for backward compatibility
                .submodules = ds::makeVector(scenarioDeviceList | boost::adaptors::transformed([&](const auto& x) {
                                               return dsModel::SubmoduleId{x};
                                             })),
                .dsDevices = ds::makeVector(scenarioDeviceList | boost::adaptors::transformed([&](const auto& x) {
                                              return dsModel::DSDeviceId{x};
                                            })),
                .zone = zone->id(),
                .application = zoneApplication(scenario.group),
            }});
      }
    }
    boost::sort(result, [](const auto& a, const auto& b) { return idView(a) < idView(b); });
    return result;
  }

  [[nodiscard]] std::vector<dsModel::Scenario> userDefinedActions() const override {
    auto result = std::vector<dsModel::Scenario>{};
    for (auto&& udaNode : m_udasNode->getChildNodes()) {
      if (auto&& idNode = udaNode->getPropertyByName("id")) {
        if (auto&& nameName = udaNode->getPropertyByName("name")) {
          const auto id = idNode->getValue<std::string>();
          const auto name = nameName->getValue<std::string>();
          result.push_back(dsModel::UserDefinedAction{.id = encodeId(UserDefinedActionId{unescapeHTML(id)}),
                                                      .type = "userDefinedAction",
                                                      .lastChanged = {},
                                                      .attributes = {.name = unescapeHTML(name)}});
        }
      }
    }
    boost::sort(result, [](const auto& a, const auto& b) { return idView(a) < idView(b); });
    return result;
  }

  [[nodiscard]] std::vector<dsModel::Scenario> deviceScenarios() const override {
    std::vector<dsModel::Scenario> result;

    for (const auto& deviceId : m_apartment.deviceIds()) {
      const auto& device = m_apartment.device(deviceId);
      auto scenarios = device->scenarios();
      std::move(scenarios.begin(), scenarios.end(), std::back_inserter(result));
    }

    boost::sort(result, [](const auto& a, const auto& b) { return idView(a) < idView(b); });
    return result;
  }

  [[nodiscard]] std::vector<dsModel::Scenario> scenarios() const override {
    auto result = std::vector<dsModel::Scenario>{};
    boost::copy(applicationZoneScenarios(), std::back_inserter(result));
    boost::copy(userDefinedActions(), std::back_inserter(result));
    boost::copy(deviceScenarios(), std::back_inserter(result));
    boost::sort(result, [](const auto& a, const auto& b) { return idView(a) < idView(b); });
    return result;
  }

  /// specification: https://digitalstrom-zrh.atlassian.net/wiki/spaces/SA/pages/484671578/Standard+Scenarios
  void invokeStandardScenario(const dsModel::InvokeStandardScenarioArgs& args) override {
    DS_NOTICE("Invoke standard scenario", args);
    auto callScene = [](auto&& target, int scene, bool force = false) {
      target.callScene(CallOrigin::json, SceneAccessCategory::manual, scene, "", force);
    };
    auto undoScene = [](auto&& target, int scene) { target.undoScene(CallOrigin::json, SceneAccessCategory::manual, scene, ""); };

    auto zoneApplication = [&](Zone& zone) {
      if (static_cast<const std::string&>(args.application).empty()) {
        auto& group = *zone.getGroup(GroupIDBroadcast);
        auto actionId = ds::requireEnum<dsModel::ZoneActionId>(ds::str(args.actionId));
        group.callScene(CallOrigin::json, SceneAccessCategory::manual, zoneScene(actionId), "", false);
      } else {
        auto application = ds::requireEnum<dsModel::ZoneApplicationType>(ds::str(args.application));
        switch (application) {
          case dsModel::ZoneApplicationType::lights: {
            auto& group = *zone.getGroup(groupNumber(application));
            if (ds::tryToEnum<dsModel::CommonActionId>(ds::str(args.actionId)) != std::nullopt) {
              auto actionId = ds::requireEnum<dsModel::CommonActionId>(ds::str(args.actionId));
              group.callScene(CallOrigin::json, SceneAccessCategory::manual, commonScene(actionId), "", false);
            } else if (args.actionId == dsModel::ActionId{"forcedOff"}) {
              auto actionId = ds::requireEnum<dsModel::LightsActionId>(ds::str(args.actionId));
              group.callScene(CallOrigin::json, SceneAccessCategory::manual, lightsScene(actionId), "", true);
            } else {
              auto actionId = ds::requireEnum<dsModel::LightsActionId>(ds::str(args.actionId));
              group.callScene(CallOrigin::json, SceneAccessCategory::manual, lightsScene(actionId), "", false);
            }
            break;
          }
          case dsModel::ZoneApplicationType::shades: {
            auto& group = *zone.getGroup(groupNumber(application));
            if (ds::tryToEnum<dsModel::CommonActionId>(ds::str(args.actionId)) != std::nullopt) {
              auto actionId = ds::requireEnum<dsModel::CommonActionId>(ds::str(args.actionId));
              group.callSceneBlinds(CallOrigin::json, SceneAccessCategory::manual, commonScene(actionId), "", false);
            } else {
              auto actionId = ds::requireEnum<dsModel::ShadesActionId>(ds::str(args.actionId));
              // Shades actions need to use `callAction` function, therefore their
              // actionId needs to be translated to "app.moveUp", "app.moveDown", etc.
              auto groupCallActionId = shadesActionId(actionId);
              group.callAction(CallOrigin::json, SceneAccessCategory::manual, groupCallActionId);
            }
            break;
          }
          case dsModel::ZoneApplicationType::awnings: {
            auto& group = *zone.getGroup(groupNumber(application)); // This also returns `Blinds` group
            if (ds::tryToEnum<dsModel::CommonActionId>(ds::str(args.actionId)) != std::nullopt) {
              auto actionId = ds::requireEnum<dsModel::CommonActionId>(ds::str(args.actionId));
              group.callSceneBlinds(CallOrigin::json, SceneAccessCategory::manual, commonScene(actionId), "", false, true);
            } else {
              auto actionId = ds::requireEnum<dsModel::AwningsActionId>(ds::str(args.actionId));
              auto groupCallActionId = awningsActionId(actionId);
              group.callAction(CallOrigin::json, SceneAccessCategory::manual, groupCallActionId, true);
            }
            break;
          }
          case dsModel::ZoneApplicationType::audio: {
            auto& group = *zone.getGroup(groupNumber(application));
            if (ds::tryToEnum<dsModel::CommonActionId>(ds::str(args.actionId)) != std::nullopt) {
              auto actionId = ds::requireEnum<dsModel::CommonActionId>(ds::str(args.actionId));
              group.callScene(CallOrigin::json, SceneAccessCategory::manual, commonScene(actionId), "", false);
            } else {
              auto actionId = ds::requireEnum<dsModel::AudioActionId>(ds::str(args.actionId));
              group.callScene(CallOrigin::json, SceneAccessCategory::manual, audioScene(actionId), "", false);
            }
            break;
          }
          case dsModel::ZoneApplicationType::video: {
            auto& group = *zone.getGroup(groupNumber(application));
            if (ds::tryToEnum<dsModel::CommonActionId>(ds::str(args.actionId)) != std::nullopt) {
              auto actionId = ds::requireEnum<dsModel::CommonActionId>(ds::str(args.actionId));
              group.callScene(CallOrigin::json, SceneAccessCategory::manual, commonScene(actionId), "", false);
            } else {
              auto actionId = ds::requireEnum<dsModel::VideoActionId>(ds::str(args.actionId));
              group.callScene(CallOrigin::json, SceneAccessCategory::manual, videoScene(actionId), "", false);
            }
            break;
          }
          case dsModel::ZoneApplicationType::ventilation: {
            auto& group = *zone.getGroup(groupNumber(application));
            if (ds::tryToEnum<dsModel::CommonActionId>(ds::str(args.actionId)) != std::nullopt) {
              auto actionId = ds::requireEnum<dsModel::CommonActionId>(ds::str(args.actionId));
              group.callScene(CallOrigin::json, SceneAccessCategory::manual, commonScene(actionId), "", false);
            } else {
              auto actionId = ds::requireEnum<dsModel::VentilationActionId>(ds::str(args.actionId));
              group.callScene(CallOrigin::json, SceneAccessCategory::manual, ventilationScene(actionId), "", false);
            }
            break;
          }
          case dsModel::ZoneApplicationType::recirculation: {
            auto& group = *zone.getGroup(groupNumber(application));
            if (ds::tryToEnum<dsModel::CommonActionId>(ds::str(args.actionId)) != std::nullopt) {
              auto actionId = ds::requireEnum<dsModel::CommonActionId>(ds::str(args.actionId));
              group.callScene(CallOrigin::json, SceneAccessCategory::manual, commonScene(actionId), "", false);
            } else {
              auto actionId = ds::requireEnum<dsModel::RecirculationActionId>(ds::str(args.actionId));
              group.callScene(CallOrigin::json, SceneAccessCategory::manual, recirculationScene(actionId), "", false);
            }
            break;
          }
          case dsModel::ZoneApplicationType::security: [[fallthrough]];
          case dsModel::ZoneApplicationType::access: [[fallthrough]];
          case dsModel::ZoneApplicationType::heating: [[fallthrough]];
          case dsModel::ZoneApplicationType::cooling: [[fallthrough]];
          case dsModel::ZoneApplicationType::temperature: [[fallthrough]];
          case dsModel::ZoneApplicationType::window: [[fallthrough]];
          case dsModel::ZoneApplicationType::joker: [[fallthrough]];
          case dsModel::ZoneApplicationType::none: [[fallthrough]];
          default: DS_FAIL_REQUIRE("Cannot invoke application.", application, args);
        }
      }
    };

    const auto& context = args.context;
    if (context == dsModel::InvokeContextId{"applicationApartment"}) {
      DS_REQUIRE(static_cast<const std::string&>(args.zone).empty(), "Zone must be empty for apartment scenarios",
                 args);
      DS_REQUIRE(static_cast<const std::string&>(args.area).empty(), "Area must be empty for apartment scenarios",
                 args);
      auto callApartmentScene = [&](int scene, bool force = false) { callScene(*m_apartment.getGroup(0), scene, force); };
      auto undoApartmentScene = [&](int scene) { undoScene(*m_apartment.getGroup(0), scene); };
      if (auto application = ds::tryToEnum<dsModel::ApartmentApplicationType>(ds::str(args.application))) {
        auto actionId = ds::requireEnum<dsModel::VentilationActionId>(ds::str(args.actionId));
        switch (*application) {
          case dsModel::ApartmentApplicationType::none: break;
          case dsModel::ApartmentApplicationType::ventilationUnit:
            m_apartment.getGroup(64)->callScene(CallOrigin::json, SceneAccessCategory::manual, ventilationScene(actionId), "", false);
            return;
        }
        DS_FAIL_REQUIRE("No scene for apartment action.", actionId, args);
      } else if (auto actionId = ds::tryToEnum<dsModel::ApartmentActionType>(ds::str(args.actionId))) {
        if (!
            // `access` application used to be required for `absent` and `present`. Preserve compatibility.
            ((*actionId == dsModel::ApartmentActionType::absent &&
              args.application == dsModel::ApplicationId{"access"}) ||
             (*actionId == dsModel::ApartmentActionType::present &&
              args.application == dsModel::ApplicationId{"access"}))) {
          DS_REQUIRE(static_cast<const std::string&>(args.application).empty(),
                     "Application must be empty for apartment scenarios", args);
        }
        if (*actionId == dsModel::ApartmentActionType::frost) {
          m_apartment.pushEvent(std::make_shared<Event>(EventName::FrostProtection, Properties{{"value", "1"}}));
          return;
        } else if (*actionId == dsModel::ApartmentActionType::frostEnd) {
          m_apartment.pushEvent(std::make_shared<Event>(EventName::FrostProtection, Properties{{"value", "2"}}));
          return;
        }
        if (*actionId == dsModel::ApartmentActionType::fireMute) {
          m_apartment.getModelMaintenance()->muteFire(CallOrigin::unknown);
          return;
        }
        switch (*actionId) {
          case dsModel::ApartmentActionType::frost: [[fallthrough]];
          case dsModel::ApartmentActionType::frostEnd: [[fallthrough]];
          case dsModel::ApartmentActionType::fireMute: [[fallthrough]];
          case dsModel::ApartmentActionType::none: break;
          case dsModel::ApartmentActionType::absent: return callApartmentScene(SceneAbsent);
          case dsModel::ApartmentActionType::present: return callApartmentScene(ScenePresent);
          case dsModel::ApartmentActionType::sleeping: return callApartmentScene(SceneSleeping);
          case dsModel::ApartmentActionType::wakeup: return callApartmentScene(SceneWakeUp);
          case dsModel::ApartmentActionType::doorBell: return callApartmentScene(SceneBell);
          case dsModel::ApartmentActionType::panic: return callApartmentScene(ScenePanic);
          case dsModel::ApartmentActionType::panicEnd: return undoApartmentScene(ScenePanic);
          case dsModel::ApartmentActionType::fire: return callApartmentScene(SceneFire);
          case dsModel::ApartmentActionType::fireEnd: return undoApartmentScene(SceneFire);
          case dsModel::ApartmentActionType::alarm1: return callApartmentScene(SceneAlarm);
          case dsModel::ApartmentActionType::alarm2: return callApartmentScene(SceneAlarm2);
          case dsModel::ApartmentActionType::alarm3: return callApartmentScene(SceneAlarm3);
          case dsModel::ApartmentActionType::alarm4: return callApartmentScene(SceneAlarm4);
          case dsModel::ApartmentActionType::alarm1End: return undoApartmentScene(SceneAlarm);
          case dsModel::ApartmentActionType::alarm2End: return undoApartmentScene(SceneAlarm2);
          case dsModel::ApartmentActionType::alarm3End: return undoApartmentScene(SceneAlarm3);
          case dsModel::ApartmentActionType::alarm4End: return undoApartmentScene(SceneAlarm4);
          case dsModel::ApartmentActionType::smoke: return callApartmentScene(SceneSmoke);
          case dsModel::ApartmentActionType::smokeEnd: return undoApartmentScene(SceneSmoke);
          case dsModel::ApartmentActionType::water: return callApartmentScene(SceneWater);
          case dsModel::ApartmentActionType::waterEnd: return undoApartmentScene(SceneWater);
          case dsModel::ApartmentActionType::gas: return callApartmentScene(SceneGas);
          case dsModel::ApartmentActionType::gasEnd: return undoApartmentScene(SceneGas);
          case dsModel::ApartmentActionType::wind: return callApartmentScene(SceneWindActive);
          case dsModel::ApartmentActionType::windEnd: return callApartmentScene(SceneWindInactive);
          case dsModel::ApartmentActionType::rain: return callApartmentScene(SceneRainActive);
          case dsModel::ApartmentActionType::rainEnd: return callApartmentScene(SceneRainInactive);
          case dsModel::ApartmentActionType::hail: return callApartmentScene(SceneHailActive, true);
          case dsModel::ApartmentActionType::hailEnd: return callApartmentScene(SceneHailInactive, true);
          case dsModel::ApartmentActionType::pollution: return callApartmentScene(SceneHazardousAir);
          case dsModel::ApartmentActionType::pollutionEnd: return undoApartmentScene(SceneHazardousAir);
        }
        DS_FAIL_REQUIRE("No scene for apartment action.", actionId, args);
      }
      // all zone applications can be invoked also in the context of the whole apartment
      zoneApplication(*m_apartment.getZone(0));

    } else if (context == dsModel::InvokeContextId{"applicationZone"}) {
      auto zoneId = ds::requireInt(args.zone);
      DS_REQUIRE(static_cast<const std::string&>(args.area).empty(), "Area must be empty for apartment scenarios", args);
      DS_REQUIRE(zoneId != 0, "Zone must be provided for zone scenarios", args);
      zoneApplication(*m_apartment.getZone(zoneId));

    } else if (context == dsModel::InvokeContextId{"applicationArea"}) {
      auto zoneId = ds::requireInt(args.zone);
      DS_REQUIRE(zoneId != 0, "Zone must be provided for zone scenarios", args);
      auto& zone = *m_apartment.getZone(zoneId);
      auto area = ds::requireInt(args.area);
      auto application = ds::requireEnum<dsModel::AreaApplicationType>(ds::str(args.application));
      switch (application) {
        case dsModel::AreaApplicationType::lights: {
          auto actionId = ds::requireEnum<dsModel::LightsActionId>(ds::str(args.actionId));
          zone.getGroup(GroupIDYellow)->callScene(CallOrigin::json, SceneAccessCategory::manual, lightsAreaScene(actionId, area), "", false);
          break;
        }
        case dsModel::AreaApplicationType::shades: {
          auto actionId = ds::requireEnum<dsModel::ShadesActionId>(ds::str(args.actionId));
          zone.getGroup(GroupIDGray)->callScene(CallOrigin::json, SceneAccessCategory::manual, shadesAreaScene(actionId, area), "", false);
          break;
        }
        case dsModel::AreaApplicationType::none:
        default: DS_FAIL_REQUIRE("Unexpected area application.", application);
      }

    } else if (context == dsModel::InvokeContextId{"applicationDevice"}) {
      DS_REQUIRE(static_cast<const std::string&>(args.application).empty(),
                 "Application must be empty for device scenarios", args);
      DS_REQUIRE(static_cast<const std::string&>(args.area).empty(), "Area must be empty for device scenarios", args);
      DS_REQUIRE(static_cast<const std::string&>(args.zone).empty(), "Zone must be empty for device scenarios", args);
      auto actionId = ds::requireEnum<dsModel::DeviceActionId>(ds::str(args.actionId));
      auto& device = *m_apartment.getDeviceByDSID(str2dsuid(ds::str(args.dsDevice)));
      device.callScene(CallOrigin::json, SceneAccessCategory::manual, deviceScene(actionId), "", false);
    } else if (context == dsModel::InvokeContextId{"applicationCluster"}) {
      DS_REQUIRE(static_cast<const std::string&>(args.area).empty(), "Area must be empty for cluster scenarios", args);
      DS_REQUIRE(static_cast<const std::string&>(args.zone).empty(), "Zone must be empty for cluster scenarios", args);

      auto& cluster = *m_apartment.cluster(dsModel::ClusterId{args.cluster});
      DS_REQUIRE(cluster.isInDsModel(), "Cluster does not exist in apartment", args);

      auto application = ds::requireEnum<dsModel::ClusterApplicationType>(ds::str(args.application));
      switch (application) {
        case dsModel::ClusterApplicationType::lights: {
          DS_REQUIRE(cluster.getApplicationType() == ApplicationType::lights, "Cluster has different application type",
                     cluster.getApplicationType(), args);
          if (ds::tryToEnum<dsModel::CommonActionId>(ds::str(args.actionId)) != std::nullopt) {
            auto actionId = ds::requireEnum<dsModel::CommonActionId>(ds::str(args.actionId));
            cluster.callScene(CallOrigin::json, SceneAccessCategory::manual, commonScene(actionId), "", false);
          } else if (args.actionId == dsModel::ActionId{"forcedOff"}) {
            auto actionId = ds::requireEnum<dsModel::LightsActionId>(ds::str(args.actionId));
            cluster.callScene(CallOrigin::json, SceneAccessCategory::manual, lightsScene(actionId), "", true);
          } else {
            auto actionId = ds::requireEnum<dsModel::LightsActionId>(ds::str(args.actionId));
            cluster.callScene(CallOrigin::json, SceneAccessCategory::manual, lightsScene(actionId), "", false);
          }
          break;
        }
        case dsModel::ClusterApplicationType::shades: {
          DS_REQUIRE(cluster.getApplicationType() == ApplicationType::blinds, "Cluster has different application type",
                     cluster.getApplicationType(), args);
          if (ds::tryToEnum<dsModel::CommonActionId>(ds::str(args.actionId)) != std::nullopt) {
            auto actionId = ds::requireEnum<dsModel::CommonActionId>(ds::str(args.actionId));
            cluster.callSceneBlinds(CallOrigin::json, SceneAccessCategory::manual, commonScene(actionId), "", false);
          } else {
            auto actionId = ds::requireEnum<dsModel::ShadesActionId>(ds::str(args.actionId));
            auto groupCallActionId = shadesActionId(actionId);
            cluster.callAction(CallOrigin::json, SceneAccessCategory::manual, groupCallActionId);
          }
          break;
        }
        case dsModel::ClusterApplicationType::awnings: {
          DS_REQUIRE(cluster.getApplicationType() == ApplicationType::blinds, "Cluster has different application type",
                     cluster.getApplicationType(), args);
          if (ds::tryToEnum<dsModel::CommonActionId>(ds::str(args.actionId)) != std::nullopt) {
            auto actionId = ds::requireEnum<dsModel::CommonActionId>(ds::str(args.actionId));
            cluster.callSceneBlinds(CallOrigin::json, SceneAccessCategory::manual, commonScene(actionId), "", false, true);
          } else {
            auto actionId = ds::requireEnum<dsModel::AwningsActionId>(ds::str(args.actionId));
            auto groupCallActionId = awningsActionId(actionId);
            cluster.callAction(CallOrigin::json, SceneAccessCategory::manual, groupCallActionId, true);
          }
          break;
        }
        case dsModel::ClusterApplicationType::audio: {
          DS_REQUIRE(cluster.getApplicationType() == ApplicationType::audio, "Cluster has different application type",
                     cluster.getApplicationType(), args);
          if (ds::tryToEnum<dsModel::CommonActionId>(ds::str(args.actionId)) != std::nullopt) {
            auto actionId = ds::requireEnum<dsModel::CommonActionId>(ds::str(args.actionId));
            cluster.callScene(CallOrigin::json, SceneAccessCategory::manual, commonScene(actionId), "", false);
          } else {
            auto actionId = ds::requireEnum<dsModel::AudioActionId>(ds::str(args.actionId));
            cluster.callScene(CallOrigin::json, SceneAccessCategory::manual, audioScene(actionId), "", false);
          }
          break;
        }
        case dsModel::ClusterApplicationType::video: {
          DS_REQUIRE(cluster.getApplicationType() == ApplicationType::video, "Cluster has different application type",
                     cluster.getApplicationType(), args);
          if (ds::tryToEnum<dsModel::CommonActionId>(ds::str(args.actionId)) != std::nullopt) {
            auto actionId = ds::requireEnum<dsModel::CommonActionId>(ds::str(args.actionId));
            cluster.callScene(CallOrigin::json, SceneAccessCategory::manual, commonScene(actionId), "", false);
          } else {
            auto actionId = ds::requireEnum<dsModel::VideoActionId>(ds::str(args.actionId));
            cluster.callScene(CallOrigin::json, SceneAccessCategory::manual, videoScene(actionId), "", false);
          }
          break;
        }
        case dsModel::ClusterApplicationType::ventilation: {
          DS_REQUIRE(cluster.getApplicationType() == ApplicationType::ventilation,
                     "Cluster has different application set", cluster.getApplicationType(), args);
          if (ds::tryToEnum<dsModel::CommonActionId>(ds::str(args.actionId)) != std::nullopt) {
            auto actionId = ds::requireEnum<dsModel::CommonActionId>(ds::str(args.actionId));
            cluster.callScene(CallOrigin::json, SceneAccessCategory::manual, commonScene(actionId), "", false);
          } else {
            auto actionId = ds::requireEnum<dsModel::VentilationActionId>(ds::str(args.actionId));
            cluster.callScene(CallOrigin::json, SceneAccessCategory::manual, ventilationScene(actionId), "", false);
          }
          break;
        }
        case dsModel::ClusterApplicationType::recirculation: {
          DS_REQUIRE(cluster.getApplicationType() == ApplicationType::recirculation,
                     "Cluster has different application set", cluster.getApplicationType(), args);
          if (ds::tryToEnum<dsModel::CommonActionId>(ds::str(args.actionId)) != std::nullopt) {
            auto actionId = ds::requireEnum<dsModel::CommonActionId>(ds::str(args.actionId));
            cluster.callScene(CallOrigin::json, SceneAccessCategory::manual, commonScene(actionId), "", false);
          } else {
            auto actionId = ds::requireEnum<dsModel::RecirculationActionId>(ds::str(args.actionId));
            cluster.callScene(CallOrigin::json, SceneAccessCategory::manual, recirculationScene(actionId), "", false);
          }
          break;
        }
        case dsModel::ClusterApplicationType::heating: [[fallthrough]];
        case dsModel::ClusterApplicationType::cooling: [[fallthrough]];
        case dsModel::ClusterApplicationType::none: [[fallthrough]];
        default: DS_FAIL_REQUIRE("Unexpected cluster application.", application);
      }
    } else {
      DS_FAIL_REQUIRE("Unknown context for invoke scenario call.", context, args);
    }
  }

  dsModel::ScenarioId createScenario(const dsModel::Scenario& data) override {
    auto maybe = std::get_if<dsModel::ApplicationZoneScenario>(&data);
    DS_REQUIRE(maybe, "Scenario type must be 'applicationZoneScenario'", data);
    auto& attributes = maybe->attributes;
    DS_REQUIRE(!attributes.name.empty(), "New name cannot be empty.");

    auto& zone = *m_apartment.zone(attributes.zone);
    int groupNr = groupNumber(attributes.application);
    int scene = attributes.actionId == dsModel::CommonActionId{} ? zone.getGroup(groupNr)->availablePreset()
                                                                 : commonScene(attributes.actionId);
    const auto id = ApplicationZoneScenarioId{.zone = strToInt(attributes.zone), .group = groupNr, .scene = scene};
    DS_NOTICE("Create scenario", data, scene, id);
    const auto encodedId = encodeId(id);
    decodeId(encodedId); // Checks if scenarioId is correct, like zoneId != 0, etc.

    m_structureManipulator.sceneSetName(group(id), id.scene, attributes.name);

    return encodeId(encodedId);
  }

  void deleteScenario(const dsModel::ScenarioId& id) override {
    DS_NOTICE("Delete scenario", id);
    auto decodedScenarioId = DecodedScenarioId{};
    try {
      decodedScenarioId = requireScenarioId(id);
    } catch (const std::exception& e) {
      DS_WARNING("Deleting non-existing scenario.", id);
      return; // success if resource does not exist
    }
    std::visit(
        ds::Visitor{
            [&](const ApplicationZoneScenarioId& x) { m_structureManipulator.sceneSetName(group(x), x.scene, ""); },
            [&](const UserDefinedActionId& x) { DS_FAIL_REQUIRE("TODO(someday)."); },
            [&](const DeviceScenarioId& x) { DS_FAIL_REQUIRE("TODO(soon)."); },
        },
        decodedScenarioId);
  }

  void setScenarioName(const dsModel::ScenarioId& id, const std::string& value) override {
    DS_NOTICE("Set scenario name", id, value);
    DS_REQUIRE(!value.empty(), "New name cannot be empty.");
    std::visit(
        ds::Visitor{
            [&](const ApplicationZoneScenarioId& x) { m_structureManipulator.sceneSetName(group(x), x.scene, value); },
            [&](const UserDefinedActionId& x) { DS_FAIL_REQUIRE("TODO(someday)."); },
            [&](const DeviceScenarioId& x) { DS_FAIL_REQUIRE("TODO(someday)."); },
        },
        requireScenarioId(id));
  }

  void invokeScenario(const dsModel::ScenarioId& id) override {
    DS_NOTICE("Invoke scenario", id);
    std::visit(
        ds::Visitor{
            [&](const ApplicationZoneScenarioId& x) { group(x)->callScene(CallOrigin::json, SceneAccessCategory::manual, x.scene, "", false); },
            [&](const UserDefinedActionId& x) {
              m_apartment.pushEvent(std::make_shared<Event>(EventName::HighLevelEvent,
                                                              Properties{{"id", static_cast<std::string>(x)}}));
            },
            [&](const DeviceScenarioId& x) {
              m_apartment.device(x.device)->callAction(CallOrigin::unknown, SceneAccessCategory::unknown, ds::str(x.action), {}, {});
            },
        },
        requireScenarioId(id));
  }

  void persistScenario(const dsModel::ScenarioId& id) override {
    DS_NOTICE("Persist scenario", id);
    std::visit(
        ds::Visitor{
            [&](const ApplicationZoneScenarioId& x) {
              m_structureManipulator.saveScene(*group(x), CallOrigin::json, x.scene, {});
            },
            [&](const UserDefinedActionId& x) { DS_FAIL_REQUIRE("Persist of user defined action is not possible."); },
            [&](const DeviceScenarioId& x) { DS_FAIL_REQUIRE("Persist of device action is not possible."); },
        },
        requireScenarioId(id));
  }
};
} // namespace

std::vector<ApplicationZoneScenarioId> zoneApplicationScenarioIds(dss::Zone& zone) {
  std::vector<ApplicationZoneScenarioId> result{};
  for (const auto& group : zone.getGroups()) {
    if (group->getID() == 0) {
      continue;
    }
    for (const auto& scene : group->namedScenes()) {
      DS_CONTEXT(zone, *group, scene);
      try {
        if (!isCommonScenario(scene)) {
          continue;
        }
        result.push_back(
            ApplicationZoneScenarioId{.zone = group->getZoneID(), .group = group->getID(), .scene = scene});
      } catch (const std::exception& e) {
        DS_ERROR("Failed to get application zone scenario.");
      }
    }
  }
  return result;
}

std::vector<dsModel::ScenarioId> zoneScenarioIds(dss::Zone& zone) {
  std::vector<dsModel::ScenarioId> result;
  for (const auto& scenario : zoneApplicationScenarioIds(zone)) {
    result.push_back(encodeId(encodeId(scenario)));
  }
  return result;
}

ScenarioId decodeId(const dsModel::ScenarioId& id) {
  std::stringstream ss(id);
  std::string strType;
  std::getline(ss, strType, '-');

  std::string strId;
  std::getline(ss, strId);
  DS_REQUIRE(!strId.empty(), "Wrong scenario ID format", id, strId);

  if (strType == "applicationZone") {
    return dsModel::ApplicationZoneScenarioId{id};
  }
  if (strType == "userDefined") {
    return dsModel::UserDefinedActionId{id};
  }
  if (strType == "device") {
    return dsModel::DeviceScenarioId{id};
  }
  DS_FAIL_REQUIRE("Wrong scenario ID type");
}

DecodedScenarioId decodeDecodeId(const dsModel::ScenarioId& id) {
  return std::visit(ds::Visitor{[&](const std::monostate&) -> DecodedScenarioId {
                                  DS_FAIL_REQUIRE("Invalid ApplicationZoneScenarioId.", id);
                                },
                                [&](const auto& x) -> DecodedScenarioId { return decodeId(x); }},
                    decodeId(id));
}

dsModel::ScenarioId encodeId(const ScenarioId& id) {
  return dsModel::ScenarioId{std::visit(ds::Visitor{[](const std::monostate& x) { return dsModel::ScenarioId{}; },
                                                    [](const auto& x) { return dsModel::ScenarioId{x}; }},
                                        id)};
}

ApplicationZoneScenarioId decodeId(const dsModel::ApplicationZoneScenarioId& id) {
  std::stringstream ss(id);
  std::string strType;
  std::getline(ss, strType, '-');
  DS_REQUIRE(strType == "applicationZone", "Wrong scenario ID format");

  std::string strZone;
  std::getline(ss, strZone, '-');
  DS_REQUIRE(boost::starts_with(strZone, "z"), "Wrong scenario ID format", id);
  auto zone = strToInt(strZone.substr(1));
  DS_REQUIRE(zone != 0, "Cannot use scenario ID for zone 0");

  std::string strGroup;
  std::getline(ss, strGroup, '-');
  DS_REQUIRE(boost::starts_with(strGroup, "g"), "Wrong scenario ID format", id);
  auto group = strToInt(strGroup.substr(1));
  DS_REQUIRE(group != 0, "Cannot use scenario ID for group 0");

  std::string strScene;
  std::getline(ss, strScene);
  DS_REQUIRE(boost::starts_with(strScene, "s"), "Wrong scenario ID format", id);
  auto scene = strToInt(strScene.substr(1));
  DS_REQUIRE(isCommonScenario(scene), "Wrong scenario ID - not a common scene", scene);
  return ApplicationZoneScenarioId{.zone = zone, .group = group, .scene = scene};
}

dsModel::ApplicationZoneScenarioId encodeId(const ApplicationZoneScenarioId& x) {
  return dsModel::ApplicationZoneScenarioId{ds::str("applicationZone-z", x.zone, "-g", x.group, "-s", x.scene)};
}

UserDefinedActionId decodeId(const dsModel::UserDefinedActionId& id) {
  std::stringstream ss(id);
  std::string strType;
  std::getline(ss, strType, '-');
  DS_REQUIRE(strType == "userDefined", "Wrong scenario ID format");

  std::string strId;
  std::getline(ss, strId, '-');
  DS_REQUIRE(!strId.empty(), "Wrong scenario ID format", id);
  return UserDefinedActionId{strId};
}

dsModel::UserDefinedActionId encodeId(const UserDefinedActionId& x) {
  return dsModel::UserDefinedActionId{ds::str("userDefined-", x)};
}

DeviceScenarioId decodeId(const dsModel::DeviceScenarioId& id) {
  std::stringstream ss(id);
  std::string strType;
  std::getline(ss, strType, '-');
  DS_REQUIRE(strType == "device", "Wrong scenario ID format");

  std::string strId;
  std::getline(ss, strId, '-');
  DS_REQUIRE(!strId.empty(), "Wrong scenario ID format", id);
  auto deviceId = dsModel::DSDeviceId{strId};

  std::string strAction;
  std::getline(ss, strAction, '-');
  DS_REQUIRE(!strAction.empty(), "Wrong scenario ID format", id);
  return DeviceScenarioId{.device = deviceId, .action = dsModel::DeviceScenarioActionId{strAction}};
}

dsModel::DeviceScenarioId encodeId(const DeviceScenarioId& x) {
  return dsModel::DeviceScenarioId{ds::str("device-", x.device, "-", x.action)};
}

std::unique_ptr<dsModel::ScenariosApi> makeDsScenarios(ds::asio::ResourceBase& parent, Apartment& apartment) {
  return std::make_unique<DsScenarios>(parent, apartment);
}

} // namespace dss
