/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "resource.h"

#include <ds/vector.h>

#include <boost/algorithm/string.hpp>
#include <boost/range/adaptor/filtered.hpp>
#include <boost/range/adaptor/transformed.hpp>
#include <boost/range/algorithm/find_if.hpp>
#include <boost/range/algorithm_ext/erase.hpp>
#include <regex>

static ds::LogChannel dsLogChannel("dsAsioResource");

namespace ds::asio {
ds::LogChannel& resourceLogChannel = dsLogChannel;

/// Global registry of resource context.
///
/// No synchronization, no removal. It is populated at startup and then constant.
std::vector<std::reference_wrapper<ResourceContext>>& resourceContexts() {
  static std::vector<std::reference_wrapper<ResourceContext>> instance;
  return instance;
}

ds::asio::ResourceContext& resourceContext(std::string_view path) {
  for (const auto& context : resourceContexts()) {
    if (context.get().invoke([&]() -> bool {
          return boost::find_if<boost::return_found_end>(
              context.get().resources(), [&](const auto& x) { return boost::starts_with(path, x.get().path()); });
        })) {
      return context;
    }
  }
  DS_FAIL_REQUIRE("Resource context not found.", path);
}

ResourceContext::ResourceContext(ResourceContextData& data, boost::asio::io_service& io, ds::System& system,
                                 std::thread::id threadId)
    : m_data{data},
      io{io},
      system{system},
      threadId{threadId},
      m_root{new ResourceBase{ResourceBase::Private{}, *this, "", ""}} {
  resourceContexts().push_back(*this);
}

ResourceContext::ResourceContext(ResourceContextData& data, boost::asio::io_service& io, ds::System& system)
    : ResourceContext(data, io, system, ds::asio::invoke(io, [] { return std::this_thread::get_id(); })) {}

ResourceContext::~ResourceContext() = default;

bool ResourceContext::isSynchronized() const {
  return std::this_thread::get_id() == threadId;
};

void ResourceContext::assertSynchronized() const {
  if (!isSynchronized()) {
    if (!std::uncaught_exceptions()) {
      // This assert helps debugging.
      // But it is too common that exception paths are not correctly synchronized.
      // And we lose the orinal exception that is the problem being searched.
      DS_FAIL_ASSERT("Execution is not synchronized to resource thread.");
    } else {
      DS_ERROR("Execution is not synchronized to resource thread.", std::uncaught_exceptions());
    }
  }
}

bool ResourceContext::isBlockingAllowed() const {
  return !isSynchronized();
}

void ResourceContext::assertBlockingAllowed() const {
  DS_ASSERT(isBlockingAllowed(), "Blocking is not allowed when execution is synchronized to resource thread.");
}

const std::vector<std::reference_wrapper<ResourceBase>>& ResourceContext::resources() {
  assertSynchronized();
  return m_data.resources;
}

void ResourceContext::addResource(ResourceBase& resource) {
  m_data.resources.push_back(resource);
}

void ResourceContext::removeResource(ResourceBase& resource) {
  boost::remove_erase_if(m_data.resources, [&](const auto& x) { return &x.get() == &resource; });
}

const ResourceInside& IResourceBase::inside() {
  static constexpr ds::asio::ResourceInside instance{};
  return instance;
}

ds::JsonFragment IResourceBase::jsonConfig() const {
  return {};
}

void IResourceBase::assertSynchronized() const {
  DS_CONTEXT(*this);
  context().assertSynchronized();
}

void IResourceBase::assertBlockingAllowed() const {
  DS_CONTEXT(*this);
  context().assertBlockingAllowed();
}

std::vector<std::string> IResourceBase::childIds() const {
  assertSynchronized();
  const auto m_path = this->path();
  const auto pathSize = m_path.size();
  return ds::makeVectorNoReserve(
      context().resources() | boost::adaptors::filtered([&](const auto& x) {
        const auto& path = x.get().path();
        return boost::starts_with(path, m_path) && path.size() > m_path.size() &&
               path.find('/', m_path.size() + 1) == std::string::npos;
      }) |
      boost::adaptors::transformed([&](const auto& x) -> std::string { return x.get().path().substr(pathSize + 1); }));
}

std::ostream& operator<<(std::ostream& ostream, const IResourceBase& x) {
  return ostream << x.path();
}

ResourceBase::ResourceBase(ResourceBase& parent, std::string&& id)
    : ResourceBase(Private{}, parent.context(), composePath(parent.path(), id), std::move(id)) {
  if (!isIdNormalized(m_id)) {
    // TODO(someday): DS_ASSERT?
    DS_WARNING("Special characters in id are not allowed", m_id, *this);
    DS_FAIL_REQUIRE("Special characters in id are not allowed", m_id, *this);
  }
}

ResourceBase::ResourceBase(Private, ResourceContext& context, std::string&& path, std::string&& id)
    : m_context{context}, m_path{std::move(path)}, m_id{std::move(id)} {
  if (!m_path.empty()) {
    // root resource is not synchronized and not in resources
    assertSynchronized();
    m_context.addResource(*this);
  }
}

ResourceBase::~ResourceBase() {
  DS_INFO_ENTER(path(), std::uncaught_exceptions());
  if (!m_path.empty()) {
    // root resource is not synchronized and not in resources
    assertSynchronized();
    m_context.removeResource(*this);
  }
}

bool ResourceBase::isIdNormalized(const std::string& id) {
  static const auto regex = std::regex{"[\\w-]*"};
  return std::regex_match(id, regex);
}

std::string ResourceBase::normalizeId(const std::string& id) {
  static const auto regex = std::regex{"\\W"};
  return std::regex_replace(id, regex, "-");
}

std::string ResourceBase::composePath(std::string_view parentPath, const std::string& id) {
  return ds::str(parentPath, "/", id);
}

} // namespace ds::asio

DS_JSON_DEFINE_STRUCT((ds::asio), EmptyResourceConfig);
