#include <ds/catch/catch.h>
#include <ds/log.h>
#include <boost/asio/wait_traits.hpp>
#include <cmath>
#include "clock-mock.h"

namespace {

static const char* TAGS = "[dsClock][dsAsio][ds]";

using namespace std::literals;

SCENARIO("dsClock", TAGS) {
  THEN("is not steady") { REQUIRE(!ds::Clock::is_steady); }

  THEN("Clock::now equals to steady_clock") {
    REQUIRE((std::chrono::steady_clock::now().time_since_epoch() - ds::Clock::now().time_since_epoch() + 1s) < 2s);
  }

  THEN("ClockWaitTraits equals to steady_clock") {
    REQUIRE(ds::ClockWaitTraits::to_wait_duration(2s) ==
            boost::asio::wait_traits<std::chrono::steady_clock>::to_wait_duration(2s));
  }

  WHEN("ClockMocker is activated") {
    ds::ClockMocker clockMocker;

    THEN("clock is reset to 0") { REQUIRE(ds::Clock::now().time_since_epoch() == 0s); }

    THEN("clock can be advanced") {
      auto start = ds::Clock::now();
      clockMocker.advance(1s);
      auto diff = ds::Clock::now() - start;
      REQUIRE(diff == 1s);
    }

    THEN("ClockWaitTraits is low number for a few calls") {
      REQUIRE(ds::ClockWaitTraits::to_wait_duration(1ms) < 1ms);
      REQUIRE(ds::ClockWaitTraits::to_wait_duration(1ms) < 1ms);
      REQUIRE(ds::ClockWaitTraits::to_wait_duration(1ms) < 1ms);
      REQUIRE(ds::ClockWaitTraits::to_wait_duration(1ms) < 1ms);
    }

    GIVEN("manyToWaitDurationCalls") {
      constexpr auto manyToWaitDurationCalls = 100;

      THEN("ClockWaitTraits is low if called many times with advance.") {
        for (auto i = 0; i < manyToWaitDurationCalls; i++) {
          clockMocker.advance({});
          REQUIRE(ds::ClockWaitTraits::to_wait_duration(1ms) < 1ms);
        }
      }
    }
  }

  WHEN("ClockMocker is activated and deactivated") {
    { ds::ClockMocker clockMocker; }

    THEN("Clock::now equals to steady_clock::now") {
      REQUIRE((std::chrono::steady_clock::now().time_since_epoch() - ds::Clock::now().time_since_epoch() + 1s) < 2s);
    }

    THEN("ClockWaitTraits equals to steady_clock") {
      REQUIRE(ds::ClockWaitTraits::to_wait_duration(2s) ==
              boost::asio::wait_traits<std::chrono::steady_clock>::to_wait_duration(2s));
    }
  }
}

} // namespace
