/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "str.h"
#include <ds/catch/catch.h>
#include <boost/serialization/strong_typedef.hpp>
#include <locale>
#include <set>
#include <stdexcept>
#include "log.h"
#include "reflect.h"


namespace {
static const char* TAGS = "[dsStr][ds]";
using namespace std::literals;

BOOST_STRONG_TYPEDEF(std::string, Id);

enum class Enum { A, B };
std::ostream& operator<<(std::ostream& stream, Enum x) {
  switch (x) {
    case Enum::A:
      return stream << "a";
    case Enum::B:
      return stream << "b";
  }
  return stream << "?";
}

struct X { int i; char c; unsigned char uc; std::string s;};
struct Y {
  X x;
  std::vector<X> xs;
  Enum enum_;
};

} // namespace

DS_REFLECT_ANONYMOUS_STRUCT(X, (i)(c)(uc)(s));
DS_REFLECT_ANONYMOUS_STRUCT(Y, (x)(xs)(enum_));

namespace {

class Clock {
public:
  using rep = int;
  using period = std::milli;
  using duration = std::chrono::duration<rep, period>;
  using time_point = std::chrono::time_point<Clock>;
  static inline time_point now() noexcept { return time_point{500ms}; }
};

TEST_CASE("dsStr", TAGS) {
  SECTION("all arguments are serialized and concatenated") {
    CHECK(ds::str() == "");
    CHECK(ds::str("a") == "a");
    CHECK(ds::str("a", "b") == "ab");
    CHECK(ds::str("a", 4, "b") == "a4b");
    CHECK(ds::str("a", -4, "b") == "a-4b");
  }

  SECTION("ostream operator << is used for serialization") { CHECK(ds::str("-", Enum::A, Enum::B, "-") == "-ab-"); }

  SECTION("unsigned and signed chars are serialized as numbers") {
    CHECK(ds::str(static_cast<unsigned char>('a')) == "97");
    CHECK(ds::str(static_cast<signed char>('a')) == "97");
    CHECK(ds::str(static_cast<uint8_t>('a')) == "97");
    CHECK(ds::str(static_cast<int8_t>('a')) == "97");
  }

  SECTION("chars are serialized as characters") {
    CHECK(ds::str('a') == "a");
    CHECK(ds::str(static_cast<char>('a')) == "a");
  }

  SECTION("std::unique_ptr is serialized as value or null") {
    CHECK(ds::str(std::make_unique<int>(7)) == "7");
    CHECK(ds::str(std::unique_ptr<int>()) == "null");
  }

  SECTION("std::shared_ptr is serialized as value or null") {
    CHECK(ds::str(std::make_shared<int>(7)) == "7");
    CHECK(ds::str(std::shared_ptr<int>()) == "null");
  }

  SECTION("std::shared_ptr is serialized as value or null") {
    CHECK(ds::str(std::make_shared<int>(7)) == "7");
    CHECK(ds::str(std::shared_ptr<int>()) == "null");
  }

  SECTION("std::optional is serialized as value or null") {
    CHECK(ds::str(std::optional<int>{7}) == "7");
    CHECK(ds::str(std::optional<int>{}) == "null");
  }

  SECTION("std:exceptions references are serialized with what()") {
    auto e = std::runtime_error("a");
    CHECK(ds::str(static_cast<const std::exception&>(e)) == "a");
    CHECK(ds::str(static_cast<std::exception&>(e)) == "a");
  }

  SECTION("boost::system::error_code is serialized as message") {
    CHECK(ds::str(make_error_code(boost::system::errc::timed_out)) == "Connection timed out");
  }

  SECTION("std::chrono::duration is serialized as number in seconds") { CHECK(ds::str(500ms) == "0.5"); }

  SECTION("std::chrono::time_point is serialized as duration since epoch") { CHECK(ds::str(Clock::now()) == "0.5"); }

  SECTION("std::vector<T> is serialized") {
    CHECK(ds::str(std::vector<int>()) == "[]");
    CHECK(ds::str(std::vector<int>({1})) == "[1]");
    CHECK(ds::str(std::vector<int>({1, 2})) == "[1,2]");
    CHECK(ds::str(std::vector<std::string>()) == "[]");
    CHECK(ds::str(std::vector<std::string>({"a"})) == "[a]");
    CHECK(ds::str(std::vector<std::string>({"a", "b"})) == "[a,b]");
    CHECK(ds::str(std::vector<std::vector<int>>({{1}, {}})) == "[[1],[]]");
  }

  SECTION("std::set<T> is serialized") {
    CHECK(ds::str(std::set<int>()) == "[]");
    CHECK(ds::str(std::set<int>({1})) == "[1]");
    CHECK(ds::str(std::set<int>({2, 1})) == "[1,2]");
    CHECK(ds::str(std::set<std::string>()) == "[]");
    CHECK(ds::str(std::set<std::string>({"a"})) == "[a]");
    CHECK(ds::str(std::set<std::string>({"a", "b"})) == "[a,b]");
  }

  SECTION("std::pair<T1, T2> is serialized") {
    CHECK(ds::str(std::make_pair(0, "a")) == "(0,a)");
    CHECK(ds::str(std::make_pair(std::make_pair(1, 2), "a")) == "((1,2),a)");
  }

  SECTION("struct with reflection is serialized") {
    X x = {};
    x.i = 8;
    x.c = 'f';
    x.uc = 9;
    x.s = "hi";
    Y y = {};
    y.x = x;
    y.xs.push_back(x);
    y.xs.push_back(x);
    y.xs.back().i = 0;
    y.enum_ = Enum::B;

    CHECK(ds::str(y) ==
          R"({"x":{"i":8,"c":f,"uc":9,"s":hi},"xs":[{"i":8,"c":f,"uc":9,"s":hi},{"c":f,"uc":9,"s":hi}],"enum_":b})");
  }

  SECTION("(const char *) nullptr is serialized as null") { CHECK(ds::str((const char*)nullptr) == "null"); }

  SECTION("BOOST_STRONG_TYPEDEF(std::string, Id) id is serialized as string") { CHECK(ds::str(Id{"hi"}) == "hi"); }

  SECTION("intToString") {
    CHECK(ds::intToString(10) == "10");
    CHECK(ds::intToString(10, 10) == "10");
    CHECK(ds::intToString(10, 16) == "0xa");
    CHECK_NOTHROW(ds::intToString(std::numeric_limits<long long int>::max()));
    CHECK_NOTHROW(ds::intToString(std::numeric_limits<long long int>::min()));
  }

  SECTION("test tryToInt") {
    CHECK(0 == ds::tryToInt("0"));
    CHECK(0 == ds::tryToInt("-0"));
    CHECK(1 == ds::tryToInt("1"));
    CHECK(-1 == ds::tryToInt("-1"));
    CHECK(std::numeric_limits<int>::max() == ds::tryToInt(ds::intToString(std::numeric_limits<int>::max())));
    CHECK(std::numeric_limits<int>::min() == ds::tryToInt(ds::intToString(std::numeric_limits<int>::min())));

    CHECK(!ds::tryToInt("asdf"));
    CHECK(!ds::tryToInt(""));
    CHECK(!ds::tryToInt(" "));
    CHECK(!ds::tryToInt(" 1"));
    CHECK(!ds::tryToInt("\t1"));
    CHECK(0 == ds::tryToInt("00"));
    CHECK(1 == ds::tryToInt("01"));
    CHECK(10 == ds::tryToInt("010"));
    CHECK(0 == ds::tryToInt("-00"));
    CHECK(-1 == ds::tryToInt("-01"));
    CHECK(-10 == ds::tryToInt("-010"));
    CHECK(!ds::tryToInt("1a"));
    CHECK(!ds::tryToInt("0x10"));
  }

  SECTION("requireInt") {
    CHECK(1 == ds::requireInt("1"));
    CHECK_THROWS_FIND(ds::requireInt("asdf"), "Conversion to integer failed.");
  }

  SECTION("test tryToDouble") {
    CHECK(0 == *ds::tryToDouble("0"));
    CHECK(0 == *ds::tryToDouble("-0"));
    CHECK(1 == *ds::tryToDouble("1"));
    CHECK(-1 == *ds::tryToDouble("-1"));
    CHECK(10 == *ds::tryToDouble("1E1"));
    CHECK(1.5 == *ds::tryToDouble("1.5"));
    CHECK(0.5 == *ds::tryToDouble("0.5"));
    CHECK(0.0 == *ds::tryToDouble("0e1"));
    CHECK(0.0 == *ds::tryToDouble("0E1"));
    CHECK(5.0 == *ds::tryToDouble("0.5e1"));
    CHECK(5.0 == *ds::tryToDouble("0.5E1"));
    CHECK(-1.5 == *ds::tryToDouble("-1.5"));
    CHECK(-0.5 == *ds::tryToDouble("-0.5"));
    CHECK(-0.0 == *ds::tryToDouble("-0e1"));
    CHECK(-0.0 == *ds::tryToDouble("-0E1"));
    CHECK(-5.0 == *ds::tryToDouble("-0.5e1"));
    CHECK(-5.0 == *ds::tryToDouble("-0.5E1"));

    CHECK(ds::tryToDouble("asdf") == std::nullopt);
    CHECK(ds::tryToDouble("") == std::nullopt);
    CHECK(ds::tryToDouble(" ") == std::nullopt);
    CHECK(ds::tryToDouble(" 1") == std::nullopt);
    CHECK(ds::tryToDouble("\t1") == std::nullopt);
    CHECK(ds::tryToDouble("00") == std::nullopt);
    CHECK(ds::tryToDouble("01") == std::nullopt);
    CHECK(ds::tryToDouble("1a") == std::nullopt);
    CHECK(ds::tryToDouble("0x10") == std::nullopt);
    CHECK(ds::tryToDouble("0X10") == std::nullopt);
    CHECK(ds::tryToDouble("0x") == std::nullopt);
    CHECK(ds::tryToDouble("0X") == std::nullopt);
    CHECK(ds::tryToDouble("1.5.3") == std::nullopt);
  }

  SECTION("requireDouble") {
    CHECK(1 == ds::requireDouble("1"));
    CHECK_THROWS_FIND(ds::requireDouble("asdf"), "Conversion to double failed.");
  }

  SECTION("intToString") {
    CHECK(ds::intToString(10) == "10");
    CHECK(ds::intToString(10, 10) == "10");
    CHECK(ds::intToString(10, 16) == "0xa");
    CHECK_NOTHROW(ds::intToString(std::numeric_limits<long long int>::max()));
    CHECK_NOTHROW(ds::intToString(std::numeric_limits<long long int>::min()));
  }
}

} // namespace
