/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <unistd.h>

#include "logger.h"

#include "ds485-modify.h"

#include "ds485/dsbusinterface.h"
#include "vdc-messages.pb.h"
#include "model/apartment.h"
#include "model/dsmeter.h"

#define BROADCAST_SLEEP_MICROSECONDS 50000          // 50ms
#define BROADCAST_CLUSTER_SLEEP_MICROSECONDS 150000 // 150ms

static ds::LogChannel dsLogChannel("ds485Modify");

namespace dss {

//================================================== DSStructureModifyingBusInterface

std::string DSStructureModifyingBusInterface::name() const { return dsLogChannel.name(); }

void DSStructureModifyingBusInterface::setZoneID(const dsuid_t& dsm, const devid_t device, const int zone) {
  DS_CONTEXT(dsm, device, zone);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  std::shared_ptr<DSMeter> pMeter = DSS::getInstance()->getApartment().getDSMeterByDSID(dsm);
  int ret;
  if (pMeter->getApiVersion() >= 0x301) {
    ret = DeviceProperties_set_zone_sync(m_DSMApiHandle, dsm, device, zone, 30);
  } else {
    ret = DeviceProperties_set_zone(m_DSMApiHandle, dsm, device, zone);
  }
  DS_NOTICE("Set device zone.", ret);
  DSBusInterface::checkResultCode(ret);
} // setZoneID

void DSStructureModifyingBusInterface::createZone(const dsuid_t& dsm, const int zone) {
  DS_CONTEXT(dsm, zone);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  int ret = ZoneModify_add(m_DSMApiHandle, dsm, zone);
  DS_NOTICE("Create zone.", ret);
  DSBusInterface::checkResultCode(ret);
} // createZone

void DSStructureModifyingBusInterface::removeZone(const dsuid_t& dsm, const int zone) {
  DS_CONTEXT(dsm, zone);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  DSS::getInstance()->assertBlockingAllowed();
  int ret = ZoneModify_remove(m_DSMApiHandle, dsm, zone);
  DS_NOTICE("Remove zone.", ret);

  if (dsm == DSUID_BROADCAST) {
    DSBusInterface::checkBroadcastResultCode(ret);
  } else {
    DSBusInterface::checkResultCode(ret);
  }
} // removeZone

void DSStructureModifyingBusInterface::addToGroup(const dsuid_t& dsm, const int group, const int device) {
  DS_CONTEXT(dsm, group, device);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  std::shared_ptr<DSMeter> pMeter = DSS::getInstance()->getApartment().getDSMeterByDSID(dsm);
  int ret = 0;
  if (pMeter->getApiVersion() >= 0x301) {
    ret = DeviceGroupMembershipModify_add_sync(m_DSMApiHandle, dsm, device, group, 30);
  } else {
    ret = DeviceGroupMembershipModify_add(m_DSMApiHandle, dsm, device, group);
  }
  DS_NOTICE("Add to group.", ret);
  sleep(1); // #2578 prevent dS485 bus flooding with multiple requests
  DSBusInterface::checkResultCode(ret);
} // addToGroup

void DSStructureModifyingBusInterface::removeFromGroup(const dsuid_t& dsm, const int group, const int device) {
  DS_CONTEXT(dsm, group, device);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  if (dsm == DSUID_NULL) {
    throw BusApiError("dsm is NULL");
  }
  int ret = DeviceGroupMembershipModify_remove(m_DSMApiHandle, dsm, device, group);
  DS_NOTICE("Remove from group.", ret);
  DSBusInterface::checkResultCode(ret);
} // removeFromGroup

void DSStructureModifyingBusInterface::removeDeviceFromDSMeter(const dsuid_t& dsm, const int device) {
  DS_CONTEXT(dsm, device);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  if (dsm == DSUID_NULL) {
    return;
  }

  int ret = CircuitRemoveDevice_by_id(m_DSMApiHandle, dsm, device);
  DS_NOTICE("Remove device from dsm.", ret);
  DSBusInterface::checkResultCode(ret);
} // removeDeviceFromDSMeter

void DSStructureModifyingBusInterface::removeDeviceFromDSMeters(const dsuid_t& device) {
  DS_CONTEXT(device);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  DSS::getInstance()->assertBlockingAllowed();
  int ret = CircuitRemoveDevice_by_dsuid(m_DSMApiHandle, DSUID_BROADCAST, device);
  DS_NOTICE("Remove device from dsms.", ret);
  DSBusInterface::checkBroadcastResultCode(ret);
} // removeDeviceFromDSMeters

void DSStructureModifyingBusInterface::sceneSetName(
    uint16_t zone, uint8_t group, uint8_t scene, const std::string& name) {
  DS_CONTEXT(zone, group, scene, name);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  std::string nameStr = truncateUTF8String(name, 20);
  DSS::getInstance()->assertBlockingAllowed();
  int ret = ZoneGroupSceneProperties_set_name(
      m_DSMApiHandle, DSUID_BROADCAST, zone, group, scene, (unsigned char*)nameStr.c_str());
  DS_NOTICE("Set scene name.", ret);
  DSBusInterface::checkBroadcastResultCode(ret);
} // sceneSetName

void DSStructureModifyingBusInterface::deviceSetName(dsuid_t dsm, devid_t device, const std::string& name) {
  DS_CONTEXT(dsm, device, name);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  std::string nameStr = truncateUTF8String(name, 20);
  int ret = DeviceProperties_set_name(m_DSMApiHandle, dsm, device, (unsigned char*)nameStr.c_str());
  DS_NOTICE("Set device name.", ret);
  DSBusInterface::checkResultCode(ret);
} // deviceSetName

void DSStructureModifyingBusInterface::meterSetName(dsuid_t dsm, const std::string& name) {
  DS_CONTEXT(dsm, name);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  std::string nameStr = truncateUTF8String(name, 20);
  int ret = dSMProperties_set_name(m_DSMApiHandle, dsm, (unsigned char*)nameStr.c_str());
  DS_NOTICE("Set meter name.", ret);
  DSBusInterface::checkResultCode(ret);
} // meterSetName

void DSStructureModifyingBusInterface::setProperty(
    const dsuid_t& dsm, const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement>& properties) {
  return setVdcProperty(dsm, dsm, properties);
}

vdcapi::Message DSStructureModifyingBusInterface::getProperty(
    const dsuid_t& dsm, const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement>& query) {
  return getVdcProperty(dsm, dsm, query);
}

void DSStructureModifyingBusInterface::setFireMuteEnabled(dsuid_t dsm, bool value) {
  DS_CONTEXT(dsm, value);
  DS_INFO_ENTER("");
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  uint8_t oldFlags;
  DSBusInterface::checkResultCode(dSMProperties_get_flags(m_DSMApiHandle, dsm, &oldFlags));
  auto flags = [&] {
    std::bitset<8> x{oldFlags};
    x[5] = value;
    return x.to_ulong();
  }();
  if (flags != oldFlags) {
    DS_NOTICE("Set fire mute enabled.", flags, oldFlags);
    DSBusInterface::checkResultCode(dSMProperties_set_flags(m_DSMApiHandle, dsm, flags));
  }
}

void DSStructureModifyingBusInterface::removeVdc(const dsuid_t& vdsmID, const dsuid_t& vdcID) {
  DS_CONTEXT(vdsmID, vdcID);
  DS_INFO_ENTER("");
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  DSBusInterface::checkResultCode(RemoveVdc(m_DSMApiHandle, vdsmID, vdcID));
}

void DSStructureModifyingBusInterface::createGroup(uint16_t zone, uint8_t group, ApplicationType applicationType,
    uint32_t applicationConfig, const std::string& name) {
  DS_CONTEXT(zone, group, applicationType, applicationConfig, name);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  DS_NOTICE("Create group.");
  DSS::getInstance()->assertBlockingAllowed();
  int ret = ZoneGroupModify_add(m_DSMApiHandle, DSUID_BROADCAST, zone, group, static_cast<int>(applicationType));
  DSBusInterface::checkBroadcastResultCode(ret);
  usleep(BROADCAST_SLEEP_MICROSECONDS);

  ret = ZoneGroupProperties_set_application(
      m_DSMApiHandle, DSUID_BROADCAST, zone, group, static_cast<int>(applicationType), applicationConfig);
  DSBusInterface::checkBroadcastResultCode(ret);
  usleep(BROADCAST_SLEEP_MICROSECONDS);

  std::string nameStr = truncateUTF8String(name, 20);
  ret = ZoneGroupProperties_set_name(m_DSMApiHandle, DSUID_BROADCAST, zone, group, (unsigned char*)nameStr.c_str());
  DSBusInterface::checkBroadcastResultCode(ret);
  usleep(BROADCAST_SLEEP_MICROSECONDS);
} // createGroup

void DSStructureModifyingBusInterface::createCluster(
    uint8_t group, ApplicationType applicationType, uint32_t applicationConfig, const std::string& name) {
  DS_CONTEXT(group, applicationType, applicationConfig, name);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  DS_NOTICE("Create cluster.");

  // This does not work for zone id = 0
  // int ret = ZoneGroupModify_add(m_DSMApiHandle, DSUID_BROADCAST, 0, group, _standardGroupID);
  // DSBusInterface::checkBroadcastResultCode(ret);
  // usleep(BROADCAST_SLEEP_MICROSECONDS);

  DSS::getInstance()->assertBlockingAllowed();
  int ret = ClusterProperties_set_application(
      m_DSMApiHandle, DSUID_BROADCAST, group, static_cast<int>(applicationType), applicationConfig);
  DSBusInterface::checkBroadcastResultCode(ret);
  usleep(BROADCAST_CLUSTER_SLEEP_MICROSECONDS);

  std::string nameStr = truncateUTF8String(name, 20);
  ret = ClusterProperties_set_name(m_DSMApiHandle, DSUID_BROADCAST, group, (unsigned char*)nameStr.c_str());
  DSBusInterface::checkBroadcastResultCode(ret);
  usleep(BROADCAST_CLUSTER_SLEEP_MICROSECONDS);

} // createCluster

void DSStructureModifyingBusInterface::removeCluster(uint8_t cluster) {
  DS_CONTEXT(cluster);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  DS_NOTICE("Remove cluster.");

  // This does not work for zone id = 0
  // int ret = ZoneGroupModify_remove(m_DSMApiHandle, DSUID_BROADCAST, 0, cluster);
  // DSBusInterface::checkBroadcastResultCode(ret);
  // usleep(BROADCAST_SLEEP_MICROSECONDS);

  DSS::getInstance()->assertBlockingAllowed();
  int ret = ClusterProperties_set_application(m_DSMApiHandle, DSUID_BROADCAST, cluster, 0, 0);
  DSBusInterface::checkBroadcastResultCode(ret);
  usleep(BROADCAST_CLUSTER_SLEEP_MICROSECONDS);

  std::string nameStr = truncateUTF8String("", 20);
  ret = ClusterProperties_set_name(m_DSMApiHandle, DSUID_BROADCAST, cluster, (unsigned char*)nameStr.c_str());
  DSBusInterface::checkBroadcastResultCode(ret);
  usleep(BROADCAST_CLUSTER_SLEEP_MICROSECONDS);

  ret = ClusterProperties_set_location_class(m_DSMApiHandle, DSUID_BROADCAST, cluster, 0, 0, 0);
  DSBusInterface::checkBroadcastResultCode(ret);
  usleep(BROADCAST_CLUSTER_SLEEP_MICROSECONDS);
} // removeCluster

void DSStructureModifyingBusInterface::groupSetApplication(
    uint16_t zone, uint8_t group, ApplicationType applicationType, uint32_t applicationConfig) {
  DS_CONTEXT(zone, group, applicationType, applicationConfig);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  int ret;
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }

  DSS::getInstance()->assertBlockingAllowed();
  ret = ZoneGroupProperties_set_application(
      m_DSMApiHandle, DSUID_BROADCAST, zone, group, static_cast<int>(applicationType), applicationConfig);
  DS_NOTICE("Set group application.");
  DSBusInterface::checkBroadcastResultCode(ret);
  usleep(BROADCAST_SLEEP_MICROSECONDS);
} // groupSetStandardID

void DSStructureModifyingBusInterface::groupSetName(uint16_t zone, uint8_t group, const std::string& name) {
  DS_CONTEXT(zone, group, name);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  int ret;
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  std::string nameStr = truncateUTF8String(name, 20);
  DSS::getInstance()->assertBlockingAllowed();
  ret = ZoneGroupProperties_set_name(m_DSMApiHandle, DSUID_BROADCAST, zone, group, (unsigned char*)nameStr.c_str());
  DS_NOTICE("Set group name.");
  DSBusInterface::checkBroadcastResultCode(ret);
  usleep(BROADCAST_SLEEP_MICROSECONDS);
} // groupSetName

void DSStructureModifyingBusInterface::removeGroup(uint16_t zone, uint8_t group) {
  DS_CONTEXT(zone, group);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  DSS::getInstance()->assertBlockingAllowed();
  int ret = ZoneGroupModify_remove(m_DSMApiHandle, DSUID_BROADCAST, zone, group);
  DS_NOTICE("Remove group.");
  DSBusInterface::checkBroadcastResultCode(ret);
} // removeGroup

void DSStructureModifyingBusInterface::setButtonSetsLocalPriority(
    const dsuid_t& dsm, const devid_t device, bool setsPriority) {
  DS_CONTEXT(dsm, device, setsPriority);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  int ret = DeviceProperties_set_button_set_local_priority(m_DSMApiHandle, dsm, device, setsPriority ? 1 : 0);
  DS_NOTICE("Set button setsLocalPriority.");
  DSBusInterface::checkResultCode(ret);
} // setButtonSetsLocalPriority

void DSStructureModifyingBusInterface::setButtonCallsPresent(
    const dsuid_t& dsm, const devid_t device, bool callsPresent) {
  DS_CONTEXT(dsm, device, callsPresent);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  int ret = DeviceProperties_set_button_set_no_coming_home_call(m_DSMApiHandle, dsm, device, callsPresent ? 0 : 1);
  DS_NOTICE("Set button callsPresent.");
  DSBusInterface::checkResultCode(ret);
} // setButtonCallsPresent

void DSStructureModifyingBusInterface::setZoneHeatingConfig(
    const dsuid_t& dsm, const uint16_t zone, const ZoneHeatingConfigSpec_t spec) {
  DS_CONTEXT(dsm, zone, spec);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }

  // A lot of code expects that 0 initialized config is valid configuration for OFF mode.
  // But (v)dsm refuse to accept the config if `emergencyValue`
  // is not in range <100, 200> even in OFF mode.
  // Although (v)dsm never uses `emergencyValue` in OFF mde.
  auto emergencyValue = spec.EmergencyValue;
  if (emergencyValue < 100 || emergencyValue > 200) {
    emergencyValue = 150;
  }

  DSS::getInstance()->assertBlockingAllowed();
  int ret = ControllerHeating_set_config(m_DSMApiHandle, dsm, zone, static_cast<uint8_t>(spec.mode), spec.Kp, spec.Ts,
      spec.Ti, spec.Kd, spec.Imin, spec.Imax, spec.Ymin, spec.Ymax, spec.AntiWindUp, spec.KeepFloorWarm,
      spec.SourceZoneId, spec.Offset, spec.ManualValue, emergencyValue);
  DS_NOTICE("Set zone heating config.");
  if (dsm == DSUID_BROADCAST) {
    DSBusInterface::checkBroadcastResultCode(ret);
  } else {
    DSBusInterface::checkResultCode(ret);
  }
  usleep(BROADCAST_SLEEP_MICROSECONDS);
} // setZoneHeatingConfig

void DSStructureModifyingBusInterface::setZoneHeatingOperationModes(
    const dsuid_t& dsm, const uint16_t zone, const ZoneHeatingOperationModeSpec_t spec) {
  DS_CONTEXT(dsm, zone, spec);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  DSS::getInstance()->assertBlockingAllowed();
  int ret = ControllerHeating_set_operation_modes(m_DSMApiHandle, dsm, zone, spec.opModes[0], spec.opModes[1],
      spec.opModes[2], spec.opModes[3], spec.opModes[4], spec.opModes[5], spec.opModes[6], spec.opModes[7],
      spec.opModes[8], spec.opModes[9], spec.opModes[10], spec.opModes[11], spec.opModes[12], spec.opModes[13],
      spec.opModes[14], spec.opModes[15]);
  DS_NOTICE("Set zone heating operation modes.");
  if (dsm == DSUID_BROADCAST) {
    DSBusInterface::checkBroadcastResultCode(ret);
    usleep(BROADCAST_SLEEP_MICROSECONDS);
  } else {
    DSBusInterface::checkResultCode(ret);
  }

  if (m_pModelMaintenance) {
    ModelEvent* pEvent = new ModelEventWithDSID(ModelEvent::etControllerValues, dsm);
    pEvent->addParameter(zone);
    pEvent->setSingleObjectParameter(std::make_shared<ZoneHeatingOperationModeSpec_t>(spec));
    m_pModelMaintenance->addModelEvent(pEvent);
  }
} // setZoneHeatingOperationModes

void DSStructureModifyingBusInterface::setZoneSensor(
    const uint16_t zone, SensorType sensorType, const dsuid_t& sensorDsuid) {
  DS_CONTEXT(zone, sensorType, sensorDsuid);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }

  DSS::getInstance()->assertBlockingAllowed();
  int ret = ZoneProperties_set_zone_sensor(
      m_DSMApiHandle, DSUID_BROADCAST, zone, static_cast<uint8_t>(sensorType), sensorDsuid);
  DS_NOTICE("Set zone sensor.");
  DSBusInterface::checkBroadcastResultCode(ret);
  usleep(BROADCAST_SLEEP_MICROSECONDS);
  if (m_pModelMaintenance) {
    m_pModelMaintenance->addModelEvent(new ModelEvent(ModelEvent::etModelDirty));
  }
}

void DSStructureModifyingBusInterface::setCircuitPowerStateConfig(
    const dsuid_t& dsm, const int index, const int setThreshold, const int resetThreshold) {
  DS_CONTEXT(dsm, index, setThreshold, resetThreshold);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }

  int ret = CircuitPowerState_config(m_DSMApiHandle, dsm, index, setThreshold, resetThreshold);
  DS_NOTICE("Set circuit power state config.");
  DSBusInterface::checkResultCode(ret);
}

void DSStructureModifyingBusInterface::setMeteringConfiguration(const dsuid_t& dsm,
                                                                const MeteringConfigurationSpec_t& spec) {
  DS_CONTEXT(dsm, spec.category, spec.scaleFactor);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  DS_REQUIRE(m_DSMApiHandle != NULL, "Bus not ready");

  int ret =
      dSMProperties_set_metering_category(m_DSMApiHandle, dsm, static_cast<uint8_t>(spec.category), spec.scaleFactor);
  DS_NOTICE("Set metering category.");
  DSBusInterface::checkResultCode(ret);
}

void DSStructureModifyingBusInterface::resetZoneSensor(const uint16_t zone, SensorType sensorType) {
  DS_CONTEXT(zone, sensorType);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }

  DSS::getInstance()->assertBlockingAllowed();
  int ret = ZoneProperties_reset_zone_sensor(m_DSMApiHandle, DSUID_BROADCAST, zone, static_cast<uint8_t>(sensorType));
  DS_NOTICE("Reset zone sensor.");
  DSBusInterface::checkBroadcastResultCode(ret);
  usleep(BROADCAST_SLEEP_MICROSECONDS);
  if (m_pModelMaintenance) {
    m_pModelMaintenance->addModelEvent(new ModelEvent(ModelEvent::etModelDirty));
  }
}

void DSStructureModifyingBusInterface::setModelMaintenace(ModelMaintenance* modelMaintenance) {
  m_pModelMaintenance = modelMaintenance;
}

void DSStructureModifyingBusInterface::clusterSetName(uint8_t cluster, const std::string& name) {
  DS_CONTEXT(cluster, name);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  int ret;
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  std::string nameStr = truncateUTF8String(name, 20);
  DSS::getInstance()->assertBlockingAllowed();
  ret = ClusterProperties_set_name(m_DSMApiHandle, DSUID_BROADCAST, cluster, (unsigned char*)nameStr.c_str());
  DS_NOTICE("Set cluster name.");
  DSBusInterface::checkBroadcastResultCode(ret);
  usleep(BROADCAST_CLUSTER_SLEEP_MICROSECONDS);
}

void DSStructureModifyingBusInterface::clusterSetApplication(
    uint8_t cluster, ApplicationType applicationType, uint32_t applicationConfig) {
  DS_CONTEXT(cluster, applicationType, applicationConfig);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  int ret;
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }

  DSS::getInstance()->assertBlockingAllowed();
  ret = ClusterProperties_set_application(
      m_DSMApiHandle, DSUID_BROADCAST, cluster, static_cast<int>(applicationType), applicationConfig);
  DS_NOTICE("Set cluster application.");
  DSBusInterface::checkBroadcastResultCode(ret);
  usleep(BROADCAST_CLUSTER_SLEEP_MICROSECONDS);
}

void DSStructureModifyingBusInterface::clusterSetProperties(
    uint8_t cluster, uint16_t location, uint16_t floor, uint16_t protectionClass) {
  DS_CONTEXT(cluster, location, floor, protectionClass);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  int ret;
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  DSS::getInstance()->assertBlockingAllowed();
  ret =
      ClusterProperties_set_location_class(m_DSMApiHandle, DSUID_BROADCAST, cluster, location, floor, protectionClass);
  DS_NOTICE("Set cluster properties.");
  DSBusInterface::checkBroadcastResultCode(ret);
  usleep(BROADCAST_CLUSTER_SLEEP_MICROSECONDS);
}

void DSStructureModifyingBusInterface::clusterSetLockedScenes(uint8_t cluster, const std::vector<int>& lockedScenes) {
  DS_CONTEXT(cluster, lockedScenes);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  int ret;
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  std::array<uint8_t, 16> lockedScenesMap = {};
  for (int scene : lockedScenes) {
    lockedScenesMap[scene / 8] |= 1 << (scene % 8);
  }
  DSS::getInstance()->assertBlockingAllowed();
  ret = ClusterProperties_set_scene_lock(m_DSMApiHandle, DSUID_BROADCAST, cluster, lockedScenesMap.data());
  DS_NOTICE("Set cluster locked scenes.");
  DSBusInterface::checkBroadcastResultCode(ret);
  usleep(BROADCAST_CLUSTER_SLEEP_MICROSECONDS);
}

void DSStructureModifyingBusInterface::clusterSetConfigurationLock(uint8_t cluster, bool configurationLock) {
  DS_CONTEXT(cluster, configurationLock);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  int ret;
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  DSS::getInstance()->assertBlockingAllowed();
  ret = ClusterProperties_set_configuration_lock(m_DSMApiHandle, DSUID_BROADCAST, cluster, (configurationLock ? 1 : 0));
  DS_NOTICE("Set cluster configuraiton lock.");
  DSBusInterface::checkBroadcastResultCode(ret);
  usleep(BROADCAST_CLUSTER_SLEEP_MICROSECONDS);
}

void DSStructureModifyingBusInterface::clusterSetOperationLock(uint8_t cluster, bool operationLock) {
  DS_CONTEXT(cluster, operationLock);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  int ret;
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }
  DSS::getInstance()->assertBlockingAllowed();
  ret = ZoneGroupActionRequest_action_extra_command(m_DSMApiHandle, DSUID_BROADCAST, 0, cluster, 0, (operationLock ? 1 : 0));
  DS_NOTICE("Set cluster operation lock.");
  DSBusInterface::checkBroadcastResultCode(ret);

  if (m_pModelMaintenance) {
    ModelEvent* pEvent = new ModelEvent(ModelEvent::etOperationLock);
    pEvent->addParameter(0);
    pEvent->addParameter(cluster);
    pEvent->addParameter(operationLock);
    pEvent->addParameter(CallOrigin::dsmApi);
    m_pModelMaintenance->addModelEvent(pEvent);
  }
  usleep(BROADCAST_CLUSTER_SLEEP_MICROSECONDS);
}

void DSStructureModifyingBusInterface::clusterSetStopTimeout(
                                              uint8_t cluster,
                                              uint8_t timeoutSeconds) {
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }

  DSS::getInstance()->assertBlockingAllowed();
  int ret = ClusterProperties_set_stop_timeout(m_DSMApiHandle,
                                               DSUID_BROADCAST, cluster,
                                               timeoutSeconds);
  DSBusInterface::checkBroadcastResultCode(ret);
}

void DSStructureModifyingBusInterface::zoneSetStopTimeout(
                                              uint16_t zone,
                                              uint8_t group,
                                              uint8_t timeoutSeconds) {
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw BusApiError("Bus not ready");
  }

  DSS::getInstance()->assertBlockingAllowed();
  int ret = ZoneGroupProperties_set_stop_timeout(m_DSMApiHandle,
                                                 DSUID_BROADCAST, zone, group,
                                                 timeoutSeconds);
  DSBusInterface::checkBroadcastResultCode(ret);
}

} // namespace dss
