/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "hardware-abstraction.h"

#include <iostream>
#include <ostream>
#include <trompeloeil.hpp>
#include "businterface-mock.h"
#include "cache/device-cache.h"
#include "ds/catch/catch.h"
#include "ds/json.h"
#include "ds/log.h"
#include "event-mock.h"
#include "model/apartment.h"
#include "model/device-mock.h"
#include "model/modelmaintenance-mock.h"
#include "util.h"

static ds::LogChannel dsLogChannel("halClimateStateMachine", ds::LogSeverity::DEBUG);
static const char* TAGS = "[dss][halClimateStateMachine1]";

using namespace dss;
using namespace trompeloeil;
using Catch::Generators::values;
using trompeloeil::_;

namespace {

struct Area {
  int area;
  int inc;
  int dec;
};

SCENARIO("BL-KM300 device", TAGS) {
  auto pApartment = makeApartment(nullptr);
  auto& apt = *pApartment;

  BusInterfaceMock busInterface;
  EventSinkMock eventSink;
  ModelMaintenanceMock modelMaintenance;

  apt.setBusInterface(&busInterface);
  apt.setEventSink(eventSink);
  apt.setModelMaintenance(&modelMaintenance);

  DeviceSpec spec;
  spec.FunctionID = functionID100(DeviceClasses_t::DEVICE_CLASS_BL, FID100::output, FID100keys::_1);
  spec.ProductID = 0x012C;

  auto dev1 = createDevice(apt, {.hardwareType = HardwareType::BL_KM300});
  auto dsuid1 = dev1->getDSID();
  dev1->setOutputChannels({ChannelType::waterFlowRate});

  ds::sqlite3::Database d(":memory:");
  auto halPtr = makeHardwareAbstraction(&apt, &d);
  auto& hal = *halPtr;
  hal.getCache().empty();

  auto& climateState = hal.getStateCache();
  climateState.empty();
  ds::DateTime now = std::chrono::system_clock::now();

  GIVEN("the cache is empty") {
    WHEN("a scene is called") {
      REQUIRE(hal.callScene(dsuid1, Scene1, false, now) == false);
      THEN("no action is taken") {}
    }
  }

  GIVEN("the cache is populated") {
    {
      auto cache = hal.getCache().tryLookup(dsuid1);
      cache.scene.sceneTable[Scene1].flags.dontCare = false;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::waterFlowRate].value = 1.0;
      cache.scene.sceneTable[Scene2].flags.dontCare = false;
      cache.scene.sceneTable[Scene2].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene2].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene2].channels[ChannelType::waterFlowRate].value = 0.5;
      cache.scene.sceneTable[Scene4].channels[ChannelType::waterFlowRate].value = 0;

      cache.scene.sceneTable[SceneAutoOff].flags.dontCare = false;
      cache.scene.sceneTable[SceneAutoOff].flags.dimTime = DimTime::dimTime1;
      cache.scene.sceneTable[SceneAutoOff].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[SceneAutoOff].channels[ChannelType::waterFlowRate].value = 0.0;
      hal.getCache().update(dsuid1, cache);
    }

    WHEN("No state is known") {
      THEN("after callScene state is known") {
        REQUIRE(hal.callScene(dsuid1, Scene1, false, now) == true);
        auto state = climateState.tryLookup(dsuid1);
        DS_DEBUG("Just debugging here");
        REQUIRE(state.channels[ChannelType::waterFlowRate].current);
        REQUIRE(*state.channels[ChannelType::waterFlowRate].current == 1.0);
        REQUIRE(!state.channels[ChannelType::waterFlowRate].target);
        REQUIRE(state.lastChanged > state.lastChanged.min());
      }
      THEN("INC do not make the state defined") {
        REQUIRE(hal.callScene(dsuid1, SceneInc, false, now) == true);
        auto state = climateState.tryLookup(dsuid1);
        REQUIRE(!state.channels[ChannelType::waterFlowRate].current);
        REQUIRE(!state.channels[ChannelType::waterFlowRate].target);
      }
      THEN("DEC do not make the state defined") {
        REQUIRE(hal.callScene(dsuid1, SceneDec, false, now) == true);
        auto state = climateState.tryLookup(dsuid1);
        REQUIRE(!state.channels[ChannelType::waterFlowRate].current);
        REQUIRE(!state.channels[ChannelType::waterFlowRate].target);
      }
    }
    WHEN("Current device state is known") {
      {
        auto state = climateState.tryLookup(dsuid1);
        state.channels[ChannelType::waterFlowRate].current = 0.5;
        state.lastSceneCalled = Scene2;
        state.beforeLastSceneCalled = Scene3;
        climateState.update(dsuid1, state);
      }
      THEN("after callScene state is known") {
        REQUIRE(hal.callScene(dsuid1, Scene1, false, now) == true);
        auto state = climateState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::waterFlowRate].current);
        REQUIRE(*state.channels[ChannelType::waterFlowRate].current == 1.0);
        REQUIRE(state.lastChanged > state.lastChanged.min());
      }
      THEN("callSceneMin does not change the output") {
        REQUIRE(hal.callSceneMin(dsuid1, Scene1, now) == false);
        auto state = climateState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::waterFlowRate].current);
        REQUIRE(roundDigits(*state.channels[ChannelType::waterFlowRate].current, 2) == 0.5);
      }
    }
  }

  GIVEN("the cache is populated for undo scene") {
    {
      auto cache = hal.getCache().tryLookup(dsuid1);
      cache.scene.sceneTable[Scene1].flags.dontCare = false;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::waterFlowRate].value = 50;

      cache.scene.sceneTable[Scene2].flags.dontCare = false;
      cache.scene.sceneTable[Scene2].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene2].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene2].channels[ChannelType::waterFlowRate].value = 25;

      cache.scene.sceneTable[Scene3].flags.dontCare = false;
      cache.scene.sceneTable[Scene3].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene3].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene3].channels[ChannelType::waterFlowRate].value = 75;
      hal.getCache().update(dsuid1, cache);
    }

    WHEN("No state is known") {
      THEN("after undoScene does not make the state defined") {
        REQUIRE(hal.undoScene(dsuid1, -1, now) == false);
        auto state = climateState.tryLookup(dsuid1);
        REQUIRE(!state.channels[ChannelType::waterFlowRate].current);
      }
      THEN("after undoSceneNr does not make the state defined") {
        REQUIRE(hal.undoScene(dsuid1, Scene1, now) == false);
        auto state = climateState.tryLookup(dsuid1);
        REQUIRE(!state.channels[ChannelType::waterFlowRate].current);
      }
    }
    WHEN("Current device state is known") {
      {
        auto state = climateState.tryLookup(dsuid1);
        state.channels[ChannelType::waterFlowRate].current = 100;
        state.lastSceneCalled = Scene2;
        state.beforeLastSceneCalled = Scene3;
        climateState.update(dsuid1, state);
      }
      THEN("after undoScene previous scene is called") {
        REQUIRE(hal.undoScene(dsuid1, -1, now) == true);
        auto state = climateState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::waterFlowRate].current);
        REQUIRE(*state.channels[ChannelType::waterFlowRate].current == 75);
        REQUIRE(*state.lastSceneCalled == Scene3);
        REQUIRE(*state.beforeLastSceneCalled == Scene2);
      }
      THEN("after undoSceneNr with matching sceneId previous scene is called") {
        REQUIRE(hal.undoScene(dsuid1, Scene2, now) == true);
        auto state = climateState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::waterFlowRate].current);
        REQUIRE(*state.channels[ChannelType::waterFlowRate].current == 75);
        REQUIRE(*state.lastSceneCalled == Scene3);
        REQUIRE(*state.beforeLastSceneCalled == Scene2);
      }
      THEN("after undoSceneNr with not matching sceneId previous scene is not called") {
        REQUIRE(hal.undoScene(dsuid1, Scene44, now) == false);
        auto state = climateState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::waterFlowRate].current);
        REQUIRE(*state.channels[ChannelType::waterFlowRate].current == 100);
        REQUIRE(*state.lastSceneCalled == Scene2);
        REQUIRE(*state.beforeLastSceneCalled == Scene3);
      }
    }
  }

  GIVEN("A scene with dontCare") {
    {
      auto cache = hal.getCache().tryLookup(dsuid1);
      cache.scene.sceneTable[Scene1].flags.dontCare = true;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::waterFlowRate].value = 100.0;
      hal.getCache().update(dsuid1, cache);
    }
    {
      auto state = climateState.tryLookup(dsuid1);
      state.channels[ChannelType::waterFlowRate].current = 50;
      state.lastSceneCalled = 13;
      state.localPriority = false;
      climateState.update(dsuid1, state);
    }
    WHEN("After the scene is called") {
      REQUIRE(hal.callScene(dsuid1, Scene1, false, now) == false);
      THEN("state has not changed") {
        auto state = climateState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::waterFlowRate].current);
        REQUIRE(*state.channels[ChannelType::waterFlowRate].current == 50);
        REQUIRE(!state.channels[ChannelType::waterFlowRate].target);
        REQUIRE(*state.lastSceneCalled == 13);
      }
    }
  }

  GIVEN("A setOutputValue call") {
    {
      auto state = climateState.tryLookup(dsuid1);
      state.channels[ChannelType::waterFlowRate].current = 50;
      state.channels[ChannelType::waterFlowRate].target = 55;
      state.lastSceneCalled = 13;
      state.localPriority = true;
      climateState.update(dsuid1, state);
    }
    WHEN("After setOutputValue is called") {
      REQUIRE(hal.setOutputValue(dsuid1, 245, now) == true);
      THEN("state is changed") {
        auto state = climateState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::waterFlowRate].current);
        REQUIRE(roundDigits(*state.channels[ChannelType::waterFlowRate].current, 2) == 96.08);
        REQUIRE(!state.channels[ChannelType::waterFlowRate].target);
        REQUIRE(!state.lastSceneCalled);
        REQUIRE(state.localPriority == false);
      }
    }
  }

  GIVEN("The device is set to switched mode") {
    {
      auto cache = hal.getCache().tryLookup(dsuid1);
      cache.property[DeviceProperty::outputMode] = static_cast<int>(OutputMode::KM_Switch);
      cache.property[DeviceProperty::switchedThreshold] = 128;
      cache.scene.sceneTable[Scene1].flags.dontCare = false;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::waterFlowRate].value = 100;
      cache.scene.sceneTable[Scene2].flags.dontCare = false;
      cache.scene.sceneTable[Scene2].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene2].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene2].channels[ChannelType::waterFlowRate].value = 25;
      hal.getCache().update(dsuid1, cache);
    }
    WHEN("after callScene with value above the threshold") {
      {
        auto state = climateState.tryLookup(dsuid1);
        state.channels[ChannelType::waterFlowRate].current = 25;
        state.channels[ChannelType::waterFlowRate].level = 0;
        state.lastSceneCalled = 13;
        climateState.update(dsuid1, state);
      }
      REQUIRE(hal.callScene(dsuid1, Scene1, false, now) == true);
      THEN("the output is on (level1)") {
        auto ret = hal.getState(dsuid1);
        REQUIRE(ret.outputs[ChannelType::waterFlowRate].value == 100);
        REQUIRE(ret.outputs[ChannelType::waterFlowRate].level == 1);
      }
    }
    WHEN("after callScene with value below the threshold") {
      {
        auto state = climateState.tryLookup(dsuid1);
        state.channels[ChannelType::waterFlowRate].current = 75;
        state.channels[ChannelType::waterFlowRate].level = 1;
        state.lastSceneCalled = 13;
        climateState.update(dsuid1, state);
      }
      REQUIRE(hal.callScene(dsuid1, Scene2, false, now) == true);
      THEN("the output is off (level0)") {
        auto ret = hal.getState(dsuid1);
        REQUIRE(ret.outputs[ChannelType::waterFlowRate].value == 25);
        REQUIRE(ret.outputs[ChannelType::waterFlowRate].level == 0);
      }
    }
    WHEN("device switch threshold is reconfigured") {
      {
        auto state = climateState.tryLookup(dsuid1);
        state.channels[ChannelType::waterFlowRate].current = 75;
        state.channels[ChannelType::waterFlowRate].level = 1;
        state.lastSceneCalled = 13;
        climateState.update(dsuid1, state);
      }
      REQUIRE_NOTHROW(hal.updateDeviceConfig(dsuid1, 3, 0x43, 205));
      THEN("the output level is reset") {
        auto ret = hal.getState(dsuid1);
        REQUIRE(ret.outputs[ChannelType::waterFlowRate].level == 0);
      }
    }
  }

  GIVEN("a climate device is controlled by writing to the runtime bank") {
    THEN("writing the output value will change it") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(dsuid1, 64, 0x00, 0x34));
      auto state = climateState.tryLookup(dsuid1);
      REQUIRE(state.channels[ChannelType::waterFlowRate].current);
      REQUIRE(roundDigits(*state.channels[ChannelType::waterFlowRate].current, 2) == 20.39);
      REQUIRE(!state.terminatesAt);
      auto ret = hal.getState(dsuid1);
    }
  }

  GIVEN("a climate device is switched and controlled by writing to the runtime bank") {
    {
      auto cache = hal.getCache().tryLookup(dsuid1);
      cache.property[DeviceProperty::outputMode] = static_cast<int>(OutputMode::KL_Switch);
      cache.property[DeviceProperty::switchedThreshold] = 128;
      hal.getCache().update(dsuid1, cache);
    }
    {
      auto state = climateState.tryLookup(dsuid1);
      state.channels[ChannelType::waterFlowRate].current = 0;
      state.channels[ChannelType::waterFlowRate].level = 0;
      state.lastSceneCalled = 0;
      climateState.update(dsuid1, state);
    }
    THEN("writing the output value will change it and the switched state") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(dsuid1, 64, 0x00, 0xac));
      auto state = climateState.tryLookup(dsuid1);
      REQUIRE(state.channels[ChannelType::waterFlowRate].current);
      REQUIRE(roundDigits(*state.channels[ChannelType::waterFlowRate].current, 2) == 67.45);
      REQUIRE(!state.terminatesAt);
      auto ret = hal.getState(dsuid1);
      REQUIRE(ret.outputs[ChannelType::waterFlowRate].level == 1);
    }
  }
}

} // namespace
