/*
 *  Copyright (c) 2020 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "hal-device.h"

#include "util.h"
#include "model/device.h"
#include "ds-api/route-segments.h"

namespace dss::hal {

static ds::LogChannel dsLogChannel("halClimateDevice");

void HalClimateDevice::write(const Device& device, const std::string& parameter, double value) const {
  // "/scene/5/brightness" -> 128, 5
  // "/scene/6/dontCare"   -> 128, 134 (=128+6)
  // "/scene/7/brightness/dontCare" -> 128, 135 (=128+7)
  DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (auto route = dsApi::matchRoute("/scene/{sceneNr}/{channelOrFlag}", parameter)) {
    auto& scene = route.args["sceneNr"];
    auto& param = route.args["channelOrFlag"];
    unsigned int sceneNr = strToUInt(scene);
    DS_REQUIRE(sceneNr <= MaxSceneNumber, "Scene number is out of range.", sceneNr);
    SceneEntry& sceneCache = devCache.scene.sceneTable[sceneNr];
    if (param == "waterFlowRate") {
      auto channelSpec = getOutputChannelIdAndSize("waterFlowRate");
      device.setDeviceConfig(128, sceneNr, convertToOutputChannelValue(channelSpec.first, value), false);
      SceneChannel& waterFlowRate = sceneCache.channels[ChannelType::waterFlowRate];
      waterFlowRate.value = value;
      sceneCache.lastChanged = std::chrono::system_clock::now();
      devCache.lastChanged = sceneCache.lastChanged;
    } else {
      DS_FAIL_REQUIRE("Unknown parameter requested", parameter);
    }
    m_sceneCacheDb.update(device.getDSID(), devCache);
  } else {
    return HalLightDevice::write(device, parameter,value);
  }
}

double HalClimateDevice::readDouble(const Device& device, const std::string& parameter) const {
  double returnValue = 0;
  DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (auto route = dsApi::matchRoute("/scene/{sceneNr}/{channelOrFlag}", parameter)) {
    auto& scene = route.args["sceneNr"];
    auto& param = route.args["channelOrFlag"];
    unsigned int sceneNr = strToUInt(scene);
    DS_REQUIRE(sceneNr <= MaxSceneNumber, "Scene number is out of range.", sceneNr);
    SceneEntry& sceneCache = devCache.scene.sceneTable[sceneNr];
    if (param == "waterFlowRate") {
      auto value = device.getDeviceConfig(128, sceneNr);
      SceneChannel& waterFlowRate = sceneCache.channels[ChannelType::waterFlowRate];
      auto channelSpec = getOutputChannelIdAndSize("waterFlowRate");
      returnValue = convertFromOutputChannelValue(channelSpec.first, value);
      waterFlowRate.value = returnValue;
      sceneCache.lastChanged = std::chrono::system_clock::now();
      devCache.lastChanged = sceneCache.lastChanged;
      DS_DEBUG("Dev cache to save: ", value, returnValue, devCache);
    } else {
      DS_FAIL_REQUIRE("Unknown parameter requested", parameter);
    }
    m_sceneCacheDb.update(device.getDSID(), devCache);
  } else {
    return HalLightDevice::readDouble(device, parameter);
  }
  return returnValue;
}

const std::vector<ChannelType> HalClimateDevice::getSceneChannels(const Device& device) const {
  return std::vector<ChannelType>({ChannelType::waterFlowRate});
}

DeviceState HalClimateDevice::getState(const Device& device) const {
  auto state = m_StateCache.tryLookup(device.getDSID());
  DeviceState ret{};
  ret.lastChanged = state.lastChanged;
  ret.terminatesAt = state.terminatesAt;

  //Determine group membership and from there determine channel, brightness or powerState
  bool isVentilation = device.isInGroup(GroupIDVentilation) || device.isInGroup(GroupIDRecirculation) || device.isInGroup(GroupIDGlobalAppDsVentilation) || device.isInGroup(GroupIDGlobalAppDsRecirculation);
  ChannelType type = getDefaultChannel();
  if (isVentilation) {
    type = ChannelType::airFlowIntensity;
  }

  for (auto& ch : state.channels) {
    if (ch.second.current) {
      ret.outputs[type].value = convertValueToOtherChannelType(getDefaultChannel(), *ch.second.current, type);
    }
    ret.outputs[type].targetValue = ch.second.target ? convertValueToOtherChannelType(getDefaultChannel(), *ch.second.target, type) : ret.outputs[type].value;
    ret.outputs[type].level = ch.second.level;
  }
  
  return ret;
}

const std::vector<DeviceProperty> HalClimateDevice::getDeviceProperties(const Device& device) const {
  return std::vector<DeviceProperty>({DeviceProperty::outputMode, DeviceProperty::switchedThreshold});
}

} // namespace dss::hal
