/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "hardware-abstraction.h"

#include <trompeloeil.hpp>
#include "businterface-mock.h"
#include "cache/device-cache.h"
#include "ds/catch/catch.h"
#include "ds/json.h"
#include "ds/log.h"
#include "event-mock.h"
#include "model/apartment.h"
#include "model/device-mock.h"
#include "model/modelmaintenance-mock.h"
#include "util.h"

static ds::LogChannel dsLogChannel("halStateMachineShade", ds::LogSeverity::DEBUG);
static const char* TAGS = "[dss][halStateMachineShade]";

using namespace dss;
using namespace trompeloeil;
using Catch::Generators::values;
using trompeloeil::_;
using namespace std::literals;

namespace {

struct Area {
  int area;
  int inc;
  int dec;
};

SCENARIO("Standard Shade device", TAGS) {
  auto pApartment = makeApartment(nullptr);
  auto& apt = *pApartment;
  BusInterfaceMock busInterface;
  EventSinkMock eventSink;
  ModelMaintenanceMock modelMaintenance;

  apt.setBusInterface(&busInterface);
  apt.setEventSink(eventSink);
  apt.setModelMaintenance(&modelMaintenance);

  ALLOW_CALL(eventSink, pushEvent(_));

  auto blind = createDevice(apt, {.hardwareType = HardwareType::GR_KL220});

  ds::sqlite3::Database d(":memory:");
  auto halPtr = makeHardwareAbstraction(&apt, &d);
  auto& hal = *halPtr;
  hal.getCache().empty();

  auto& lightState = hal.getStateCache();
  lightState.empty();
  ds::DateTime now = std::chrono::system_clock::now();

  GIVEN("the cache is empty") {
    WHEN("a scene is called") {
      REQUIRE(hal.callScene(blind->getDSID(), Scene1, false, now) == false);
      THEN("no action is taken") {}
    }
  }

  GIVEN("the cache is populated") {
    {
      auto cache = hal.getCache().tryLookup(blind->getDSID());
      cache.property[DeviceProperty::posTimeUp] = 20 * 100;
      cache.property[DeviceProperty::posTimeDown] = 20 * 100;
      cache.property[DeviceProperty::minDim] = 10;
      cache.scene.sceneTable[Scene1].flags.dontCare = false;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::shadePositionOutside].value = 50.0;
      cache.scene.sceneTable[Scene2].channels[ChannelType::shadePositionOutside].value = 46.0;
      hal.getCache().update(blind->getDSID(), cache);
    }

    WHEN("No state is known") {
      THEN("after callScene state is known") {
        REQUIRE(hal.callScene(blind->getDSID(), Scene1, false, now) == true);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(*state.channels[ChannelType::shadePositionOutside].current == 50);
        REQUIRE(!state.channels[ChannelType::shadePositionOutside].target);
        REQUIRE(state.lastChanged > state.lastChanged.min());
      }
      THEN("INC do not make the state defined") {
        REQUIRE(hal.callScene(blind->getDSID(), SceneInc, false, now) == true);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(!state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(!state.channels[ChannelType::shadePositionOutside].target);
      }
      THEN("DEC do not make the state defined") {
        REQUIRE(hal.callScene(blind->getDSID(), SceneDec, false, now) == true);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(!state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(!state.channels[ChannelType::shadePositionOutside].target);
      }
    }
    WHEN("Current device state is known") {
      auto position =
          GENERATE(values<std::pair<double, int>>({{0.0, 10000}, {25.0, 5000}, {75.0, 5000}, {100.0, 10000}}));
      {
        auto state = lightState.tryLookup(blind->getDSID());
        state.channels[ChannelType::shadePositionOutside].current = position.first;
        state.lastSceneCalled = Scene2;
        state.beforeLastSceneCalled = Scene3;
        lightState.update(blind->getDSID(), state);
      }
      REQUIRE(hal.callScene(blind->getDSID(), Scene1, false, now) == true);
      THEN("after callScene shade is moving") {
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(*state.channels[ChannelType::shadePositionOutside].current == position.first);
        REQUIRE(*state.channels[ChannelType::shadePositionOutside].target == 50);
        REQUIRE(
            std::chrono::duration_cast<std::chrono::milliseconds>(*state.terminatesAt - state.lastChanged).count() ==
            round(20000.0 * std::abs((50.0 - position.first) / 100.0)));
      }
      THEN("after callScene and " + intToString(position.second / 2) + "ms shade is still moving") {
        auto state = lightState.tryLookup(blind->getDSID());
        auto ret = hal.getState(blind->getDSID());
        REQUIRE(ret.outputs[ChannelType::shadePositionOutside].value == position.first);
        REQUIRE(ret.outputs[ChannelType::shadePositionOutside].targetValue == 50);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::moving);
        REQUIRE(std::chrono::duration_cast<std::chrono::milliseconds>(*ret.terminatesAt - ret.lastChanged).count() ==
                round(20000.0 * std::abs((50.0 - position.first) / 100.0)));
      }
      THEN("after callScene and " + intToString(position.second) + "ms shade has stopped at final position") {
        auto state = lightState.tryLookup(blind->getDSID());
        auto changed = state.lastChanged;
        REQUIRE(hal.completeMoving(blind->getDSID(), changed + std::chrono::milliseconds(position.second)) == true);
        auto ret = hal.getState(blind->getDSID());
        REQUIRE(ret.outputs[ChannelType::shadePositionOutside].value == 50);
        REQUIRE(ret.outputs[ChannelType::shadePositionOutside].targetValue == 50);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(!ret.terminatesAt);
        state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(*state.channels[ChannelType::shadePositionOutside].current == 50);
        REQUIRE(!state.channels[ChannelType::shadePositionOutside].target);
      }
      THEN("after callScene(STOP) after " + intToString(position.second / 2) +
           "ms shade has stopped at middle position") {
        auto state = lightState.tryLookup(blind->getDSID());
        auto changed = state.lastChanged;
        REQUIRE_NOTHROW(hal.callScene(blind->getDSID(), SceneStop, false,
                                      changed + std::chrono::milliseconds(position.second / 2)));
        auto ret = hal.getState(blind->getDSID());
        REQUIRE(ret.outputs[ChannelType::shadePositionOutside].value == position.first + (50 - position.first) / 2);
        REQUIRE(ret.outputs[ChannelType::shadePositionOutside].targetValue ==
                position.first + (50 - position.first) / 2);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(!ret.terminatesAt);
        state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(*state.channels[ChannelType::shadePositionOutside].current ==
                position.first + (50 - position.first) / 2);
        REQUIRE(!state.channels[ChannelType::shadePositionOutside].target);
      }
    }
    WHEN("Current device state is zero") {
      {
        auto state = lightState.tryLookup(blind->getDSID());
        state.channels[ChannelType::shadePositionOutside].current = 0.0;
        state.lastSceneCalled = Scene2;
        state.beforeLastSceneCalled = Scene3;
        lightState.update(blind->getDSID(), state);
      }
      THEN("INC does not change the value") {
        REQUIRE(hal.callScene(blind->getDSID(), SceneInc, false, now) == true);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(state.dimmStatus == DeviceDimmingStatus::none);
        REQUIRE(roundDigits(*state.channels[ChannelType::shadePositionOutside].current, 2) == 0);
        REQUIRE(!state.channels[ChannelType::shadePositionOutside].target);
      }
      THEN("DEC not change the value") {
        REQUIRE(hal.callScene(blind->getDSID(), SceneDec, false, now) == true);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(state.dimmStatus == DeviceDimmingStatus::none);
        REQUIRE(roundDigits(*state.channels[ChannelType::shadePositionOutside].current, 2) == 0);
        REQUIRE(!state.channels[ChannelType::shadePositionOutside].target);
      }
      THEN("INC changes the value after callSceneMin") {
        REQUIRE(hal.callSceneMin(blind->getDSID(), Scene1, now) == true);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(*state.channels[ChannelType::shadePositionOutside].current > 0);

        REQUIRE(hal.callScene(blind->getDSID(), SceneInc, false, now + std::chrono::milliseconds(250)) == true);
        state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(state.channels[ChannelType::shadePositionOutside].target);
        REQUIRE(state.terminatesAt);
        REQUIRE(state.dimmingTimeout);

        REQUIRE(hal.callScene(blind->getDSID(), SceneInc, false, now + std::chrono::milliseconds(500)) == true);
        state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(state.channels[ChannelType::shadePositionOutside].target);
        REQUIRE(state.terminatesAt);
        REQUIRE(state.dimmingTimeout);

        REQUIRE(hal.callScene(blind->getDSID(), SceneInc, false, now + std::chrono::milliseconds(750)) == true);
        state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(state.channels[ChannelType::shadePositionOutside].target);
        REQUIRE(state.terminatesAt);
        REQUIRE(state.dimmingTimeout);

        REQUIRE(hal.callScene(blind->getDSID(), SceneInc, false, now + std::chrono::milliseconds(1000)) == true);
        state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(state.channels[ChannelType::shadePositionOutside].target);
        REQUIRE(state.terminatesAt);
        REQUIRE(state.dimmingTimeout);

        REQUIRE(hal.callScene(blind->getDSID(), SceneInc, false, now + std::chrono::milliseconds(1250)) == true);
        state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(state.channels[ChannelType::shadePositionOutside].target);
        REQUIRE(state.terminatesAt);
        REQUIRE(state.dimmingTimeout);

        REQUIRE(hal.callScene(blind->getDSID(), SceneInc, false, now + std::chrono::milliseconds(1500)) == true);
        state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(state.channels[ChannelType::shadePositionOutside].target);
        REQUIRE(state.terminatesAt);
      }
    }
    WHEN("Current device state is max") {
      {
        auto state = lightState.tryLookup(blind->getDSID());
        state.channels[ChannelType::shadePositionOutside].current = 100.0;
        state.lastSceneCalled = Scene2;
        state.beforeLastSceneCalled = Scene3;
        lightState.update(blind->getDSID(), state);
      }
      THEN("INC does not change the value") {
        REQUIRE(hal.callScene(blind->getDSID(), SceneInc, false, now) == true);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(roundDigits(*state.channels[ChannelType::shadePositionOutside].current, 2) == 100);
        REQUIRE(state.channels[ChannelType::shadePositionOutside].target);
        REQUIRE(roundDigits(*state.channels[ChannelType::shadePositionOutside].target, 2) == 100);
      }
      THEN("DEC changes the value") {
        REQUIRE(hal.callScene(blind->getDSID(), SceneDec, false, now) == true);

        auto status = hal.getState(blind->getDSID());
        REQUIRE(status.status == dsModel::OutputStatusStatus::dimming);
        REQUIRE(roundDigits(status.outputs[ChannelType::shadePositionOutside].targetValue, 2) < 100);
        REQUIRE(hal.stopDimming(blind->getDSID(), SceneDec, now + std::chrono::milliseconds(1500)) == true);
        status = hal.getState(blind->getDSID());
        REQUIRE(status.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(status.outputs[ChannelType::shadePositionOutside].value, 2) < 100);
      }
    }
    WHEN("Dimming process") {
      auto [first, cont, area] = GENERATE(values<std::tuple<int, int, int>>({{SceneInc, SceneInc, Scene1},
                                                                             {SceneDec, SceneDec, Scene1},
                                                                             {SceneDecA1, SceneDimArea, SceneA11},
                                                                             {SceneDecA2, SceneDimArea, SceneA21},
                                                                             {SceneDecA3, SceneDimArea, SceneA31},
                                                                             {SceneDecA4, SceneDimArea, SceneA41},
                                                                             {SceneIncA1, SceneDimArea, SceneA11},
                                                                             {SceneIncA2, SceneDimArea, SceneA21},
                                                                             {SceneIncA3, SceneDimArea, SceneA31},
                                                                             {SceneIncA4, SceneDimArea, SceneA41}}));
      {
        auto state = lightState.tryLookup(blind->getDSID());
        state.channels[ChannelType::shadePositionOutside].current = 50.0;
        state.lastSceneCalled = Scene2;
        state.beforeLastSceneCalled = Scene3;
        lightState.update(blind->getDSID(), state);

        auto cache = hal.getCache().tryLookup(blind->getDSID());

        cache.scene.sceneTable[area].flags.dontCare = false;
        cache.scene.sceneTable[area].flags.dimTime = DimTime::noDimTime;
        cache.scene.sceneTable[area].flags.ignoreLocalPrio = false;
        cache.scene.sceneTable[area].channels[ChannelType::shadePositionOutside].value = 50.0;
        hal.getCache().update(blind->getDSID(), cache);
      }
      THEN("Dimming is 4 fine steps, then continuous") {
        auto status = hal.getState(blind->getDSID());
        auto state = lightState.tryLookup(blind->getDSID());
        auto scene = first;
        int j = 0;
        for (; j < 20; ++j) {
          auto changed = now + std::chrono::milliseconds(j * 250);
          REQUIRE(hal.callScene(blind->getDSID(), scene, false, changed) == true);
          if (scene == first) {
            scene = cont;
          }
          status = hal.getState(blind->getDSID());
          REQUIRE(status.status == dsModel::OutputStatusStatus::dimming);
          state = lightState.tryLookup(blind->getDSID());
          REQUIRE(state.dimmStatus == DeviceDimmingStatus::stepping);
          REQUIRE(*state.lastDimmCommand == first);
        }
        {
          auto changed = now + std::chrono::milliseconds(j++ * 250);
          REQUIRE(hal.callScene(blind->getDSID(), cont, false, changed) == true);
          status = hal.getState(blind->getDSID());
          state = lightState.tryLookup(blind->getDSID());
          REQUIRE(std::chrono::duration_cast<std::chrono::milliseconds>(*status.terminatesAt - status.lastChanged)
                      .count() == 500);
          REQUIRE(status.status == dsModel::OutputStatusStatus::dimming);
          REQUIRE(state.dimmStatus == DeviceDimmingStatus::dimming);
          REQUIRE(*state.lastDimmCommand == first);
        }
        {
          auto changed = now + std::chrono::milliseconds(j++ * 250);
          REQUIRE(hal.callScene(blind->getDSID(), cont, false, changed) == true);
          status = hal.getState(blind->getDSID());
          state = lightState.tryLookup(blind->getDSID());
          REQUIRE(std::chrono::duration_cast<std::chrono::milliseconds>(*status.terminatesAt - status.lastChanged)
                      .count() == 500);
          REQUIRE(status.status == dsModel::OutputStatusStatus::dimming);
          REQUIRE(state.dimmStatus == DeviceDimmingStatus::dimming);
          REQUIRE(*state.lastDimmCommand == first);
        }
        {
          auto changed = now + std::chrono::milliseconds((j + 2) * 250);
          REQUIRE(hal.stopDimming(blind->getDSID(), first, changed) == true);
          status = hal.getState(blind->getDSID());
          state = lightState.tryLookup(blind->getDSID());
          REQUIRE(!status.terminatesAt);
          REQUIRE(status.status == dsModel::OutputStatusStatus::ok);
          REQUIRE(state.dimmStatus == DeviceDimmingStatus::none);
          REQUIRE(!state.lastDimmCommand);
        }
      }
    }
    WHEN("Dimming process changes direction") {
      auto [first, cont, opposite, area] =
          GENERATE(values<std::tuple<int, int, int, int>>({{SceneInc, SceneInc, SceneDec, Scene1},
                                                           {SceneDec, SceneDec, SceneInc, Scene1},
                                                           {SceneDecA1, SceneDimArea, SceneIncA1, SceneA11},
                                                           {SceneDecA2, SceneDimArea, SceneIncA2, SceneA21},
                                                           {SceneDecA3, SceneDimArea, SceneIncA3, SceneA31},
                                                           {SceneDecA4, SceneDimArea, SceneIncA4, SceneA41},
                                                           {SceneIncA1, SceneDimArea, SceneDecA1, SceneA11},
                                                           {SceneIncA2, SceneDimArea, SceneDecA2, SceneA21},
                                                           {SceneIncA3, SceneDimArea, SceneDecA3, SceneA31},
                                                           {SceneIncA4, SceneDimArea, SceneDecA4, SceneA41}}));
      {
        auto state = lightState.tryLookup(blind->getDSID());
        state.channels[ChannelType::shadePositionOutside].current = 50.0;
        state.lastSceneCalled = area;
        state.beforeLastSceneCalled = Scene3;
        lightState.update(blind->getDSID(), state);

        auto cache = hal.getCache().tryLookup(blind->getDSID());

        cache.scene.sceneTable[area].flags.dontCare = false;
        cache.scene.sceneTable[area].flags.dimTime = DimTime::noDimTime;
        cache.scene.sceneTable[area].flags.ignoreLocalPrio = false;
        cache.scene.sceneTable[area].channels[ChannelType::shadePositionOutside].value = 50.0;
        hal.getCache().update(blind->getDSID(), cache);
      }
      THEN("changing direction works") {
        auto changed = std::chrono::system_clock::now();
        auto status = hal.getState(blind->getDSID());
        auto state = lightState.tryLookup(blind->getDSID());
        changed += std::chrono::milliseconds(250);
        REQUIRE(hal.callScene(blind->getDSID(), first, false, changed) == true);
        status = hal.getState(blind->getDSID());
        changed = status.lastChanged;
        REQUIRE(
            std::chrono::duration_cast<std::chrono::milliseconds>(*status.terminatesAt - status.lastChanged).count() ==
            1010);
        REQUIRE(status.status == dsModel::OutputStatusStatus::dimming);
        state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.dimmStatus == DeviceDimmingStatus::stepping);
        REQUIRE(*state.lastDimmCommand == first);
        {
          changed += std::chrono::milliseconds(250);
          status = hal.getState(blind->getDSID());
          state = lightState.tryLookup(blind->getDSID());
          REQUIRE(std::chrono::duration_cast<std::chrono::milliseconds>(*status.terminatesAt - status.lastChanged)
                      .count() == 1010);
          REQUIRE(status.status == dsModel::OutputStatusStatus::dimming);
          REQUIRE(state.dimmStatus == DeviceDimmingStatus::stepping);
          REQUIRE(hal.callScene(blind->getDSID(), cont, false, changed) == true);
          status = hal.getState(blind->getDSID());
          state = lightState.tryLookup(blind->getDSID());
          REQUIRE(std::chrono::duration_cast<std::chrono::milliseconds>(*status.terminatesAt - status.lastChanged)
                      .count() == 760);
          REQUIRE(status.status == dsModel::OutputStatusStatus::dimming);
          REQUIRE(state.dimmStatus == DeviceDimmingStatus::stepping);
          REQUIRE(*state.lastDimmCommand == first);
        }

        {
          changed += std::chrono::milliseconds(250);
          status = hal.getState(blind->getDSID());
          state = lightState.tryLookup(blind->getDSID());
          REQUIRE(status.terminatesAt);
          REQUIRE(status.status == dsModel::OutputStatusStatus::dimming);
          REQUIRE(state.dimmStatus == DeviceDimmingStatus::stepping);
          REQUIRE(hal.callScene(blind->getDSID(), opposite, false, changed) == true);
          status = hal.getState(blind->getDSID());
          state = lightState.tryLookup(blind->getDSID());
          REQUIRE(std::chrono::duration_cast<std::chrono::milliseconds>(*status.terminatesAt - status.lastChanged)
                      .count() == 1010);
          REQUIRE(status.status == dsModel::OutputStatusStatus::dimming);
          REQUIRE(state.dimmStatus == DeviceDimmingStatus::stepping);
          REQUIRE(*state.lastDimmCommand == opposite);
        }
      }
    }
  }
  GIVEN("the cache is populated for undo scene") {
    {
      auto cache = hal.getCache().tryLookup(blind->getDSID());
      cache.property[DeviceProperty::posTimeUp] = 20 * 100;
      cache.property[DeviceProperty::posTimeDown] = 20 * 100;
      cache.scene.sceneTable[Scene1].flags.dontCare = false;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::shadePositionOutside].value = 50.0;

      cache.scene.sceneTable[Scene2].flags.dontCare = false;
      cache.scene.sceneTable[Scene2].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene2].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene2].channels[ChannelType::shadePositionOutside].value = 25.0;

      cache.scene.sceneTable[Scene3].flags.dontCare = false;
      cache.scene.sceneTable[Scene3].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene3].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene3].channels[ChannelType::shadePositionOutside].value = 75.0;
      hal.getCache().update(blind->getDSID(), cache);
    }

    WHEN("No state is known") {
      THEN("after undoScene does not make the state defined") {
        REQUIRE(hal.undoScene(blind->getDSID(), -1, now) == false);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(!state.channels[ChannelType::shadePositionOutside].current);
      }
      THEN("after undoSceneNr does not make the state defined") {
        REQUIRE(hal.undoScene(blind->getDSID(), Scene1, now) == false);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(!state.channels[ChannelType::shadePositionOutside].current);
      }
    }
    WHEN("Current device state is known") {
      {
        auto state = lightState.tryLookup(blind->getDSID());
        state.channels[ChannelType::shadePositionOutside].current = 100.0;
        state.lastSceneCalled = Scene2;
        state.beforeLastSceneCalled = Scene3;
        lightState.update(blind->getDSID(), state);
      }
      THEN("after undoScene previous scene is called") {
        REQUIRE(hal.undoScene(blind->getDSID(), -1, now) == true);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].target);
        REQUIRE(*state.channels[ChannelType::shadePositionOutside].target == 75);
        REQUIRE(state.terminatesAt);
        REQUIRE(*state.lastSceneCalled == Scene3);
        REQUIRE(*state.beforeLastSceneCalled == Scene2);
      }
      THEN("after undoSceneNr with matching sceneId previous scene is called") {
        REQUIRE(hal.undoScene(blind->getDSID(), Scene2, now) == true);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].target);
        REQUIRE(*state.channels[ChannelType::shadePositionOutside].target == 75);
        REQUIRE(state.terminatesAt);
        REQUIRE(*state.lastSceneCalled == Scene3);
        REQUIRE(*state.beforeLastSceneCalled == Scene2);
      }
      THEN("after undoSceneNr with not matching sceneId previous scene is not called") {
        REQUIRE(hal.undoScene(blind->getDSID(), Scene44, now) == false);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(*state.channels[ChannelType::shadePositionOutside].current == 100);
        REQUIRE(!state.channels[ChannelType::shadePositionOutside].target);
        REQUIRE(!state.terminatesAt);
        REQUIRE(*state.lastSceneCalled == Scene2);
        REQUIRE(*state.beforeLastSceneCalled == Scene3);
      }
    }
  }

  GIVEN("Device is not locally prioritized") {
    {
      auto state = lightState.tryLookup(blind->getDSID());
      state.channels[ChannelType::shadePositionOutside].current = 75.0;
      state.localPriority = false;
      lightState.update(blind->getDSID(), state);
    }
    WHEN("A localOn scene is called") {
      {
        auto cache = hal.getCache().tryLookup(blind->getDSID());
        cache.property[DeviceProperty::posTimeUp] = 20 * 100;
        cache.property[DeviceProperty::posTimeDown] = 20 * 100;
        cache.scene.sceneTable[SceneLocalOn].flags.dontCare = false;
        cache.scene.sceneTable[SceneLocalOn].flags.specialHandling = true;
        cache.scene.sceneTable[SceneLocalOn].flags.dimTime = DimTime::noDimTime;
        cache.scene.sceneTable[SceneLocalOn].flags.ignoreLocalPrio = false;
        cache.scene.sceneTable[SceneLocalOn].channels[ChannelType::shadePositionOutside].value = 50.0;
        hal.getCache().update(blind->getDSID(), cache);
      }
      THEN("Without force, it becomes prioritized") {
        REQUIRE(hal.callScene(blind->getDSID(), SceneLocalOn, false, now) == true);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].target);
        REQUIRE(*state.channels[ChannelType::shadePositionOutside].target == 50.0);
        REQUIRE(state.terminatesAt);
        REQUIRE(state.localPriority == true);
      }
    }
    WHEN("A localOn scene is called, but not configured to be prioritized") {
      {
        auto cache = hal.getCache().tryLookup(blind->getDSID());
        cache.property[DeviceProperty::posTimeUp] = 20 * 100;
        cache.property[DeviceProperty::posTimeDown] = 20 * 100;
        cache.scene.sceneTable[SceneLocalOn].flags.dontCare = false;
        cache.scene.sceneTable[SceneLocalOn].flags.specialHandling = false;
        cache.scene.sceneTable[SceneLocalOn].flags.dimTime = DimTime::noDimTime;
        cache.scene.sceneTable[SceneLocalOn].flags.ignoreLocalPrio = false;
        cache.scene.sceneTable[SceneLocalOn].channels[ChannelType::shadePositionOutside].value = 50.0;
        hal.getCache().update(blind->getDSID(), cache);
      }
      THEN("It does not becomes prioritized") {
        REQUIRE(hal.callScene(blind->getDSID(), SceneLocalOn, false, now) == true);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].target);
        REQUIRE(*state.channels[ChannelType::shadePositionOutside].target == 50.0);
        REQUIRE(state.terminatesAt);
        REQUIRE(state.localPriority == false);
      }
    }
  }

  GIVEN("Device is locally prioritized") {
    {
      auto state = lightState.tryLookup(blind->getDSID());
      state.channels[ChannelType::shadePositionOutside].current = 75.0;
      state.localPriority = true;
      state.lastSceneCalled = Scene1;
      lightState.update(blind->getDSID(), state);
    }
    WHEN("A normal scene is called") {
      {
        auto cache = hal.getCache().tryLookup(blind->getDSID());
        cache.property[DeviceProperty::posTimeUp] = 20 * 100;
        cache.property[DeviceProperty::posTimeDown] = 20 * 100;
        cache.scene.sceneTable[Scene1].flags.dontCare = false;
        cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
        cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
        cache.scene.sceneTable[Scene1].channels[ChannelType::shadePositionOutside].value = 50.0;
        hal.getCache().update(blind->getDSID(), cache);
      }
      THEN("Without force, it does not change") {
        REQUIRE(hal.callScene(blind->getDSID(), Scene1, false, now) == false);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(*state.channels[ChannelType::shadePositionOutside].current == 75.0);
        REQUIRE(!state.terminatesAt);
        REQUIRE(state.localPriority == true);
      }
      THEN("With force, it does change") {
        REQUIRE(hal.callScene(blind->getDSID(), Scene1, true, now) == true);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].target);
        REQUIRE(*state.channels[ChannelType::shadePositionOutside].target == 50.0);
        REQUIRE(state.terminatesAt);
        REQUIRE(state.localPriority == false);
      }
    }
    WHEN("A dimming scene is called") {
      {
        auto cache = hal.getCache().tryLookup(blind->getDSID());
        cache.property[DeviceProperty::posTimeUp] = 20 * 100;
        cache.property[DeviceProperty::posTimeDown] = 20 * 100;
        cache.scene.sceneTable[Scene1].channels[ChannelType::shadePositionOutside].value = 50.0;
        hal.getCache().update(blind->getDSID(), cache);
      }
      THEN("Without force, it does not change") {
        REQUIRE(hal.callScene(blind->getDSID(), SceneInc, false, now) == false);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(*state.channels[ChannelType::shadePositionOutside].current == 75.0);
        REQUIRE(!state.terminatesAt);
        REQUIRE(state.localPriority == true);
      }
      THEN("With force, it does change") {
        REQUIRE(hal.callScene(blind->getDSID(), SceneInc, true, now) == true);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.dimmStatus == DeviceDimmingStatus::stepping);
        REQUIRE(state.terminatesAt);
        REQUIRE(state.localPriority == true);
      }
    }
    WHEN("A scene with ignoreLocalProp is called") {
      {
        auto cache = hal.getCache().tryLookup(blind->getDSID());
        cache.scene.sceneTable[Scene1].flags.dontCare = false;
        cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
        cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = true;
        cache.scene.sceneTable[Scene1].channels[ChannelType::shadePositionOutside].value = 50.0;
        hal.getCache().update(blind->getDSID(), cache);
      }
      THEN("Without force, it does change") {
        REQUIRE(hal.callScene(blind->getDSID(), Scene1, false, now) == true);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].target);
        REQUIRE(*state.channels[ChannelType::shadePositionOutside].target == 50.0);
        REQUIRE(state.terminatesAt);
        REQUIRE(state.localPriority == false);
        REQUIRE(state.terminatesAt);
      }
      THEN("With force, it does change") {
        REQUIRE(hal.callScene(blind->getDSID(), Scene1, true, now) == true);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].target);
        REQUIRE(*state.channels[ChannelType::shadePositionOutside].target == 50.0);
        REQUIRE(state.terminatesAt);
        REQUIRE(state.localPriority == false);
      }
    }
    WHEN("A local off scene is called") {
      {
        auto cache = hal.getCache().tryLookup(blind->getDSID());
        cache.scene.sceneTable[SceneLocalOff].flags.dontCare = false;
        cache.scene.sceneTable[SceneLocalOff].flags.dimTime = DimTime::noDimTime;
        cache.scene.sceneTable[SceneLocalOff].flags.ignoreLocalPrio = false;
        cache.scene.sceneTable[SceneLocalOff].channels[ChannelType::shadePositionOutside].value = 50.0;
        hal.getCache().update(blind->getDSID(), cache);
      }
      THEN("Without force, it becomes prioritized") {
        REQUIRE(hal.callScene(blind->getDSID(), SceneLocalOff, false, now) == true);
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].target);
        REQUIRE(*state.channels[ChannelType::shadePositionOutside].target == 50.0);
        REQUIRE(state.terminatesAt);
        REQUIRE(state.localPriority == false);
      }
    }
  }

  GIVEN("A scene with dontCare") {
    {
      auto cache = hal.getCache().tryLookup(blind->getDSID());
      cache.scene.sceneTable[Scene1].flags.dontCare = true;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::shadePositionOutside].value = 100.0;
      hal.getCache().update(blind->getDSID(), cache);
    }
    {
      auto state = lightState.tryLookup(blind->getDSID());
      state.channels[ChannelType::shadePositionOutside].current = 50.0;
      state.lastSceneCalled = 13;
      state.localPriority = false;
      lightState.update(blind->getDSID(), state);
    }
    WHEN("After the scene is called") {
      REQUIRE(hal.callScene(blind->getDSID(), Scene1, false, now) == false);
      THEN("state has not changed") {
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(*state.channels[ChannelType::shadePositionOutside].current == 50);
        REQUIRE(!state.channels[ChannelType::shadePositionOutside].target);
        REQUIRE(*state.lastSceneCalled == 13);
      }
    }
  }

  GIVEN("A setOutputValue call") {
    {
      auto state = lightState.tryLookup(blind->getDSID());
      state.channels[ChannelType::shadePositionOutside].current = 50.0;
      state.channels[ChannelType::shadePositionOutside].target = 55.0;
      state.lastSceneCalled = 13;
      state.localPriority = true;
      lightState.update(blind->getDSID(), state);
    }
    WHEN("After setOutputValue is called") {
      REQUIRE(hal.setOutputValue(blind->getDSID(), 0xD3, now) == true);
      THEN("state is changed") {
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
        REQUIRE(state.channels[ChannelType::shadePositionOutside].target);
        REQUIRE(roundDigits(*state.channels[ChannelType::shadePositionOutside].target, 2) == 82.75);
        REQUIRE(!state.lastSceneCalled);
        REQUIRE(state.localPriority == false);
      }
    }
  }

  GIVEN("A setLocalPrioCall") {
    {
      auto state = lightState.tryLookup(blind->getDSID());
      state.channels[ChannelType::shadePositionOutside].current = 50.0;
      state.channels[ChannelType::shadePositionOutside].target = 55.0;
      state.lastSceneCalled = Scene42;
      state.localPriority = false;
      lightState.update(blind->getDSID(), state);
    }
    WHEN("After setLocalPrioCall is called with matching sceneId") {
      REQUIRE_NOTHROW(hal.setSceneLocalPrio(blind->getDSID(), Scene42));
      THEN("state is changed") {
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.localPriority == true);
      }
    }
    WHEN("After setLocalPrioCall is called with matching sceneId") {
      REQUIRE_NOTHROW(hal.setSceneLocalPrio(blind->getDSID(), Scene43));
      THEN("state is changed") {
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.localPriority == false);
      }
    }
  }

  GIVEN("A operationLock call") {
    WHEN("After operationLock(lock) device is locked") {
      {
        auto state = lightState.tryLookup(blind->getDSID());
        state.operationLocked = false;
        lightState.update(blind->getDSID(), state);
      }
      REQUIRE_NOTHROW(hal.setOperationLock(blind->getDSID(), true));
      THEN("state is changed") {
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.operationLocked == true);
        auto status = hal.getState(blind->getDSID());
        REQUIRE(status.status == dsModel::OutputStatusStatus::blocked);
      }
    }
    WHEN("After operationLock(lock) device is un-locked") {
      {
        auto state = lightState.tryLookup(blind->getDSID());
        state.operationLocked = true;
        lightState.update(blind->getDSID(), state);
      }
      REQUIRE_NOTHROW(hal.setOperationLock(blind->getDSID(), false));
      THEN("state is changed") {
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.operationLocked == false);
        auto status = hal.getState(blind->getDSID());
        REQUIRE(status.status == dsModel::OutputStatusStatus::ok);
      }
    }
  }

  GIVEN("a shade device is controlled by writing to the runtime bank") {
    THEN("writing to fist byte of position start to move") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(blind->getDSID(), 64, 0x02, 0x34));
      auto state = lightState.tryLookup(blind->getDSID());
      REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
      REQUIRE(roundDigits(*state.channels[ChannelType::shadePositionOutside].current, 2) == 0.08);
    }
    THEN("writing to second byte of position start to move") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(blind->getDSID(), 64, 0x03, 0x34));
      auto state = lightState.tryLookup(blind->getDSID());
      REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
      REQUIRE(roundDigits(*state.channels[ChannelType::shadePositionOutside].current, 2) == 20.31);
    }
    THEN("writing to angle byte changes angle") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(blind->getDSID(), 64, 0x04, 0x34));
      auto state = lightState.tryLookup(blind->getDSID());
      REQUIRE(state.channels[ChannelType::shadeOpeningAngleOutside].current);
      REQUIRE(roundDigits(*state.channels[ChannelType::shadeOpeningAngleOutside].current, 2) == 20.39);
    }
  }

  GIVEN("a shade device is controlled by writing to the runtime bank with existing current state") {
    {
      auto state = lightState.tryLookup(blind->getDSID());
      state.channels[ChannelType::shadePositionOutside].current = 50;
      lightState.update(blind->getDSID(), state);
    }
    THEN("writing to fist byte of position start to move") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(blind->getDSID(), 64, 0x02, 0x34));
      auto state = lightState.tryLookup(blind->getDSID());
      REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
      REQUIRE(roundDigits(*state.channels[ChannelType::shadePositionOutside].current, 2) == 50.08);
    }
    THEN("writing to fist byte of position start to move") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(blind->getDSID(), 64, 0x03, 0x34));
      auto state = lightState.tryLookup(blind->getDSID());
      REQUIRE(state.channels[ChannelType::shadePositionOutside].current);
      REQUIRE(roundDigits(*state.channels[ChannelType::shadePositionOutside].current, 2) == 20.31);
    }
  }

  GIVEN("a moving shade device read by reading from the untime bank with existing current state") {
    {
      auto state = lightState.tryLookup(blind->getDSID());
      state.channels[ChannelType::shadePositionOutside].current = 50;
      state.channels[ChannelType::shadePositionOutside].target = 100;
      state.terminatesAt = now + std::chrono::milliseconds(1500);
      lightState.update(blind->getDSID(), state);
    }
    THEN("writing to fist byte of position start to move") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(blind->getDSID(), 64, 0x02, 0x34));
      auto state = lightState.tryLookup(blind->getDSID());
      REQUIRE(state.channels[ChannelType::shadePositionOutside].target);
      REQUIRE(roundDigits(*state.channels[ChannelType::shadePositionOutside].target, 2) == 50.08);
    }
    THEN("writing to fist byte of position start to move") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(blind->getDSID(), 64, 0x03, 0x34));
      auto state = lightState.tryLookup(blind->getDSID());
      REQUIRE(state.channels[ChannelType::shadePositionOutside].target);
      REQUIRE(roundDigits(*state.channels[ChannelType::shadePositionOutside].target, 2) == 20.31);
    }
  }

  GIVEN("a shade device that participates in scene1 calls") {
    auto cache = hal.getCache().tryLookup(blind->getDSID());
    cache.scene.sceneTable[Scene1].flags.dontCare = false;
    hal.getCache().update(blind->getDSID(), cache);

    auto state = lightState.tryLookup(blind->getDSID());
    state.sceneInRAMBank = Scene4;
    lightState.update(blind->getDSID(), state);

    WHEN("scene1 is called") {
      REQUIRE(hal.callScene(blind->getDSID(), Scene1, false, now) == true);
      THEN("sceneInRAMBank is set to scene1") {
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.sceneInRAMBank == Scene1);
      }
    }

    WHEN("SceneInc is called") {
      REQUIRE(hal.callScene(blind->getDSID(), SceneInc, false, now) == true);
      THEN("sceneInRAMBank is cleared") {
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(!state.sceneInRAMBank);
      }
    }
  }

  GIVEN("a shade device that we don't know yet if it participates in scene1 calls") {
    auto cache = hal.getCache().tryLookup(blind->getDSID());
    cache.scene.sceneTable[Scene1].flags.dontCare.reset();
    hal.getCache().update(blind->getDSID(), cache);

    auto state = lightState.tryLookup(blind->getDSID());
    state.sceneInRAMBank = Scene4;
    lightState.update(blind->getDSID(), state);

    WHEN("scene1 is called") {
      REQUIRE(hal.callScene(blind->getDSID(), Scene1, false, now) == false);

      THEN("sceneInRAMBank is cleared") {
        // stay with the safe statement, which is we don't know
        // try to minimize this by reading out dontCare flags first especially for 3.8.7 devices
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(!state.sceneInRAMBank);
      }
    }
  }

  GIVEN("a shade device that does not participate in scene1 calls") {
    auto cache = hal.getCache().tryLookup(blind->getDSID());
    cache.scene.sceneTable[Scene1].flags.dontCare = true;
    hal.getCache().update(blind->getDSID(), cache);

    auto state = lightState.tryLookup(blind->getDSID());
    state.sceneInRAMBank = Scene4;
    lightState.update(blind->getDSID(), state);

    WHEN("scene1 is called") {
      REQUIRE(hal.callScene(blind->getDSID(), Scene1, false, now) == false);

      THEN("sceneInRAMBank remains unchanged") {
        // stay with the safe statement, which is we don't know
        // try to minimize this by reading out dontCare flags first especially for 3.8.7 devices
        auto state = lightState.tryLookup(blind->getDSID());
        REQUIRE(state.sceneInRAMBank == Scene4);
      }
    }
  }
}

// TODO merge with functions in cache-updater
static std::string toPath(int sceneId, ChannelType channelType) {
  return std::string("/scene/") + std::to_string(sceneId) + "/" + ds::fromEnum<std::string>(channelType);
}

SCENARIO("FW 3.8.7 workaround: channel oputput values can't be read from flash", std::string(TAGS) + "[fw387]") {
  auto pApartment = makeApartment(nullptr);
  auto& apt = *pApartment;
  BusInterfaceMock busInterface;
  EventSinkMock eventSink;
  ModelMaintenanceMock modelMaintenance;

  apt.setBusInterface(&busInterface);
  apt.setEventSink(eventSink);
  apt.setModelMaintenance(&modelMaintenance);

  ds::sqlite3::Database d(":memory:");
  auto halPtr = makeHardwareAbstraction(&apt, &d);
  auto& hal = *halPtr;
  hal.getCache().empty();

  constexpr ds::DateTime now{1505731033777ms};

  GIVEN("Shade device with 3.8.7 FW bug") {
    auto blind = createDevice(apt, DeviceMockConf{.hardwareType = HardwareType::GR_KL220, .revisionId = 0x387});

    auto sceneCache = hal.getCache().tryLookup(blind->getDSID());

    WHEN("Output levels of both channels for Scene1 are unknown") {
      auto& scene1Cfg = sceneCache.scene.sceneTable[Scene1];
      scene1Cfg.flags.dontCare = false; // callScene function will skip if device is not part of scene
      scene1Cfg.channels[ChannelType::shadePositionOutside] = {};
      scene1Cfg.channels[ChannelType::shadeOpeningAngleOutside] = {};
      hal.getCache().update(blind->getDSID(), sceneCache);

      THEN("CallScene will wake-up the cache-update thread") {
        REQUIRE_CALL(eventSink, pushEvent(_)); // only one event
        REQUIRE(hal.callScene(blind->getDSID(), Scene1, false, now) == true);
      }
    }

    WHEN("Cache is up-to-date for Scene1") {
      auto& scene1Cfg = sceneCache.scene.sceneTable[Scene1];
      scene1Cfg.flags.dontCare = false;
      scene1Cfg.channels[ChannelType::shadePositionOutside] = {.value = 50.0, .dontCare = false};
      scene1Cfg.channels[ChannelType::shadeOpeningAngleOutside] = {.value = 50.0, .dontCare = false};
      hal.getCache().update(blind->getDSID(), sceneCache);

      THEN("CallScene will not wake-up the cache-update thread") {
        REQUIRE(hal.callScene(blind->getDSID(), Scene1, false, now) == true);
      }
    }

    WHEN("Output level of 2nd channel(AngleOutside) for Scene1 yet missing") {
      auto& scene1Cfg = sceneCache.scene.sceneTable[Scene1];
      scene1Cfg.flags.dontCare = false;
      scene1Cfg.channels[ChannelType::shadePositionOutside] = {.value = 50.0, .dontCare = false};
      scene1Cfg.channels[ChannelType::shadeOpeningAngleOutside] = {};
      hal.getCache().update(blind->getDSID(), sceneCache);

      THEN("CallScene will wake-up the cache-update thread") {
        REQUIRE_CALL(eventSink, pushEvent(_));
        REQUIRE(hal.callScene(blind->getDSID(), Scene1, false, now) == true);
      }
    }

    WHEN("First channel is dontCare but 2nd channel needs to be read out") {
      auto& scene1Cfg = sceneCache.scene.sceneTable[Scene1];
      scene1Cfg.flags.dontCare = false;
      scene1Cfg.channels[ChannelType::shadePositionOutside] = {.dontCare = true};
      scene1Cfg.channels[ChannelType::shadeOpeningAngleOutside] = {};
      hal.getCache().update(blind->getDSID(), sceneCache);

      THEN("CallScene should wake-up the cache-update thread") {
        REQUIRE_CALL(eventSink, pushEvent(_));
        REQUIRE(hal.callScene(blind->getDSID(), Scene1, false, now) == true);

        auto state = hal.getStateCache().tryLookup(blind->getDSID());
        REQUIRE(state.sceneInRAMBank == Scene1);
      }
    }

    WHEN("First channel is fully cached and the other output has its dontCare set") {
      auto& scene1Cfg = sceneCache.scene.sceneTable[Scene1];
      scene1Cfg.flags.dontCare = false; // callScene function will skip if device is not part of scene
      scene1Cfg.channels[ChannelType::shadePositionOutside] = {.value = 50.0, .dontCare = false};
      scene1Cfg.channels[ChannelType::shadeOpeningAngleOutside] = {.dontCare = true};
      hal.getCache().update(blind->getDSID(), sceneCache);

      THEN("CallScene will not wake-up the cache-update thread") {
        REQUIRE(hal.callScene(blind->getDSID(), Scene1, false, now) == true);
      }
    }

    WHEN("readDoubleRAM is called for non-active scene") {
      THEN("call returns NaN and cache is not modified") {
        REQUIRE(std::isnan(hal.readDouble(blind->getDSID(), toPath(Scene1, ChannelType::shadePositionOutside))));
        REQUIRE(std::isnan(hal.readDouble(blind->getDSID(), toPath(Scene1, ChannelType::shadeOpeningAngleOutside))));

        auto sceneCache = hal.getCache().tryLookup(blind->getDSID());
        REQUIRE(!sceneCache.scene.sceneTable[Scene1].channels[ChannelType::shadePositionOutside].value);
        REQUIRE(!sceneCache.scene.sceneTable[Scene2].channels[ChannelType::shadeOpeningAngleOutside].value);
      }
    }

    WHEN("readDoubleRAM is called for active scene") {
      auto state = hal.getStateCache().tryLookup(blind->getDSID());
      state.sceneInRAMBank = Scene1;
      hal.getStateCache().update(blind->getDSID(), state);

      auto& scene1Cfg = sceneCache.scene.sceneTable[Scene1];
      scene1Cfg.flags.dontCare = false;
      scene1Cfg.channels[ChannelType::shadePositionOutside] = {.dontCare = false};
      scene1Cfg.channels[ChannelType::shadeOpeningAngleOutside] = {.dontCare = false};
      hal.getCache().update(blind->getDSID(), sceneCache);

      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, CfgClassRuntime, 2)).RETURN(0xf);  // pos low
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, CfgClassRuntime, 3)).RETURN(0x7f); // pos hi
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, CfgClassRuntime, 4)).RETURN(0x4c); // angle 30%

      THEN("call succeeds and values are cached") {
        REQUIRE(hal.readDouble(blind->getDSID(), toPath(Scene1, ChannelType::shadePositionOutside)) ==
                Approx(50.0).epsilon(0.5));
        REQUIRE(hal.readDouble(blind->getDSID(), toPath(Scene1, ChannelType::shadeOpeningAngleOutside)) ==
                Approx(30.0).epsilon(0.5));

        auto sceneCache = hal.getCache().tryLookup(blind->getDSID());
        REQUIRE(sceneCache.scene.sceneTable[Scene1].channels[ChannelType::shadePositionOutside].value ==
                Approx(50.0).epsilon(0.5));
        REQUIRE(sceneCache.scene.sceneTable[Scene1].channels[ChannelType::shadeOpeningAngleOutside].value ==
                Approx(30.0).epsilon(0.5));
      }
    }

    WHEN("readDoubleRAM is called for scene we don't know if device participate") {
      // shade of gray, should sceneInRAMBank be set if dontCare is yet not known?
      // currently we don't set it in that case, but cache update was faster if we allowed it

      auto state = hal.getStateCache().tryLookup(blind->getDSID());
      state.sceneInRAMBank = Scene1;
      hal.getStateCache().update(blind->getDSID(), state);

      auto& scene1Cfg = sceneCache.scene.sceneTable[Scene1];
      scene1Cfg.flags.dontCare.reset();
      scene1Cfg.channels[ChannelType::shadePositionOutside] = {.dontCare = false};
      scene1Cfg.channels[ChannelType::shadeOpeningAngleOutside] = {.dontCare = false};
      hal.getCache().update(blind->getDSID(), sceneCache);

      THEN("call returns NaN and cache is not modified") {
        REQUIRE(std::isnan(hal.readDouble(blind->getDSID(), toPath(Scene1, ChannelType::shadePositionOutside))));
        REQUIRE(std::isnan(hal.readDouble(blind->getDSID(), toPath(Scene1, ChannelType::shadeOpeningAngleOutside))));

        auto sceneCache = hal.getCache().tryLookup(blind->getDSID());
        REQUIRE(!sceneCache.scene.sceneTable[Scene1].channels[ChannelType::shadePositionOutside].value);
        REQUIRE(!sceneCache.scene.sceneTable[Scene2].channels[ChannelType::shadeOpeningAngleOutside].value);
      }
    }

    WHEN("readDoubleRAM is called for dontCare scene") {
      auto state = hal.getStateCache().tryLookup(blind->getDSID());
      state.sceneInRAMBank = Scene1;
      hal.getStateCache().update(blind->getDSID(), state);

      auto& scene1Cfg = sceneCache.scene.sceneTable[Scene1];
      scene1Cfg.flags.dontCare = true;
      scene1Cfg.channels[ChannelType::shadePositionOutside] = {.dontCare = false};
      scene1Cfg.channels[ChannelType::shadeOpeningAngleOutside] = {.dontCare = false};
      hal.getCache().update(blind->getDSID(), sceneCache);

      THEN("call returns NaN and cache is not modified") {
        REQUIRE(std::isnan(hal.readDouble(blind->getDSID(), toPath(Scene1, ChannelType::shadePositionOutside))));
        REQUIRE(std::isnan(hal.readDouble(blind->getDSID(), toPath(Scene1, ChannelType::shadeOpeningAngleOutside))));

        auto sceneCache = hal.getCache().tryLookup(blind->getDSID());
        REQUIRE(!sceneCache.scene.sceneTable[Scene1].channels[ChannelType::shadePositionOutside].value);
        REQUIRE(!sceneCache.scene.sceneTable[Scene2].channels[ChannelType::shadeOpeningAngleOutside].value);
      }
    }

    WHEN("readDoubleRAM is called for dontCare channel") {
      auto state = hal.getStateCache().tryLookup(blind->getDSID());
      state.sceneInRAMBank = Scene1;
      hal.getStateCache().update(blind->getDSID(), state);

      auto& scene1Cfg = sceneCache.scene.sceneTable[Scene1];
      scene1Cfg.flags.dontCare = false;
      scene1Cfg.channels[ChannelType::shadePositionOutside] = {.dontCare = true};
      scene1Cfg.channels[ChannelType::shadeOpeningAngleOutside] = {.dontCare = false};
      hal.getCache().update(blind->getDSID(), sceneCache);

      THEN("call returns NaN and cache is not modified, but other channels work") {
        REQUIRE_CALL(busInterface.device, getDeviceConfig(_, CfgClassRuntime, 4)).RETURN(0x4c); // angle 30%

        REQUIRE(hal.readDouble(blind->getDSID(), toPath(Scene1, ChannelType::shadeOpeningAngleOutside)) ==
                Approx(30.0).epsilon(0.5));
        REQUIRE(std::isnan(hal.readDouble(blind->getDSID(), toPath(Scene1, ChannelType::shadePositionOutside))));

        auto sceneCache = hal.getCache().tryLookup(blind->getDSID());
        REQUIRE(std::isnan(hal.readDouble(blind->getDSID(), toPath(Scene1, ChannelType::shadePositionOutside))));
        REQUIRE(sceneCache.scene.sceneTable[Scene1].channels[ChannelType::shadeOpeningAngleOutside].value ==
                Approx(30.0).epsilon(0.5));
      }
    }
  }
}

} // namespace
