/*
    Copyright (c) 2009 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#pragma once

#include <ds/log.h>
/*
 * Usage:
 * class FooBar {
 *  __DECL_LOG_CHANNEL__
 *
 * foobar.cpp
 * __DEFINE_LOG_CHANNEL__(yourClassName, logLevel)
 *
 * Adds function 'log' with personalized 'FooBar' channel
 * to your class
 */
#define __DEFINE_LOG_CHANNEL__(_NAME_, level) LogChannel _NAME_::dsLogChannel(#_NAME_, level);
#define __DECL_LOG_CHANNEL__                                                    \
  static LogChannel dsLogChannel;                                               \
  static void log(std::string message, const aLogSeverity severity) {           \
    dsLogChannel.log(logSeverityToDsLogSeverity(severity), std::move(message)); \
  }

namespace dss {

enum class aLogSeverity {
  // numeric values published in property tree
  lsDebug = 0,
  lsInfo = 1,
  lsNotice = 2,
  lsWarning = 3,
  lsError = 4,
  lsFatal = 5
};

static constexpr aLogSeverity lsDebug = aLogSeverity::lsDebug;
static constexpr aLogSeverity lsInfo = aLogSeverity::lsInfo;
static constexpr aLogSeverity lsNotice = aLogSeverity::lsNotice;
static constexpr aLogSeverity lsWarning = aLogSeverity::lsWarning;
static constexpr aLogSeverity lsError = aLogSeverity::lsError;
static constexpr aLogSeverity lsFatal = aLogSeverity::lsFatal;

aLogSeverity dsLogSeverityToLogSeverity(ds::LogSeverity x);
ds::LogSeverity logSeverityToDsLogSeverity(aLogSeverity x);

class LogChannel : public ds::LogChannel {
  // Subclass ds::LogChannel to be able to mix ds logging macros with DSS logging.
public:
  LogChannel(std::string name, aLogSeverity severity = lsNotice)
      : ds::LogChannel(std::move(name), logSeverityToDsLogSeverity(severity)) {}

  aLogSeverity getMinimumSeverity() const { return dsLogSeverityToLogSeverity(severity()); }
  void setSeverity(aLogSeverity x) { ds::LogChannel::setSeverity(logSeverityToDsLogSeverity(x)); }
  const std::string& getName() const { return name(); }
  bool maylog(aLogSeverity _severity) const { return _severity >= getMinimumSeverity(); }
};

class Logger {
public:
  static Logger* getInstance();

  void log(std::string message, const aLogSeverity _severity = lsNotice);
  static void log(const LogChannel& channel, std::string message, const aLogSeverity _severity = lsDebug);

  LogChannel* getLogChannel() { return &m_defaultLogChannel; }

private:
  Logger();
  LogChannel m_defaultLogChannel;
}; // Logger

// implementation details

inline ds::LogSeverity logSeverityToDsLogSeverity(aLogSeverity x) {
  switch (x) {
    case lsFatal:
      return ds::LogSeverity::FATAL;
    case lsError:
      return ds::LogSeverity::ERROR;
    case lsWarning:
      return ds::LogSeverity::WARNING;
    case lsNotice:
      return ds::LogSeverity::NOTICE;
    case lsInfo:
      return ds::LogSeverity::INFO;
    case lsDebug:
      return ds::LogSeverity::DEBUG;
  }
  DS_FAIL_REQUIRE("Cannot convert dss log severity", static_cast<int>(x));
}
}
