/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "metering.h"

#include <ds/catch/catch.h>
#include <ds/system-mock.h>
#include <model/apartment-mock.h>

#include <cmath>
#include <trompeloeil.hpp>

#include "archives.h"
#include "rrd-interface.h"

using namespace trompeloeil;
using namespace std::literals;

using namespace dss;
using namespace dss::metering;

using system_clock = std::chrono::system_clock;

static ds::LogChannel dsLogChannel("dssMeteringMeteringTest");

namespace {

static const char* TAGS = "[dSS][metering][core]";

struct ArchivesMock : public trompeloeil::mock_interface<dss::metering::Archives> {
  IMPLEMENT_CONST_MOCK0(getConfig);
  IMPLEMENT_CONST_MOCK0(restartCached);
  IMPLEMENT_CONST_MOCK1(filepath);
  IMPLEMENT_CONST_MOCK1(flush);
  IMPLEMENT_CONST_MOCK1(withoutUpdates);
  IMPLEMENT_CONST_MOCK1(hasArchive);
  IMPLEMENT_CONST_MOCK2(checkDBFlush);
  IMPLEMENT_MOCK1(create);
  IMPLEMENT_MOCK1(lastValue);
  IMPLEMENT_MOCK1(setRRDCachedPath);
  IMPLEMENT_MOCK1(setMeteringStorageLocation);
  IMPLEMENT_MOCK1(tuneDBPowerSettings);
  IMPLEMENT_MOCK2(addValue);
  IMPLEMENT_MOCK2(validateDBSeries);
  IMPLEMENT_MOCK6(xport);
};

SCENARIO("add/retrieve data", TAGS) {
  auto propertySystem = dss::PropertySystem{};
  auto configNode = propertySystem.createProperty("/config");
  dss::ApartmentMock apartment{nullptr, &propertySystem};

  auto createController = [&apartment](const dsModel::ControllerId& controllerId) {
    static int ct{};
    auto controller = apartment.allocateDSMeter(dss::Dsuid{controllerId});
    controller->setName(ds::str("controller", ct++));
    controller->setIsPresent(true);
    controller->setIsConnected(true); // isPresent checks isConnected internally
    controller->setCapability_HasMetering(true);
    return controller;
  };

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto controller1 = createController(controllerId1);

  auto system = ds::SystemMock();
  auto archives = std::make_shared<ArchivesMock>();
  auto metering = makeMockedMetering(apartment, archives, system);

  ALLOW_CALL(*archives, filepath(_)).RETURN(ds::str(_1, ".rrd"));

  auto addConsumptionController = [&metering, &archives = *archives](auto& controller, ArchiveValue&& archiveValue) {
    REQUIRE_CALL(archives, hasArchive(_)).RETURN(true); // -> m_meters.push_back
    REQUIRE_CALL(archives, lastValue(_)).RETURN(archiveValue);
    metering->getEnergy(controller.getDSID());
  };

  auto addProductionController = [&metering, &archives = *archives](auto& controller, ArchiveValue&& archiveValue) {
    // controller.setCapability_HasMeteringProducerEnabled(false);
    REQUIRE_CALL(archives, hasArchive(_)).RETURN(true);
    REQUIRE_CALL(archives, lastValue(_)).RETURN(archiveValue);
    metering->getProducedEnergy(controller.getDSID());
  };

  SECTION("getSeries") {
    GIVEN("valid archives") {
      REQUIRE_CALL(*archives, hasArchive(ArchiveId{controllerId1, Category::consumption})).RETURN(true);
      REQUIRE_CALL(*archives, getConfig()).RETURN(std::vector{MeteringConfig{1s, 600}});
      REQUIRE_CALL(*archives, flush(std::vector{ArchiveId{controllerId1, Category::consumption}}));
      REQUIRE_CALL(*archives, xport(1, 601, 1, std::vector{ArchiveId{controllerId1, Category::consumption}},
                                    SeriesTypes::power, false))
          .RETURN([]() -> ValueRange {
            std::vector<Value> ret;
            while (ret.size() < 600) {
              ret.push_back(Value{1, ds::DateTime{1s}}); // all valid data, incl. last entry
            }
            return ValueRange{.step = 1, .values = std::move(ret)};
          }());

      WHEN("single valid controller") {
        auto range = metering->getSeries({Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                         {ds::DateTime{1s}, ds::DateTime{600s}}, 0);
        CHECK(range.values.size() == 600);
      }

      WHEN("add ghost controllers") {
        // metering disabled -> no data
        auto controllerId2 = dsModel::ControllerId{"1111111111111111111111111111111200"};
        REQUIRE_CALL(*archives, hasArchive(ArchiveId{controllerId2, Category::consumption})).RETURN(false);
        auto controller2 = apartment.allocateDSMeter(dss::Dsuid{controllerId2});

        // metering but no data
        auto controllerId3 = dsModel::ControllerId{"1111111111111111111111111111111300"};
        auto controller3 = apartment.allocateDSMeter(dss::Dsuid{controllerId3});
        controller3->setCapability_HasMetering(true);
        REQUIRE_CALL(*archives, hasArchive(ArchiveId{controllerId3, Category::consumption})).RETURN(false);

        THEN("same as above") {
          auto range = metering->getSeries({Dsuid{controllerId1}, Dsuid{controllerId2}, Dsuid{controllerId3}}, 1s,
                                           SeriesTypes::power, false, {ds::DateTime{1s}, ds::DateTime{600s}}, 0);
          CHECK(range.values.size() == 600);
        }
      }
    }
    WHEN("empty list of dsmid given") {
      apartment.removeDSMeter(dss::Dsuid{controllerId1});
      THEN("no data returned") {
        auto range = metering->getSeries({}, 1s, SeriesTypes::power, false, {ds::DateTime{1s}, ds::DateTime{600s}}, 0);
        CHECK(range == ValueRange{});
      }
    }
  }

  GIVEN("new measurements") {
    // first only sets baseline to compute delta, only RAM
    REQUIRE_CALL(*archives, hasArchive(_)).RETURN(false);
    metering->addConsumedEnergy(*controller1, {75, 100, ds::DateTime(100s)});

    WHEN("archive already exists") {
      THEN("value is added") {
        REQUIRE_CALL(*archives, hasArchive(_)).RETURN(true);
        REQUIRE_CALL(*archives, lastValue(_))
            .RETURN(ArchiveValue{.power = 0.0, .energy = 50.0, .lastChanged = ds::DateTime{0s}});
        REQUIRE_CALL(*archives,
                     addValue(ArchiveId{Dsuid{"1111111111111111111111111111111100"s}, Category::consumption},
                              ArchiveValue{.power = 75.0, .energy = 250.0, .lastChanged = ds::DateTime{105s}}));
        metering->addConsumedEnergy(*controller1, {75, 300, ds::DateTime(105s)});
      }
    }
    WHEN("archive needs creation") {
      REQUIRE_CALL(*archives, hasArchive(_)).RETURN(false).TIMES(2); // getEnergy, getOrCreateCachedSeries
      REQUIRE_CALL(*archives, create(_));
      REQUIRE_CALL(*archives, addValue(ArchiveId{Dsuid{"1111111111111111111111111111111100"s}, Category::consumption},
                                       ArchiveValue{.power = 75, .energy = 200.0, .lastChanged = ds::DateTime{101s}}));
      metering->addConsumedEnergy(*controller1, {75, 300, ds::DateTime(101s)});
    }

    SECTION("jumps in consumed energy detected") {
      WHEN("timestamp of new measurement predates previous") {
        THEN("nothing added to archive") {
          ALLOW_CALL(*archives, hasArchive(_)).RETURN(false); // property tree update
          metering->addConsumedEnergy(*controller1, {75, 300, ds::DateTime(99s)});
        }
      }
      WHEN("new energy decreases") {
        THEN("nothing added to archive") {
          ALLOW_CALL(*archives, hasArchive(_)).RETURN(false); // property tree update
          metering->addConsumedEnergy(*controller1, {75, 99, ds::DateTime(101s)});
        }
      }
      THEN("nothing added to archive if limit exceeded") {
        ALLOW_CALL(*archives, hasArchive(_)).RETURN(false); // property tree update
        metering->addConsumedEnergy(*controller1, {75, 8000 + 100, ds::DateTime(101s)});
      }
      THEN("archive is updated if below limit") {
        REQUIRE_CALL(*archives, create(_));
        REQUIRE_CALL(*archives, addValue(_, _));
        ALLOW_CALL(*archives, hasArchive(_)).RETURN(false); // property tree update
        metering->addConsumedEnergy(*controller1, {75, 8000 + 99, ds::DateTime(101s)});
      }
      WHEN("updates have been lost") {
        THEN("nothing added to archive if more then 5s") {
          ALLOW_CALL(*archives, hasArchive(_)).RETURN(false); // property tree update
          metering->addConsumedEnergy(*controller1, {75, 300, ds::DateTime(106s)});
        }
      }
    }
  }

  GIVEN("getEnergy") {
    WHEN("archive does not exist") {
      REQUIRE_CALL(*archives, hasArchive(_)).RETURN(false);
      THEN("no data return") { CHECK(!metering->getEnergy(dss::Dsuid{controllerId1})); }
    }
    WHEN("archive exists") {
      REQUIRE_CALL(*archives, hasArchive(_)).RETURN(true);

      WHEN("accessed first time") {
        REQUIRE_CALL(*archives, lastValue(_))
            .RETURN(ArchiveValue{.power = 0.0, .energy = 50.0, .lastChanged = ds::DateTime{10s}});
        THEN("data returned from DB") {
          CHECK(metering->getEnergy(dss::Dsuid{controllerId1}) ==
                ArchiveValue{.power = 0, .energy = 50.0, .lastChanged = ds::DateTime(10s)});
        }
      }
      WHEN("lastValue fails") {
        REQUIRE_CALL(*archives, lastValue(_)).RETURN(std::nullopt);
        THEN("return is nullopt") { CHECK(!metering->getEnergy(dss::Dsuid{controllerId1})); }
      }
    }
  }

  GIVEN("getLastReading") {
    using dsModel::ApartmentMeteringOrigin;
    using dsModel::ControllerMeteringOrigin;
    using dsModel::MeteringType;
    addConsumptionController(*controller1, {20, 3600, ds::DateTime(10s)});

    SECTION("ControllerOrigin") {
      CHECK(metering->getLastReading(ControllerMeteringOrigin{.id = controllerId1}, MeteringType::energyMetering) ==
            ArchiveValue{20, 3600, ds::DateTime{10s}});
      CHECK(metering->getLastReading(ControllerMeteringOrigin{.id = controllerId1}, MeteringType::energyMetering,
                                     EnergyUnit::Ws) == ArchiveValue{20, 3600, ds::DateTime{10s}});
      CHECK(metering->getLastReading(ControllerMeteringOrigin{.id = controllerId1}, MeteringType::energyMetering,
                                     EnergyUnit::Wh) == ArchiveValue{20, 1, ds::DateTime{10s}});
    }
    SECTION("ApartmentOrigin") {
      auto controllerId2 = dsModel::ControllerId{"1111111111111111111111111111111200"};
      auto controller2 = createController(controllerId2);
      addConsumptionController(*controller2, {10, 1800, ds::DateTime(10s)});

      CHECK(metering->getLastReading(ApartmentMeteringOrigin{}, MeteringType::energyMetering) ==
            ArchiveValue{30, 5400, ds::DateTime{10s}});
      CHECK(metering->getLastReading(ApartmentMeteringOrigin{}, MeteringType::powerMetering) ==
            ArchiveValue{30, 5400, ds::DateTime{10s}});
    }
  }

  SECTION("15min average uploads") {
    REQUIRE_CALL(*archives, hasArchive(_)).RETURN(true);
    REQUIRE_CALL(*archives, getConfig()).RETURN(std::vector{MeteringConfig{15min, 1000}});
    REQUIRE_CALL(*archives, flush(std::vector{ArchiveId{controllerId1, Category::consumption}}));

    WHEN("no data present") {
      REQUIRE_CALL(*archives,
                   xport(_, _, std::chrono::duration_cast<std::chrono::seconds>(15min).count(),
                         std::vector{ArchiveId{controllerId1, Category::consumption}}, SeriesTypes::energyDelta, true))
          .RETURN(ValueRange{.start = 1, .end = 901, .step = 900, .values = {}}); // empty
      auto ret = metering->get15minConsumed({dss::Dsuid{controllerId1}}, SeriesTypes::energyDelta);
      CHECK(ret == 0.0);
    }
    WHEN("NaN value found") {
      THEN("exception thrown") { // rely on getSeries to filter them out
        REQUIRE_CALL(*archives, xport(_, _, std::chrono::duration_cast<std::chrono::seconds>(15min).count(),
                                      std::vector{ArchiveId{controllerId1, Category::consumption}},
                                      SeriesTypes::energyDelta, true))
            .RETURN(ValueRange{1, 901, 900, {{std::nan("myNaN"), ds::DateTime{0s}}}});
        CHECK_THROWS(metering->get15minConsumed({dss::Dsuid{controllerId1}}, SeriesTypes::energyDelta));
      }
    }
  }

  SECTION("meterings listing") {
    auto& meter = *controller1;

    using dsModel::ApartmentMeteringOrigin;
    using dsModel::ControllerMeteringOrigin;
    using dsModel::MeteringOrigin;
    using dsModel::MeteringType;

    GIVEN("consumption meter") {
      addConsumptionController(*controller1, {75, 100, ds::DateTime(100s)});

      SECTION("normal case") {
        CHECK(metering->meterings() ==
              std::vector{
                  std::pair<MeteringOrigin, MeteringType>{ControllerMeteringOrigin{.id = controllerId1},
                                                          MeteringType::powerMetering},
                  std::pair<MeteringOrigin, MeteringType>{ControllerMeteringOrigin{.id = controllerId1},
                                                          MeteringType::energyMetering},
                  std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::powerMetering},
                  std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::energyMetering}});
      }
      SECTION("meter is not present") {
        meter.setIsPresent(false);
        CHECK(metering->meterings() == std::vector<std::pair<MeteringOrigin, MeteringType>>{});
      }
      SECTION("meter does not have metering capability") {
        meter.setCapability_HasMetering(false);
        CHECK(metering->meterings() == std::vector<std::pair<MeteringOrigin, MeteringType>>{});
      }
    }
    GIVEN("production meter") {
      addConsumptionController(*controller1, {75, 100, ds::DateTime(100s)});
      addProductionController(*controller1, {75, 100, ds::DateTime(101s)});

      meter.setCapability_HasMeteringProducer(true);
      meter.setCapability_HasMeteringProducerEnabled(true);

      SECTION("normal case") {
        CHECK(
            metering->meterings() ==
            std::vector{
                std::pair<MeteringOrigin, MeteringType>{ControllerMeteringOrigin{.id = controllerId1},
                                                        MeteringType::powerMetering},
                std::pair<MeteringOrigin, MeteringType>{ControllerMeteringOrigin{.id = controllerId1},
                                                        MeteringType::energyMetering},
                std::pair<MeteringOrigin, MeteringType>{ControllerMeteringOrigin{.id = controllerId1},
                                                        MeteringType::powerProducedMetering},
                std::pair<MeteringOrigin, MeteringType>{ControllerMeteringOrigin{.id = controllerId1},
                                                        MeteringType::energyProducedMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::powerMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::energyMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::powerProducedMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{},
                                                        MeteringType::energyProducedMetering}});
      }
      SECTION("meter does not have metering capability") {
        meter.setCapability_HasMetering(false);
        CHECK(
            metering->meterings() ==
            std::vector{
                std::pair<MeteringOrigin, MeteringType>{ControllerMeteringOrigin{.id = controllerId1},
                                                        MeteringType::powerProducedMetering},
                std::pair<MeteringOrigin, MeteringType>{ControllerMeteringOrigin{.id = controllerId1},
                                                        MeteringType::energyProducedMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::powerProducedMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{},
                                                        MeteringType::energyProducedMetering}});
      }
      SECTION("meter does not have metering producer enabled") {
        meter.setCapability_HasMeteringProducerEnabled(false);
        CHECK(metering->meterings() ==
              std::vector{
                  std::pair<MeteringOrigin, MeteringType>{ControllerMeteringOrigin{.id = controllerId1},
                                                          MeteringType::powerMetering},
                  std::pair<MeteringOrigin, MeteringType>{ControllerMeteringOrigin{.id = controllerId1},
                                                          MeteringType::energyMetering},
                  std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::powerMetering},
                  std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::energyMetering}});
      }
    }
  }
}

SCENARIO("update/retrieve from cache", TAGS) {
  auto propertySystem = dss::PropertySystem{};
  auto configNode = propertySystem.createProperty("/config");
  dss::ApartmentMock apartment{nullptr, &propertySystem};

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto controller1 = apartment.allocateDSMeter(dss::Dsuid{controllerId1});
  controller1->setName("controller1");
  controller1->setCapability_HasMetering(true);

  auto system = ds::SystemMock();
  auto archives = std::make_shared<ArchivesMock>();
  auto metering = makeMockedMetering(apartment, archives, system);

  ALLOW_CALL(*archives, filepath(_)).RETURN(ds::str(_1, ".rrd"));

  GIVEN("empty rrd / this test is never modifying to round-robin archive") {
    WHEN("1st sensor value is received") {
      REQUIRE_CALL(*archives, hasArchive(_)).RETURN(false);
      metering->addConsumedEnergy(*controller1, {.power = 0, .energy = 10, .timestamp = ds::DateTime(300s)});
      THEN("total energy counter is not updated yet") {
        // cache is not populated upon 1st value, also rrd file does not exist
        REQUIRE_CALL(*archives, hasArchive(_)).RETURN(false);
        CHECK(!metering->getEnergy(str2dsuid(controllerId1)).has_value());
      }

      WHEN("2nd sensor value is reiceved") {
        REQUIRE_CALL(*archives, hasArchive(_)).RETURN(true);
        REQUIRE_CALL(*archives, lastValue(_))
            .RETURN(ArchiveValue{.power = 0.0, .energy = 0.0, .lastChanged = ds::DateTime{0s}});
        REQUIRE_CALL(*archives, addValue(_, ArchiveValue{.power = 1, .energy = 2, .lastChanged = ds::DateTime(301s)}));
        metering->addConsumedEnergy(*controller1, {.power = 1, .energy = 12, .timestamp = ds::DateTime(301s)});
        THEN("energy counter is increased by (2nd - 1st) sensor value") {
          CHECK(metering->getEnergy(str2dsuid(controllerId1)) ==
                ArchiveValue{.power = 1, .energy = 2.0, .lastChanged = ds::DateTime(301s)});
        }

        WHEN("sensor value is smaller then previous sensor value") {
          // sets new baseline
          metering->addConsumedEnergy(*controller1, {.power = 1, .energy = 2, .timestamp = ds::DateTime(302s)});

          THEN("energy counter is not changed") {
            CHECK(metering->getEnergy(str2dsuid(controllerId1)) ==
                  ArchiveValue{.power = 1, .energy = 2.0, .lastChanged = ds::DateTime(301s)});
          }

          WHEN("next sensor value is bigger then previous") {
            REQUIRE_CALL(*archives,
                         addValue(_, ArchiveValue{.power = 2, .energy = 3.0, .lastChanged = ds::DateTime(303s)}));
            metering->addConsumedEnergy(*controller1,
                                        {.power = 2, .energy = 3, .timestamp = ds::DateTime(303s)}); // 1 - 0
            THEN("energy counter is increased again") {
              CHECK(metering->getEnergy(str2dsuid(controllerId1)) ==
                    ArchiveValue{.power = 2, .energy = 3.0, .lastChanged = ds::DateTime(303s)});
            }
          }
        }
      }
    }
  }

  GIVEN("non-empty rrd / this test is never modifying to round-robin archive") {
    REQUIRE_CALL(*archives, hasArchive(ArchiveId{controllerId1, Category::consumption})).RETURN(true);
    REQUIRE_CALL(*archives, lastValue(_))
        .RETURN(ArchiveValue{.power = 0.0, .energy = 12.5, .lastChanged = ds::DateTime{300s}});

    THEN("last value is known") {
      CHECK(metering->getEnergy(str2dsuid(controllerId1)) ==
            ArchiveValue{.power = 0, .energy = 12.5, .lastChanged = ds::DateTime(300s)});
    }
  }

  GIVEN("archive with UINT_MAX") {
    REQUIRE_CALL(*archives, hasArchive(ArchiveId{controllerId1, Category::consumption})).RETURN(true);
    REQUIRE_CALL(*archives, lastValue(_))
        .RETURN(ArchiveValue{.power = 0.0, .energy = UINT_MAX, .lastChanged = ds::DateTime{0s}});
    REQUIRE_CALL(*archives, addValue(_, _)).LR_SIDE_EFFECT(CHECK(_2.energy == static_cast<double>(UINT_MAX) + 1));

    THEN("energy counter is increased by (2nd - 1st) sensor value") {
      metering->addConsumedEnergy(*controller1, {.power = 1, .energy = 12, .timestamp = ds::DateTime(300s)});
      metering->addConsumedEnergy(*controller1, {.power = 1, .energy = 13, .timestamp = ds::DateTime(301s)});
    }
  }
}

SCENARIO("energy production", TAGS) {
  auto propertySystem = dss::PropertySystem{};
  auto configNode = propertySystem.createProperty("/config");
  dss::ApartmentMock apartment{nullptr, &propertySystem};

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto controller1 = apartment.allocateDSMeter(dss::Dsuid{controllerId1});
  controller1->setName("controller1");
  controller1->setCapability_HasMetering(true);

  auto system = ds::SystemMock();
  auto archives = std::make_shared<ArchivesMock>();
  auto metering = makeMockedMetering(apartment, archives, system);

  ALLOW_CALL(*archives, filepath(_)).RETURN(ds::str(_1, ".rrd"));

  SECTION("addProducedEnergy") {
    // online sets baseline
    metering->addProducedEnergy(*controller1, {75, 100, ds::DateTime(100s)}); // baseline

    // second call will start adding values
    REQUIRE_CALL(*archives, hasArchive(_)).RETURN(true);
    REQUIRE_CALL(*archives, lastValue(_))
        .RETURN(ArchiveValue{.power = 0.0, .energy = 50.0, .lastChanged = ds::DateTime{0s}});
    REQUIRE_CALL(*archives, addValue(ArchiveId{Dsuid{"1111111111111111111111111111111100"s}, Category::production},
                                     ArchiveValue{.power = 115.0, .energy = 260.0, .lastChanged = ds::DateTime{101s}}));
    metering->addProducedEnergy(*controller1, {115, 310, ds::DateTime(101s)});

    THEN("getEnergyValue last value is known") {
      CHECK(metering->getProducedEnergy(Dsuid(controllerId1)) ==
            ArchiveValue{.power = 115, .energy = 260, .lastChanged = ds::DateTime(101s)});
    }

    SECTION("jumps in consumed energy detected") {
      WHEN("timestamp of new measurement predates previous") {
        THEN("nothing added to archive") {
          metering->addProducedEnergy(*controller1, {75, 10000 + 310, ds::DateTime(99s)});
        }
      }
      WHEN("timestamp of new measurement equals previous") {
        THEN("nothing added to archive") {
          metering->addProducedEnergy(*controller1, {75, 49999 + 310, ds::DateTime(101s)});
        }
      }
      WHEN("new timestamp is a few millisconds less than a full second") {
        THEN("value is added to archive") {
          REQUIRE_CALL(*archives, addValue(_, _));
          metering->addProducedEnergy(*controller1, {75, gsl::narrow_cast<unsigned int>(670.0 * 49999 / 1000 + 310),
                                                     ds::DateTime(101s + 670ms)});
        }
      }
      WHEN("when more then 5s pass bewtween 2 updates") {
        THEN("nothing added to archive") {
          metering->addProducedEnergy(*controller1, {75, 49999 + 310, ds::DateTime(101s + 6s)});
        }
      }
      THEN("nothing added to archive if limit exceeded") {
        metering->addProducedEnergy(*controller1, {75, 50000 + 310, ds::DateTime(102s)});
      }
      THEN("archive is updated if below limit") {
        REQUIRE_CALL(*archives, addValue(_, _));
        metering->addProducedEnergy(*controller1, {75, 49999 + 310, ds::DateTime(102s)});
      }
      WHEN("an update is lost") {
        auto x = GENERATE(2s, 3s, 4s, 5s);
        THEN("limit is multiplied by the delay upon next update") {
          INFO(ds::str(x));
          REQUIRE_CALL(*archives, addValue(_, _));
          metering->addProducedEnergy(
              *controller1, {75, gsl::narrow_cast<double>(x.count() * 50000 + 310 - 1), ds::DateTime(101s + x)});
        }
      }
    }
  }

  WHEN("getProducedSeries") {
    REQUIRE_CALL(*archives, hasArchive(_))
        .LR_SIDE_EFFECT(CHECK(_1.first == Dsuid{controllerId1}))
        .LR_SIDE_EFFECT(CHECK(_1.second == Category::production))
        .RETURN(true);
    REQUIRE_CALL(*archives, getConfig()).RETURN(std::vector{MeteringConfig{1s, 600}});
    REQUIRE_CALL(*archives, flush(std::vector{ArchiveId{Dsuid{controllerId1}, Category::production}}));
    REQUIRE_CALL(*archives, xport(1, 601, 1, std::vector{ArchiveId(controllerId1, Category::production)},
                                  SeriesTypes::power, false))
        .RETURN([]() -> ValueRange {
          std::vector<Value> ret;
          while (ret.size() < 600) {
            ret.push_back(Value{1, ds::DateTime{1s}}); // all valid data, incl. last entry
          }
          return ValueRange{.step = 1, .values = std::move(ret)};
        }());

    WHEN("single valid controller") {
      auto range = metering->getProducedSeries({Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                               {ds::DateTime{1s}, ds::DateTime{600s}}, 0);
      CHECK(range.values.size() == 600);
    }
  }
}

SCENARIO("getSeries start/end", TAGS) {
  auto propertySystem = dss::PropertySystem{};
  auto configNode = propertySystem.createProperty("/config");
  dss::ApartmentMock apartment{nullptr, &propertySystem};

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto controller1 = apartment.allocateDSMeter(dss::Dsuid{controllerId1});
  controller1->setName("controller1");
  controller1->setCapability_HasMetering(true);

  auto system = ds::SystemMock();
  auto archives = std::make_shared<ArchivesMock>();
  auto metering = makeMockedMetering(apartment, archives, system);

  ALLOW_CALL(*archives, filepath(_)).RETURN(ds::str(_1, ".rrd"));
  ALLOW_CALL(*archives, flush(std::vector{ArchiveId{Dsuid{controllerId1}, Category::consumption}}));

  REQUIRE_CALL(*archives, getConfig())
      .TIMES(AT_LEAST(1))
      .RETURN(std::vector{MeteringConfig{1s, 600}, MeteringConfig{1min, 720}, MeteringConfig{15min, 2976},
                          MeteringConfig{24h, 370}, MeteringConfig{7 * 24h, 260}, MeteringConfig{30 * 24h, 60}});

  REQUIRE_CALL(*archives, hasArchive(ArchiveId{controllerId1, Category::consumption})).RETURN(true);

  GIVEN("start/end") {
    WHEN("resolution=1") {
      REQUIRE_CALL(*archives, xport(1, 601, 1, _, _, _))
          .RETURN(ValueRange{.step = 1, .values = std::vector<Value>(600)});
      auto values = metering->getSeries({Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                        {ds::DateTime{1s}, ds::DateTime{600s}}, 0);
      WHEN("values < max values for given resolution") {
        REQUIRE_CALL(*archives, xport(1, 601, 1, _, _, _))
            .RETURN(ValueRange{.step = 1, .values = std::vector<Value>(10, {1, ds::DateTime{1s}})});
        auto range = metering->getSeries({Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                         {ds::DateTime{1s}, ds::DateTime{600s}}, 10);
        CHECK(range.values.size() == 10); // no change, more values than valueCount is returned
      }
    }

    WHEN("resolution=60") {
      REQUIRE_CALL(*archives, xport(0, 660, 60, _, _, _))
          .RETURN(ValueRange{.step = 60, .values = std::vector<Value>(720, {1, ds::DateTime{1s}})});
      auto range = metering->getSeries({Dsuid{controllerId1}}, 60s, SeriesTypes::power, false,
                                       {ds::DateTime{1s}, ds::DateTime{600s}}, 0);
      CHECK(range.values.size() == 720);
    }
  }

  GIVEN("start/values") {
    WHEN("resolution=1") {
      REQUIRE_CALL(*archives, xport(1, 102, 1, _, _, _))
          .RETURN(ValueRange{.step = 1, .values = std::vector<Value>(100, {1, ds::DateTime{1s}})});
      auto range = metering->getSeries({Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                       {ds::DateTime{1s}, std::nullopt}, 100);
      CHECK(range.step == 1);
      CHECK(range.values.size() == 100);

      WHEN("unlimited values") {
        auto now = system_clock::now();

        // bug: end time is derived from now, start is not adjusted
        REQUIRE_CALL(*archives, xport(1, _, 1, _, _, _))
            .LR_SIDE_EFFECT(CHECK(system_clock::to_time_t(now) <= _2))
            .RETURN(ValueRange{.step = 1, .values = std::vector<Value>(600, {1, ds::DateTime{1s}})});
        auto range = metering->getSeries({Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                         {ds::DateTime{1s}, std::nullopt}, 0);
        CHECK(range.step == 1);
        CHECK(range.values.size() == 600); // real check is start/end in REQUIRE_CALL
      }

      WHEN("more values then available") {
        REQUIRE_CALL(*archives, xport(1, _, 1, _, _, _))
            .RETURN(ValueRange{.step = 1, .values = std::vector<Value>(600)});
        auto range = metering->getSeries({Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                         {ds::DateTime{1s}, std::nullopt}, 1000);
        CHECK(range.step == 1);
        CHECK(range.values.size() == 599); // off-by-one due to null Value
      }
    }

    WHEN("resolution=60") {
      REQUIRE_CALL(*archives, xport(0, 6060, 60, _, _, _))
          .RETURN(ValueRange{.step = 60, .values = std::vector<Value>(100, {1, ds::DateTime{1s}})});
      auto range = metering->getSeries({Dsuid{controllerId1}}, 60s, SeriesTypes::power, false,
                                       {ds::DateTime{1s}, std::nullopt}, 100);
      CHECK(range.step == 60);
      CHECK(range.values.size() == 100);

      WHEN("unlimited values") {
        auto now = system_clock::now();

        // bug: end is derived form now(), not relative to end
        REQUIRE_CALL(*archives, xport(0, _, 60, _, _, _))
            .LR_SIDE_EFFECT(CHECK(system_clock::to_time_t(now) - (720 + 1) * 60 < _2))
            .RETURN(ValueRange{.step = 60, .values = std::vector<Value>(720, {1, ds::DateTime{1s}})});
        auto range = metering->getSeries({Dsuid{controllerId1}}, 60s, SeriesTypes::power, false,
                                         {ds::DateTime{1s}, std::nullopt}, 0);
        CHECK(range.values.size() == 720); // real check is start/end in REQUIRE_CALL
        CHECK(range.step == 60);
      }
    }
  }

  GIVEN("end/values") {
    WHEN("resolution=1") {
      REQUIRE_CALL(*archives, xport(500, 601, 1, _, _, _))
          .RETURN(ValueRange{.step = 1, .values = std::vector<Value>(100, {1, ds::DateTime{1s}})});
      auto range = metering->getSeries({Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                       {std::nullopt, ds::DateTime{600s}}, 100);
      CHECK(range.step == 1);
      CHECK(range.values.size() == 100);

      WHEN("request more values then avail") {
        REQUIRE_CALL(*archives, xport(0, 601, 1, _, _, _))
            .RETURN(ValueRange{.step = 1, .values = std::vector<Value>(600, {1, ds::DateTime{1s}})});
        auto range = metering->getSeries({Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                         {std::nullopt, ds::DateTime{600s}}, 1000);
        CHECK(range.step == 1);
        CHECK(range.values.size() == 600);
      }
    }
    WHEN("resolution=60") {
      REQUIRE_CALL(*archives, xport(960, 7020, 60, _, _, _))
          .RETURN(ValueRange{.step = 60, .values = std::vector<Value>(100, {1, ds::DateTime{1s}})});
      auto range = metering->getSeries({Dsuid{controllerId1}}, 60s, SeriesTypes::power, false,
                                       {std::nullopt, ds::DateTime{7000s}}, 100);
      CHECK(range.step == 60);
      CHECK(range.values.size() == 100);

      WHEN("unlimited values") {
        auto now = system_clock::now();

        // bug: start is derived form now(), not relative to end
        REQUIRE_CALL(*archives, xport(_, 43320, 60, _, _, _))
            .LR_SIDE_EFFECT(CHECK(system_clock::to_time_t(now) - (720 + 1) * 60 < _1))
            .RETURN(ValueRange{.step = 60, .values = std::vector<Value>(720, {1, ds::DateTime{1s}})});
        auto range = metering->getSeries({Dsuid{controllerId1}}, 60s, SeriesTypes::power, false,
                                         {std::nullopt, ds::DateTime{43260s}}, 0);
        CHECK(range.step == 60);
        CHECK(range.values.size() == 720);
      }
    }
  }

  WHEN("intermediate resolution") {
    WHEN("resolution=2s") {
      CHECK_THROWS(metering->getSeries({Dsuid{controllerId1}}, 2s, SeriesTypes::power, false,
                                       {std::nullopt, ds::DateTime{600s}}, INT_MAX));
    }
    WHEN("resolution=61s") {
      CHECK_THROWS(metering->getSeries({Dsuid{controllerId1}}, 61s, SeriesTypes::power, false,
                                       {std::nullopt, ds::DateTime{600s}}, INT_MAX));
    }
    WHEN("resolution=15min+1s") {
      CHECK_THROWS(metering->getSeries({Dsuid{controllerId1}}, 15min + 1s, SeriesTypes::power, false,
                                       {std::nullopt, ds::DateTime{600s}}, INT_MAX));
    }
    WHEN("resolution=24h+1s") {
      CHECK_THROWS(metering->getSeries({Dsuid{controllerId1}}, 24h + 1s, SeriesTypes::power, false,
                                       {std::nullopt, ds::DateTime{600s}}, INT_MAX));
    }
    WHEN("resolution=24h+1s") {
      CHECK_THROWS(metering->getSeries({Dsuid{controllerId1}}, 7 * 24h + 1s, SeriesTypes::power, false,
                                       {std::nullopt, ds::DateTime{600s}}, INT_MAX));
    }
    WHEN("resolution=30d+1s") {
      CHECK_THROWS(metering->getSeries({Dsuid{controllerId1}}, 30 * 24h + 1s, SeriesTypes::power, false,
                                       {std::nullopt, ds::DateTime{600s}}, INT_MAX));
    }
  }
}

SCENARIO("total power", TAGS) {
  auto propertySystem = dss::PropertySystem{};
  auto configNode = propertySystem.createProperty("/config");
  dss::ApartmentMock apartment{nullptr, &propertySystem};

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto controller1 = apartment.allocateDSMeter(dss::Dsuid{controllerId1});

  auto system = ds::SystemMock();
  auto archives = std::make_shared<ArchivesMock>();
  auto metering = makeMockedMetering(apartment, archives, system);

  ALLOW_CALL(*archives, filepath(_)).RETURN(ds::str(_1, ".rrd"));

  GIVEN("no metering trippled in yet") {
    THEN("getApartmentPower") {
      CHECK(metering->getConsumedPowerApartment() == 0);
      CHECK(metering->getProducedPowerApartment() == 0);
    }
  }

  GIVEN("consumption/production power") {
    REQUIRE_CALL(*archives, hasArchive(_)).TIMES(AT_LEAST(4)).RETURN(false);
    ALLOW_CALL(*archives, create(_));
    ALLOW_CALL(*archives, addValue(_, _));

    auto now = system_clock::now();
    metering->addProducedEnergy(*controller1, {75, 100, now - 1s}); // first baseline is set
    metering->addProducedEnergy(*controller1, {75, 120, now});      // 2nd call populates cache

    metering->addConsumedEnergy(*controller1, {25, 100, now - 1s});
    metering->addConsumedEnergy(*controller1, {25, 120, now});

    THEN("getApartmentPower") {
      CHECK(metering->getConsumedPowerApartment() == 25);
      CHECK(metering->getProducedPowerApartment() == 75);
    }

    GIVEN("2nd controller") {
      auto controllerId2 = dsModel::ControllerId{"1111111111111111111111111111111200"};
      auto controller2 = apartment.allocateDSMeter(dss::Dsuid{controllerId2});

      metering->addProducedEnergy(*controller2, {75, 100, now - 1s});
      metering->addProducedEnergy(*controller2, {75, 120, now});

      metering->addConsumedEnergy(*controller2, {25, 100, now - 1s});
      metering->addConsumedEnergy(*controller2, {25, 120, now});

      THEN("getApartmentPower doubles") {
        CHECK(metering->getConsumedPowerApartment() == 50);
        CHECK(metering->getProducedPowerApartment() == 150);
      }
    }
  }
}

SCENARIO("get instant power", TAGS) {
  auto propertySystem = dss::PropertySystem{};
  auto configNode = propertySystem.createProperty("/config");
  dss::ApartmentMock apartment{nullptr, &propertySystem};

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto controller1 = apartment.allocateDSMeter(dss::Dsuid{controllerId1});

  auto system = ds::SystemMock();
  auto archives = std::make_shared<ArchivesMock>();
  auto metering = makeMockedMetering(apartment, archives, system);

  ALLOW_CALL(*archives, filepath(_)).RETURN(ds::str(_1, ".rrd"));

  GIVEN("current data") {
    REQUIRE_CALL(*archives, hasArchive(_)).TIMES(AT_LEAST(4)).RETURN(false);
    ALLOW_CALL(*archives, create(_));
    ALLOW_CALL(*archives, addValue(_, _));

    auto now = std::chrono::system_clock::now();
    metering->addConsumedEnergy(*controller1, {0, 1000, now - 1s});
    metering->addProducedEnergy(*controller1, {0, 100, now - 1s}); // first baseline is set

    metering->addConsumedEnergy(*controller1, {25, 2000, now});
    metering->addProducedEnergy(*controller1, {75, 200, now}); // 2nd call populates cache

    THEN("instant power is set") {
      CHECK(75 == metering->getProducedPower(dss::Dsuid{controllerId1}));
      CHECK(25 == metering->getConsumedPower(dss::Dsuid{controllerId1}));
    }
    THEN("instant energy is set") {
      CHECK(metering->getLastEnergy(dss::Dsuid{controllerId1}) == TotalEnergy{.production = 100, .consumption = 1000});
    }
  }

  GIVEN("old sensor data") {
    REQUIRE_CALL(*archives, hasArchive(_)).TIMES(AT_LEAST(4)).RETURN(false);
    ALLOW_CALL(*archives, create(_));
    ALLOW_CALL(*archives, addValue(_, _));

    auto now = std::chrono::system_clock::now();
    metering->addConsumedEnergy(*controller1, {0, 1000, now - 10s});
    metering->addProducedEnergy(*controller1, {0, 100, now - 10s}); // first baseline is set

    metering->addConsumedEnergy(*controller1, {25, 2000, now - 9s});
    metering->addProducedEnergy(*controller1, {75, 200, now - 9s}); // 2nd call populates cache

    WHEN("no update for 3s") {
      THEN("instant power is 0") {
        CHECK(metering->getConsumedPower(dss::Dsuid{controllerId1}) == 0);
        CHECK(metering->getProducedPower(dss::Dsuid{controllerId1}) == 0);
      }
      THEN("last energy is preserved") {
        CHECK(metering->getLastEnergy(dss::Dsuid{controllerId1}) ==
              TotalEnergy{.production = 100, .consumption = 1000});
      }
    }
  }
}

SCENARIO("migrate files", TAGS) {
  auto propertySystem = dss::PropertySystem{};
  auto configNode = propertySystem.createProperty("/config");
  dss::ApartmentMock apartment{nullptr, &propertySystem};

  auto system = ds::SystemMock();
  auto archives = std::make_shared<ArchivesMock>();
  auto metering = makeMockedMetering(apartment, archives, system);

  ALLOW_CALL(*archives, filepath(_)).RETURN(ds::str(_1, ".rrd"));

  GIVEN("files in archive folder") {
    REQUIRE_CALL(system, lsRegularFiles(_, ".rrd"s))
        .RETURN(std::vector{"/foo/1111111111111111111111111111111100.rrd"s,
                            "/foo/1111111111111111111111111111111100-production.rrd"s});
    CHECK(metering->searchArchivesInFolder() ==
          std::vector{ArchiveId{"1111111111111111111111111111111100"s, Category::consumption},
                      ArchiveId{"1111111111111111111111111111111100"s, Category::production}});
  }

  GIVEN("legacy dsid archive") {
    REQUIRE_CALL(system, lsRegularFiles(_, ".rrd"s)).RETURN(std::vector{"/foo/3504175fe000000000016be7.rrd"s});

    THEN("file is rename") {
      REQUIRE_CALL(system, existsFile(_))
          .LR_SIDE_EFFECT(CHECK(_1 == "/invalid-store/(3504175fe00000000000000000016be700,consumption).rrd"sv))
          .RETURN(false);
      REQUIRE_CALL(system, rename(_, _))
          .LR_SIDE_EFFECT(CHECK(_1 == "/foo/3504175fe000000000016be7.rrd"sv))
          .LR_SIDE_EFFECT(CHECK(_2 == "/invalid-store/(3504175fe00000000000000000016be700,consumption).rrd"sv));
      CHECK(metering->searchArchivesInFolder() ==
            std::vector{ArchiveId{"3504175fe00000000000000000016be700"s, Category::consumption}});
    }
    WHEN("newer dsuid file exists") {
      THEN("file is lieft untouched") {
        REQUIRE_CALL(system, existsFile(_))
            .LR_SIDE_EFFECT(CHECK(_1 == "/invalid-store/(3504175fe00000000000000000016be700,consumption).rrd"sv))
            .RETURN(true);
        CHECK(metering->searchArchivesInFolder() ==
              std::vector{ArchiveId{"3504175fe00000000000000000016be700"s, Category::consumption}});
      }
    }
  }

  GIVEN("invalid file") {
    REQUIRE_CALL(system, lsRegularFiles(_, ".rrd"s)).RETURN(std::vector{"/foo/foo.rrd"s});
    CHECK(metering->searchArchivesInFolder() == std::vector<ArchiveId>{});
  }

  SECTION("files receive live updates") {
    auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
    GIVEN("old file") {
      REQUIRE_CALL(*archives, withoutUpdates(_)).RETURN(true);
      REQUIRE_CALL(*archives, create(_));
      CHECK(metering->recreateDeadArchives(std::vector{ArchiveId{controllerId1, Category::consumption}}) == true);
    }
    GIVEN("valid file") {
      REQUIRE_CALL(*archives, withoutUpdates(_)).RETURN(false);
      CHECK(metering->recreateDeadArchives(std::vector{ArchiveId{controllerId1, Category::consumption}}) == false);
    }
  }
  SECTION("check file format ") {
    auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
    auto y = GENERATE(Category::consumption, Category::production);
    GIVEN("archive is up-to-date") {
      REQUIRE_CALL(*archives, hasArchive(_)).RETURN(true);
      REQUIRE_CALL(*archives, validateDBSeries(_, _)).RETURN(true);
      metering->validateAndConvertArchive({Dsuid{controllerId1}, y});
    }
    GIVEN("archive is invalid") {
      REQUIRE_CALL(*archives, hasArchive(_)).RETURN(true);
      REQUIRE_CALL(*archives, validateDBSeries(_, _)).RETURN(false);
      REQUIRE_CALL(system, remove(_));
      metering->validateAndConvertArchive({Dsuid{controllerId1}, y});
    }
    GIVEN("archive needs conversion ") {
      REQUIRE_CALL(*archives, hasArchive(_)).RETURN(true);
      REQUIRE_CALL(*archives, validateDBSeries(_, _)).LR_SIDE_EFFECT(_2 = true).RETURN(true);
      REQUIRE_CALL(*archives, tuneDBPowerSettings(_));
      metering->validateAndConvertArchive({Dsuid{controllerId1}, y});
    }
  }
}

} // namespace
