/*
    Copyright (c) 2009,2010,2011 digitalSTROM.org, Zurich, Switzerland

    Authors: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>
             Christian Hitz, aizo AG <christian.hitz@aizo.com>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#include "metering.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <digitalSTROM/dsuid.h>
#include <ds/asio/io-service.h>
#include <ds/asio/timer.h>
#include <ds/log.h>
#include <sys/types.h>
#include <time.h>

#include <algorithm>
#include <boost/filesystem.hpp>
#include <cmath>
#include <functional>
#include <iterator>
#include <numeric>

#include "businterface.h"
#include "dss.h"
#include "metering/archives.h"
#include "metering/config.h"
#include "metering/controller.h"
#include "metering/open-metering-system.h"
#include "metering/power-states.h"
#include "metering/rrd-interface.h"
#include "metering/socket-server.h"
#include "model/apartment.h"
#include "model/dsmeter.h"
#include "model/modelmaintenance.h"
#include "propertysystem.h"
#include "security/security.h"

static ds::LogChannel dsLogChannel("dssMeteringMetering");

using namespace std::literals;

using namespace dss;
using namespace dss::metering;

using system_clock = std::chrono::system_clock;

namespace {

auto toCategory = [](dsModel::MeteringType meteringType) {
  switch (meteringType) {
    case dsModel::MeteringType::powerMetering:
    case dsModel::MeteringType::energyMetering: 
    case dsModel::MeteringType::energyConsumedTariff1Metering:
    case dsModel::MeteringType::energyConsumedTariff2Metering: return Category::consumption;
    case dsModel::MeteringType::energyProducedMetering:
    case dsModel::MeteringType::powerProducedMetering:
    case dsModel::MeteringType::energyProducedTariff1Metering:
    case dsModel::MeteringType::energyProducedTariff2Metering: return Category::production;
    case dsModel::MeteringType::none:
    default: DS_FAIL_REQUIRE("bad metering type");
  }
};


class MeteringImpl : public MeteringSubsystem {
  Apartment& m_apartment;
  ds::System& m_system;
  ElectricityMeteringConfig m_electricityMeteringConfig;
  std::string m_MeteringStorageLocation{"/invalid-store/"};
  MeteringBusInterface* m_pMeteringBusInterface = NULL;

  std::shared_ptr<Archives> m_archives;
  std::shared_ptr<PowerStates> m_powerStates = makePowerStates(m_apartment);
  std::unique_ptr<Direction> m_consumption;
  std::unique_ptr<Direction> m_production;
  ds::Clock::time_point m_nextPollAt;
  std::unique_ptr<OpenMeteringSystem> m_openMeteringSystem;

public:
  void checkDBConsistency() override;
  std::vector<ArchiveId> searchArchivesInFolder() const override;
  bool recreateDeadArchives(const std::vector<ArchiveId>& archiveIds) const override;
  bool validateAndConvertArchive(const ArchiveId& archiveId) override;

  metering::Archives& getArchives() override { return *m_archives; }

  void doStart() override;
  void execute() override;
  void initialize() override;
  void pollLoop(boost::asio::io_service& io, ds::asio::Timer& timer);

public:
  MeteringImpl(DSS* _pDSS, Apartment& apartment, std::shared_ptr<Archives>&& archives, ds::System& system);
  ~MeteringImpl() override = default;

  const std::vector<MeteringConfig> getConfig() const override { return m_archives->getConfig(); }
  void setMeteringBusInterface(MeteringBusInterface* _value) override { m_pMeteringBusInterface = _value; }

  std::optional<unsigned long> getConsumedPower(const dsuid_t& meterId) const override {
    return gsl::narrow<unsigned int>(m_consumption->getInstantPower(meterId));
  }
  unsigned int getConsumedPowerApartment() const override { return m_consumption->getApartmentPower(); }

  std::optional<unsigned long> getProducedPower(const dsuid_t& meterId) const override {
    return gsl::narrow<unsigned int>(m_production->getInstantPower(meterId));
  }
  unsigned int getProducedPowerApartment() const override { return m_production->getApartmentPower(); }

  std::optional<ArchiveValue> getEnergy(const dsuid_t& meterId) const override {
    return m_consumption->getEnergy(meterId);
  }
  std::optional<ArchiveValue> getProducedEnergy(const dsuid_t& meterId) const override {
    return m_production->getEnergy(meterId);
  }

  double get15minConsumed(std::vector<Dsuid> meterIds, metering::SeriesTypes type) override;
  double get15minProduced(std::vector<Dsuid> meterIds, metering::SeriesTypes type) override;

  TotalEnergy getLastEnergy(const dsuid_t& meterId) const override {
    return {m_production->getLastEnergy(meterId), m_consumption->getLastEnergy(meterId)};
  }

  std::shared_ptr<State> createState(MeteringSystem meteringSystem, PowerStateConfig&& powerStateConfig) override {
    return m_powerStates->createState(meteringSystem, std::move(powerStateConfig));
  }
  void removeState(MeteringSystem meteringSystem, const PowerStateConfig& powerStateConfig) override {
    return m_powerStates->removeState(meteringSystem, powerStateConfig);
  }
  void triggerState(MeteringSystem meteringSystem, TotalPower&& totalPower) override {
    m_powerStates->trigger(meteringSystem, std::move(totalPower));
  }

  void addConsumedEnergy(DSMeter& meter, ElectricityMeasurement&& energyEvent) override;
  void addProducedEnergy(DSMeter& meter, ElectricityMeasurement&& energyMeasurement) override {
    m_production->addEnergy(meter.getDSID(), std::move(energyMeasurement));
  }

  void addReading(metering::P1Reading&& p1Reading) override {
    if (!m_openMeteringSystem) {
      m_openMeteringSystem = makeOpenMeteringSystem(*m_archives, m_electricityMeteringConfig);
    }
    m_openMeteringSystem->addReading(std::move(p1Reading));
  }

  ValueRange getSeries(std::vector<Dsuid> meterIds, std::chrono::seconds resolution, metering::SeriesTypes _type,
                       bool _energyInWh, const TimeDuration& timeDuration, int _valueCount) override {
    return m_consumption->getSeries(meterIds, resolution, _type, _energyInWh, timeDuration, _valueCount);
  }

  ValueRange getProducedSeries(std::vector<Dsuid> meterIds, std::chrono::seconds resolution,
                               metering::SeriesTypes _type, bool _energyInWh, const TimeDuration& timeDuration,
                               int _valueCount) override {
    return m_production->getSeries(meterIds, resolution, _type, _energyInWh, timeDuration, _valueCount);
  }

  std::vector<std::pair<dsModel::MeteringOrigin, dsModel::MeteringType>> meterings() const override {
    std::vector<std::pair<dsModel::MeteringOrigin, dsModel::MeteringType>> result;

    auto toControllerMeteringOrigin = [](const dsuid_t& dsuid) {
      return dsModel::ControllerMeteringOrigin{.id = dsModel::ControllerId{ds::str(dsuid)}};
    };

    auto meterPresent = [&apartment = m_apartment](const dsuid_t dsuid) {
      // dsModel::metering technical name DS_REQUIRE meter to be present...
      if (auto meter = apartment.tryGetDSMeterByDSID(dsuid); meter && meter->isPresent()) {
        return meter;
      }
      return decltype(apartment.tryGetDSMeterByDSID(dsuid)){};
    };

    bool hasConsumption{false};
    for (const auto& meterId : m_consumption->getActiveMeters()) {
      // futeless check for HasMetering, if it sends metering data
      if (auto meter = meterPresent(meterId); !meter || !meter->getCapability_HasMetering()) {
        continue;
      }
      result.push_back({toControllerMeteringOrigin(meterId), dsModel::MeteringType::powerMetering});
      result.push_back({toControllerMeteringOrigin(meterId), dsModel::MeteringType::energyMetering});
      hasConsumption = true;
    }

    bool hasProduction{false};
    for (const auto& meterId : m_production->getActiveMeters()) {
      // from dsModel::Meterings::meterings -- propably drop checking flags, unless instantly react to user input,
      if (auto meter = meterPresent(meterId); !meter || !meter->getCapability_HasMeteringProducerEnabled()) {
        continue;
      }

      result.push_back({toControllerMeteringOrigin(meterId), dsModel::MeteringType::powerProducedMetering});
      result.push_back({toControllerMeteringOrigin(meterId), dsModel::MeteringType::energyProducedMetering});
      hasProduction = true;
    }

    if (hasConsumption) {
      result.push_back({dsModel::ApartmentMeteringOrigin{}, dsModel::MeteringType::powerMetering});
      result.push_back({dsModel::ApartmentMeteringOrigin{}, dsModel::MeteringType::energyMetering});
    }

    if (hasProduction) {
      result.push_back({dsModel::ApartmentMeteringOrigin{}, dsModel::MeteringType::powerProducedMetering});
      result.push_back({dsModel::ApartmentMeteringOrigin{}, dsModel::MeteringType::energyProducedMetering});
    }

    return result;
  }

  std::optional<ArchiveValue> getLastReading(dsModel::MeteringOrigin meteringOrigin, dsModel::MeteringType meteringType,
                                             EnergyUnit energyUnit) const override {
    std::optional<ArchiveValue> result{};

    auto controllerMeterings = [meterings = meterings()](dsModel::MeteringType meteringType) {
      std::decay_t<decltype(meterings)> result;
      for (auto metering : meterings) {
        if (!std::holds_alternative<dsModel::ControllerMeteringOrigin>(metering.first)) {
          continue;
        }
        if (metering.second != meteringType) {
          continue;
        }
        result.push_back(metering);
      }
      return result;
    };

    auto accumulateControllers = [](auto& controllerDirection, auto&& meterings) {
      ArchiveValue result{};

      for (const auto& metering : meterings) {
        auto dsuid = Dsuid{std::get<dsModel::ControllerMeteringOrigin>(metering.first).id};
        if (auto reading = controllerDirection.getEnergy(dsuid)) {
          result.power += reading->power;
          result.energy += reading->energy;
          result.lastChanged = ds::max(result.lastChanged, reading->lastChanged);
        }
      }
      return result;
    };

    std::visit(::ds::Visitor{[&](auto& origin) { DS_FAIL_REQUIRE("not implemented"); },
                             [&](dsModel::ApartmentMeteringOrigin& origin) {
                               auto meterings = controllerMeterings(meteringType);
                               if (toCategory(meteringType) == Category::production) {
                                 result = accumulateControllers(*m_production, meterings);
                               } else {
                                 result = accumulateControllers(*m_consumption, meterings);
                               }
                             },
                             [&, this](dsModel::ControllerMeteringOrigin& origin) {
                               if (toCategory(meteringType) == Category::production) {
                                 result = m_production->getEnergy(Dsuid{origin.id});
                               } else {
                                 result = m_consumption->getEnergy(Dsuid{origin.id});
                               }
                             }},
               meteringOrigin);

    if (energyUnit == EnergyUnit::Wh) {
      result->energy /= 3600;
    }

    return result;
  }

  ValueRange getSeriesValue(dsModel::MeteringOrigin meteringOrigin, dsModel::MeteringType meteringType,
                            std::chrono::seconds resolution, int valueCount = 0,
                            bool energyInWh = false) const override {
    ValueRange result{};

    auto toSeriesType = [](dsModel::MeteringType meteringType) {
      switch (meteringType) {
        case dsModel::MeteringType::powerMetering: 
        case dsModel::MeteringType::powerProducedMetering: return SeriesTypes::power;
        case dsModel::MeteringType::energyMetering: 
        case dsModel::MeteringType::energyProducedMetering: 
        case dsModel::MeteringType::energyConsumedTariff1Metering:
        case dsModel::MeteringType::energyConsumedTariff2Metering:
        case dsModel::MeteringType::energyProducedTariff1Metering: 
        case dsModel::MeteringType::energyProducedTariff2Metering: return SeriesTypes::energy;
        case dsModel::MeteringType::none:
        default: DS_FAIL_REQUIRE("bad metering type");
      }
    };

    std::visit(
        ::ds::Visitor{[&](auto& origin) { DS_FAIL_REQUIRE("not implemented"); },
                      [&](dsModel::ApartmentMeteringOrigin& origin) { DS_FAIL_REQUIRE("not implemented"); },
                      [&, this](dsModel::ControllerMeteringOrigin& origin) {
                        if (toCategory(meteringType) == Category::production) {
                          result = m_production->getSeries({Dsuid{origin.id}}, resolution, toSeriesType(meteringType),
                                                           energyInWh, TimeDuration{}, valueCount);
                        } else {
                          result = m_consumption->getSeries({Dsuid{origin.id}}, resolution, toSeriesType(meteringType),
                                                            energyInWh, TimeDuration{}, valueCount);
                        }
                      }},
        meteringOrigin);
    return result;
  }
}; // MeteringImpl

MeteringImpl::MeteringImpl(DSS* _pDSS, Apartment& apartment, std::shared_ptr<Archives>&& archives, ds::System& system)
    : MeteringSubsystem(_pDSS),
      m_apartment(apartment),
      m_system(system),
      m_archives(std::move(archives)),
      m_consumption(makeDirection(m_archives.get(), Category::consumption, m_electricityMeteringConfig)),
      m_production(makeDirection(m_archives.get(), Category::production, m_electricityMeteringConfig)) {}

void MeteringImpl::addConsumedEnergy(DSMeter& meter, ElectricityMeasurement&& energyMeasurement) {
  auto meteringCache = MeteringCache{
      .power = gsl::narrow<int>(energyMeasurement.power), // property tree exported
      .lastSensorValue = gsl::narrow_cast<unsigned long>(energyMeasurement.energy),
      .lastChanged = dss::DateTime{energyMeasurement.timestamp},
  };

  m_consumption->addEnergy(meter.getDSID(), std::move(energyMeasurement));

  if (auto energyMetering = m_consumption->getEnergy(meter.getDSID())) {
    meteringCache.energyWs = energyMetering->energy;
    meteringCache.energyWh = energyMetering->energy / 3600;

    // backend data for property tree, etc
    meter.updateMeteringCache(std::move(meteringCache));
  }
}

double MeteringImpl::get15minConsumed(std::vector<Dsuid> meterIds, metering::SeriesTypes type) {
  return m_consumption->get15MinuteAverage(meterIds, type, Unit::Wh);
}

double MeteringImpl::get15minProduced(std::vector<Dsuid> meterIds, metering::SeriesTypes type) {
  return m_production->get15MinuteAverage(meterIds, type, Unit::Wh);
}

// subsystem interface
void MeteringImpl::initialize() {
  Subsystem::initialize();
  auto config = getDSS().getPropertySystem().createProperty(getConfigPropertyBasePath());
  m_MeteringStorageLocation =
      config->getOrCreateChildValue<std::string>("storageLocation", getDSS().getDataDirectory() + "metering/");
  m_MeteringStorageLocation = addTrailingBackslash(m_MeteringStorageLocation);
  DS_NOTICE(m_MeteringStorageLocation);
  m_archives->setRRDCachedPath(
      config->getOrCreateChildValue<std::string>("rrdDaemonAddress", "unix:/var/run/rrdcached.sock"));

  if (isEnabled()) {
    if (!boost::filesystem::is_directory(m_MeteringStorageLocation)) {
      throw std::runtime_error("Metering directory " +
                               boost::filesystem::system_complete(m_MeteringStorageLocation).string() +
                               " does not exist!");
    }

    if (!rwAccess(m_MeteringStorageLocation)) {
      throw std::runtime_error("Metering directory " +
                               boost::filesystem::system_complete(m_MeteringStorageLocation).string() +
                               " is not readable/writable!");
    }
  }

  m_electricityMeteringConfig = m_system.conf().readOr<ElectricityMeteringConfig>("electricityMetering", {});
  DS_NOTICE("config", ds::hideArg(m_electricityMeteringConfig));
  m_archives->setMeteringStorageLocation(std::string(m_MeteringStorageLocation));
}

bool MeteringImpl::recreateDeadArchives(const std::vector<ArchiveId>& archiveIds) const {
  bool restartCached{false};
  for (const auto& archiveId : archiveIds) {
    if (m_archives->withoutUpdates(archiveId)) {
      DS_WARNING("Metering DB ", m_archives->filepath(archiveId), " is out of date. removing ");
      m_archives->create(archiveId); // bogus, file is not removed, TODO add unit test to verify
      restartCached = true;
    }
  }
  return restartCached;
}

std::vector<ArchiveId> MeteringImpl::searchArchivesInFolder() const {
  constexpr auto len = "1111111111111111111111111111111100"sv.length();

  std::vector<ArchiveId> archiveIds{};

  for (auto file : m_system.lsRegularFiles(m_MeteringStorageLocation, ".rrd"s)) {
    auto stem = std::string(boost::filesystem::path(file).stem().c_str());
    DS_DEBUG(stem);

    dsuid_t dsuid;
    try {
      if (stem.size() <= 24) {
        dsid_t dsid = str2dsid(stem);
        dsuid = dsuid_from_dsid(&dsid);

        auto oldFile = std::exchange(
            file, ds::str(m_MeteringStorageLocation, m_archives->filepath({dsuid, Category::consumption})));

        if (m_system.existsFile(file)) {
          DS_WARNING("Failed to rename old dsid based archive, new file already exists");
          // m_system.remove(file);
        } else {
          m_system.rename(oldFile, file);
          DS_NOTICE("Migrated metering data from ", oldFile, " to ", file);
        }
      } else if (stem.length() >= len) {
        dsuid = str2dsuid(stem.substr(0, len));
      } else {
        DS_NOTICE("invalid archive file", file);
        continue;
      }

    } catch (std::exception& e) {
      DS_NOTICE("failed to migrate", file, e.what());
      continue;
    }

    Category category =
        (stem.rfind("-production"sv) != std::string::npos) ? Category::production : Category::consumption;
    DS_NOTICE(dsuid, category);
    archiveIds.push_back(ArchiveId{dsuid, category});
  }
  return archiveIds;
}

void MeteringImpl::checkDBConsistency() {
  auto now = std::chrono::system_clock::now();

  auto archiveIds = std::vector<ArchiveId>();

  auto fillMeterCategory = [&archiveIds](const auto& meters, Category category) {
    std::transform(meters.begin(), meters.end(), std::back_inserter(archiveIds), [category](const auto& dsuid) {
      return ArchiveId{dsuid, category};
    });
  };

  fillMeterCategory(m_consumption->getActiveMeters(), Category::consumption);
  fillMeterCategory(m_production->getActiveMeters(), Category::production);

  auto restartCached = recreateDeadArchives(archiveIds);
  if (restartCached) {
    // no need to flush
    m_archives->restartCached();
    return;
  }

  auto needsFlush = std::any_of(std::cbegin(archiveIds), std::cend(archiveIds),
                                [&, now](const auto& archiveId) { return m_archives->checkDBFlush(now, archiveId); });

  if (needsFlush) {
    m_archives->flush(archiveIds);
  }
}

void MeteringImpl::pollLoop(boost::asio::io_service& io, ds::asio::Timer& timer) {
  if (m_Terminated) {
    DS_NOTICE("pollLoop: terminating");
    return;
  }

  auto startedAt = ds::Clock::now();
  if (m_nextPollAt != std::chrono::floor<std::chrono::seconds>(startedAt)) {
    DS_NOTICE("pollLoop: missed wakeup", startedAt, m_nextPollAt, ds::hideArg(startedAt - m_nextPollAt));
  }

  DS_ASSERT(m_pMeteringBusInterface != NULL, "");
  try {
    auto now = system_clock::now();
    m_consumption->pollTimestamp(now);
    m_production->pollTimestamp(now);
    m_pMeteringBusInterface->requestMeterData();
  } catch (dss::BusApiError& apiError) {
    log("Metering::execute: Couldn't get metering data: " + std::string(apiError.what()));
  }

  triggerState(MeteringSystem::apartmentSummation,
               {.production = m_production->getApartmentPower(), .consumption = m_consumption->getApartmentPower()});
  checkDBConsistency();

  auto now = ds::Clock::now();
  if (now - startedAt > 1s) {
    DS_WARNING("pollLoop: exceeded 1s budget", now - startedAt);
  }

  m_nextPollAt = std::chrono::ceil<std::chrono::seconds>(now);
  timer.expiresFromNow(m_nextPollAt - now, [this, &io, &timer] { pollLoop(io, timer); });
}

bool MeteringImpl::validateAndConvertArchive(const ArchiveId& archiveId) {
  DS_ASSERT(m_archives->hasArchive(archiveId));
  try {
    bool tunePowerMaxSetting = false;
    bool validate = m_archives->validateDBSeries(archiveId, tunePowerMaxSetting);
    if (!validate) {
      DS_NOTICE("remove invalid archive", archiveId)
      m_system.remove(m_archives->filepath(archiveId));
      return false;
    } else if (tunePowerMaxSetting) {
      DS_NOTICE("archive needs tuning", archiveId);
      m_archives->tuneDBPowerSettings(archiveId); // non-const
    }
    return true;
  } catch (const ds::Exception& e) {
    DS_WARNING(e.what());
    return false;
  }
  return true;
}

void MeteringImpl::execute() {
  while (DSS::getInstance()->getModelMaintenance().isInitializing()) {
    sleepSeconds(1);
  }
  DS_NOTICE("pollLoop: modelReady received");

  auto archiveIds = searchArchivesInFolder();
  DS_NOTICE("found", ds::hideArg(archiveIds.size()), "archives");

  auto changes = recreateDeadArchives(archiveIds); // TODO remove file not re-create
  if (changes) {
    // file was removed/recreated
    archiveIds = searchArchivesInFolder();
  }

  for (auto it = archiveIds.cbegin(); it != archiveIds.end(); it++) {
    const auto& archiveId = *it;
    DS_NOTICE("checking:", archiveId);
    if (!validateAndConvertArchive(*it)) {
      DS_NOTICE("skip archive", *it);
      archiveIds.erase(it);
    }
  }

  ds::asio::IoService ioService; // `ds::asio::IoService` requires its construct to run in the same thread
  boost::asio::io_service::work work(ioService); // block in `run()` until `m_stopping`

  ds::asio::Timer timer(ioService);
  pollLoop(ioService, timer);
  auto socketServer = metering::makeSocketServer(ioService, m_electricityMeteringConfig.socketPath, *this);
  socketServer->start();

  while (!m_Terminated) {
    try {
      ioService.run();
      break; // run() exited normally
    } catch (std::exception& e) {
      DS_ERROR("io_service", e.what());
    }
  }
}

void MeteringImpl::doStart() {
  log("Writing files to: " + m_MeteringStorageLocation);
  run();
} // start

} // namespace

namespace dss {

std::shared_ptr<MeteringSubsystem> makeMetering(DSS* dss, Apartment& apartment) {
  return std::make_shared<MeteringImpl>(dss, apartment, makeArchives(makeRRDInterface(), ds::system()), ds::system());
}

std::shared_ptr<MeteringSubsystem> makeMockedMetering(Apartment& apartment,
                                                        std::shared_ptr<metering::Archives> archives,
                                                        ds::System& system) {
  return std::make_shared<MeteringImpl>(nullptr, apartment, std::move(archives), system);
}

} // namespace dss
