#include "power-states.h"

#include <numeric>
#include <tuple>

#include "metering.h"
#include "model/dsmeter.h"
#include "model/state.h"

static ds::LogChannel dsLogChannel("dssMeteringPowerStates");

using namespace dss;
using namespace dss::metering;

namespace {

static auto createStateName = [](MeteringSystem meteringSystem, const PowerStateConfig& config) {
  switch (meteringSystem) {
    case MeteringSystem::apartmentSummation:
      return ds::str("apartment-power-", config.category, "-on", config.setThreshold, "-off", config.resetThreshold);
    case MeteringSystem::digitalStromMeter:
      DS_FAIL_REQUIRE("use legacy code-path to generate state-name, dsuid needed");
    case MeteringSystem::openMeteringSystem:
    default:
      return ds::str(meteringSystem, "-", config.category, "-on", config.setThreshold, "-off", config.resetThreshold);
  }
};

class PowerStatesImpl : public PowerStates {
  std::vector<std::tuple<std::shared_ptr<State>, MeteringSystem, PowerStateConfig, int>> m_states{};
  Apartment& m_apartment;

public:
  PowerStatesImpl(Apartment& apartment) : m_apartment(apartment) {}
  std::shared_ptr<State> createState(MeteringSystem meteringSystem, PowerStateConfig&& powerStateConfig) override;
  void removeState(MeteringSystem meteringSystem, const PowerStateConfig& powerStateConfig) override;
  void trigger(MeteringSystem meteringSystem, TotalPower&& totalPower) override;
  void checkHysteresis(unsigned int value, PowerStateConfig& config, std::shared_ptr<State> state);
};

std::shared_ptr<State> PowerStatesImpl::createState(MeteringSystem meteringSystem,
                                                    PowerStateConfig&& powerStateConfig) {
  DS_INFO_ENTER(meteringSystem, powerStateConfig);
  auto it = std::find_if(
      std::begin(m_states), std::end(m_states), [meteringSystem, config = powerStateConfig](const auto& element) {
        return std::get<PowerStateConfig>(element) == config && std::get<MeteringSystem>(element) == meteringSystem;
      });
  if (it != std::end(m_states)) {
    auto& [state, meterSysten, config, references] = *it;
    // in clang variables from structured binding are only aliases they are not
    // variables, thus you cannot capture it in DS_REQUIRE lambda by reference
    auto scopeIssue = references++;
    DS_REQUIRE(scopeIssue > 0);
    DS_INFO_LEAVE();
    return state;
  }

  auto state = m_apartment.allocateState(StateType_Circuit, createStateName(meteringSystem, powerStateConfig), {});
  m_states.push_back(std::make_tuple(state, meteringSystem, std::move(powerStateConfig), 1));
  DS_INFO_LEAVE();
  return state;
}

void PowerStatesImpl::removeState(MeteringSystem meteringSystem, const PowerStateConfig& powerStateConfig) {
  DS_INFO_ENTER(meteringSystem, powerStateConfig);
  auto it = std::find_if(
      std::begin(m_states), std::end(m_states), [meteringSystem, config = powerStateConfig](const auto& element) {
        return std::get<PowerStateConfig>(element) == config && std::get<MeteringSystem>(element) == meteringSystem;
      });
  DS_REQUIRE(it != std::end(m_states));

  auto& [state, _, config, references] = *it;
  auto scopeIssue = references--;
  DS_REQUIRE(scopeIssue > 0);

  if (references == 0) {
    DS_NOTICE("removeState", std::get<MeteringSystem>(*it), config, state->getName());
    m_apartment.removeState(state); // NOTE: state is reference if you erase, iterator becomes invalid
    m_states.erase(it);
  }
  DS_INFO_LEAVE();
}

void PowerStatesImpl::checkHysteresis(unsigned int meteringValue, PowerStateConfig& config,
                                      std::shared_ptr<State> state) {
  if (state->getState() != State_Active) {
    if (meteringValue > config.setThreshold) {
      DS_NOTICE("  ->active", state);
      state->setState(CallOrigin::system, State_Active);
    }
  } else {
    if (meteringValue < config.resetThreshold) {
      DS_NOTICE("->inactive", state);
      state->setState(CallOrigin::system, State_Inactive);
    }
  }
}

void PowerStatesImpl::trigger(MeteringSystem meteringSystem, TotalPower&& totalPower) {
  DS_DEBUG(meteringSystem, totalPower);

  for (auto [state, meteringSystemInner, config, _] : m_states) {
    if (meteringSystemInner != meteringSystem) {
      continue;
    }

    DS_DEBUG(config, state, state->getState());
    if (config.category == Category::consumption) {
      checkHysteresis(totalPower.consumption, config, state);
    } else {
      checkHysteresis(totalPower.production, config, state);
    }
  }
}

} // namespace

namespace dss::metering {

std::shared_ptr<PowerStates> makePowerStates(Apartment& apartment) {
  return std::make_shared<PowerStatesImpl>(apartment);
}
} // namespace dss::metering
