/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "addressablemodelitem.h"

#include <memory>

#include "businterface.h"
#include "ds/log.h"
#include "model/apartment.h"
#include "model/modelconst.h"
#include "propertysystem.h"

static ds::LogChannel dsLogChannel("modelAddressableModelItem");

namespace dss {

//================================================== AddressableModelItem

AddressableModelItem::AddressableModelItem(Apartment* _pApartment) : m_pApartment(_pApartment) {} // ctor

void AddressableModelItem::increaseValue(const CallOrigin _origin, const SceneAccessCategory _category) {
  if (!SceneAccess::checkAccess(this, _category)) {
    DS_DEBUG("AddressableModelItem: increaseValue blocked");
    return;
  }
  m_pApartment->getActionRequestInterface()->callScene(this, SceneInc, {_origin, _category, ""}, false);
} // increaseValue

void AddressableModelItem::decreaseValue(const CallOrigin _origin, const SceneAccessCategory _category) {
  if (!SceneAccess::checkAccess(this, _category)) {
    DS_DEBUG("AddressableModelItem: decreaseValue blocked");
    return;
  }
  m_pApartment->getActionRequestInterface()->callScene(this, SceneDec, {_origin, _category, ""}, false);
} // decreaseValue

void AddressableModelItem::setValue(const CallOrigin _origin, const SceneAccessCategory _category,
                                    const uint8_t _value, const std::string _token) {
  if (!SceneAccess::checkAccess(this, _category)) {
    DS_DEBUG("AddressableModelItem: setValue blocked");
    return;
  }
  m_pApartment->getActionRequestInterface()->setValue(this, _value, Caller{_origin, _category, _token});
} // setValue

void AddressableModelItem::callScene(const CallOrigin _origin, const SceneAccessCategory _category,
                                     const int _sceneNr, const std::string _token, const bool _force) {
  if (!SceneAccess::checkAccess(this, _category)) {
    DS_DEBUG("AddressableModelItem: callScene blocked");
    return;
  }
  if (_origin == CallOrigin::json) {
    if (_force == false && (!SceneAccess::checkStates(this, _sceneNr))) {
      DS_DEBUG("AddressableModelItem: callScene from web service blocked");
      return;
    }
  }
  m_pApartment->getActionRequestInterface()->callScene(this, _sceneNr, {_origin, _category, _token}, _force);
} // callScene

void AddressableModelItem::callSceneMin(const CallOrigin _origin, const SceneAccessCategory _category,
                                        const int _sceneNr, const std::string _token) {
  if (!SceneAccess::checkAccess(this, _category)) {
    DS_DEBUG("AddressableModelItem: callScene blocked");
    return;
  }
  if (_origin == CallOrigin::json) {
    if (!SceneAccess::checkStates(this, _sceneNr)) {
      DS_DEBUG("AddressableModelItem: callSceneMin from web service blocked");
      return;
    }
  }
  m_pApartment->getActionRequestInterface()->callSceneMin(this, _sceneNr, Caller{_origin, _category, _token});
} // callSceneMin

void AddressableModelItem::saveScene(const CallOrigin _origin, const int _sceneNr, const std::string _token) {
  m_pApartment->getActionRequestInterface()->saveScene(this, _sceneNr, Caller{_origin, SceneAccessCategory{}, _token});
} // saveScene

void AddressableModelItem::undoScene(const CallOrigin _origin, const SceneAccessCategory _category,
                                     const int _sceneNr, const std::string _token) {
  if (!SceneAccess::checkAccess(this, _category)) {
    DS_DEBUG("AddressableModelItem: undoScene blocked");
    return;
  }
  m_pApartment->getActionRequestInterface()->undoScene(this, _sceneNr, Caller{_origin, _category, _token});
} // undoScene

void AddressableModelItem::undoSceneLast(const CallOrigin _origin, const SceneAccessCategory _category,
                                         const std::string _token) {
  if (!SceneAccess::checkAccess(this, _category)) {
    DS_DEBUG("AddressableModelItem: undoSceneLast blocked");
    return;
  }
  m_pApartment->getActionRequestInterface()->undoSceneLast(this, Caller{_origin, _category, _token});
} // undoSceneLast

void AddressableModelItem::blink(const CallOrigin _origin, const SceneAccessCategory _category,
                                 const std::string _token) {
  if (!SceneAccess::checkAccess(this, _category)) {
    DS_DEBUG("AddressableModelItem: blink blocked");
    return;
  }
  m_pApartment->getActionRequestInterface()->blink(this, Caller{_origin, _category, _token});
} // blink

void AddressableModelItem::increaseOutputChannelValue(const CallOrigin _origin, const SceneAccessCategory _category,
                                                      const ChannelType _channel, const std::string _token) {
  if (!SceneAccess::checkAccess(this, _category)) {
    DS_DEBUG("AddressableModelItem: increaseOutputChannelValue blocked");
    return;
  }
  m_pApartment->getActionRequestInterface()->increaseOutputChannelValue(this, _channel,
                                                                        Caller{_origin, _category, _token});
} // increaseOutputChannelValue

void AddressableModelItem::decreaseOutputChannelValue(const CallOrigin _origin, const SceneAccessCategory _category,
                                                      const ChannelType _channel, const std::string _token) {
  if (!SceneAccess::checkAccess(this, _category)) {
    DS_DEBUG("AddressableModelItem: decreaseOutputChannelValue blocked");
    return;
  }
  m_pApartment->getActionRequestInterface()->decreaseOutputChannelValue(this, _channel,
                                                                        Caller{_origin, _category, _token});
} // decreaseOutputChannelValue

void AddressableModelItem::stopOutputChannelValue(const CallOrigin _origin, const SceneAccessCategory _category,
                                                  const ChannelType _channel, const std::string _token) {
  if (!SceneAccess::checkAccess(this, _category)) {
    DS_DEBUG("AddressableModelItem: stopOutputChannelValue blocked");
    return;
  }
  m_pApartment->getActionRequestInterface()->stopOutputChannelValue(this, _channel, Caller{_origin, _category, _token});
} // stopOutputChannelValue

void AddressableModelItem::pushSensor(const CallOrigin _origin, const SceneAccessCategory _category,
                                      const dsuid_t _sourceID, SensorType _sensorType, const double _sensorValueFloat,
                                      const std::string _token) {
  if (!SceneAccess::checkAccess(this, _category)) {
    DS_DEBUG("AddressableModelItem: pushSensor blocked");
    return;
  }
  m_pApartment->getActionRequestInterface()->pushSensor(this, _sourceID, _sensorType, _sensorValueFloat,
                                                        Caller{_origin, _category, _token});
}

} // namespace dss
