/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/


#ifndef DEVICEREFERENCE_H
#define DEVICEREFERENCE_H

#include <memory>

#include "ds485types.h"
#include "deviceinterface.h"

namespace dss {

  class Apartment;
  class Device;

    /** Internal reference to a device.
   * A DeviceReference is virtually interchangable with a device. It is used in places
     where a reference to a device is needed.
   */
  class DeviceReference : public IDeviceInterface {
  private:
    dsuid_t m_DSID;
    const Apartment* m_Apartment;
  public:
    DeviceReference(const dsuid_t _dsid, const Apartment* _apartment);
    DeviceReference(std::shared_ptr<const Device> _device, const Apartment* _apartment);
    virtual ~DeviceReference() = default;

    /** Returns a reference to the referenced device
     * @note This accesses the apartment which has to be valid.
     */
    std::shared_ptr<Device> getDevice();
    /** @copydoc getDevice() */
    std::shared_ptr<const Device> getDevice() const;
    /** Returns the DSID of the referenced device */
    dsuid_t getDSID() const;

    /** Compares two device references.
     * Device references are considered equal if their DSID match. */
    bool operator==(const DeviceReference& _other) const {
      return (m_DSID == _other.m_DSID);
    }

    virtual void setValue(const CallOrigin _origin, const SceneAccessCategory _category, uint8_t _value,
                          const std::string _token);

    virtual void increaseValue(const CallOrigin _origin, const SceneAccessCategory _category);
    virtual void decreaseValue(const CallOrigin _origin, const SceneAccessCategory _category);

    /** Returns wheter the device is turned on.
     * @note The detection is soly based on the last called scene. As soon as we've
     * got submetering data this should reflect the real state.
     */
    virtual bool isOn() const;

    virtual void callScene(const CallOrigin _origin, const SceneAccessCategory _category, const int _sceneNr,
                           std::string _token, const bool _force);
    virtual void callSceneMin(const CallOrigin _origin, const SceneAccessCategory _category, const int _sceneNr,
                              const std::string _token);
    virtual void saveScene(const CallOrigin _origin, const int _sceneNr, const std::string _token);
    virtual void undoScene(const CallOrigin _origin, const SceneAccessCategory _category, const int _sceneNr,
                           const std::string _token);
    virtual void undoSceneLast(const CallOrigin _origin, const SceneAccessCategory _category,
                               const std::string _token);

    virtual void nextScene(const CallOrigin _origin, const SceneAccessCategory _category);
    virtual void previousScene(const CallOrigin _origin, const SceneAccessCategory _category);

    virtual void blink(const CallOrigin _origin, const SceneAccessCategory _category, const std::string _token);

    virtual void increaseOutputChannelValue(const CallOrigin _origin, const SceneAccessCategory _category,
                                            const ChannelType channel, const std::string _token);
    virtual void decreaseOutputChannelValue(const CallOrigin _origin, const SceneAccessCategory _category,
                                            const ChannelType channel, const std::string _token);
    virtual void stopOutputChannelValue(const CallOrigin _origin, const SceneAccessCategory _category,
                                        const ChannelType channel, const std::string _token);

    virtual void pushSensor(const CallOrigin _origin, const SceneAccessCategory _category, dsuid_t _sourceID,
                            SensorType _sensorType, double _sensorValueFloat, const std::string _token);

    virtual unsigned long getPowerConsumption();
  }; // DeviceReference

  std::ostream& operator<<(std::ostream& ostream, const DeviceReference& x);

} // namespace dss

#endif // DEVICEREFERENCE_H
