/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/
#include "status.h"

#include "ds/random.h"
#include "ds/log.h"

#include "base.h"
#include "status-field.h"
#include "group.h"
#include "apartment.h"
#include "dss.h"

static ds::LogChannel dsLogChannel("modelStatus");

namespace dss {

std::chrono::seconds Status::PUSH_PERIOD = std::chrono::minutes(90);

Status::Status(Group& group)
    : m_group(group)
    , m_periodicPushTimer(group.getApartment().getDss()->getIoService())
    , m_malfunctionField(*this, StatusFieldType::MALFUNCTION)
    , m_serviceField(*this, StatusFieldType::SERVICE) {
  // Start periodic push loop to recover from broadcast failures,
  // to spread the value to new devices.
  // It makes no sense to push immediately,
  // we are likely to even not have dsm-api connection now.
  asyncPushLoopRestart(PUSH_PERIOD);
}

Status::~Status() = default;

StatusField& Status::getField(StatusFieldType type) {
  switch (type) {
    case StatusFieldType::MALFUNCTION: return m_malfunctionField;
    case StatusFieldType::SERVICE: return m_serviceField;
  }
  DS_FAIL_REQUIRE("Unknow status field type.", type);
};

StatusSensorBitset Status::getValueAsBitset() const {
  StatusSensorBitset out;
  out |= m_malfunctionField.getValueAsBitset();
  out |= m_serviceField.getValueAsBitset();
  return out;
}

void Status::push() {
  DS_NOTICE("push ", ds::hideArg(*this));
  // moderate push calls
  asyncPushLoopRestart(std::chrono::seconds(1));
}

void Status::asyncPushLoopRestart(ds::asio::Timer::duration delay) {
  DS_DEBUG("asyncPushLoopRestart", *this, delay);

  m_periodicPushTimer.randomlyExpiresFromNowPercentDown(delay, 25, [this] {
    // getIoService must never block
    IDeviceInterface::async(
        [this, &group = m_group, value = getValueAsBitset().to_ulong()]() {
          DS_NOTICE("asyncPushLoopRestart call", ds::hideArg(*this), value);
          DSS::getInstance()->assertBlockingAllowed();
          group.pushSensor(CallOrigin::system, SceneAccessCategory::unknown, DSUID_NULL, SensorType::Status, value, "");
        },
        [this, groupId = m_group.getID()](std::optional<std::exception> exception) {
          DS_DEBUG("asyncPushLoopRestart handler", ds::hideArg(*this), exception);
          if (exception) {
            DS_ERROR("asyncPushLoopRestart failed what:", exception->what());
          }
          if (groupId == GroupIDGlobalAppDsVentilation) {
            DS_NOTICE("asyncPushLoopRestart handler", ds::hideArg(*this), "reschedule", PUSH_PERIOD);
            asyncPushLoopRestart(PUSH_PERIOD); // async loop
          }
        });
  });
}

std::ostream& operator<<(std::ostream& stream, const Status &x) {
    return stream << x.getGroup();
}

} // namespace dss
