/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ds/log.h>

#include <boost/range/algorithm/sort.hpp>
#include <boost/range/algorithm/unique.hpp>
#include <boost/range/algorithm_ext/erase.hpp>

#include "base.h"
#include "businterface.h"
#include "dss.h"
#include "event.h"
#include "includerapidjson.h"
#include "logger.h"
#include "model/apartment.h"
#include "model/device.h"
#include "model/dsmeter.h"
#include "model/group.h"
#include "model/modelconst.h"
#include "model/modelmaintenance.h"
#include "model/set.h"
#include "propertysystem.h"
#include "util.h"
#include "zone.h"
using rapidjson::Document;

static ds::LogChannel dsLogChannel("modelZone");

namespace dss {

static const std::string HeatingSuffix("heatingOperation_");

//================================================== Zone

Zone::Zone(const int _id, Apartment* _pApartment)
    : m_ZoneID(_id), m_HeatingOperationMode{}, m_pApartment(_pApartment), m_floorID(APARTMENT_FLOOR_ID) {}

Zone::~Zone() {
  // we don't own our dsMeters
  m_DSMeters.clear();
} // dtor

bool Zone::isInDsModel() const {
  return getID() != ZONE_ID_BROADCAST; // zone 0 is not really a zone but whole apartment
}

dsModel::ZoneId Zone::id() const {
  DS_REQUIRE(isInDsModel(), "Accessing zone that does not exist in ds model.", *this);
  return dsModel::ZoneId{ds::str(m_ZoneID)};
}

std::vector<dsModel::SubmoduleId> Zone::submoduleIds() const {
  auto result = std::vector<dsModel::SubmoduleId>{};
  std::lock_guard g1(m_pApartment->getMutex());
  for (const auto& devicesReference : m_Devices) {
    auto device = devicesReference.getDevice();
    if (device->isInDsModel()) {
      result.push_back(device->submoduleId());
    }
  }
  return result;
}

ModelMaintenance* Zone::tryGetModelMaintenance() {
  return m_pApartment ? m_pApartment->getModelMaintenance() : nullptr;
}

void Zone::setName(const std::string& name) {
  if (m_Name == name) {
    return;
  }
  DS_INFO_ENTER(name);
  DeviceContainer::setName(name);
  auto event = std::make_shared<Event>(EventName::ZoneChanged);
  event->setProperty("id", ds::str(m_ZoneID));
  event->setProperty("name", ds::str(m_Name));
  if (DSS::hasInstance()) {
    DSS::getInstance()->raiseEvent(event);
  }
}

Set Zone::getDevices() const {
  std::lock_guard g1(m_pApartment->getMutex());
  return Set(m_Devices);
} // getDevices

void Zone::addDevice(DeviceReference _device) {
  std::lock_guard g1(m_pApartment->getMutex());
  std::shared_ptr<const Device> dev = _device.getDevice();
  int oldZoneID = dev->getZoneID();
  if (oldZoneID == m_ZoneID && contains(m_Devices, _device)) {
    DS_DEBUG_LEAVE("Nothing to do.");
    return;
  }

  if ((oldZoneID != -1) && (oldZoneID != 0)) {
    try {
      std::shared_ptr<Zone> oldZone = dev->getApartment().getZone(oldZoneID);
      oldZone->removeDevice(_device);
    } catch (std::runtime_error&) {
    }
  }
  if (!contains(m_Devices, _device)) {
    m_Devices.push_back(_device);
  } else {
    // don't warn about multiple additions to zone 0
    if (m_ZoneID != ZONE_ID_BROADCAST) {
      Logger::getInstance()->log("Zone::addDevice: DUPLICATE DEVICE Detected Zone: " + intToString(m_ZoneID) +
                                     " device: " + dsuid2str(_device.getDSID()),
                                 lsWarning);
    }
  }
  _device.getDevice()->setZoneID(m_ZoneID);
} // addDevice

void Zone::addGroup(std::shared_ptr<Group> _group) {
  // TODO(soon): remove again before R1705
  if (_group->getID() > GroupIDGlobalAppDsAwnings) {
    return;
  }
  // until here
  if (_group->getZoneID() != m_ZoneID) {
    throw std::runtime_error("Zone::addGroup: ZoneID of _group does not match own");
  }
  m_Groups.push_back(_group);
  if (m_pPropertyNode != NULL) {
    _group->publishToPropertyTree();
  }

} // addGroup

void Zone::removeGroup(std::shared_ptr<Group> _group) {
  std::vector<std::shared_ptr<Group> >::iterator it = find(m_Groups.begin(), m_Groups.end(), _group);
  if (it != m_Groups.end()) {
    m_Groups.erase(it);
  }
  if (m_pPropertyNode != NULL) {
    PropertyNodePtr groupNode = m_pPropertyNode->getProperty("groups/group" + intToString(_group->getID()));
    if (groupNode != NULL) {
      groupNode->getParentNode()->removeChild(groupNode);
    }
  }
} // removeGroup

void Zone::removeDevice(DeviceReference _device) {
  std::lock_guard g1(m_pApartment->getMutex());
  auto pos = find(m_Devices.begin(), m_Devices.end(), _device);
  if (pos != m_Devices.end()) {
    if (m_pPropertyNode != NULL) {
      for (std::vector<std::shared_ptr<Group> >::const_iterator ipGroup = m_Groups.begin(), e = m_Groups.end();
           ipGroup != e; ++ipGroup) {
        PropertyNodePtr groupNode = m_pPropertyNode->getProperty("groups/group" + intToString((*ipGroup)->getID()) +
                                                                 "/devices/" + dsuid2str(_device.getDSID()));
        if (groupNode != NULL) {
          groupNode->getParentNode()->removeChild(groupNode);
        }
      }
    }
    m_Devices.erase(pos);
  }
} // removeDevice

std::shared_ptr<Group> Zone::getGroup(const std::string& _name) const {
  for (std::vector<std::shared_ptr<Group> >::const_iterator ipGroup = m_Groups.begin(), e = m_Groups.end();
       ipGroup != e; ++ipGroup) {
    if ((*ipGroup)->getName() == _name) {
      return *ipGroup;
    }
  }
  return std::shared_ptr<Group>();
} // getGroup

std::weak_ptr<Group> Zone::tryGetGroup(const int id) const {
  for (auto&& group : m_Groups) {
    if (group->getID() == id) {
      return group;
    }
  }
  return std::weak_ptr<Group>();
} // getGroup

std::shared_ptr<Group> Zone::getGroup(const int id) const {
  if (auto&& group = tryGetGroup(id).lock()) {
    return std::move(group);
  }
  return std::shared_ptr<Group>(); // TODO(someday): throw
} // getGroup

int Zone::getID() const {
  return m_ZoneID;
}

void Zone::setZoneID(const int _value) {
  m_ZoneID = _value;
} // setZoneID

void Zone::addToDSMeter(std::shared_ptr<DSMeter> _dsMeter) {
  // make sure the zone is connected to the dsMeter
  if (find_if(m_DSMeters.begin(), m_DSMeters.end(), [&_dsMeter](std::weak_ptr<const DSMeter>& dsm) {
        if (std::shared_ptr<const DSMeter> meter = dsm.lock()) {
          return (_dsMeter.get() == meter.get());
        }
        return false;
      }) == m_DSMeters.end()) {
    m_DSMeters.push_back(_dsMeter);
    if (_dsMeter->getPropertyNode() != NULL) {
      PropertyNodePtr alias = _dsMeter->getPropertyNode()->createProperty("zones/" + intToString(m_ZoneID));
      alias->alias(m_pPropertyNode);
    }
  }
} // addToDSMeter

void Zone::removeFromDSMeter(std::shared_ptr<DSMeter> _dsMeter) {
  m_DSMeters.erase(remove_if(m_DSMeters.begin(), m_DSMeters.end(), [&_dsMeter](std::weak_ptr<const DSMeter>& dsm) {
    if (std::shared_ptr<const DSMeter> meter = dsm.lock()) {
      return (_dsMeter.get() == meter.get());
    }
    return false;
  }));
  if (_dsMeter->getPropertyNode() != NULL) {
    PropertyNodePtr alias = _dsMeter->getPropertyNode()->getProperty("zones/" + intToString(m_ZoneID));
    if (alias != NULL) {
      alias->getParentNode()->removeChild(alias);
    }
  }
  if (temperatureController() == _dsMeter) {
    m_temperatureController.reset();
  }
} // removeFromDSMeter

bool Zone::registeredOnDSMeter(std::shared_ptr<const DSMeter> _dsMeter) const {
  return find_if(m_DSMeters.begin(), m_DSMeters.end(), [&_dsMeter](const std::weak_ptr<const DSMeter>& dsm) {
           if (std::shared_ptr<const DSMeter> meter = dsm.lock()) {
             return (_dsMeter.get() == meter.get());
           }
           return false;
         }) != m_DSMeters.end();
} // registeredOnDSMeter

bool Zone::isRegisteredOnAnyMeter() const {
  return !m_DSMeters.empty();
} // isRegisteredOnAnyMeter

unsigned long Zone::getPowerConsumption() {
  return getDevices().getPowerConsumption();
} // getPowerConsumption

void Zone::nextScene(const CallOrigin _origin, const SceneAccessCategory _category) {
  getGroup(GroupIDBroadcast)->nextScene(_origin, _category);
} // nextScene

void Zone::previousScene(const CallOrigin _origin, const SceneAccessCategory _category) {
  getGroup(GroupIDBroadcast)->previousScene(_origin, _category);
} // previousScene

std::vector<std::shared_ptr<AddressableModelItem> > Zone::splitIntoAddressableItems() {
  std::vector<std::shared_ptr<AddressableModelItem> > result;
  result.push_back(getGroup(GroupIDBroadcast));
  return result;
} // splitIntoAddressableItems

void Zone::publishToPropertyTree() {
  if (m_pPropertyNode == NULL) {
    if (m_pApartment->getPropertyNode() != NULL) {
      m_pPropertyNode = m_pApartment->getPropertyNode()->createProperty("zones/zone" + intToString(m_ZoneID));
      m_pPropertyNode->createProperty("ZoneID")->setIntegerValue(m_ZoneID);
      m_pPropertyNode->createProperty("name")->linkToProxy(
          PropertyProxyMemberFunction<Zone, std::string>(*this, &Zone::getName, &Zone::setName));
      m_pPropertyNode->createProperty("floorID")->linkToProxy(PropertyProxyReference<int, int>(m_floorID, false));

      m_pPropertyNode->createProperty("devices/");

      for (std::shared_ptr<Group> pGroup : m_Groups) {
        pGroup->publishToPropertyTree();
      }
    }
  }
} // publishToPropertyTree

void Zone::removeFromPropertyTree() {
  if (m_pPropertyNode != NULL) {
    m_pPropertyNode->getParentNode()->removeChild(m_pPropertyNode);
    m_pPropertyNode.reset();
  }
} // removeFromPropertyTree

const TemperatureConfig& Zone::temperatureConfig() const {
  return m_temperatureConfig;
}

const ZoneHeatingStatus& Zone::getHeatingStatus() const {
  return m_HeatingStatus;
}

const ZoneSensorStatus& Zone::getSensorStatus() const {
  return m_SensorStatus;
}

void Zone::sanitizeHeatingProperties() {
  DS_CONTEXT(*this);
  for (size_t i = 0; i < m_temperatureConfig.targetTemperatures.size(); i++) {
    auto& value = m_temperatureConfig.targetTemperatures.at(i);
    if (value < 3.0 || value > 35) {
      const auto& defaultValue = TemperatureConfig::DEFAULT_TARGET_TEMPERATURES.at(i);
      DS_ERROR("Sanitizing insane target temperature.", value, defaultValue, i);
      value = defaultValue;
    }
  }
  for (size_t i = 0; i < m_temperatureConfig.fixedValues.size(); i++) {
    auto& value = m_temperatureConfig.fixedValues.at(i);
    if (value < -100.0 || value > 100.0) {
      const auto& defaultValue = TemperatureConfig::DEFAULT_FIXED_VALUES.at(i);
      DS_ERROR("Sanitizing insane fixed value.", value, defaultValue, i);
      value = defaultValue;
    }
  }
}

void Zone::setTemperatureConfig(TemperatureConfig config) {
  m_temperatureConfig = std::move(config);
  sanitizeHeatingProperties();
  m_temperatureConfig.m_propertiesValid = true;
  raiseOperationModeChanged();
}

void Zone::setHeatingSetpoints(const TemperatureConfig& config) {
  m_temperatureConfig.targetTemperatures = config.targetTemperatures;
  sanitizeHeatingProperties();
  m_temperatureConfig.m_setPointsValid = true;
}

void Zone::setHeatingFixValues(const TemperatureConfig& config) {
  m_temperatureConfig.fixedValues = config.fixedValues;
  sanitizeHeatingProperties();
  m_temperatureConfig.m_setPointsValid = true;
}

void Zone::setTemperatureController(const std::shared_ptr<const DSMeter>& meter) {
  m_temperatureController = meter;
  m_temperatureControllerTs = DateTime();
}

std::vector<dsModel::ZoneApplicationType> Zone::applications() const {
  std::vector<dsModel::ZoneApplicationType> result;
  bool awningsOnlyInZone = awningsOnly();
  std::lock_guard g1(m_pApartment->getMutex());

  // only PID and passive colling temperature configurations are available in dsModel

  auto isTemperatureInDsModel = [&] {
    int lastCalledScene = getHeatingOperationMode();
    if (lastCalledScene == HeatingOperationModeIDPassiveCoolingOn ||
        lastCalledScene == HeatingOperationModeIDPassiveCoolingOff) {
      return true;
    }
    if (temperatureConfig().m_mode == HeatingControlMode::PID) {
      return true;
    }
    return false;
  };
  for (const auto& d : m_Devices) {
    if (d.getDevice()->isLegacyAwning() || (d.getDevice()->isBlindsButton() && awningsOnlyInZone)) {
      result.push_back(dsModel::ZoneApplicationType::awnings);
    } else if (auto x = ds::tryToEnum<dsModel::ZoneApplicationType>(d.getDevice()->applicationType())) {
      if (x == dsModel::ZoneApplicationType::temperature && !isTemperatureInDsModel()) {
        continue;
      }
      result.push_back(*x);
    }
  }
  // Following is a solution for Sonos devices. Those are not integrated in dsDevice model,
  // but logic in backend javascript sets m_connectedDevices flag if any are found in the zone.
  if (this->getGroup(ColorIDCyan)->hasConnectedDevices()) {
    result.push_back(dsModel::ZoneApplicationType::audio);
  }
  boost::erase(result, boost::unique<boost::return_found_end>(boost::sort(result)));
  return result;
}

bool Zone::awningsOnly() const {
  bool anyAwning = false;
  std::lock_guard g1(m_pApartment->getMutex());
  for (const auto& d : m_Devices) {
    if (d.getDevice()->isShadeDevice() && !d.getDevice()->isBlindsButton()) {
      return false;
    } else if (!anyAwning && d.getDevice()->isLegacyAwning()) {
      anyAwning = true;
    }
  }
  return anyAwning;
}

bool Zone::isTemperatureConfigValid() const {
  return m_temperatureConfig.m_propertiesValid;
}

bool Zone::isHeatingSetpointsValid() const {
  return m_temperatureConfig.m_setPointsValid;
}

void Zone::setHeatingConfigSpec(const ZoneHeatingConfigSpec& spec) {
  m_temperatureConfig.setSpec(spec);
  sanitizeHeatingProperties();
  m_temperatureConfig.m_propertiesValid = true;
  raiseOperationModeChanged();
}

ZoneHeatingConfigSpec Zone::getHeatingConfigSpec() {
  return m_temperatureConfig.spec();
}

void Zone::setHeatingControlState(int _ctrlState) {
  if (m_temperatureConfig.m_HeatingControlState != _ctrlState) {
    Logger::getInstance()->log("Zone::setHeatingControlState: zone " + getName() + " (" + intToString(getID()) +
                                   "): change state from " + intToString(m_temperatureConfig.m_HeatingControlState) +
                                   " to " + intToString(_ctrlState),
                               lsNotice);
    m_temperatureConfig.m_HeatingControlState = _ctrlState;
    raiseOperationModeChanged();
  }
}

void Zone::setHeatingControlOperationMode(const ZoneHeatingOperationModeSpec_t& operationModeValues) {
  for (int i = 0; i <= HeatingOperationModeIDMax; ++i) {
    m_temperatureConfig.targetTemperatures[i] =
        sensorValueToDouble(SensorType::RoomTemperatureSetpoint, operationModeValues.opModes[i]);
  }
  sanitizeHeatingProperties();
  m_temperatureConfig.m_setPointsValid = true;
  raiseOperationModeChanged();
}

void Zone::setHeatingFixedOperationMode(const ZoneHeatingOperationModeSpec_t& operationModeValues) {
  for (int i = 0; i <= HeatingOperationModeIDMax; ++i) {
    m_temperatureConfig.fixedValues[i] =
        sensorValueToDouble(SensorType::RoomTemperatureControlVariable, operationModeValues.opModes[i]);
  }
  sanitizeHeatingProperties();
  m_temperatureConfig.m_setPointsValid = true;
  raiseOperationModeChanged();
}

void Zone::setHeatingOperationMode(const ZoneHeatingOperationModeSpec_t& operationModeValues) {
  switch (m_temperatureConfig.m_mode) {
    case HeatingControlMode::PID: setHeatingControlOperationMode(operationModeValues); break;
    case HeatingControlMode::FIXED: setHeatingFixedOperationMode(operationModeValues); break;
    case HeatingControlMode::OFF:           // no break
    case HeatingControlMode::ZONE_FOLLOWER: // no break
    case HeatingControlMode::MANUAL:        // no break
    default: break;
  }
}

ZoneHeatingOperationModeSpec_t Zone::getHeatingControlOperationModeValues() const {
  ZoneHeatingOperationModeSpec_t retVal;

  for (int i = 0; i <= HeatingOperationModeIDMax; ++i) {
    try {
      retVal.opModes[i] =
          doubleToSensorValue(SensorType::RoomTemperatureSetpoint, m_temperatureConfig.targetTemperatures[i]);
    } catch (DSSException& ex) {
      retVal.opModes[i] = 0;
      Logger::getInstance()->log(
          "Zone::getHeatingControlOperationModeValues: ZoneID " + intToString(m_ZoneID) + "error on value conversion",
          lsError);
    }
  }

  return retVal;
}

ZoneHeatingOperationModeSpec_t Zone::getHeatingFixedOperationModeValues() const {
  ZoneHeatingOperationModeSpec_t retVal;

  for (int i = 0; i <= HeatingOperationModeIDMax; ++i) {
    try {
      retVal.opModes[i] =
          doubleToSensorValue(SensorType::RoomTemperatureControlVariable, m_temperatureConfig.fixedValues[i]);
    } catch (DSSException& ex) {
      retVal.opModes[i] = 0;
      Logger::getInstance()->log(
          "Zone::getHeatingFixedOperationModeValues: ZoneID " + intToString(m_ZoneID) + "error on value conversion",
          lsError);
    }
  }

  return retVal;
}

ZoneHeatingOperationModeSpec_t Zone::getHeatingOperationModeValues() const {
  switch (m_temperatureConfig.m_mode) {
    case HeatingControlMode::PID: return getHeatingControlOperationModeValues();
    case HeatingControlMode::FIXED: return getHeatingFixedOperationModeValues();
    case HeatingControlMode::OFF:
    case HeatingControlMode::ZONE_FOLLOWER:
    case HeatingControlMode::MANUAL:
    default: return ZoneHeatingOperationModeSpec_t();
  }
}

void Zone::setHeatingOperationMode(int _operationMode) {
  if (_operationMode > HeatingOperationModeIDMax) {
    Logger::getInstance()->log("Zone::setHeatingOperationMode: zone-id:" + intToString(m_ZoneID) +
                                   " mode: " + intToString(_operationMode) + " is invalid",
                               lsWarning);
    return;
  }
  m_HeatingOperationMode = _operationMode;
  if (m_pApartment) {
    m_pApartment->emitStatusChanged();
  }
}

int Zone::getHeatingOperationMode() const {
  return m_HeatingOperationMode;
}

void Zone::setTemperature(double _value, const DateTime& _ts) {
  m_SensorStatus.temperatureValue.emplace(_value);
  m_SensorStatus.temperatureValueTS = _ts;
}

void Zone::resetTemperature() {
  m_SensorStatus.temperatureValue.reset();
  m_SensorStatus.temperatureValueTS = DateTime::NullDate;
}

void Zone::setNominalValue(double _value, const DateTime& _ts) {
  m_HeatingStatus.nominalValue.emplace(_value);
  m_HeatingStatus.nominalValueTS = _ts;

  if (m_pApartment) {
    m_pApartment->emitStatusChanged();
  }
}

void Zone::resetNominalValue() {
  m_HeatingStatus.nominalValue.reset();
  m_HeatingStatus.nominalValueTS = DateTime::NullDate;
}

void Zone::setControlValue(double _value, const DateTime& _ts) {
  m_HeatingStatus.controlValue.emplace(_value);
  m_HeatingStatus.controlValueTS = _ts;
}

void Zone::resetControlValue() {
  m_HeatingStatus.controlValue.reset();
  m_HeatingStatus.controlValueTS = DateTime::NullDate;
}

void Zone::setHumidityValue(double _value, const DateTime& _ts) {
  m_SensorStatus.humidityValue.emplace(_value);
  m_SensorStatus.humidityValueTS = _ts;
}

void Zone::resetHumidityValue() {
  m_SensorStatus.humidityValue.reset();
  m_SensorStatus.humidityValueTS = DateTime::NullDate;
}

void Zone::setBrightnessValue(double _value, const DateTime& _ts) {
  m_SensorStatus.brightnessValue.emplace(_value);
  m_SensorStatus.brightnessValueTS = _ts;
}

void Zone::resetBrightnessValue() {
  m_SensorStatus.brightnessValue.reset();
  m_SensorStatus.brightnessValueTS = DateTime::NullDate;
}

void Zone::setCO2ConcentrationValue(double _value, const DateTime& _ts) {
  m_SensorStatus.co2ConcentrationValue.emplace(_value);
  m_SensorStatus.co2ConcentrationValueTS = _ts;
}

void Zone::resetCO2ConcentrationValue() {
  m_SensorStatus.co2ConcentrationValue.reset();
  m_SensorStatus.co2ConcentrationValueTS = DateTime::NullDate;
}

bool Zone::isAllowedSensorType(SensorType _sensorType) {
  if (m_ZoneID == 0) {
    // #dsd-811
    // Earlier only outdoor sensors were allowed to be assigned to zone 0,
    // however this is no longer the case.
    // Outdoor sensors should never be assigned to any zone, meaning that
    // we do not support assigning any sensor to zone 0 anymore.
    return false;
  } else {
    switch (_sensorType) {
      case SensorType::TemperatureIndoors:
      case SensorType::BrightnessIndoors:
      case SensorType::HumidityIndoors:
      case SensorType::CO2Concentration: return true;
      case SensorType::ActivePower:
      case SensorType::OutputCurrent:
      case SensorType::ElectricMeter:
      case SensorType::TemperatureOutdoors:
      case SensorType::BrightnessOutdoors:
      case SensorType::HumidityOutdoors:
      case SensorType::AirPressure:
      case SensorType::GustSpeed:
      case SensorType::GustDirection:
      case SensorType::WindSpeed:
      case SensorType::WindDirection:
      case SensorType::Precipitation:
      case SensorType::COConcentration:
      case SensorType::SoundPressureLevel:
      case SensorType::RoomTemperatureSetpoint:
      case SensorType::RoomTemperatureControlVariable:
      case SensorType::Status: ///< Status sensor value. Value bits are defined by \re StatusFieldType
      case SensorType::Reserved1:
      case SensorType::Reserved2:
      case SensorType::OutputCurrent16A:
      case SensorType::ActivePowerVA:
      case SensorType::Temperature:   // no break
      case SensorType::Brightness:    // no break
      case SensorType::Humidity:      // no break
      case SensorType::WaterQuantity: // no break
      case SensorType::WaterFlowRate: // no break
      case SensorType::NotUsed:
      case SensorType::UnknownType:
      case SensorType::SunAzimuth:
      case SensorType::SunElevation:
      default: return false;
    }
  }
}

void Zone::setSensor(const Device& _device, SensorType _sensorType) {
  const auto& sensor = _device.getSensorByType(_sensorType);

  if (!isAllowedSensorType(sensor.m_sensorType)) {
    throw std::runtime_error("Assignment of sensor type " + sensorTypeName(sensor.m_sensorType) + " is not allowed!");
  }

  for (auto&& ms : m_MainSensors) {
    if (ms.m_sensorType == sensor.m_sensorType) {
      ms.m_DSUID = _device.getDSID();
      ms.m_sensorIndex = sensor.m_sensorIndex;
      return;
    }
  }

  m_MainSensors.push_back(MainZoneSensor_t());
  m_MainSensors.back().m_DSUID = _device.getDSID();
  m_MainSensors.back().m_sensorIndex = sensor.m_sensorIndex;
  m_MainSensors.back().m_sensorType = sensor.m_sensorType;
}

void Zone::setSensor(const MainZoneSensor_t& _mainZoneSensor) {
  if (!isAllowedSensorType(_mainZoneSensor.m_sensorType)) {
    throw std::runtime_error("Assignment of sensor type " + sensorTypeName(_mainZoneSensor.m_sensorType) +
                             " is not allowed!");
  }

  for (auto&& ms : m_MainSensors) {
    if (ms.m_sensorType == _mainZoneSensor.m_sensorType) {
      ms.m_DSUID = _mainZoneSensor.m_DSUID;
      ms.m_sensorIndex = _mainZoneSensor.m_sensorIndex;
      return;
    }
  }
  m_MainSensors.push_back(_mainZoneSensor);
}

void Zone::resetSensor(SensorType _sensorType) {
  if (!isAllowedSensorType(_sensorType)) {
    return;
  }

  auto it = std::remove_if(m_MainSensors.begin(), m_MainSensors.end(),
                           [&](MainZoneSensor_t& sensorDesc) { return sensorDesc.m_sensorType == _sensorType; });
  m_MainSensors.erase(it, m_MainSensors.end());
}

bool Zone::isSensorAssigned(SensorType _sensorType) const {
  for (auto&& ms : m_MainSensors) {
    if (ms.m_sensorType == _sensorType) {
      return true;
    }
  }
  return false;
}

std::vector<SensorType> Zone::getUnassignedSensorTypes() const {
  SensorType sensorTypesIndoor[] = {SensorType::TemperatureIndoors, SensorType::BrightnessIndoors,
                                    SensorType::HumidityIndoors, SensorType::CO2Concentration};

  std::vector<SensorType> ret;
  /* no zone autoassignment for outdoor sensors*/
  if (m_ZoneID == 0) {
    return ret;
  }

  ret.assign(sensorTypesIndoor, sensorTypesIndoor + ds::arraySize(sensorTypesIndoor));
  auto it =
      std::remove_if(ret.begin(), ret.end(), [this](SensorType& sensorType) { return isSensorAssigned(sensorType); });
  ret.erase(it, ret.end());

  return ret;
}

std::vector<SensorType> Zone::getAssignedSensorTypes(const Device& _device) const {
  std::vector<SensorType> ret;
  for (auto&& s : m_MainSensors) {
    if (s.m_DSUID == _device.getDSID()) {
      ret.push_back(s.m_sensorType);
    }
  }
  return ret;
}

bool Zone::isDeviceZoneMember(const DeviceReference& _device) const {
  std::lock_guard g1(m_pApartment->getMutex());
  return contains(m_Devices, _device);
}

void Zone::removeInvalidZoneSensors() {
  // bring sensors without a device to the end of the vector
  auto it = std::remove_if(m_MainSensors.begin(), m_MainSensors.end(), [this](MainZoneSensor_t& sensorDesc) {
    return !isDeviceZoneMember(DeviceReference(sensorDesc.m_DSUID, m_pApartment));
  });

  // now remove the tail
  m_MainSensors.erase(it, m_MainSensors.end());
}

std::shared_ptr<Device> Zone::getAssignedSensorDevice(SensorType _sensorType) const {
  for (auto&& s : m_MainSensors) {
    if (s.m_sensorType == _sensorType) {
      return getDevices().getByDSID(s.m_DSUID).getDevice();
    }
  }
  return std::shared_ptr<Device>();
}

bool Zone::isZoneSensor(const Device& _device, SensorType _sensorType) const {
  std::shared_ptr<Device> zoneSensor = getAssignedSensorDevice(_sensorType);
  if (!zoneSensor) {
    return false;
  }

  return (_device.getDSID() == zoneSensor->getDSID());
}

void Zone::raiseOperationModeChanged() {
  if (auto&& modelMaintenance = tryGetModelMaintenance()) {
    modelMaintenance->addModelEvent(new ModelEvent(ModelEvent::etModelOperationModeChanged));
  }
}

int Zone::getFloorID() const {
  return m_floorID;
}

void Zone::setFloorID(int id) {
  m_floorID = id;
}

const std::array<double, 16> TemperatureConfig::DEFAULT_TARGET_TEMPERATURES{{
    8,  // off(0)
    22, // comfort(1)
    20, // economy(2)
    18, // notUsed(3)
    17, // night(4)
    16, // holiday(5)
    23, // passiveCoolingOnSignal(6)
    35, // passiveCoolingOffSignal(7)
    20, // nothing(8)
    35, // coolingOff(9)
    23, // cooling(10)
    25, // coolingEconomy(11)
    27, // coolingNotUsed(12)
    28, // coolingNight(13)
    29, // coolingHoliday(14)
    20,
}};

const std::array<double, 16> TemperatureConfig::DEFAULT_FIXED_VALUES{{
    0,    // off(0)
    100,  // comfort(1)
    50,   // economy(2)
    0,    // notUsed(3)
    0,    // night(4)
    0,    // holiday(5)
    0,    // passiveCoolingOnSignal(6)
    0,    // passiveCoolingOffSignal(7)
    0,    // nothing(8)
    0,    // coolingOff(9)
    -100, // cooling(10)
    -50,  // coolingEconomy(11)
    0,    // coolingNotUsed(12)
    0,    // coolingNight(13)
    0,    // coolingHoliday(14)
    0,
}};

bool TemperatureConfig::isEqual(const ZoneHeatingConfigSpec_t& config,
                                const ZoneHeatingOperationModeSpec_t& operationMode) const {
  bool configEqual =
      ((config.mode == m_mode) && (config.Kp == m_Kp) && (config.Ts == m_Ts) && (config.Ti == m_Ti) &&
       (config.Kd == m_Kd) && (config.Imin == m_Imin) && (config.Imax == m_Imax) && (config.Ymin == m_Ymin) &&
       (config.Ymax == m_Ymax) && (config.AntiWindUp == m_AntiWindUp) && (config.KeepFloorWarm == m_KeepFloorWarm) &&
       (config.SourceZoneId == m_HeatingMasterZone) && (config.Offset == m_CtrlOffset) &&
       (config.EmergencyValue == m_EmergencyValue) && (config.ManualValue == m_ManualValue));

  // the operation mode is important only in control and fixed mode
  bool operationModeEqual = true;
  try {
    if (m_mode == HeatingControlMode::PID) {
      for (int i = 0; i <= HeatingOperationModeIDMax; ++i) {
        if (doubleToSensorValue(SensorType::RoomTemperatureSetpoint, targetTemperatures[i]) !=
            operationMode.opModes[i]) {
          operationModeEqual = false;
          break;
        }
      }
    } else if (m_mode == HeatingControlMode::FIXED) {
      for (int i = 0; i <= HeatingOperationModeIDMax; ++i) {
        if (doubleToSensorValue(SensorType::RoomTemperatureControlVariable, fixedValues[i]) !=
            operationMode.opModes[i]) {
          operationModeEqual = false;
          break;
        }
      }
    }
  } catch (DSSException& ex) {
    operationModeEqual = false;
    Logger::getInstance()->log("ZoneHeatingProperties::isEqual: error on value conversion", lsError);
  }

  return configEqual && operationModeEqual;
}

static void parseTargetTemperaturesImpl(TemperatureConfig& config, const rapidjson::Value& d) {
  DS_REQUIRE(d.IsObject(), "Error during Json parsing");

  // try to get all valid passed values
  for (int i = 0; i <= HeatingOperationModeIDMax; ++i) {
    std::string strIdx = ds::str(i);
    if (d.HasMember(strIdx.c_str())) {
      DS_REQUIRE(d[strIdx.c_str()].IsNumber());
      config.targetTemperatures[i] = d[strIdx.c_str()].GetDouble();
      config.m_setPointsValid = true;
    }
  }
}

void TemperatureConfig::parseTargetTemperatures(const std::string& jsonObject) {
  Document d;
  d.Parse(jsonObject.c_str());
  parseTargetTemperaturesImpl(*this, d);
}

static void parseFixedValuesImpl(TemperatureConfig& config, const rapidjson::Value& d) {
  DS_REQUIRE(d.IsObject(), "Error during Json parsing");

  // try to get all valid passed values
  for (int i = 0; i <= HeatingOperationModeIDMax; ++i) {
    std::string strIdx = ds::str(i);
    if (d.HasMember(strIdx.c_str())) {
      DS_REQUIRE(d[strIdx.c_str()].IsNumber());
      config.fixedValues[i] = d[strIdx.c_str()].GetDouble();
      config.m_setPointsValid = true;
    }
  }
}

void TemperatureConfig::parseFixedValues(const std::string& jsonObject) {
  Document d;
  d.Parse(jsonObject.c_str());
  parseFixedValuesImpl(*this, d);
}

static void parseControlModeImpl(TemperatureConfig& config, const rapidjson::Value& d) {
  DS_REQUIRE(d.IsObject(), "Error during Json parsing");

  if (d.HasMember("emergencyValue")) {
    DS_REQUIRE(d["emergencyValue"].IsNumber());
    config.m_EmergencyValue = static_cast<int>(d["emergencyValue"].GetDouble()) + 100;
  }
  if (d.HasMember("ctrlKp")) {
    DS_REQUIRE(d["ctrlKp"].IsNumber());
    config.m_Kp = static_cast<uint16_t>(d["ctrlKp"].GetDouble() * 40);
  }
  if (d.HasMember("ctrlTs")) {
    DS_REQUIRE(d["ctrlTs"].IsNumber());
    config.m_Ts = static_cast<int>(d["ctrlTs"].GetDouble());
  }
  if (d.HasMember("ctrlTi")) {
    DS_REQUIRE(d["ctrlTi"].IsNumber());
    config.m_Ti = static_cast<int>(d["ctrlTi"].GetDouble());
  }
  if (d.HasMember("ctrlKd")) {
    DS_REQUIRE(d["ctrlKd"].IsNumber());
    config.m_Kd = static_cast<int>(d["ctrlKd"].GetDouble());
  }
  if (d.HasMember("ctrlImin")) {
    DS_REQUIRE(d["ctrlImin"].IsNumber());
    config.m_Imin = static_cast<int16_t>(d["ctrlImin"].GetDouble() * 40);
  }
  if (d.HasMember("ctrlImax")) {
    DS_REQUIRE(d["ctrlImax"].IsNumber());
    config.m_Imax = static_cast<int16_t>(d["ctrlImax"].GetDouble() * 40);
  }
  if (d.HasMember("ctrlYmin")) {
    DS_REQUIRE(d["ctrlYmin"].IsNumber());
    config.m_Ymin = static_cast<int>(d["ctrlYmin"].GetDouble()) + 100;
  }
  if (d.HasMember("ctrlYmax")) {
    DS_REQUIRE(d["ctrlYmax"].IsNumber());
    config.m_Ymax = static_cast<int>(d["ctrlYmax"].GetDouble()) + 100;
  }
  if (d.HasMember("ctrlAntiWindUp")) {
    DS_REQUIRE(d["ctrlAntiWindUp"].IsBool());
    config.m_AntiWindUp = d["ctrlAntiWindUp"].GetBool() ? 1 : 0;
  }
}

void TemperatureConfig::parseControlMode(const std::string& jsonObject) {
  Document d;
  d.Parse(jsonObject.c_str());
  parseControlModeImpl(*this, d);
}

static void parseFollowerModeImpl(TemperatureConfig& config, const rapidjson::Value& d) {
  DS_REQUIRE(d.IsObject(), "Error during Json parsing");

  if (d.HasMember("referenceZone")) {
    DS_REQUIRE(d["referenceZone"].IsNumber());
    config.m_HeatingMasterZone = static_cast<int>(d["referenceZone"].GetDouble());
  }
  if (d.HasMember("ctrlOffset")) {
    DS_REQUIRE(d["ctrlOffset"].IsNumber());
    config.m_CtrlOffset = static_cast<int>(d["ctrlOffset"].GetDouble());
  }
}

void TemperatureConfig::parseFollowerMode(const std::string& jsonObject) {
  Document d;
  d.Parse(jsonObject.c_str());
  parseFollowerModeImpl(*this, d);
}

static void parseManualModeImpl(TemperatureConfig& config, const rapidjson::Value& d) {
  DS_REQUIRE(d.IsObject(), "Error during Json parsing");

  if (d.HasMember("controlValue")) {
    DS_REQUIRE(d["controlValue"].IsNumber());
    config.m_ManualValue = static_cast<int>(d["controlValue"].GetDouble()) + 100;
  }
}

void TemperatureConfig::parseManualMode(const std::string& jsonObject) {
  Document d;
  d.Parse(jsonObject.c_str());
  parseManualModeImpl(*this, d);
}

void TemperatureConfig::parseFields(const std::string& jsonObject) {
  Document d;
  d.Parse(jsonObject.c_str());

  DS_REQUIRE(d.IsObject());

  if (d.HasMember("mode")) {
    const auto& modeString = d["mode"].GetString();
    const auto& mode = heatingControlModeFromName(modeString);
    DS_REQUIRE(mode, "Mode is invalid", modeString);
    m_mode = *mode;
  }

  if (d.HasMember("targetTemperatures")) {
    parseTargetTemperaturesImpl(*this, d["targetTemperatures"]);
  }
  if (d.HasMember("fixedValues")) {
    parseFixedValuesImpl(*this, d["fixedValues"]);
  }
  if (d.HasMember("controlMode")) {
    parseControlModeImpl(*this, d["controlMode"]);
  }
  if (d.HasMember("zoneFollowerMode")) {
    parseFollowerModeImpl(*this, d["zoneFollowerMode"]);
  }
  if (d.HasMember("manualMode")) {
    parseManualModeImpl(*this, d["manualMode"]);
  }
}

void TemperatureConfig::setSpec(const ZoneHeatingConfigSpec spec) {
  m_mode = spec.mode;
  m_Kp = spec.Kp;
  m_Ts = spec.Ts;
  m_Ti = spec.Ti;
  m_Kd = spec.Kd;
  m_Imin = spec.Imin;
  m_Imax = spec.Imax;
  m_Ymin = spec.Ymin;
  m_Ymax = spec.Ymax;
  m_AntiWindUp = spec.AntiWindUp;
  m_KeepFloorWarm = spec.KeepFloorWarm;
  m_HeatingMasterZone = spec.SourceZoneId;
  m_CtrlOffset = spec.Offset;
  m_EmergencyValue = spec.EmergencyValue;
  m_ManualValue = spec.ManualValue;
}

ZoneHeatingConfigSpec TemperatureConfig::spec() const {
  ZoneHeatingConfigSpec x;
  x.mode = m_mode;
  x.Kp = m_Kp;
  x.Ts = m_Ts;
  x.Ti = m_Ti;
  x.Kd = m_Kd;
  x.Imin = m_Imin;
  x.Imax = m_Imax;
  x.Ymin = m_Ymin;
  x.Ymax = m_Ymax;
  x.AntiWindUp = m_AntiWindUp;
  x.KeepFloorWarm = m_KeepFloorWarm;
  x.SourceZoneId = m_HeatingMasterZone;
  x.Offset = m_CtrlOffset;
  x.EmergencyValue = m_EmergencyValue;
  x.ManualValue = m_ManualValue;
  return x;
}

std::ostream& operator<<(std::ostream& ostream, const Zone& x) {
  return ostream << x.getName() << "(" << x.getID() << ")";
}
} // namespace dss
