#include "parser.h"

#include <ds/log.h>

#include <regex>

#include "../base.h"

static ds::LogChannel dsLogChannel("p1Parser");

namespace {

const char startComms = '/';
const char endData = '!';
const char carriageReturn = '\r';
const char lineFeed = '\n';

const int p1TimeStampLenth = 12;
const int kWtoWRatio = 1000;
const int crcLengthBytes = 4;

const std::regex timeStampRegex(R"(0-0:1\.0\.0\(([0-9]+)S\))");
const std::regex meterIdRegex(R"(0-0:96\.1\.1\(([0-9]+)\))");
const std::regex powerConsumedRegex(R"(1-0:1\.7\.0\(([0-9]+\.[0-9]+)\*kW\))");
const std::regex powerProducedRegex(R"(1-0:2\.7\.0\(([0-9]+\.[0-9]+)\*kW\))");
const std::regex t1EnergyConsumedRegex(R"(1-0:1\.8\.1\(([0-9]+\.[0-9]+)\*kWh\))");
const std::regex t1EnergyProduceRegex(R"(1-0:2\.8\.1\(([0-9]+\.[0-9]+)\*kWh\))");
const std::regex t2EnergyConsumedRegex(R"(1-0:1\.8\.2\(([0-9]+\.[0-9]+)\*kWh\))");
const std::regex t2EnergyProducedRegex(R"(1-0:2\.8\.2\(([0-9]+\.[0-9]+)\*kWh\))");

} // namespace

namespace dss::p1 {

ds::DateTime MeterParser::getTimeStamp(std::string p1TimeStamp) {
  if (p1TimeStamp.length() != p1TimeStampLenth) {
    DS_WARNING("TimeStamp from p1 malformed");
    return ds::DateTime();
  }

  std::string dsTimeStamp = "20" + p1TimeStamp.substr(0, 2) + "-" + p1TimeStamp.substr(2, 2) + "-" +
                            p1TimeStamp.substr(4, 2) + "T" + p1TimeStamp.substr(6, 2) + ":" + p1TimeStamp.substr(8, 2) +
                            ":" + p1TimeStamp.substr(10, 2) + ".000Z";

  ds::DateTime timeStamp = ds::toDateTime(dsTimeStamp);
  return timeStamp;
}

void MeterParser::parseLine(const std::string& line, dss::p1::MeterReading& currentReading) {
  std::smatch matches;

  if (std::regex_search(line, matches, timeStampRegex)) {
    if (matches.size() == 2) {
      DS_DEBUG("Got timestamp", matches[1]);
      currentReading.timestamp = getTimeStamp(matches[1].str());
      return;
    }
  }

  if (std::regex_search(line, matches, meterIdRegex)) {
    if (matches.size() == 2) {
      DS_DEBUG("Got meterId", matches[1]);
      currentReading.meterId = matches[1].str();
      return;
    }
  }

  if (std::regex_search(line, matches, powerConsumedRegex)) {
    if (matches.size() == 2) {
      DS_DEBUG("Got power consumed", matches[1]);
      currentReading.powerConsumed = std::stod(matches[1].str()) * kWtoWRatio;
      return;
    }
  }

  if (std::regex_search(line, matches, powerProducedRegex)) {
    if (matches.size() == 2) {
      DS_DEBUG("Got power produced", matches[1]);
      currentReading.powerProduced = std::stod(matches[1].str()) * kWtoWRatio;
      return;
    }
  }

  if (std::regex_search(line, matches, t1EnergyConsumedRegex)) {
    if (matches.size() == 2) {
      DS_DEBUG("Got T1 energy consumed", matches[1]);
      currentReading.t1EnergyConsumed = std::stod(matches[1].str());
      return;
    }
  }

  if (std::regex_search(line, matches, t1EnergyProduceRegex)) {
    if (matches.size() == 2) {
      DS_DEBUG("Got T1 energy produced", matches[1]);
      currentReading.t1EnergyProduced = std::stod(matches[1].str());
      return;
    }
  }

  if (std::regex_search(line, matches, t2EnergyConsumedRegex)) {
    if (matches.size() == 2) {
      DS_DEBUG("Got T2 energy consumed", matches[1]);
      currentReading.t2EnergyConsumed = std::stod(matches[1].str());
      return;
    }
  }

  if (std::regex_search(line, matches, t2EnergyProducedRegex)) {
    if (matches.size() == 2) {
      DS_DEBUG("Got T2 energy consumed", matches[1]);
      currentReading.t2EnergyProduced = std::stod(matches[1].str());
      return;
    }
  }
}

std::optional<MeterReading> MeterParser::parseFrame(const std::vector<char>& buffer) {
  DS_DEBUG("parseBuffer");
  int crcRemaining = 0;
  std::string oneLine = "";
  std::string crc = "";
  dss::p1::MeterReading currentReading;
  int endOfFrame = 0;

  for (unsigned i = 0; i < buffer.size(); i++) {
    char c = buffer.at(i);
    switch (c) {
      case startComms: {
        break;
      }
      case endData: {
        endOfFrame = i;
        crcRemaining = crcLengthBytes;
        break;
      }
      case carriageReturn: {
        break;
      }
      case lineFeed: {
        parseLine(oneLine, currentReading);
        oneLine.clear();
        break;
      }
      default: {
        if (crcRemaining > 0) {
          crcRemaining--;
          crc.push_back(c);
          if (crcRemaining == 0) {
            std::vector<char> frame(&buffer[0], &buffer[endOfFrame + 1]);
            bool valid = vaildateCrc(frame, crc);
            if (valid) {
              m_crcErrCount = 0;
              return currentReading;
            } else {
              DS_WARNING("CRC check failed");
              m_crcErrCount++;
              return std::nullopt;
            }
          }
        } else {
          oneLine.push_back(c);
        }
      }
    }
  }
  return std::nullopt;
}

bool MeterParser::vaildateCrc(const std::vector<char>& buffer, std::string crc) {
  unsigned long result = crc16(reinterpret_cast<const unsigned char*>(buffer.data()), buffer.size(), 0xA001);

  unsigned long crcToCompare = std::stoul("0x" + crc, nullptr, 16);

  DS_DEBUG("CRC Validation", result, crcToCompare);
  return result == crcToCompare;
}

} // namespace dss::p1
