#include "serial-toolkit.h"

#include <ds/json.h>
#include <ds/log.h>

#include <filesystem>

static ds::LogChannel dsLogChannel("p1Serial");

const std::string serialPortPathMatcher = "/ttyUSB";
const int crcThreshold = 10;
const int bufferThreshold = 4096;
const int staleThresholdInMillis = 300000;

namespace dss::p1 {

void SenderQueue::send(MeterReading&& meterReading) {
  // we don't re-use socket client yet
  std::make_shared<dss::p1::SocketClient>(m_io, m_socketPath)->send(ds::toJson(meterReading));
}

void FrameBuffer::processFrame() {
  DS_DEBUG(m_frame.size(), m_sum, "frame:\n", ds::hideArg(std::string_view(m_frame.data(), m_frame.size())));
  if (auto reading = m_parser.parseFrame(m_frame)) {
    DS_NOTICE(reading);
    m_senderQueue.send(std::move(*reading));
  };
}

void FrameBuffer::push(data_t::const_iterator begin, data_t::const_iterator end) {
  for (auto current = begin; current != end; current++) {
    if (*current == '/') {
      processFrame();
      m_frame.clear();
      m_sum = 0;
    }
    m_frame.push_back(*current); // OOM problem if frame end is never found !!!
    if (m_frame.size() > bufferThreshold) {
      DS_WARNING("Lots of non P1 data received, flushing buffer");
      m_frame.clear();
      m_overflowCount++;
    }
    m_sum++;
  }
}

void SerialConnection::start() {
  if (isDataStale()) {
    DS_WARNING("No data received on port, retrying next", m_activePortPath);
    m_lastDataReceived = std::chrono::system_clock::now();
  } else if (m_isReading) {
    DS_NOTICE("serial console is readable");
    return;
  }

  if (m_activePortPath.empty()) {
    m_activePortPath = m_config.serialDevice;
  } else {
    m_activePortPath = *getNextPort();
  }
  DS_NOTICE("Attempting connection on port", m_activePortPath);

  try {
    m_serialPort = std::make_shared<boost::asio::serial_port>(m_io, m_activePortPath);
  } catch (std::exception& exception) {
    DS_WARNING("serialConsole", exception.what(), m_activePortPath);
    return;
  }

  m_serialPort->set_option(boost::asio::serial_port_base::baud_rate(m_config.baudRate));
  m_serialPort->set_option(boost::asio::serial_port_base::character_size(8));
  m_serialPort->set_option(boost::asio::serial_port_base::stop_bits(boost::asio::serial_port_base::stop_bits::one));
  m_serialPort->set_option(boost::asio::serial_port_base::parity(boost::asio::serial_port_base::parity::none));
  m_serialPort->set_option(
      boost::asio::serial_port_base::flow_control(boost::asio::serial_port_base::flow_control::none));

  DS_NOTICE("Successfully connected on port", m_activePortPath);

  doRead();
}

void SerialConnection::doRead() {
  m_isReading = true;
  m_serialPort->async_read_some(boost::asio::buffer(m_readBuffer), [this](const auto& error_code, size_t size) {
    m_isReading = false; // single threaded
    if (!error_code) {
      m_lastDataReceived = std::chrono::system_clock::now();
      // DS_NOTICE(size, m_readBuffer.size());
      m_frameBuffer.push(m_readBuffer.begin(), m_readBuffer.begin() + size);
      if (m_frameBuffer.getCrcErrorCount() > crcThreshold || m_frameBuffer.getOverflowCount() > 0) {
        DS_WARNING("Non P1 serial port detected, switching to next", m_activePortPath);
        // restart serial port on next
        std::optional<std::string> newPort = getNextPort();
        if (newPort) {
          DS_NOTICE("Attempting connection to serial port.", *newPort);
          try {
            m_serialPort = std::make_shared<boost::asio::serial_port>(m_io, *newPort);

            m_serialPort->set_option(boost::asio::serial_port_base::baud_rate(m_config.baudRate));
            m_serialPort->set_option(boost::asio::serial_port_base::character_size(8));
            m_serialPort->set_option(
                boost::asio::serial_port_base::stop_bits(boost::asio::serial_port_base::stop_bits::one));
            m_serialPort->set_option(
                boost::asio::serial_port_base::parity(boost::asio::serial_port_base::parity::none));
            m_serialPort->set_option(
                boost::asio::serial_port_base::flow_control(boost::asio::serial_port_base::flow_control::none));
          } catch (std::exception& exception) {
            DS_WARNING("serialConsole", exception.what(), *newPort);
          }

          DS_NOTICE("Successfully connected to serial port.", *newPort);
          m_activePortPath = *newPort;
          m_frameBuffer.clearErrors();
        }
      }
      doRead();
    }
    // DS_NOTICE("doRead (frame-split)", error_code, size);
    // std::cerr << std::string_view(m_readBuffer.data(), size) << "\n\n";
  });
}

std::vector<std::string> SerialConnection::getSerialPorts() {
  std::vector<std::string> serialPorts;
  std::string path = "/dev";
  for (const auto& entry : std::filesystem::directory_iterator(path)) {
    std::string entryString = entry.path();

    if (entryString.find(serialPortPathMatcher) != std::string::npos) {
      DS_DEBUG("Discovered serial port.", entryString);
      serialPorts.push_back(entryString);
    }
  }
  return serialPorts;
}

std::optional<std::string> SerialConnection::getNextPort() {
  std::vector<std::string> serialPorts = getSerialPorts();
  if (serialPorts.size() == 0) {
    return std::nullopt;
  }

  bool currentStillExists = false;
  unsigned int index = 0;
  for (unsigned int i = 0; i < serialPorts.size(); i++) {
    std::string port = serialPorts.at(i);
    if (m_activePortPath.compare(port) == 0) {
      currentStillExists = true;
      index = i;
    }
  }

  if (currentStillExists) {
    if (index == (serialPorts.size() - 1)) {
      return serialPorts.at(0);
    } else {
      return serialPorts.at(index + 1);
    }
  } else {
    return serialPorts.at(0);
  }
}

bool SerialConnection::isDataStale() {
  std::chrono::time_point<std::chrono::system_clock> now = std::chrono::system_clock::now();

  auto milliseconds = std::chrono::duration_cast<std::chrono::milliseconds>(now - m_lastDataReceived);

  DS_DEBUG("Time elapsed since last received data", milliseconds.count());
  if (milliseconds.count() > staleThresholdInMillis) {
    return true;
  } else {
    return false;
  }
}

} // namespace dss::p1