/*
    Copyright (c) 2013 digitalSTROM.org, Zurich, Switzerland

    Author: Christian Hitz, aizo AG <christian.hitz@aizo.com>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "sceneaccess.h"

#include <memory>

#include "base.h"
#include "dss.h"
#include "logger.h"
#include "model/addressablemodelitem.h"
#include "model/apartment.h"
#include "model/cluster.h"
#include "model/device.h"
#include "model/group.h"
#include "model/modelconst.h"
#include "model/set.h"
#include "model/state.h"
#include "propertysystem.h"
#include "stdio.h"
#include "string.h"

using namespace std::literals;

namespace dss {

bool SceneAccess::checkAccess(const AddressableModelItem* _pTarget, const SceneAccessCategory _category) {
  if (!DSS::hasInstance()) {
    return true;
  }

  // "manual" scene calls are never blocked
  if (_category == SceneAccessCategory::manual) {
    return true;
  }

  Apartment& apartment = DSS::getInstance()->getApartment();
  PropertyNodePtr property = apartment.getPropertySystem()->getRootNode();

  {
    /*
     * Fire & Wind: Prohibit automatic actions in group shade
     *
     * Behavior changed 2015/12/10:
     * Fire should not block all groups, but only shades, see #11755
     */
    std::shared_ptr<State> fire = apartment.getNonScriptState("fire");
    std::shared_ptr<State> wind = apartment.getNonScriptState("wind");
    bool windActive = (wind && (wind->getState() == State_Active));
    bool fireActive = (fire && (fire->getState() == State_Active));

    std::string protectionMessage("Execution blocked: ");
    if (fireActive && windActive) {
      protectionMessage += "fire and wind alarms are active";
    } else if (fireActive) {
      protectionMessage += "fire is active";
    } else if (windActive) {
      protectionMessage += "wind is active";
    }

    if (windActive || fireActive) {
      const Group* pGroup = dynamic_cast<const Group*>(_pTarget);
      if (pGroup != NULL) {
        if (ApplicationType::blinds == pGroup->getApplicationType()) {
          throw SceneAccessException(protectionMessage);
        }
      }
      const Device* pDevice = dynamic_cast<const Device*>(_pTarget);
      if (pDevice != NULL) {
        if (pDevice->isInGroup(static_cast<int>(DeviceClasses_t::DEVICE_CLASS_GR))) {
          throw SceneAccessException(protectionMessage);
        }
      }
    }
  }

  {
    /*
     * Partial Wind (in user groups): Prevent automatic actions in relevant user group.
     */
    std::shared_ptr<Cluster> gr;
    for (int i = GroupIDAppUserMin; i <= GroupIDAppUserMax; ++i) {
      try {
        gr = apartment.getCluster(i);
      } catch (ItemNotFoundException& e) {
        continue;
      }
      if (gr && gr->isValid() && gr->getApplicationType() == ApplicationType::blinds) {
        std::shared_ptr<State> wind;
        try {
          wind = apartment.getNonScriptState("wind.group" + intToString(i));
        } catch (ItemNotFoundException& e) {
          continue;
        }
        if (wind != NULL && (wind->getState() == State_Active)) {
          const Group* pGroup = dynamic_cast<const Group*>(_pTarget);
          if (pGroup != NULL) {
            // addressing the same cluster with an active wind alarm
            if (pGroup->getID() == i) {
              throw SceneAccessException("Execution blocked: wind is active");
            }
            // addressing group 2 in a single zone (or all zones) with devices
            // that are member of a cluster with an active wind alarm
            if (pGroup->getID() == static_cast<int>(DeviceClasses_t::DEVICE_CLASS_GR)) {
              Set devsInZones = pGroup->getDevices().getByGroup(i);
              if (devsInZones.length() > 0) {
                throw SceneAccessException("Execution blocked: wind is active");
              }
            }
          }
          const Device* pDevice = dynamic_cast<const Device*>(_pTarget);
          if (pDevice != NULL) {
            if (pDevice->isInGroup(i)) {
              throw SceneAccessException("Execution blocked: wind is active");
            }
          }
        }
      }
    }
  }

  return true;
}

bool SceneAccess::checkStates(const AddressableModelItem* _pTarget, int _sceneNr) {
  if (!DSS::hasInstance()) {
    return true;
  }

  std::string stateName;
  switch (_sceneNr) {
    case SceneWindActive: stateName = "wind"; break;
    case SceneRainActive: stateName = "rain"; break;
    case SceneHailActive: stateName = "hail"; break;
    case SceneFire: stateName = "fire"; break;
    case SceneAlarm: stateName = "alarm"; break;
    case SceneAlarm2: stateName = "alarm2"; break;
    case SceneAlarm3: stateName = "alarm3"; break;
    case SceneAlarm4: stateName = "alarm4"; break;
    default: return true;
  }

  Apartment& apartment = DSS::getInstance()->getApartment();
  std::shared_ptr<State> s = apartment.getNonScriptState(stateName);
  if (s && (s->getState() == State_Active)) {
    std::string errorMessage("Execution blocked: state " + stateName + " is already active");
    throw SceneAccessException(errorMessage);
  }

  return true;
}

SceneAccessCategory SceneAccess::stringToCategory(const std::string& _categoryString) {
  return ds::toEnum<SceneAccessCategory>(_categoryString);
}

} // namespace dss
