/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif

#include "jsonhelper.h"


#include "datetools.h"

#include "model/devicereference.h"
#include "model/device.h"
#include "model/set.h"
#include "model/group.h"
#include "model/cluster.h"
#include "model/zone.h"
#include "model/floor.h"
#include "model/state.h"
#include "model/apartment.h"
#include "dss.h"
#include "model/dsmeter.h"
#include "web/webrequests.h"

namespace dss {

void toJSON(const DeviceReference& deviceReference, JSONWriter& _json) {
  const auto devicePtr = deviceReference.getDevice();
  const auto& device = *devicePtr;
  DateTime tmp_date;

  _json.startObject();
  dsid_t dsid;
  if (dsuid_to_dsid(device.getDSID(), &dsid)) {
    _json.add("id", dsid2str(dsid));
  } else {
    _json.add("id", "");
  }
  _json.add("DisplayID", device.getDisplayID());
  _json.add("dSUID", device.getDSID());
  _json.add("GTIN", device.getGTIN());
  _json.add("name", device.getName());
  _json.add("dSUIDIndex", device.multiDeviceIndex());
  _json.add("functionID", device.getFunctionID());
  _json.add("productRevision", device.getRevisionID());
  _json.add("productID", device.getProductID());
  _json.add("hwInfo", device.getHWInfo());
  _json.add("OemStatus", device.getOemStateAsString());
  _json.add("OemEanNumber", device.getOemEanAsString());
  _json.add("OemSerialNumber", device.getOemSerialNumber());
  _json.add("OemPartNumber", device.getOemPartNumber());
  _json.add("OemProductInfoState", device.getOemProductInfoStateAsString());
  _json.add("OemProductURL", device.getOemProductURL());
  _json.add("OemInternetState", device.getOemInetStateAsString());
  _json.add("OemIsIndependent", device.getOemIsIndependent());
  _json.add("modelFeatures", device.getModelFeatures());
  _json.add("isVdcDevice", device.isVdcDevice());
  _json.add("supportedBasicScenes", device.getSupportedBasicScenes());
  _json.add("buttonUsage", device.getButtonUsageAsString());
  if (device.isVdcDevice()) {
    const auto& vdsd = device.getVdsdSpec();
    _json.add("VdcHardwareModelGuid", vdsd.hardwareModelGuid);
    _json.add("VdcModelUID", vdsd.modelUID);
    _json.add("VdcModelVersion", vdsd.modelVersion);
    _json.add("VdcVendorGuid", vdsd.vendorGuid);
    _json.add("VdcOemGuid", vdsd.oemGuid);
    _json.add("VdcConfigURL", vdsd.configURL);
    _json.add("VdcHardwareGuid", vdsd.hardwareGuid);
    _json.add("VdcHardwareInfo", vdsd.model);
    _json.add("VdcHardwareVersion", vdsd.hardwareVersion);
    _json.add("hasActions", bool(device.getHasActions()));
  }
  if (device.isValveDevice()) {
    _json.add("ValveType", device.getValveTypeAsString());
  }

  if (device.isPresent()) {
    dsid_t dsid;
    if (dsuid_to_dsid(device.getDSMeterDSID(), &dsid)) {
      _json.add("meterDSID", dsid2str(dsid));
    } else {
      _json.add("meterDSID", "");
    }
    _json.add("meterDSUID", device.getDSMeterDSID());
    std::string dSMName;
    try {
      dSMName = device.getApartment().getDSMeterByDSID(device.getDSMeterDSID())->getName();
    } catch (std::runtime_error&) {
    }
    _json.add("meterName", dSMName);
    _json.add("busID", device.getShortAddress());
  } else {
    dsid_t dsid;
    if (dsuid_to_dsid(device.getLastKnownDSMeterDSID(), &dsid)) {
      _json.add("meterDSID", dsid2str(dsid));
    } else {
      _json.add("meterDSID", "");
    }
    _json.add("meterDSUID", device.getLastKnownDSMeterDSID());
    std::string dSMName;
    try {
      dSMName = device.getApartment().getDSMeterByDSID(device.getDSMeterDSID())->getName();
    } catch (std::runtime_error&) {
    }
    _json.add("meterName", dSMName);
    _json.add("busID", device.getLastKnownShortAddress());
  }
  _json.add("zoneID", device.getZoneID());
  _json.add("isPresent", device.isPresent());
  _json.add("isValid", device.isValid());

  tmp_date = device.getLastDiscovered();
  _json.add("lastDiscovered", tmp_date);

  tmp_date = device.getFirstSeen();
  _json.add("firstSeen", tmp_date);

  tmp_date = device.getInactiveSince();
  _json.add("inactiveSince", tmp_date);

  _json.add("on", device.isOn());
  _json.add("locked", device.getIsLockedInDSM());
  _json.add("configurationLocked", device.isConfigLocked());
  _json.add("ignoreOperationLock", device.getIgnoreOperationLock());
  _json.add("outputMode", device.getOutputMode());
  _json.add("buttonID", device.getButtonID());
  _json.add("buttonActiveGroup", device.getButtonActiveGroup());
  _json.add("buttonGroupMembership", device.getButtonActiveGroup()); // deprecated
  _json.add("buttonInputMode", device.getButtonInputMode());
  _json.add("buttonInputIndex", device.getButtonInputIndex());
  _json.add("buttonInputCount", device.getButtonInputCount());
  _json.add("AKMInputProperty", device.getAKMInputProperty());

  _json.startArray("groups");
  for (int g = 0; g < device.getGroupsCount(); g++) {
    _json.add(device.getGroupIdByIndex(g));
  }
  _json.endArray();

  auto&& binaryInputs = device.getBinaryInputs();
  _json.add("binaryInputCount", (int)binaryInputs.size());
  _json.startArray("binaryInputs");
  for (auto&& it = binaryInputs.begin(); it != binaryInputs.end(); ++it) {
    _json.startObject();
    _json.add("targetGroup", (*it)->m_targetGroupId);
    _json.add("inputType", static_cast<int>((*it)->m_inputType));
    _json.add("inputId", static_cast<int>((*it)->m_inputId));
    _json.add("state", (*it)->getState().getState());
    _json.endObject();
  }
  _json.endArray();
  _json.add("sensorInputCount", (int)(device.getSensorCount()));

  _json.startArray("sensors");
  for (const auto& sensor : device.getSensors()) {
    _json.startObject();
    _json.add("type", static_cast<int>(sensor->m_sensorType));
    _json.add("valid", sensor->m_sensorValueValidity);
    _json.add("value", sensor->m_sensorValueFloat);
    _json.endObject();
  }
  _json.endArray();

  // check if device has invalid sensor values
  uint8_t sensorCount = device.getSensorCount();
  bool sensorFlag = true;
  for (uint8_t s = 0; s < sensorCount; s++) {
    if (!device.isSensorDataValid(s)) {
      sensorFlag = false;
      break;
    }
  }
  _json.add("sensorDataValid", sensorFlag);

  _json.startArray("outputChannels");
  for (int i = 0; i < device.getOutputChannelCount(); i++) {
    std::shared_ptr<Channel> pChannel = device.getOutputChannel(i);
    _json.startObject();
    _json.add("channelId", pChannel->id);
    _json.add("channelType", pChannel->type);
    _json.add("channelIndex", pChannel->index);
    _json.add("channelName", pChannel->technicalName);
    _json.endObject();
  }
  _json.endArray();

  // check if this is a "main" device, if yes render dSUIDs of all
  // partner devices
  _json.startArray("pairedDevices");
  dsuid_t dsuid = device.getDSID();
  // exclude main device in the list
  if (device.isMainDevice() && (device.getPairedDevices() > 0)) {
    for (int pd = 0; pd < (device.getPairedDevices() - 1); pd++) {
      dsuid_t next;
      dsuid_get_next_dsuid(dsuid, &next);
      _json.add(next);
      dsuid = next;
    }
  }
  _json.endArray();

  _json.endObject();
} // toJSON(DeviceReference)

  void toJSON(std::shared_ptr<const DSMeter> _dsMeter, JSONWriter& _json) {
    _json.startObject();
    _json.add("name", _dsMeter->getName());
    dsid_t dsid;
    if (dsuid_to_dsid(_dsMeter->getDSID(), &dsid)) {
      _json.add("dsid", dsid2str(dsid));
    } else {
      _json.add("dsid", "");
    }
    _json.add("dSUID", _dsMeter->getDSID());
    _json.add("DisplayID", _dsMeter->getDisplayID());
    _json.add("hwVersion", 0);
    _json.add("hwVersionString", _dsMeter->getHardwareVersion());
    _json.add("swVersion", _dsMeter->getSoftwareVersion());
    _json.add("armSwVersion", _dsMeter->getArmSoftwareVersion());
    _json.add("dspSwVersion", _dsMeter->getDspSoftwareVersion());
    _json.add("isUpToDate", _dsMeter->isUpToDate());
    _json.add("apiVersion", _dsMeter->getApiVersion());
    _json.add("authorized", _dsMeter->authorized());
    _json.add("hwName", _dsMeter->getHardwareName());
    _json.add("isPresent", _dsMeter->isPresent());
    _json.add("isValid", _dsMeter->isValid());
    _json.add("busMemberType", _dsMeter->getBusMemberType());
    _json.add("hasDevices", _dsMeter->getCapability_HasDevices());
    _json.add("hasMetering", _dsMeter->getCapability_HasMetering());
    _json.add("hasMeteringProducer", _dsMeter->getCapability_HasMeteringProducer());
    _json.add("hasMeteringProducerEnabled", _dsMeter->getCapability_HasMeteringProducerEnabled());
    _json.add("hasBlinking", _dsMeter->getCapability_HasBlinking());
    _json.add("VdcConfigURL", _dsMeter->getVdcConfigURL());
    _json.add("VdcModelUID", _dsMeter->getVdcModelUID());
    _json.add("VdcHardwareGuid", _dsMeter->getVdcHardwareGuid());
    _json.add("VdcHardwareModelGuid", _dsMeter->getVdcHardwareModelGuid());
    _json.add("VdcImplementationId", _dsMeter->getVdcImplementationId());
    _json.add("VdcVendorGuid", _dsMeter->getVdcVendorGuid());
    _json.add("VdcOemGuid", _dsMeter->getVdcOemGuid());
    std::bitset<8> flags = _dsMeter->getPropertyFlags();
    _json.add("ignoreActionsFromNewDevices", flags.test(4));
    _json.endObject();
  }

  void toJSON(const Set& _set, JSONWriter& _json, bool _showHidden) {
    for(int iDevice = 0; iDevice < _set.length(); iDevice++) {
      const DeviceReference& d = _set.get(iDevice);
      if (d.getDevice()->is2WaySlave() == true) {
        // do not render "slave" devices
        continue;
      }
      if (!_showHidden && !d.getDevice()->isVisible() && d.getDevice()->isPresent()) {
        // do not render hidden TNY devices except for inactive ones
        continue;
      }
      toJSON(d, _json);
    }
  } // toJSON(Set,Name)

  void toJSON(std::shared_ptr<const Group> _group, JSONWriter& _json) {
    _json.startObject();
    _json.add("id", _group->getID());
    _json.add("name", _group->getName());
    _json.add("color", _group->getColor());
    _json.add("applicationType", static_cast<int>(_group->getApplicationType()));
    _json.add("isPresent", _group->isPresent());
    _json.add("isValid", _group->isValid());
    _group->serializeApplicationConfiguration(_group->getApplicationConfiguration(), _json);

    _json.startArray("devices");
    Set devices = _group->getDevices();
    for(int iDevice = 0; iDevice < devices.length(); iDevice++) {
      if (devices[iDevice].getDevice()->is2WaySlave()) {
        // do not render "slave" devices
        continue;
      }
      if (devices[iDevice].getDevice()->isVisible() != true) {
        // do not render hidden TNY devices
        continue;
      }
      _json.add(devices[iDevice].getDSID());
    }
    _json.endArray();
    _json.endObject();
  } // toJSON(Group)

  void toJSON(std::shared_ptr<const Cluster> _cluster, JSONWriter& _json) {
    _json.startObject();
    _json.add("id", _cluster->getID());
    _json.add("name", _cluster->getName());
    _json.add("color", _cluster->getColor());
    _json.add("applicationType", static_cast<int>(_cluster->getApplicationType()));
    _json.add("isPresent", _cluster->isPresent());
    _json.add("isValid", _cluster->isValid());
    _cluster->serializeApplicationConfiguration(_cluster->getApplicationConfiguration(), _json);
    _json.add("CardinalDirection", toString(_cluster->getLocation()));
    _json.add("ProtectionClass", _cluster->getProtectionClass());
    _json.add("isAutomatic", _cluster->isAutomatic());
    _json.add("configurationLock", _cluster->isConfigurationLocked());

    _json.startArray("devices");
    Set devices = _cluster->getDevices();
    for(int iDevice = 0; iDevice < devices.length(); iDevice++) {
      if (devices[iDevice].getDevice()->is2WaySlave()) {
        // do not render "slave" devices
        continue;
      }
      if (devices[iDevice].getDevice()->isVisible() != true) {
        // do not render hidden TNY devices
        continue;
      }
      _json.add(devices[iDevice].getDSID());
    }
    _json.endArray();
    _json.startArray("lockEvents");
    const std::vector<std::string>& oplocks = _cluster->getLockEvents();
    for (size_t o = 0; o < oplocks.size(); o++) {
      _json.add(oplocks.at(o));
    }
    _json.endArray();
    _json.endObject();
  } // toJSON(Cluster)

  void toJSON(Zone& _zone, JSONWriter& _json, bool _includeDevices) {
    _json.startObject();
    _json.add("id", _zone.getID());
    _json.add("name", _zone.getName());
    _json.add("isPresent", _zone.isPresent());
    _json.add("floorId", _zone.getFloorID());

    if(_includeDevices) {
      _json.startArray("devices");
      toJSON(_zone.getDevices(), _json);
      _json.endArray();
      _json.startArray("groups");
      for (std::shared_ptr<Group> pGroup : _zone.getGroups()) {
        toJSON(pGroup, _json);
      }
      _json.endArray();
    }

    _json.endObject();
  } // toJSON(Zone)

  void toJSON(Floor& _floor, JSONWriter& _json) {
    _json.startObject();
    _json.add("id", _floor.getID());
    _json.add("order", _floor.getOrder());
    _json.add("name", _floor.getName());

    _json.startArray("zones");
    std::vector<int> zones = _floor.getZones();
    for (size_t i = 0; i < zones.size(); i++) {
      _json.add(zones.at(i));
    }
    _json.endArray();

    _json.endObject();
  } // toJSON(Zone)


  void toJSON(Apartment& _apartment, JSONWriter& _json) {
    _json.startObject("apartment");

    _json.startArray("clusters");
    std::vector<std::shared_ptr<Cluster> > clusters = _apartment.getClusters();
    for (std::shared_ptr<Cluster> pCluster : clusters) {
      if (pCluster->getApplicationType() == ApplicationType::none) {
        continue;
      }
      toJSON(static_cast<std::shared_ptr<const Cluster> >(pCluster), _json);
    }
    _json.endArray();

    _json.startArray("zones");
    std::vector<std::shared_ptr<Zone> > zones = _apartment.getZones();
    for (std::shared_ptr<Zone> pZone : zones) {
      toJSON(*pZone, _json);
    }
    _json.endArray();

    _json.startArray("floors");
    std::vector<std::shared_ptr<Floor> > floors = _apartment.getFloors();
    for (std::shared_ptr<Floor> pFloor : floors) {
      toJSON(*pFloor, _json);
    }
    _json.endArray();

    _json.endObject();
  } // toJSON(Apartment)

}
