/*
    Copyright (c) 2010,2011 digitalSTROM.org, Zurich, Switzerland

    Authors: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>
             Christian Hitz, aizo AG <christian.hitz@aizo.com>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#include "meteringrequesthandler.h"

#include <boost/lexical_cast.hpp>
#include <cmath>
#include <numeric>
#include <vector>

#include "metering/metering.h"
#include "model/apartment.h"
#include "model/dsmeter.h"
#include "setbuilder.h"
#include "web/restful.h"

static ds::LogChannel dsLogChannel("webHandlerMeteringRequest");

using namespace std::literals;
using namespace dss::metering;

using system_clock = std::chrono::system_clock;

namespace dss {

//=========================================== MeteringRequestHandler

MeteringRequestHandler::MeteringRequestHandler(Apartment& _apartment, Metering& _metering)
    : m_Apartment(_apartment), m_Metering(_metering) {}

std::string MeteringRequestHandler::getResolutions() {
  JSONWriter json;
  json.startArray("resolutions");
  for (auto [resolution, _] : m_Metering.getConfig()) {
    json.startObject();
    json.add("resolution", std::to_string(std::chrono::seconds(resolution).count()));
    json.endObject();
  }
  json.endArray();

  return json.successJSON();
}

std::string MeteringRequestHandler::getSeries() {
  JSONWriter json;
  json.startArray("series");

  for (const auto& dsMeter : m_Apartment.getDSMeters()) {
    if (!dsMeter->getCapability_HasMetering()) {
      continue;
    }

    json.startObject();
    json.add("dSUID", dsMeter->getDSID());
    json.add("type", "energy");
    json.endObject();

    json.startObject();
    json.add("dSUID", dsMeter->getDSID());
    json.add("type", "energyDelta");
    json.endObject();

    json.startObject();
    json.add("dSUID", dsMeter->getDSID());
    json.add("type", "power");
    json.endObject();
  }
  json.endArray();
  return json.successJSON();
}

std::vector<std::shared_ptr<DSMeter>> parseDsuid(Apartment& apartment, const std::string& deviceDSUIDString) {
  MeterSetBuilder builder(apartment);
  if (beginsWith(deviceDSUIDString, ".meters(")) {
    return builder.buildSet(deviceDSUIDString, true);
  } else {
    const auto&& deviceDSIDStringSet = ".meters(" + deviceDSUIDString + ")";
    return builder.buildSet(std::move(deviceDSIDStringSet), true);
  }
}

std::string MeteringRequestHandler::getValues(const RestfulRequest& _request) {
  DS_DEBUG_ENTER();
  std::vector<std::shared_ptr<DSMeter>> meters;
  std::chrono::seconds resolution;
  SeriesTypes energy;
  Unit unit{};
  auto timeDuration = TimeDuration{};
  int valueCount{0};
  bool requestMeteringSet;
  bool withProduction;

  try {
    std::string deviceDSUIDString = _request.getRequiredParameter("dsuid");
    requestMeteringSet = beginsWith(deviceDSUIDString, ".meters(");
    meters = parseDsuid(m_Apartment, deviceDSUIDString);

    std::string resolutionString = _request.getRequiredParameter("resolution");
    resolution = std::chrono::seconds(boost::lexical_cast<int>(resolutionString));

    std::string typeString = _request.getRequiredParameter("type");
    // consumption legacy name for power
    energy = (typeString == "consumption"sv) ? SeriesTypes::power : ds::requireEnum<SeriesTypes>(typeString);

    unit = (energy == SeriesTypes::power) ? Unit::W : Unit::Wh;
    if (auto unitString = _request.getParameter("unit"); !unitString.empty()) {
      unit = ds::requireEnum<Unit>(unitString);
      switch (energy) {
        case SeriesTypes::energyDelta:
        case SeriesTypes::energy:
          if (unit == Unit::W) {
            return JSONWriter::failure(ds::str("Invalid unit '", unitString, "'"));
          }
          break;
        case SeriesTypes::power:
          // for symmetry we should abort as well
          unit = Unit::W;
          break;
      }
    }

    auto parseSecond = [](auto& target, const auto& seconds1970String) {
      // boost::lexical_scope -> fail hard
      int timeStamp = strToIntDef(seconds1970String, -1);
      if (timeStamp > -1) {
        target = system_clock::from_time_t(timeStamp);
      }
    };

    if (auto startTimeString = _request.getParameter("startTime"); !startTimeString.empty()) {
      parseSecond(timeDuration.start, startTimeString);
    }

    if (auto endTimeString = _request.getParameter("endTime"); !endTimeString.empty()) {
      parseSecond(timeDuration.end, endTimeString);
    }

    valueCount = _request.tryGetParameter<int>("valueCount").value_or(0);
    withProduction = _request.tryGetParameter<bool>("withProduction").value_or(false);

  } catch (std::exception& e) {
    return JSONWriter::failure(e.what());
  }

  std::vector<dss::Dsuid> meterIds;
  std::transform(meters.begin(), meters.end(), std::back_inserter(meterIds),
                 [](const auto& meter) { return meter->getDSID(); });

  DS_DEBUG(meterIds, resolution, energy, unit, timeDuration, valueCount, withProduction)

  auto pSeries = m_Metering.getSeries(meterIds, resolution, energy, (unit == Unit::Wh), timeDuration, valueCount);
  DS_DEBUG(pSeries.values.size());
  if (pSeries.values.empty()) {
    return JSONWriter::failure(ds::str("Could not find data for '", energy, "' and resolution '", resolution, "'"));
  }

  decltype(pSeries) production;
  if (withProduction) {
    production =
        m_Metering.getProducedSeries(meterIds, resolution, energy, (unit == Unit::Wh), timeDuration, valueCount);
    DS_DEBUG(production.values.size());

    if (!production.values.empty()) {
      DS_NOTICE(production.values.size(), pSeries.values.size());
      DS_NOTICE(production.values.front().timeStamp, pSeries.values.front().timeStamp);
      DS_NOTICE(production.values.back().timeStamp, pSeries.values.back().timeStamp);

      if (production.values.front().timeStamp > pSeries.values.front().timeStamp) {
        // can happen if the second increases betweeen fetching consumption/prodcution data
        decltype(pSeries) tmp;
        std::copy(++pSeries.values.begin(), pSeries.values.end(), std::back_inserter(tmp.values));
        pSeries.values = std::move(tmp.values);
      };
      DS_ASSERT(production.values.front().timeStamp == pSeries.values.front().timeStamp);
      if (production.values.size() != pSeries.values.size()) {
        DS_ASSERT(std::abs(static_cast<int>(production.values.size() - pSeries.values.size())) == 1);
        auto restoreTS = [](auto& longer, auto& shorter) {
          auto ts = longer.values.back().timeStamp;
          shorter.values.push_back({.value = 0.0, .timeStamp = ts});
        };
        production.values.size() <= pSeries.values.size() ? restoreTS(pSeries, production)
                                                          : restoreTS(production, pSeries);
      }
      DS_ASSERT(production.values.size() == pSeries.values.size());
      DS_ASSERT(production.values.back().timeStamp == pSeries.values.back().timeStamp);
    }
  }

  JSONWriter json;
  if (requestMeteringSet) {
    json.startArray("meterID");
    for (const auto& meter : meters) {
      json.add(meter->getDSID());
    }
    json.endArray();
  } else {
    json.add("meterID", meters[0]->getDSID());
  }
  json.add("type", ds::fromEnum<std::string>(energy));
  json.add("unit", ds::fromEnum<std::string>(unit));
  json.add("resolution", std::to_string(std::chrono::seconds(resolution).count()));
  json.startArray("values");

  auto both = std::vector<std::tuple<dss::DateTime, double, double>>{};
  if (!production.values.empty()) {
    std::transform(pSeries.values.begin(), pSeries.values.end(), production.values.begin(), std::back_inserter(both),
                   [](const auto& consumption, const auto production) {
                     DS_ASSERT(consumption.timeStamp == production.timeStamp);
                     return std::make_tuple(dss::DateTime{consumption.timeStamp}, consumption.value, production.value);
                   });
  } else {
    std::transform(pSeries.values.begin(), pSeries.values.end(), std::back_inserter(both), [](const auto& consumption) {
      return std::make_tuple(dss::DateTime{consumption.timeStamp}, consumption.value, 0.0f);
    });
  }

  for (const auto& [timestamp, consumption, production] : both) {
    json.startArray();
    DateTime tmp_date = timestamp;
    json.add((int)tmp_date.secondsSinceEpoch());
    json.add(consumption);
    if (withProduction) {
      json.add(production);
    }
    json.endArray();
  }
  json.endArray();

  DS_DEBUG_LEAVE();
  return json.successJSON();
}

std::string MeteringRequestHandler::getLatest(const RestfulRequest& _request, bool aggregateMeterValues) {
  std::vector<std::shared_ptr<DSMeter>> meters;
  SeriesTypes seriesType;
  Unit unit{};
  bool withProduction;

  try {
    std::string from = _request.getRequiredParameter("from");
    meters = parseDsuid(m_Apartment, from);

    auto type = _request.getRequiredParameter("type");
    // consumption legacy name for power
    seriesType = (type == "consumption"sv) ? SeriesTypes::power : ds::requireEnum<SeriesTypes>(type);
    if ((seriesType != SeriesTypes::power) && (seriesType != SeriesTypes::energy)) {
      return JSONWriter::failure("Invalid or missing type parameter");
    }

    unit = (seriesType == SeriesTypes::energy) ? Unit::Wh : Unit::W;
    if (auto unitString = _request.getParameter("unit"); !unitString.empty()) {
      unit = ds::requireEnum<Unit>(unitString);
      if (seriesType == SeriesTypes::energy && (unit != Unit::Ws && unit != Unit::Wh)) {
#if 0
      DSD-9878: sven queries power with unit=Ws, what a douche bag
      // (seriesType == SeriesTypes::power && (unit != Unit::W)))
#endif
        return JSONWriter::failure(ds::str("Invalid unit '", unitString, "'"));
      }
    }

    withProduction = _request.tryGetParameter<bool>("withProduction").value_or(false);
  } catch (std::exception& e) {
    return JSONWriter::failure(e.what());
  }

  if (meters.empty()) {
    JSONWriter json;
    json.startArray("values");
    json.endArray();
    return json.successJSON();
  }

  auto getValue = [isEnergy = seriesType == SeriesTypes::energy,
                   energyQuotient = (unit == Unit::Ws) ? 1 : 3600](const ArchiveValue& currentValue) {
    return isEnergy ? currentValue.energy / energyQuotient : currentValue.power;
  };

  std::vector<std::tuple<dsuid_t, ArchiveValue, ArchiveValue>> values;
  for (auto& dsMeter : meters) {
    auto tryConsumption = m_Metering.getEnergy(dsMeter->getDSID());
    if (!tryConsumption) {
      continue;
    }
    auto production = ArchiveValue{.power = 0, .energy = 0};
    if (withProduction) {
      production = m_Metering.getProducedEnergy(dsMeter->getDSID()).value_or(production);
    }

    values.push_back(std::make_tuple(dsMeter->getDSID(), *tryConsumption, production));
  }

  ds::DateTime lastUpdateAll;
  if (values.size() > 0) {
    lastUpdateAll =
        std::get<1>(*std::max_element(std::cbegin(values), std::cend(values), [](const auto& left, const auto& right) {
          return std::get<1>(left).lastChanged < std::get<1>(right).lastChanged;
        })).lastChanged;
  }

  try {
    JSONWriter json;
    json.startArray("values");
    if (!aggregateMeterValues) {
      for (const auto& [dsuid, consumption, production] : values) {
        json.startObject();
        json.add("dSUID", dsuid);
        if (!withProduction) {
          json.add("dsid", dsuidToDsid(dsuid)); // SVEN uses, pre-dates production
          json.add("value", getValue(consumption));
        } else {
          json.add("consumption", getValue(consumption));
          json.add("production", getValue(production));
        }
        json.add("date", DateTime{lastUpdateAll}.toString());
        json.endObject();
      }
    } else {
      auto aggregatedValue = std::accumulate(std::cbegin(values), std::cend(values), std::make_pair(0.0, 0.0),
                                             [getValue](auto& sum, const auto& item) {
                                               sum.first += getValue(std::get<1>(item));
                                               sum.second += getValue(std::get<2>(item));
                                               return sum;
                                             });
      std::vector<dsuid_t> dsuids;
      std::transform(std::cbegin(values), std::cend(values), std::back_inserter(dsuids),
                     [](const auto& item) { return std::get<dsuid_t>(item); });
      json.startObject();
      {
        auto DS_UNIQUE_NAME(x) = json.startScopedArray("dSUID");
        for (dsuid_t dsuid : dsuids) {
          json.add(dsuid);
        }
      }
      if (!withProduction) {
        // legacy support
        {
          auto DS_UNIQUE_NAME(x) = json.startScopedArray("dsid");
          for (dsuid_t dsuid : dsuids) {
            json.add(dsuidToDsid(dsuid));
          }
        }
        json.add("value", static_cast<unsigned long long>(aggregatedValue.first));
      } else {
        json.add("consumption", aggregatedValue.first);
        json.add("production", aggregatedValue.second);
      }
      json.add("date", DateTime{lastUpdateAll}.toString());
      json.endObject();
    }

    json.endArray();
    return json.successJSON();

  } catch (std::runtime_error&) {
    return JSONWriter::failure("Could not apply properties to JSON object.");
  }
}

WebServerResponse MeteringRequestHandler::jsonHandleRequest(const RestfulRequest& _request,
                                                            std::shared_ptr<Session> _session,
                                                            const struct mg_connection* _connection) {
  if (_request.getMethod() == "getResolutions") {
    return getResolutions();
  } else if (_request.getMethod() == "getSeries") {
    return getSeries();
  } else if (_request.getMethod() == "getValues") { //?dsuid=;n=,resolution=,type=
    return getValues(_request);
  } else if (_request.getMethod() == "getLatest") {
    return getLatest(_request, false);
  } else if (_request.getMethod() == "getAggregatedValues") { //?set=;n=,resolution=;type=
    return getValues(_request);
  } else if (_request.getMethod() == "getAggregatedLatest") { //?dsuid=;n=,resolution=;type=
    return getLatest(_request, true);
  }
  throw std::runtime_error("Unhandled function");
} // handleRequest

} // namespace dss
