/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "zonerequesthandler.h"

#include <digitalSTROM/dsuid.h>
#include <ds/locale/ds-locale.h>
#include <ds/log.h>

#include "deviceinterfacerequesthandler.h"
#include "model/apartment.h"
#include "model/cluster.h"
#include "model/device.h"
#include "model/dsmeter.h"
#include "model/group.h"
#include "model/modelmaintenance.h"
#include "model/scenehelper.h"
#include "model/set.h"
#include "model/status.h"
#include "model/zone.h"
#include "stringconverter.h"
#include "structuremanipulator.h"
#include "util.h"
#include "web/restful.h"

static ds::LogChannel dsLogChannel("jsonApiZone");

namespace dss {

namespace {

  ApplicationType applicationTypeFromString(const std::string& application) {
    const auto applicationId = strToInt(application);
    if (applicationId < GroupIDYellow || applicationId > GroupIDGlobalAppMax) {
      const auto errorMessage = "Unsupported application type '" + application + "'";
      throw std::runtime_error(errorMessage);
    } else if (applicationId != GroupIDGray && applicationId != GroupIDGlobalAppDsAwnings) {
      const auto errorMessage = "Application actions undefined";
      throw std::runtime_error(errorMessage);
    }

    return static_cast<ApplicationType>(applicationId);
  } // applicationTypeFromString

} // namespace

//=========================================== ZoneRequestHandler

ZoneRequestHandler::ZoneRequestHandler(Apartment& apartment) : m_Apartment(apartment) {}

std::string ZoneRequestHandler::getLastCalledScene(std::shared_ptr<Zone> pZone, std::shared_ptr<Group> pGroup, const RestfulRequest& _request) {
  int lastScene = 0;
  if (pGroup != NULL) {
    lastScene = pGroup->getLastCalledScene();
  } else if (pZone != NULL) {
    lastScene = pZone->getGroup(0)->getLastCalledScene();
  } else {
    // should never reach here because ok, would be false
    DS_FAIL_ASSERT();
  }
  JSONWriter json;
  json.add("scene", lastScene);
  return json.successJSON();
}

std::string ZoneRequestHandler::getName(std::shared_ptr<Zone> pZone, std::shared_ptr<Group> pGroup, const RestfulRequest& _request) {
  JSONWriter json;
  json.add("name", pZone->getName());
  return json.successJSON();
}

std::string ZoneRequestHandler::setName(std::shared_ptr<Zone> pZone, std::shared_ptr<Group> pGroup, const RestfulRequest& _request) {
  auto name = escapeHTML(_request.getRequiredParameter("newName"));
  DS_REQUIRE(pZone->getID() != 0, "Changing name for zone 0 is not allowed");

  DS_NOTICE("Set zone name.", pZone, name);
  pZone->setName(name);
  return JSONWriter::success();
}

std::string ZoneRequestHandler::sceneSetName(Group& group, const RestfulRequest& _request) {
  int sceneNumber = strToIntDef(_request.getParameter("sceneNumber"), -1);
  if (sceneNumber == -1) {
    return JSONWriter::failure("Need valid parameter 'sceneNumber'");
  }
  if (!SceneHelper::isInRange(sceneNumber, group.getZoneID())) {
    return JSONWriter::failure("Parameter 'sceneNumber' out of bounds ('" + intToString(sceneNumber) + "')");
  }
  auto name = _request.getRequiredParameter("newName");
  DS_NOTICE("Set scene name.", group, sceneNumber, name);
  StructureManipulator(m_Apartment).sceneSetName(group.sharedFromThis(), sceneNumber, name);
  return JSONWriter::success();
}

std::string ZoneRequestHandler::sceneGetName(std::shared_ptr<Zone> pZone, std::shared_ptr<Group> pGroup, const RestfulRequest& _request) {
  if (pGroup == NULL) {
    return JSONWriter::failure("Need group to work");
  }
  int sceneNumber = strToIntDef(_request.getParameter("sceneNumber"), -1);
  if (sceneNumber == -1) {
    return JSONWriter::failure("Need valid parameter 'sceneNumber'");
  }
  if (!SceneHelper::isInRange(sceneNumber, pGroup->getZoneID())) {
    return JSONWriter::failure("Parameter 'sceneNumber' out of bounds ('" + intToString(sceneNumber) + "')");
  }

  std::string sceneName = pGroup->getSceneName(sceneNumber);
  JSONWriter json;
  json.add("name", sceneName);
  return json.successJSON();
}

std::string ZoneRequestHandler::getReachableScenes(
    std::shared_ptr<Zone> pZone, std::shared_ptr<Group> pGroup, const RestfulRequest& _request) {
  if (pGroup == NULL) {
    return JSONWriter::failure("Need group to work");
  }

  // for all selected groups try to find all reachable scenes and their names inside group
  Set devicesInGroup = pGroup->getDevices();
  std::vector<int> groupReachableScenes;

  for (int iDevice = 0; iDevice < devicesInGroup.length(); iDevice++) {
    auto device = devicesInGroup.get(iDevice).getDevice();

    // check only devices with buttons that are assigned to this exact group
    // TODO(soon): if device is removed from cluster the ButtonActiveGroup is set to 255 and not to button group,
    // current workaround is that we use device activeGroup if the buttonActiveGroup is not set
    if ((device->getButtonInputCount() > 0) &&
        ((device->getButtonActiveGroup() == pGroup->getID()) ||
            ((device->getButtonActiveGroup() == 255) && (device->getActiveGroup() == pGroup->getID())))) {
      const auto& deviceReachableScenes = SceneHelper::getReachableScenes(device->getButtonID(), pGroup->getID());
      std::copy(deviceReachableScenes.begin(), deviceReachableScenes.end(), std::back_inserter(groupReachableScenes));
    }
  }

  // sort the scenes by value and remove duplicates
  std::sort(groupReachableScenes.begin(), groupReachableScenes.end());
  groupReachableScenes.resize(std::distance(
      groupReachableScenes.begin(), std::unique(groupReachableScenes.begin(), groupReachableScenes.end())));

  // filter the scenes so only available scenes in this group remain
  const auto& availableScenes = pGroup->getAvailableScenes();
  std::vector<int> filteredGroupScenes;

  std::set_intersection(groupReachableScenes.begin(), groupReachableScenes.end(), availableScenes.begin(),
      availableScenes.end(), std::back_inserter(filteredGroupScenes));

  JSONWriter json;
  json.startArray("reachableScenes");
  for (auto&& sceneNr : filteredGroupScenes) {
    json.add(sceneNr);
  }
  json.endArray();

  json.startArray("userSceneNames");
  // add the resulting names to the output
  for (auto&& sceneNr : filteredGroupScenes) {
    // get the user defined scene name
    auto sceneName = pGroup->getSceneName(sceneNr);
    if (!sceneName.empty()) {
      json.startObject();
      json.add("sceneNr", sceneNr);
      json.add("sceneName", sceneName);
      json.endObject();
    }
  }
  json.endArray();

  return json.successJSON();
}

std::string ZoneRequestHandler::getTemperatureControlStatus(
        std::shared_ptr<Zone> pZone, std::shared_ptr<Group> pGroup, const RestfulRequest& _request) {
  if (pZone->getID() == 0) {
    return JSONWriter::failure("Zone id 0 is invalid");
  }

  JSONWriter json;
  addTemperatureControlStatus(json, pZone);

  return json.successJSON();
}

std::string ZoneRequestHandler::getTemperatureControlConfig(
        std::shared_ptr<Zone> pZone, std::shared_ptr<Group> pGroup, const RestfulRequest& _request) {
  if (pZone->getID() == 0) {
    return JSONWriter::failure("Zone id 0 is invalid");
  }

  JSONWriter json;
  addTemperatureControlConfig(json, pZone);

  return json.successJSON();
}

std::string ZoneRequestHandler::getTemperatureControlConfig2(
        std::shared_ptr<Zone> pZone, std::shared_ptr<Group> pGroup, const RestfulRequest& _request) {
  if (pZone->getID() == 0) {
    return JSONWriter::failure("Zone id 0 is invalid");
  }

  JSONWriter json;
  addTemperatureControlConfig2(json, pZone);

  return json.successJSON();
}

std::string ZoneRequestHandler::setTemperatureControlConfig(
        std::shared_ptr<Zone> pZone, std::shared_ptr<Group> pGroup, const RestfulRequest& _request) {
  if (pZone->getID() == 0) {
    return JSONWriter::failure("Zone id 0 is invalid");
  }

  JSONWriter json;
  ZoneHeatingConfigSpec_t hConfig = pZone->getHeatingConfigSpec();

  if (_request.hasParameter("ControlMode")) {
    uint8_t x;
    (void)_request.getParameter("ControlMode", x);
    hConfig.mode = static_cast<HeatingControlMode>(x);
  }
  if (_request.hasParameter("ReferenceZone")) {
    (void)_request.getParameter("ReferenceZone", hConfig.SourceZoneId);
  }
  if (_request.hasParameter("CtrlOffset")) {
    int offset;
    (void)_request.getParameter("CtrlOffset", offset);
    hConfig.Offset = offset;
  }
  if (_request.hasParameter("EmergencyValue")) {
    (void)_request.getParameter("EmergencyValue", hConfig.EmergencyValue);
    hConfig.EmergencyValue += 100;
  }
  if (_request.hasParameter("ManualValue")) {
    (void)_request.getParameter("ManualValue", hConfig.ManualValue);
    hConfig.ManualValue += 100;
  }
  if (_request.hasParameter("CtrlYmin")) {
    (void)_request.getParameter("CtrlYmin", hConfig.Ymin);
    hConfig.Ymin += 100;
  }
  if (_request.hasParameter("CtrlYmax")) {
    (void)_request.getParameter("CtrlYmax", hConfig.Ymax);
    hConfig.Ymax += 100;
  }
  double tempDouble = 0;
  if (_request.getParameter("CtrlKp", tempDouble)) {
    hConfig.Kp = (uint16_t)tempDouble * 40;
  }
  (void)_request.getParameter("CtrlTi", hConfig.Ti);
  (void)_request.getParameter("CtrlTs", hConfig.Ts);
  (void)_request.getParameter("CtrlKd", hConfig.Kd);
  if (_request.getParameter("CtrlImin", tempDouble)) {
    hConfig.Imin = static_cast<int16_t>(tempDouble * 40);
  }
  if (_request.getParameter("CtrlImax", tempDouble)) {
    hConfig.Imax = static_cast<int16_t>(tempDouble * 40);
  }
  bool tempBool;
  if (_request.getParameter("CtrlAntiWindUp", tempBool)) {
    hConfig.AntiWindUp = tempBool ? 1 : 0;
  }

  DS_NOTICE("Set zone heating config.", *pZone, _request.query());

  StructureManipulator manipulator(m_Apartment);
  manipulator.setZoneHeatingConfigSpec(pZone, hConfig);
  return json.successJSON();
}

std::string ZoneRequestHandler::setTemperatureControlConfig2(
        std::shared_ptr<Zone> pZone, std::shared_ptr<Group> pGroup, const RestfulRequest& _request) {
  if (pZone->getID() == 0) {
    return JSONWriter::failure("Zone id 0 is invalid");
  }

  JSONWriter json;
  auto config = pZone->temperatureConfig();

  if (const auto& fields = _request.tryGetParameter("fields")) {
    config.parseFields(*fields);
  }

  if (_request.hasParameter("mode")) {
    if (auto controlMode = heatingControlModeFromName(_request.getParameter("mode"))) {
      config.m_mode = *controlMode;
    } else {
      return JSONWriter::failure("Mode is invalid!");
    }
  }

  if (_request.hasParameter("targetTemperatures")) {
    config.parseTargetTemperatures(_request.getParameter("targetTemperatures"));
  }

  if (_request.hasParameter("fixedValues")) {
    config.parseFixedValues(_request.getParameter("fixedValues"));
  }

  if (_request.hasParameter("controlMode")) {
    config.parseControlMode(_request.getParameter("controlMode"));
  }

  if (_request.hasParameter("zoneFollowerMode")) {
    config.parseFollowerMode(_request.getParameter("zoneFollowerMode"));
  }

  if (_request.hasParameter("manualMode")) {
    config.parseManualMode(_request.getParameter("manualMode"));
  }

  DS_NOTICE("Set zone heating config.", *pZone, _request.query());

  StructureManipulator manipulator(m_Apartment);
  manipulator.setZoneHeatingConfig(pZone, config);

  return json.successJSON();
}

std::string ZoneRequestHandler::getTemperatureControlValues(
        std::shared_ptr<Zone> pZone, std::shared_ptr<Group> pGroup, const RestfulRequest& _request) {
  if (pZone->getID() == 0) {
    return JSONWriter::failure("Zone id 0 is invalid");
  }

  JSONWriter json;
  addTemperatureControlValues(json, pZone);

  return json.successJSON();
}

std::string ZoneRequestHandler::setTemperatureControlValues(
        std::shared_ptr<Zone> pZone, std::shared_ptr<Group> pGroup, const RestfulRequest& _request) {
  if (pZone->getID() == 0) {
    return JSONWriter::failure("Zone id 0 is invalid");
  }

  JSONWriter json;
  const TemperatureConfig& hProp = pZone->temperatureConfig();
  ZoneHeatingOperationModeSpec_t hOpValues = pZone->getHeatingOperationModeValues();
  SensorType SensorConversion;

  if (hProp.m_mode == HeatingControlMode::PID) {
    SensorConversion = SensorType::RoomTemperatureSetpoint;
  } else if (hProp.m_mode == HeatingControlMode::FIXED) {
    SensorConversion = SensorType::RoomTemperatureControlVariable;
  } else {
    return JSONWriter::failure("Cannot set control values in current mode");
  }

  gsl::czstring parameters[] = {"Off", "Comfort", "Economy", "NotUsed", "Night", "Holiday", "Cooling", "CoolingOff"};
  for (size_t i = 0; i < ds::arraySize(parameters); i++) {
    if (auto x = _request.tryGetParameter<double>(parameters[i])) {
      hOpValues.opModes[i] = doubleToSensorValue(SensorConversion, *x);
    }
  }

  DS_NOTICE("Set zone heating config.", *pZone, _request.query());

  // set the requested data in the zone model
  pZone->setHeatingOperationMode(hOpValues);

  // send current operation mode to all dsms
  DSS::getInstance()->assertBlockingAllowed();
  m_Apartment.getBusInterface()->getStructureModifyingBusInterface()->setZoneHeatingOperationModes(
          DSUID_BROADCAST, pZone->getID(), hOpValues);

  return json.successJSON();
}

std::string ZoneRequestHandler::getTemperatureControlInternals(
        std::shared_ptr<Zone> pZone, std::shared_ptr<Group> pGroup, const RestfulRequest& _request) {
  if (pZone->getID() == 0) {
    return JSONWriter::failure("Zone id 0 is invalid");
  }

  JSONWriter json;
  const TemperatureConfig& hProp = pZone->temperatureConfig();
  ZoneHeatingInternalsSpec_t hInternals;

  for (auto&& dsm : pZone->getDSMeters()) {
    if (std::shared_ptr<const DSMeter> meter = dsm.lock()) {
      if (meter->getCapability_HasTemperatureControl()) {

        // start the response for this dsm
        json.startObject(dsuid2str(meter->getDSID()));

        json.add("ControlMode", static_cast<int>(hProp.m_mode));
        json.add("ControlState", hProp.m_HeatingControlState);

        if (hProp.m_mode != HeatingControlMode::PID) {
          json.endObject();
          continue;
        }
        if (hProp.m_HeatingControlState != HeatingControlStateIDInternal) {
          json.endObject();
          continue;
        }
        try {
          memset(&hInternals, 0, sizeof(ZoneHeatingInternalsSpec_t));
          hInternals = m_Apartment.getBusInterface()->getStructureQueryBusInterface()->getZoneHeatingInternals(
              meter->getDSID(), pZone->getID());

          json.add("CtrlTRecent", sensorValueToDouble(SensorType::TemperatureIndoors, hInternals.Trecent));
          json.add("CtrlTReference", sensorValueToDouble(SensorType::RoomTemperatureSetpoint, hInternals.Treference));
          json.add("CtrlTError", (double)hInternals.TError * 0.025);
          json.add("CtrlTErrorPrev", (double)hInternals.TErrorPrev * 0.025);
          json.add("CtrlIntegral", (double)hInternals.Integral * 0.025);
          json.add("CtrlYp", (double)hInternals.Yp * 0.01);
          json.add("CtrlYi", (double)hInternals.Yi * 0.01);
          json.add("CtrlYd", (double)hInternals.Yd * 0.01);
          json.add("CtrlY", sensorValueToDouble(SensorType::RoomTemperatureControlVariable, hInternals.Y));
          json.add("CtrlAntiWindUp", hInternals.AntiWindUp);
        } catch (const std::exception& e) {
          DS_WARNING(e.what());
        }

        // end current dsm object
        json.endObject();
      }
    }
  }

  return json.successJSON();
}

std::string ZoneRequestHandler::setStatusField(
        std::shared_ptr<Zone> pZone, std::shared_ptr<Group> pGroup, const RestfulRequest& request) {
  DS_REQUIRE(pGroup != 0, "Missing required groupID parameter.");
  const auto& field = request.getRequiredParameter("field");
  const auto& value = request.getRequiredParameter("value");

  DS_NOTICE("Set status field.", field, value);

  pGroup->setStatusField(field, value);
  return JSONWriter::success();
}


std::string ZoneRequestHandler::setSensorSource(
        std::shared_ptr<Zone> pZone, std::shared_ptr<Group> pGroup, const RestfulRequest& _request) {
  if (pZone->getID() == 0) {
    return JSONWriter::failure("Not allowed to assign sensor for zone 0");
  }
  std::string dsuidStr = _request.getParameter("dsuid");
  if (dsuidStr.empty()) {
    return JSONWriter::failure("Missing parameter 'dsuid'");
  }
  dsuid_t dsuid = str2dsuid(dsuidStr);

  int intType = strToIntDef(_request.getParameter("sensorType"), -1);
  if (intType < 0) {
    return JSONWriter::failure("Missing or invalid parameter 'sensorType'");
  }
  std::shared_ptr<Device> dev = m_Apartment.getDeviceByDSID(dsuid);

  SensorType type = static_cast<SensorType>(intType);

  switch (type) {
    case SensorType::BrightnessOutdoors:
    case SensorType::TemperatureOutdoors:
    case SensorType::HumidityOutdoors:
    case SensorType::WindSpeed:
    case SensorType::WindDirection:
    case SensorType::GustSpeed:
    case SensorType::GustDirection:
    case SensorType::Precipitation:
    case SensorType::AirPressure:
    case SensorType::SunAzimuth:
    case SensorType::SunElevation:
      pZone = m_Apartment.getZone(0);
      break;
    case SensorType::ActivePower:
    case SensorType::OutputCurrent:
    case SensorType::ElectricMeter:
    case SensorType::OutputCurrent16A:
    case SensorType::ActivePowerVA:
    case SensorType::TemperatureIndoors:
    case SensorType::BrightnessIndoors:
    case SensorType::HumidityIndoors:
    case SensorType::CO2Concentration:
    case SensorType::COConcentration:
    case SensorType::SoundPressureLevel:
    case SensorType::RoomTemperatureSetpoint:
    case SensorType::RoomTemperatureControlVariable:
    case SensorType::Temperature:
    case SensorType::Brightness:
    case SensorType::Humidity:
    case SensorType::WaterQuantity:
    case SensorType::WaterFlowRate:
      // all valid sensor types do not need any processing
      break;
    case SensorType::Reserved1:
    case SensorType::Reserved2:
    case SensorType::NotUsed:
    case SensorType::UnknownType:
    case SensorType::Status:
    default:
      return JSONWriter::failure("Invalid parameter 'sensorType': " + intToString(static_cast<int>(type)));
      break;
  }

  DS_NOTICE("Set zone sensor.", *pZone, type, *dev);

  StructureManipulator manipulator(m_Apartment);
  manipulator.setZoneSensor(*pZone, type, dev);
  return JSONWriter::success();
}

std::string ZoneRequestHandler::clearSensorSource(
        std::shared_ptr<Zone> pZone, std::shared_ptr<Group> pGroup, const RestfulRequest& _request) {
  if (pZone->getID() == 0) {
    return JSONWriter::success();
  }
  int intType = strToIntDef(_request.getParameter("sensorType"), -1);
  if (intType < 0) {
    return JSONWriter::failure("Missing or invalid parameter 'sensorType'");
  }
  SensorType sensorType = static_cast<SensorType>(intType);

  DS_NOTICE("Reset zone sensor.", *pZone, sensorType);

  StructureManipulator manipulator(m_Apartment);
  manipulator.resetZoneSensor(*pZone, sensorType);
  return JSONWriter::success();
}

std::string ZoneRequestHandler::getAssignedSensors(
        std::shared_ptr<Zone> pZone, std::shared_ptr<Group> pGroup, const RestfulRequest& _request) {
  JSONWriter json;
  json.startArray("sensors");
  for (auto&& devSensor : pZone->getAssignedSensors()) {
    json.startObject();
    json.add("sensorType", devSensor.m_sensorType);
    json.add("dsuid", devSensor.m_DSUID);
    json.endObject();
  }
  json.endArray();
  return json.successJSON();
}

std::string ZoneRequestHandler::getSensorValues(
        std::shared_ptr<Zone> pZone, std::shared_ptr<Group> pGroup, const RestfulRequest& _request) {
  ZoneSensorStatus sensorStatus = pZone->getSensorStatus();
  JSONWriter json;

  json.add("id", pZone->getID());
  json.add("name", pZone->getName());
  json.startArray("values");

  if (sensorStatus.temperatureValueTS != DateTime::NullDate && sensorStatus.temperatureValue) {
    json.startObject();
    json.add("TemperatureValue", *sensorStatus.temperatureValue);
    json.add("TemperatureValueTime", sensorStatus.temperatureValueTS.toISO8601_ms());
    json.endObject();
  }
  if (sensorStatus.humidityValueTS != DateTime::NullDate && sensorStatus.humidityValue) {
    json.startObject();
    json.add("HumidityValue", *sensorStatus.humidityValue);
    json.add("HumidityValueTime", sensorStatus.humidityValueTS.toISO8601_ms());
    json.endObject();
  }
  if (sensorStatus.co2ConcentrationValueTS != DateTime::NullDate && sensorStatus.co2ConcentrationValue) {
    json.startObject();
    json.add("CO2concentrationValue", *sensorStatus.co2ConcentrationValue);
    json.add("CO2concentrationValueTime", sensorStatus.co2ConcentrationValueTS.toISO8601_ms());
    json.endObject();
  }
  if (sensorStatus.brightnessValueTS != DateTime::NullDate && sensorStatus.brightnessValue) {
    json.startObject();
    json.add("BrightnessValue", *sensorStatus.brightnessValue);
    json.add("BrightnessValueTime", sensorStatus.brightnessValueTS.toISO8601_ms());
    json.endObject();
  }
  json.endArray();
  return json.successJSON();
}

WebServerResponse ZoneRequestHandler::jsonHandleRequest(
        const RestfulRequest& _request, std::shared_ptr<Session> _session, const struct mg_connection* _connection) {
  StringConverter st("UTF-8", "UTF-8");
  bool ok = true;
  std::string errorMessage;
  std::string zoneName = _request.getParameter("name");
  std::string zoneIDString = _request.getParameter("id");
  std::shared_ptr<Zone> pZone;
  if (!zoneIDString.empty()) {
    int zoneID = strToIntDef(zoneIDString, -1);
    if (zoneID != -1) {
      try {
        pZone = m_Apartment.getZone(zoneID);
      } catch (std::runtime_error& e) {
        ok = false;
        errorMessage = "Could not find zone with id '" + zoneIDString + "'";
      }
    } else {
      ok = false;
      errorMessage = "Could not parse id '" + zoneIDString + "'";
    }
  } else if (!zoneName.empty()) {
    try {
      pZone = m_Apartment.getZoneByName(zoneName);
    } catch (std::runtime_error& e) {
      ok = false;
      errorMessage = "Could not find zone named '" + zoneName + "'";
    }
  } else {
    ok = false;
    errorMessage = "Need parameter name or id to identify zone";
  }
  if (ok) {
    std::shared_ptr<Group> pGroup;
    std::string groupName = _request.getParameter("groupName");
    std::string groupIDString = _request.getParameter("groupID");
    if (!groupName.empty()) {
      try {
        pGroup = pZone->getGroup(groupName);
        if (pGroup == NULL) {
          // TODO: this might better be done by the zone
          throw std::runtime_error("dummy");
        }
      } catch (std::runtime_error& e) {
        errorMessage = "Could not find group with name '" + groupName + "'";
        ok = false;
      }
    } else if (!groupIDString.empty()) {
      try {
        int groupID = strToIntDef(groupIDString, -1);
        if (groupID != -1) {
          pGroup = pZone->getGroup(groupID);
          if (pGroup == NULL) {
            // TODO: this might better be done by the zone
            throw std::runtime_error("dummy");
          }
        } else {
          errorMessage = "Could not parse group id '" + groupIDString + "'";
          ok = false;
        }
      } catch (std::runtime_error& e) {
        errorMessage = "Could not find group with ID '" + groupIDString + "'";
        ok = false;
      }
    }

    if (ok) {
      if (web::handler::isDeviceInterfaceCall(_request)) {
        std::shared_ptr<IDeviceInterface> interface;
        if (pGroup != NULL) {
          interface = pGroup;
        }
        if (ok) {
          if (interface == NULL) {
            interface = pZone;
          }
          auto deviceInterfaceRequestHandler = DeviceInterfaceRequestHandler(m_Apartment);
          return deviceInterfaceRequestHandler.handleDeviceInterfaceRequest(_request, interface, _session);
        }
      } else if (_request.getMethod() == "getActions") {
        return getActions(pZone, pGroup, _request);
      } else if (_request.getMethod() == "callAction") {
        return callAction(pZone, pGroup, _request);
      } else if (_request.getMethod() == "getLastCalledScene") {
        return getLastCalledScene(pZone, pGroup, _request);
      } else if (_request.getMethod() == "getName") {
        return getName(pZone, pGroup, _request);
      } else if (_request.getMethod() == "setName") {
        return setName(pZone, pGroup, _request);
      } else if (_request.getMethod() == "sceneSetName") {
        return sceneSetName(*pGroup, _request);
      } else if (_request.getMethod() == "sceneGetName") {
        return sceneGetName(pZone, pGroup, _request);
      } else if (_request.getMethod() == "getReachableScenes") {
        return getReachableScenes(pZone, pGroup, _request);
      } else if (_request.getMethod() == "getTemperatureControlStatus") {
        return getTemperatureControlStatus(pZone, pGroup, _request);
      } else if (_request.getMethod() == "getTemperatureControlConfig") {
        return getTemperatureControlConfig(pZone, pGroup, _request);
      } else if (_request.getMethod() == "getTemperatureControlConfig2") {
        return getTemperatureControlConfig2(pZone, pGroup, _request);
      } else if (_request.getMethod() == "setTemperatureControlConfig") {
        return setTemperatureControlConfig(pZone, pGroup, _request);
      } else if (_request.getMethod() == "setTemperatureControlConfig2") {
        return setTemperatureControlConfig2(pZone, pGroup, _request);
      } else if (_request.getMethod() == "getTemperatureControlValues") {
        return getTemperatureControlValues(pZone, pGroup, _request);
      } else if (_request.getMethod() == "setTemperatureControlValues") {
        return setTemperatureControlValues(pZone, pGroup, _request);
      } else if (_request.getMethod() == "getTemperatureControlInternals") {
        return getTemperatureControlInternals(pZone, pGroup, _request);
      } else if (_request.getMethod() == "setStatusField") {
        return setStatusField(pZone, pGroup, _request);
      } else if (_request.getMethod() == "setSensorSource") {
        return setSensorSource(pZone, pGroup, _request);
      } else if (_request.getMethod() == "clearSensorSource") {
        return clearSensorSource(pZone, pGroup, _request);
      } else if (_request.getMethod() == "getAssignedSensors") {
        return getAssignedSensors(pZone, pGroup, _request);
      } else if (_request.getMethod() == "getSensorValues") {
        return getSensorValues(pZone, pGroup, _request);
      } else {
        return JSONWriter::failure("Unhandled function");
      }
    }
  }
  if (!ok) {
    return JSONWriter::failure(errorMessage);
  } else {
    return JSONWriter::success(); // TODO: check this, we shouldn't get here
  }
} // handleRequest

void ZoneRequestHandler::addTemperatureControlStatus(JSONWriter& json, std::shared_ptr<Zone> pZone) {
  const TemperatureConfig& hProp = pZone->temperatureConfig();
  ZoneHeatingStatus hStatus = pZone->getHeatingStatus();
  ZoneSensorStatus hSensors = pZone->getSensorStatus();

  json.add("ControlMode", static_cast<int>(hProp.m_mode));
  json.add("ControlState", hProp.m_HeatingControlState);

  switch (hProp.m_mode) {
    case HeatingControlMode::OFF:
      break;
    case HeatingControlMode::PID:
      json.add("OperationMode", pZone->getHeatingOperationMode());
      json.add("TemperatureValue");
      if (hSensors.temperatureValue) {
        json.add(*hSensors.temperatureValue);
      } else {
        json.addNull();
      }
      json.add("TemperatureValueTime", hSensors.temperatureValueTS.toISO8601());
      json.add("NominalValue");
      if (hStatus.nominalValue) {
        json.add(*hStatus.nominalValue);
      } else {
        json.addNull();
      }
      json.add("NominalValueTime", hStatus.nominalValueTS.toISO8601());
      json.add("ControlValue");
      if (hStatus.controlValue) {
        json.add(*hStatus.controlValue);
      } else {
        json.addNull();
      }
      json.add("ControlValueTime", hStatus.controlValueTS.toISO8601());
      break;
    case HeatingControlMode::ZONE_FOLLOWER:
      json.add("ControlValue");
      if (hStatus.controlValue) {
        json.add(*hStatus.controlValue);
      } else {
        json.addNull();
      }
      json.add("ControlValueTime", hStatus.controlValueTS.toISO8601());
      break;
    case HeatingControlMode::FIXED:
    case HeatingControlMode::MANUAL:
      json.add("OperationMode", pZone->getHeatingOperationMode());
      json.add("ControlValue");
      if (hStatus.controlValue) {
        json.add(*hStatus.controlValue);
      } else {
        json.addNull();
      }
      break;
  }
}

void ZoneRequestHandler::addTemperatureControlConfig(JSONWriter& json, std::shared_ptr<Zone> pZone) {
  const TemperatureConfig& hProp = pZone->temperatureConfig();

  json.add("ControlMode", static_cast<int>(hProp.m_mode));
  switch (hProp.m_mode) {
    case HeatingControlMode::OFF:
      break;
    case HeatingControlMode::PID:
      json.add("EmergencyValue", hProp.m_EmergencyValue - 100);
      json.add("CtrlKp", (double)hProp.m_Kp * 0.025);
      json.add("CtrlTs", hProp.m_Ts);
      json.add("CtrlTi", hProp.m_Ti);
      json.add("CtrlKd", hProp.m_Kd);
      json.add("CtrlImin", (double)hProp.m_Imin * 0.025);
      json.add("CtrlImax", (double)hProp.m_Imax * 0.025);
      json.add("CtrlYmin", hProp.m_Ymin - 100);
      json.add("CtrlYmax", hProp.m_Ymax - 100);
      json.add("CtrlAntiWindUp", (hProp.m_AntiWindUp > 0));
      break;
    case HeatingControlMode::ZONE_FOLLOWER:
      json.add("ReferenceZone", hProp.m_HeatingMasterZone);
      json.add("CtrlOffset", hProp.m_CtrlOffset);
      break;
    case HeatingControlMode::MANUAL:
      json.add("ManualValue", hProp.m_ManualValue - 100);
      break;
    case HeatingControlMode::FIXED:
      break;
  }
}

void ZoneRequestHandler::addTemperatureControlConfig2(JSONWriter& json, std::shared_ptr<Zone> pZone) {
  const TemperatureConfig& hProp = pZone->temperatureConfig();

  if (auto&& name = heatingControlModeName(hProp.m_mode)) {
    json.add("mode", *name);
  } else {
    json.add("mode", "unknown");
  }

  json.startObject("targetTemperatures");
  for (int i = 0; i <= HeatingOperationModeIDMax; ++i) {
    json.add(ds::str(i), hProp.targetTemperatures[i]);
  }
  json.endObject();
  json.startObject("fixedValues");
  for (int i = 0; i <= HeatingOperationModeIDMax; ++i) {
    json.add(ds::str(i), hProp.fixedValues[i]);
  }
  json.endObject();

  json.startObject("controlMode");
  json.add("emergencyValue", hProp.m_EmergencyValue - 100);
  json.add("ctrlKp", (double)hProp.m_Kp * 0.025);
  json.add("ctrlTs", hProp.m_Ts);
  json.add("ctrlTi", hProp.m_Ti);
  json.add("ctrlKd", hProp.m_Kd);
  json.add("ctrlImin", (double)hProp.m_Imin * 0.025);
  json.add("ctrlImax", (double)hProp.m_Imax * 0.025);
  json.add("ctrlYmin", hProp.m_Ymin - 100);
  json.add("ctrlYmax", hProp.m_Ymax - 100);
  json.add("ctrlAntiWindUp", (hProp.m_AntiWindUp > 0));
  json.endObject();

  json.startObject("zoneFollowerMode");
  json.add("referenceZone", hProp.m_HeatingMasterZone);
  json.add("ctrlOffset", hProp.m_CtrlOffset);
  json.endObject();

  json.startObject("manualMode");
  json.add("controlValue", hProp.m_ManualValue - 100);
  json.endObject();
}

void ZoneRequestHandler::addTemperatureControlValues(JSONWriter& json, std::shared_ptr<Zone> pZone) {
  const TemperatureConfig& hProp = pZone->temperatureConfig();

  switch (hProp.m_mode) {
    case HeatingControlMode::OFF:
      break;
    case HeatingControlMode::PID:
      json.add("Off", hProp.targetTemperatures[0]);
      json.add("Comfort", hProp.targetTemperatures[1]);
      json.add("Economy", hProp.targetTemperatures[2]);
      json.add("NotUsed", hProp.targetTemperatures[3]);
      json.add("Night", hProp.targetTemperatures[4]);
      json.add("Holiday", hProp.targetTemperatures[5]);
      json.add("Cooling", hProp.targetTemperatures[6]);
      json.add("CoolingOff", hProp.targetTemperatures[7]);
      break;
    case HeatingControlMode::ZONE_FOLLOWER:
      break;
    case HeatingControlMode::FIXED:
      json.add("Off", hProp.fixedValues[0]);
      json.add("Comfort", hProp.fixedValues[1]);
      json.add("Economy", hProp.fixedValues[2]);
      json.add("NotUsed", hProp.fixedValues[3]);
      json.add("Night", hProp.fixedValues[4]);
      json.add("Holiday", hProp.fixedValues[5]);
      json.add("Cooling", hProp.fixedValues[6]);
      json.add("CoolingOff", hProp.fixedValues[7]);
      break;
    case HeatingControlMode::MANUAL:
      break;
  }
}

  std::string ZoneRequestHandler::getActions(std::shared_ptr<Zone> pZone, std::shared_ptr<Group> pGroup, const RestfulRequest& _request) {
    JSONWriter json;
    auto locale = _request.getLocale(DSS::getInstance()->localeGenerator());
    // If 'cluster' param exists, we can omit 'application'
    const auto clusterId = strToIntDef(_request.getParameter("cluster"), 0);
    if (clusterId != 0) {
      const auto& requestedApp = m_Apartment.getCluster(clusterId);
      json.startArray("actions");
      for (const auto& action : requestedApp->actions()) {
        json.startObject();
        json.add("id", action.id);
        json.add("title", dsTr(locale, action.title));
        json.endObject();
      }
      json.endArray();
      return json.successJSON();
    }

    // Otherwise, we have to parse 'application' param
    const auto appType = applicationTypeFromString(_request.getRequiredParameter("application"));
    const auto requestedApp = m_Apartment.getZone(pZone->getID())->getGroup(static_cast<int>(appType));
    json.startArray("actions");
    if (requestedApp == nullptr) {
      return JSONWriter::failure("unsupported application id");
    }
    for (const auto& action : requestedApp->actions()) {
      json.startObject();
      json.add("id", action.id);
      json.add("title", dsTr(locale, action.title));
      json.endObject();
    }
    json.endArray();

    return json.successJSON();
  } // getActions

  std::string ZoneRequestHandler::callAction(std::shared_ptr<Zone> pZone, std::shared_ptr<Group> pGroup, const RestfulRequest& _request) {
    JSONWriter json;
    const auto action = _request.getRequiredParameter("action");
    const auto category = web::handler::getCategory(_request);

    // If 'cluster' param exists, we can omit 'application'
    const auto clusterId = strToIntDef(_request.getParameter("cluster"), 0);
    if (clusterId != 0) {
      const auto requestedApp = m_Apartment.getCluster(clusterId);
      requestedApp->callAction(CallOrigin::json, category, action);
      return json.successJSON();
    }

    const auto zoneId = pZone->getID();
    const auto appType = applicationTypeFromString(_request.getRequiredParameter("application"));
    const auto requestedApp = m_Apartment.getZone(zoneId)->getGroup(static_cast<int>(appType));
    if (requestedApp == nullptr) {
      return JSONWriter::failure("unsupported application id");
    }
    requestedApp->callAction(CallOrigin::json, category, action);
    return json.successJSON();
  } // callAction

} // namespace dss
